#ifndef _DAQNAVI_DELEGATE_H
#define _DAQNAVI_DELEGATE_H

//----------------------------------------------------------------
// base class: event delegate manager for user/kernel event
//----------------------------------------------------------------
template<class TKey, class TValue>
class XEventManager
{
protected:
   typedef TKey XKey;

   struct XValue : public TValue {
      int32 Deleted;

      XValue() : Deleted(0) 
      { }
      XValue(XValue const &x) : TValue(x), Deleted(x.Deleted) 
      { }
      XValue(TValue const &t, int32 deleted = 0) : TValue(t), Deleted(deleted)
      { }
   };

   typedef std::multimap<XKey, XValue>                          XEventMap;
   typedef typename std::multimap<XKey, XValue>::iterator       iterator; 
   typedef typename std::multimap<XKey, XValue>::const_iterator const_iterator;

public:
   void Add(TKey const &key, TValue const &val)
   {
      // Does the event handler is duplicated with one of the items in the map?
      iterator itor = Find(key, val);

      // Not found!
      if (itor == m_xevents.end()) {
         XValue xv(val);
         itor = m_xevents.insert(std::make_pair(key, xv));
      }

      // Reset the 'Deleted' flag
      itor->second.Deleted = false;
   }

   void Remove(TKey const &key, TValue const &val)
   {
      // Look for the specified item
      iterator itor = Find(key, val);

      // Bingo! Found it!
      if (itor != m_xevents.end()) {
         itor->second.Deleted = true;
      }
   }

   void Remove(TKey const &key)
   {
      iterator itor = m_xevents.lower_bound(key);

      for (; itor != m_xevents.end() && itor->first == key; ++itor) {
         itor->second.Deleted = true;
      }
   }

   bool IsExist(TKey const &key, TValue const &val)
   {
      iterator itor = Find(key, val);
      return itor != m_xevents.end() && !itor->second.Deleted;
   }

   bool IsExist(TKey const &key)
   {
      const_iterator itor = m_xevents.lower_bound(key);
      for (; itor != m_xevents.end() && itor->first == key; ++itor) {
         if (!itor->second.Deleted) {
            return true;
         }
      }
      return false;
   }

   int Count(TKey const &key)
   {
      int count = 0;
      const_iterator itor = m_xevents.lower_bound(key);
      for (; itor != m_xevents.end() && itor->first == key; ++itor) {
         if (!itor->second.Deleted) {
            ++count;
         }
      } 
      return count;
   }

protected:
   iterator Find(TKey const &key, TValue const &val)
   {
      iterator itor = m_xevents.lower_bound(key);

      for (; itor != m_xevents.end() && itor->first == key; ++itor) {
         if (itor->second == val) {
            return itor;
         }
      }

      return m_xevents.end();
   }

protected:
   XEventMap  m_xevents;
};

//----------------------------------------------------------------
// user event delegate manager: 1-to-N map
//----------------------------------------------------------------
typedef void (BDAQCALL * UserEventProc)(void * sender, void * args, void *param); 
typedef struct UserEvent {
   UserEventProc Proc;
   void         *Param;

   UserEvent(UserEventProc proc = NULL, void *param = NULL) : Proc(proc), Param(param) 
   {}
   
   bool operator == (const UserEvent & r)
   {
      // ignore the 'Param' field if it is null
      if (r.Param && this->Param != r.Param) { return false; }
      return this->Proc == r.Proc;
   }
} UserEvent;

class UserEventManager : public XEventManager<std::pair<void *, EventId>, UserEvent>
{
   static XEventManager<std::pair<void *, EventId>, UserEvent> & Base() { return Instance(); }

public:
   // single-ton: one manager for all controls.
   static UserEventManager& Instance()
   {
      static UserEventManager manager;
      return manager;
   }

   static void Fire(void *sender, EventId id, void *args)
   {
      std::pair<void *, EventId> key = std::make_pair(sender, id);
      const_iterator itor = Instance().m_xevents.lower_bound(key);
      for (; itor != Instance().m_xevents.end() && itor->first == key; ++itor) {
         if (!itor->second.Deleted) {
            assert(id == itor->first.second);
            itor->second.Proc(sender, args, itor->second.Param);
         }
      }
   }

   //-------------------------------------------------------------------------------
   // helpers
   static void Add(void *sender, EventId id, UserEventProc proc, void *param)
   {
      Base().Add(std::make_pair(sender, id), UserEvent(proc, param));
   }
   static void Remove(void *sender, EventId id, UserEventProc proc, void *param)
   {
      Base().Remove(std::make_pair(sender, id), UserEvent(proc, param));
   }
   static void Remove(void *sender, EventId id)
   {
      Base().Remove(std::make_pair(sender, id));
   }
   static bool IsExist(void *sender, EventId id, UserEventProc proc, void *param)
   {
      return Base().IsExist(std::make_pair(sender, id), UserEvent(proc, param));
   }
   static bool IsExist(void *sender, EventId id)
   {
      return Base().IsExist(std::make_pair(sender, id));
   }
   static int Count(void *sender, EventId id)
   {
      return Base().Count(std::make_pair(sender, id));
   }
};

//----------------------------------------------------------------
// driver event delegate manager: 1-to-N map
//----------------------------------------------------------------
typedef void (BDAQCALL * KernEventProc)(EventId id, void *param); 

typedef struct KernEvent {
   KernEventProc Proc;
   void         *Param;

   KernEvent(KernEventProc proc = NULL, void *param = NULL) : Proc(proc), Param(param) 
   {}

   bool operator == (const KernEvent & r)
   {
      // ignore the 'Param' field if it is null
      if (r.Param && this->Param != r.Param) { return false; }
      return this->Proc == r.Proc;
   }
} KernEvent;

class KernEventManager : public XEventManager<HANDLE, KernEvent>
{
   static XEventManager<HANDLE, KernEvent> & Base() { return Instance(); }

public:
   // single-ton: one manager for all controls.
   static KernEventManager& Instance()
   {
      static KernEventManager manager;
      return manager;
   }

   static void Fire(HANDLE handle, EventId id)
   {
      const_iterator itor = Instance().m_xevents.lower_bound(handle);
      for (; itor != Instance().m_xevents.end() && itor->first == handle; ++itor) {
         if (!itor->second.Deleted) {
            itor->second.Proc(id, itor->second.Param);
         }
      }
   }

   //-------------------------------------------------------------------------------
   // helpers
   static void Add(HANDLE handle, KernEventProc proc, void *param)
   {
      Base().Add(handle, KernEvent(proc, param));
   }
   static void Remove(HANDLE handle, KernEventProc proc, void *param)
   {
      Base().Remove(handle, KernEvent(proc, param));
   }
   static void Remove(HANDLE handle)
   {
      Base().Remove(handle);
   }
   static bool IsExist(HANDLE handle, KernEventProc proc, void *param)
   {
      return Base().IsExist(handle, KernEvent(proc, param));
   }
   static bool IsExist(HANDLE handle)
   {
      return Base().IsExist(handle);
   }
   static int Count(HANDLE handle)
   {
      return Base().Count(handle);
   }
};

#endif
