#ifndef _DAQNAVI_THREAD_H
#define _DAQNAVI_THREAD_H

#if defined(_WIN32) || defined(WIN32) 
#  include <process.h>
#else
#  include <pthread.h>
#endif

class DaqThread
{
public:
   DaqThread() : m_canRun(FALSE), m_running(FALSE), m_thread(0)
   { }
   virtual ~DaqThread() 
   {
#  if defined(_WIN32) || defined(WIN32)
      if (m_thread) { CloseHandle(m_thread); }
#  endif
   }

   bool Start()
   {
      if (AtomicCompareExchange(&m_canRun, TRUE, FALSE) == FALSE) {
         m_running  = TRUE;
         if (!StartThread()) {
            m_canRun = FALSE;
            m_running  = FALSE;
         }
      }
      return !!m_canRun;
   }

   void Stop()
   {
      if (AtomicCompareExchange(&m_canRun, FALSE, TRUE) == TRUE) { 
         Wakeup();
      }
   }
   
   bool Wait(int timeout_ms)    
   {
      if (!m_thread) { return true; }
      return WaitForThread(timeout_ms);
   }

   bool CanRun()  { return !!m_canRun; }
   bool IsAlive() { return IsThreadAlive(); }

protected:
   virtual void Prerun(bool startup) {/* Do any thread related initialization/cleanup in the new thread context;*/ }
   virtual int  Run()    = 0;
   virtual void Wakeup() = 0;
   
private:
#  if defined(_WIN32) || defined(WIN32)
   typedef HANDLE ThreadHandle;
   static unsigned BDAQCALL DaqThreadProc(void *ctx) 
   { 
      DaqThread *_this = static_cast<DaqThread *>(ctx);
      unsigned ret = 0;

      // Do the thread related initialization in the new thread context
      _this->m_running = TRUE;
      _this->Prerun(true);

      // call thread body
      try {
         ret = _this->Run(); 
      } catch(...) {
         daq_trace(("Unhandled Exception! handle = 0x%x\n", _this->m_thread));
      }

      _this->Prerun(false);
      _this->m_running = FALSE;
      return ret;
   }
   bool IsThreadAlive()
   {
      DWORD exitCode;
      return m_thread != NULL && GetExitCodeThread(m_thread, &exitCode) && exitCode == STILL_ACTIVE; 
   }
   bool StartThread()
   {
      m_thread = (ThreadHandle)_beginthreadex(NULL, 0, DaqThreadProc, this, 0, NULL);
      return m_thread != NULL;
   }
   bool WaitForThread(int timeout_ms)
   {
      return WaitForSingleObject(m_thread, timeout_ms) == WAIT_OBJECT_0;
   }
#  else
   typedef pthread_t ThreadHandle;
   static void * BDAQCALL DaqThreadProc(void *ctx) 
   { 
      DaqThread *_this = static_cast<DaqThread *>(ctx);
      int ret = 0;

      // Do the thread related initialization in the new thread context
      _this->m_running = TRUE;
      _this->Prerun(true);

      // call thread body
      try {
         ret = _this->Run(); 
      } catch(...) {
         daq_trace(("Unhandled Exception! handle = 0x%x\n", m_thread));
      }

      _this->Prerun(false);
      _this->m_running = FALSE;
      return (void *)(int64)ret;
   }
   bool IsThreadAlive()
   {
      return m_thread && m_running;
   }
   bool StartThread()
   {
      return pthread_create(&m_thread, NULL, DaqThreadProc, this) == 0;
   }
   bool WaitForThread(int timeout_ms)
   {
      int sec = timeout_ms / 1000;
      int ns  = (timeout_ms - sec * 1000) * 1000 * 1000;
      timespec ts;

      if (clock_gettime(CLOCK_REALTIME, &ts) == -1) {
         sleep(sec + ns ? 1 : 0);
         return m_running;
      } else {
         ts.tv_sec  += sec;
         ts.tv_nsec += ns;
         return pthread_timedjoin_np(m_thread, NULL, &ts) == 0;
      }
   }
#  endif

private:
   volatile int32 m_canRun;    // whether the thread can run or not.
   volatile int32 m_running;   // whether the thread is running or not. 
   ThreadHandle   m_thread;
};

template <class T>  
class DaqThreadLocal  
{  
public:  
   DaqThreadLocal() 
   {  
      TlsKeyAlloc(m_tlsKey);
   }  

   T * Get() 
   {  
      return (T *)TlsValueGet(m_tlsKey);
   }  
   T * Set(T *value)
   {
      void * cur = TlsValueGet(m_tlsKey);  
      TlsValueSet(m_tlsKey, value); 
      return (T *)cur;
   }

protected:
#if defined(_WIN32) || defined(WIN32)
   typedef uint32 TlsKey;
   static void   TlsKeyAlloc(TlsKey &key)                    { key = TlsAlloc();   }
   static void * TlsValueGet(TlsKey const &key)              { return TlsGetValue(key); }
   static void   TlsValueSet(TlsKey const &key, void *value) { TlsSetValue(key, value); }
#else
   typedef pthread_key_t TlsKey;
   static void   TlsKeyAlloc(TlsKey &key)                    { pthread_key_create(&key, NULL); }
   static void * TlsValueGet(TlsKey const &key)              { return pthread_getspecific(key); }
   static void   TlsValueSet(TlsKey const &key, void *value) { pthread_setspecific(key, value); }
#endif

private:  
   TlsKey  m_tlsKey;
};  

#endif
