#ifndef _DAQNAVI_BUFFERED_AI_CTRL_IMPL_H
#define _DAQNAVI_BUFFERED_AI_CTRL_IMPL_H

class BufferedAiCtrlImpl : public AiCtrlBaseImpl
{
public:
   //
   // event:
   // Use the default implementation in DaqCtrlBase

   // method
   ErrorCode Prepare()
   {
      AutoLock<MutexLock> lock(PropLock());

      ErrorCode ret;
      if (!getInitialized()) {
         ret = ErrorFuncNotInited;
      } else if (IsRunning()) {
         ret = ErrorFuncBusy; 
      } else if (IsReady()) { 
         ret = Success;       
      } else {

         // DN4 sync-one-buffered AI
         if (!m_dn4EventRegistered && ModuleRef().getIntfType() == IntfType::DN4_OBJ) {
            RegisterKernEvent(EvtBufferedAiStopped, DN4_ProcessKernEvent);
            m_dn4EventRegistered = true;
         }

         // get scan channel count for the performance of method 'GetData'
         ModProp().Get(CFG_ScanChannelCount, m_scanCHCount);

         // configuration h/w. Only set the property we cached,
         // the others had been set when the user calls 'set' method of the property
         // 
         // set: sectionPerChan --> sectionDataCount
         ModProp().Set(CFG_SectionDataCount, m_scanChannel.getIntervalCount() * m_scanCHCount);

         // set: trigger delay count
         int32 multiplier = 1;
         if (getFeatures()->getSamplingMethod() == EqualTimeSwitch) {
            multiplier = m_scanCHCount;
         }
         for (int i = 0; i < getFeatures()->getTriggerCount(); ++i) {
            ModProp().Set(prop_id_trigDelay[i], m_triggers[i].getDelayCount() * multiplier);
            //Read back the adjustable parameters
            int32 delay = 0;
            ModProp().Get(prop_id_trigDelay[i], delay);
            m_triggers[i].adjustDelayCount(delay / multiplier);
         }

         // prepare buffered ai
         m_dataBufLen = m_scanChannel.getSamples() * m_scanCHCount;
         ret = ModuleRef().BfdAiPrepare((uint32 &)m_dataBufLen, m_dataPtr);
         setState(ret == Success ? Ready : Idle);
      }
      return ret;
   }

   // synchronous & one-shot
   ErrorCode RunOnce() 
   { 
      return StartBfdAi(false, false); 
   }

   // asynchronous, one-shot or streaming, depends on the property 'Streaming'
   ErrorCode Start()   
   { 
      return StartBfdAi(m_streaming, true); 
   }

   ErrorCode Stop()
   {
      if (IsIdle()) { return Success;  }
      setState(Stopped);
      return ModuleRef().BfdAiStop();
   }

   void Release()
   {
      // Fix bug#19706: Call driver 'Release' function anyway to release resource, if any.
      // It is harmless operation even if the device is in 'FAI IDLE' state.
      // if (IsIdle()) { return; }
      // Call driver 'Release' only if we had called the method 'Prepare'
      if (m_dataPtr) {
         ModuleRef().BfdAiRelease();
      }

      setState(Idle);
      m_dataPtr    = NULL;
      m_dataBufLen = 0;
   }

   ErrorCode GetData(int32 dataType, int32 count, void *buffer)
   {
      AutoLock<MutexLock> lock(BuffLock());

      ErrorCode ret;
      if (dataType == m_dataSize) { 
         // Get raw data
         ret = ModuleRef().BfdAiGetData((uint32 &)count, buffer, NULL, NULL, NULL, NULL);
      } else if (dataType == sizeof(double)) {
         // Get scaled data
         ret = ModuleRef().BfdAiGetData((uint32 &)count, NULL, (double*)buffer, NULL, NULL, NULL);
      } else {
         // Not supported data type
         ret = ErrorFuncNotSpted;
      }

      // Update the 'Read Position'
      m_dataReadyCleared = true;
      ModEvent().ClearFlag(EvtBufferedAiDataReady, count);
      return ret;
   }

   bool GetBufferStatus(int32 & count, int32 & offset)
   {
      EventStatus es = {sizeof(es)};
      ModEvent().GetStatus(EvtBufferedAiDataReady, es);
      count  = (es.AiDataReady.DataCount / m_scanCHCount) * m_scanCHCount;
      offset = es.AiDataReady.DataOffset;

      return false;
   }

   // property
   void * getBuffer()   
   {
      return m_dataPtr; 
   }

   int32 getBufferCapacity() 
   { 
      // Fix bug: DO NOT return 0 here due to the legacy APP maybe use this value to allocate its buffer.
      return m_dataBufLen ? m_dataBufLen : m_scanChannel.getSamples() * m_scanCHCount;  
   }

   ScanChannelImpl  * getScanChannel()    { return &m_scanChannel;}
   ConvertClockImpl * getConvertClock()   { return &m_convClock;  }
   ScanClockImpl    * getScanClock()      { return getFeatures()->getBurstScanSupported() ? &m_scanClock : NULL;   }
   
   // trigger 0, 1, ... , x
   TriggerImpl * getTrigger(int32 trigger)
   { 
      TriggerImpl *x = NULL;

      // Bug#14478
      //if ((uint32)trigger < (uint32)getFeatures()->getTriggerCount()) {
      if ((uint32)trigger < ARRAY_SIZE(m_triggers)) {
         x = &m_triggers[trigger];
      }

      return x;
   }
   
   bool getStreaming()              
   {       
      return m_streaming;     
   }
   ErrorCode setStreaming(bool value)   
   {      
      if (IsRunning()) { return ErrorFuncBusy; } 
      m_streaming = value;
      return Success;
   }

public: // ctor & dtor
   BufferedAiCtrlImpl()
   {
      m_dataPtr           = NULL;
      m_dataBufLen        = 0;
      m_dataReadyCleared  = false;
      m_streaming         = false;

      m_scanChannel.setOwner(this, false);
      m_convClock.setOwner(this);
      m_scanClock.setOwner(this);

      for (int i = 0; i < MAX_TRIG_COUNT; ++i) {
         m_triggers[i].setOwner(this, i);
      }

      m_dn4StopEvent = CreateEvent(NULL, false, false, NULL);
      m_dn4EventRegistered = false;
   }

   virtual ~BufferedAiCtrlImpl()
   {
      // clear all user event delegate
      for (unsigned i = 0; i < ARRAY_SIZE(bfd_ai_events); ++i) {
         UserEventManager::Remove(getOuter(), bfd_ai_events[i]);
      }

      if (m_dn4StopEvent) {
         CloseHandle(m_dn4StopEvent);
      }
   }

protected: 
   //------------------------------------------------------------------------
   // override the method in base class
   //------------------------------------------------------------------------
   virtual std::vector<AccessMode> WantedModes()
   {
      AccessMode spted[] = { ModeWrite, ModeWriteWithReset };
      return std::vector<AccessMode>(&spted[0], &spted[2]);
   }

   virtual bool IsSuitable(DaqModule &module)
   {
      if (AiCtrlBaseImpl::IsSuitable(module)) {
         return DaqProperty(module).GetLength(CFG_FeatureConvertClockSources) != 0;
      }
      return false;
   }

   // update the control state
   virtual void RefreshState(ControlState &curr)
   {
      if (curr == Running) {
         EventStatus es = {sizeof(es)};
         ModEvent().GetStatus(EvtBufferedAiStopped, es);
         if (es.AiDataReady.Stopped) {
            setState(Stopped);
         }
      }
   }

   // initialize/deinitialize the control 
   virtual ErrorCode InitializeControl(DeviceInformation const &x)
   {
      // let's the base do initialization at first.
      ErrorCode ret = AiCtrlBaseImpl::InitializeControl(x);
      if (ret != Success) { return ret; }

      // buffered AI not supported
      if (!getFeatures()->getBufferedAiSupported()) {
         return ErrorFuncNotSpted;
      }

      LoadProperties();

      // activate the events   
      for (unsigned i = 0; i < ARRAY_SIZE(bfd_ai_events); ++i) {
         if (UserEventManager::IsExist(getOuter(), bfd_ai_events[i])) {
            RegisterKernEvent(bfd_ai_events[i]);
         }
      }
      return Success;
   }

   virtual void DeinitializeControl()
   {
      // release buffered AI
      Release();

      // deactivate the events        
      for (unsigned i = 0; i < ARRAY_SIZE(bfd_ai_events); ++i) {
         UnregisterKernEvent(bfd_ai_events[i]);
      }

      // for DN4 sync-one-buffered AI
      if (m_dn4EventRegistered) {
         UnregisterKernEvent(EvtBufferedAiStopped, DN4_ProcessKernEvent);
         m_dn4EventRegistered = false;
      }

      // reset properties to its default value
      m_dataSize = 0;

      m_scanChannel.adjustIntervalCount(0);
      m_scanChannel.adjustSamples(0);

      for (int i = 0; i < getFeatures()->getTriggerCount(); ++i) {
         m_triggers[i].adjustDelayCount(0);
      }

      // let's the base do de-initialization at last.
      AiCtrlBaseImpl::DeinitializeControl();
   }

   //
   // Reload the properties' value from device
   virtual void ReloadProperties()
   {
      AiCtrlBaseImpl::ReloadProperties();
      LoadProperties();
   }

   void LoadProperties()
   {
      // cache the raw data size for the performance of method 'GetData'
      m_dataSize = getFeatures()->getDataSize();

      // get the property default value from device
      m_scanCHCount = m_scanChannel.getChannelCount();
      assert(m_scanCHCount);

      int32 sectPerCh = 1024;
      if (ModuleRef().getIntfType() == IntfType::DN4_OBJ) {
         ModProp().Get(CFG_RecordSectionLength, sectPerCh);
      } else {
         ModProp().Get(CFG_SectionDataCount, sectPerCh);
         sectPerCh /= m_scanCHCount;
      }

      m_scanChannel.adjustIntervalCount(sectPerCh);
      m_scanChannel.adjustSamples(sectPerCh ? sectPerCh * 2 : 1024);

      // trigger delay count
      int32 multiplier = 1;
      if (getFeatures()->getSamplingMethod() == EqualTimeSwitch) {
         multiplier = m_scanCHCount;
      }

      for (int i = 0; i < getFeatures()->getTriggerCount(); ++i) {
         int32 delay = 0;
         ModProp().Get(prop_id_trigDelay[i], delay);
         m_triggers[i].adjustDelayCount(delay / multiplier);
      }
   }

   //------------------------------------------------------------------------
   // internal methods
   //------------------------------------------------------------------------
   ErrorCode StartBfdAi(bool streaming, bool async)
   {
      // initialize if it is necessary
      if (ModuleRef().getIntfType() == IntfType::DN4_OBJ) {
         int32 sectLength = m_scanChannel.getIntervalCount();
         int32 sectCount  = streaming ? 0 : x_max(1, m_scanChannel.getSamples() / m_scanChannel.getIntervalCount());
         int32 oldX;

         ModProp().Get(CFG_RecordSectionLength, oldX);
         if (oldX != sectLength) { 
            ModProp().Set(CFG_RecordSectionLength, sectLength);
            setState(Idle);
         }

         ModProp().Get(CFG_RecordSectionCount, oldX);
         if (oldX != sectCount) {
            ModProp().Set(CFG_RecordSectionCount, sectCount);
            setState(Idle);
         }

         ModProp().Get(CFG_RecordCycles, oldX);
         if (oldX != 1) {
            ModProp().Set(CFG_RecordCycles, (int32)1);
            setState(Idle);
         }
      }

      ErrorCode ret = Prepare();
      if (ret != Success) { return ret; }

      // start buffered AI
      if (streaming) {
         async = true;
         ret = ModuleRef().BfdAiStart();
      } else {
         switch(ModuleRef().getIntfType())
         {
         default: 
            ret = ErrorFuncNotSpted; 
            break;
         case IntfType::DN3_API: 
            ret = ((DN3Api::DN3Ai &)ModuleRef()).BfdAiRunOnce(async); 
            break;
         case IntfType::DN3_OBJ: 
            ret = ((DN3Obj::DN3Ai &)ModuleRef()).BfdAiRunOnce(async); 
            break;
         case IntfType::DN4_OBJ: 
            // clear DN4 notify event state. 
            ResetEvent(m_dn4StopEvent);

            ret = ModuleRef().BfdAiStart();
            if (!async && ret == Success) {
               WaitForSingleObject(m_dn4StopEvent, INFINITE);
            }
            break;
         }
      }

      if (ret == Success) {
         setState(async ? Running : Stopped);
      }

      return ret;
   }

   // -----------------------------------------------------------------
   // Check whether the event id is supported by a counter control or not
   // -----------------------------------------------------------------
   virtual bool IsValidEvent(EventId id)
   {
      return ARRAY_INDEX_OF(bfd_ai_events, id) != -1;
   }

   // -----------------------------------------------------------------
   // driver event receiver for buffered AI
   // -----------------------------------------------------------------
   virtual void OnKernEvent(EventId id)
   {
      // Discard the un-expected events
      if (IsIdle()) { return; }

      EventStatus    es   = {sizeof(es)};
      BfdAiEventArgs args = {id};

      switch(id)
      {
      default:
         break;
      case EvtBufferedAiDataReady:  /* data ready */
         m_dataReadyCleared = false;

         ModEvent().GetStatus(id, es);
         args.Count  = (es.AiDataReady.DataCount / m_scanCHCount) * m_scanCHCount;
         args.Offset = es.AiDataReady.DataOffset;
         args.MarkCount = es.AiDataReady.MarkCount;
         if (args.Count) { 
            UserEventManager::Fire(getOuter(), id, &args); 
         }

         // clear the event flag if the used hadn't clear it by calling 'GetData'.
         // using a count '0' to clear only the 'had sent' flag, but not effect the 'read pos'.
         if (!m_dataReadyCleared) { ModEvent().ClearFlag(id); }
         break;
      case EvtBufferedAiOverrun:  /* overrun */
         ModEvent().GetStatus(id, es);
         args.Count  = es.AiOverrun.DataCount;
         args.Offset = es.AiOverrun.DataOffset;
         UserEventManager::Fire(getOuter(), id, &args);
         ModEvent().ClearFlag(id);            // clear the event flag after the user had processed it.
         break;
      case EvtBufferedAiStopped:    /* stopped */
         ModEvent().GetStatus(id, es);
         args.Count  = es.AiDataReady.DataCount;
         args.Offset = es.AiDataReady.DataOffset; 
         args.MarkCount = es.AiDataReady.MarkCount;
         UserEventManager::Fire(getOuter(), id, &args);
         // need not clear the event flag
         break;
      case EvtBufferedAiCacheOverflow:  /* cache overflow */// no status data
         UserEventManager::Fire(getOuter(), id, &args);
         ModEvent().ClearFlag(id);            // clear the event flag after the user had processed it.
         break;
      case EvtBufferedAiBurnOut: /*burn out*/// no status data
         UserEventManager::Fire(getOuter(), id, &args);
         ModEvent().ClearFlag(id);            // clear the event flag after the user had processed it.
      }
   };

   // -----------------------------------------------------------------
   // DN4 event receiver for sync-one-buffer AI
   // -----------------------------------------------------------------
   static void BDAQCALL DN4_ProcessKernEvent(EventId id, void *param)
   {
      BufferedAiCtrlImpl *_this = (BufferedAiCtrlImpl *)param;
      if (id == EvtBufferedAiStopped) {
         SetEvent(_this->m_dn4StopEvent);
      }
   }

private:
   void *   m_dataPtr;
   int32    m_dataBufLen;
   bool     m_dataReadyCleared; // whether the data ready event has been cleared in 'GetData', used by DataReadyHandler.

   bool     m_streaming;        // buffered AI running mode : streaming VS one-shot.
   int32    m_dataSize;         // for performance of 'GetData' method, cache this value, it will be initialized in 'InitializeControl'
   int32    m_scanCHCount;      // for performance of 'GetData' method, cache this value, it will be initialized in 'Prepare'
   HANDLE   m_dn4StopEvent;
   bool     m_dn4EventRegistered; // flag indicating whether the event for sync-one-buffered is registered or not.

   ScanChannelImpl  m_scanChannel;
   ConvertClockImpl m_convClock;
   ScanClockImpl    m_scanClock;
   TriggerImpl      m_triggers[MAX_TRIG_COUNT];
};

#endif
