#ifndef _DAQNAVI_WAVEFORM_AI_CTRL_IMPL_H
#define _DAQNAVI_WAVEFORM_AI_CTRL_IMPL_H

class WaveformAiCtrlImpl : public AiCtrlBaseImpl
{
public:
   //
   // event:
   // Use the default implementation in DaqCtrlBase

   // method
   ErrorCode Prepare()
   {
      AutoLock<MutexLock> lock(PropLock());
    
      ErrorCode ret;
      if (!getInitialized()) {
         ret = ErrorFuncNotInited;
      } else if (IsRunning()) {
         ret = ErrorFuncBusy;
      } else if (IsReady()){
         ret = Success;
      } else {

         // for blocked 'GetData'
         if (!m_dn4EventRegistered) {
            RegisterKernEvent(EvtBufferedAiDataReady, DN4_ProcessKernEvent);
            RegisterKernEvent(EvtBufferedAiStopped, DN4_ProcessKernEvent);
            m_dn4EventRegistered = true;
         }

         // Must wait for the AI 'stopped' before 'GetData' if:
         // 1. one buffer AI, not in re-trigger mode; 
         // 2. the only trigger mode is 'delay to stop'.
         bool delay_to_stop_en = false;
         for (int i = 0; i < getFeatures()->getTriggerCount(); ++i) {
            if (  getTrigger(i)->getSource() != SignalNone 
               && getTrigger(i)->getAction() == DelayToStop
               && getTrigger(i)->getDelayCount() != 0) {
                  delay_to_stop_en = true; 
            }
         }
         m_waitforStopped = (m_record.getCycles() == 1 && m_record.getSectionCount() != 0 && delay_to_stop_en);

         // get scan channel count for the performance of method 'GetData'
         ModProp().Get(CFG_ScanChannelCount, m_scanCHCount);

         // configuration h/w. Only set the property we cached,
         // the others had been set when the user calls 'set' method of the property
 
         // set: trigger delay count
         int32 multiplier = 1;
         if (getFeatures()->getSamplingMethod() == EqualTimeSwitch) {
            multiplier = m_scanCHCount;
         }
         for (int i = 0; i < getFeatures()->getTriggerCount(); ++i) {
            ModProp().Set(prop_id_trigDelay[i], m_triggers[i].getDelayCount() * multiplier);   
            //Read back the adjustable parameters
            int32 delay = 0;
            ModProp().Get(prop_id_trigDelay[i], delay);
            m_triggers[i].adjustDelayCount(delay / multiplier);
         }

         // prepare buffered ai
         m_dataBufLen = 0;
         ret = ModuleRef().BfdAiPrepare((uint32 &)m_dataBufLen, m_dataPtr);
         setState(ret == Success ? Ready : Idle);
      }
      return ret;
   }

   // one-shot or streaming, depends on the property 'Record.SectionCount'
   ErrorCode Start()   
   { 
      // initialize if it is necessary
      ErrorCode ret = Prepare();
      if (ret != Success) { return ret; }

      // clear DN4 notify event state. 
      ResetEvent(m_dataArrived);

      // start buffered AI
      m_recordCount = 0;
      ret = ModuleRef().BfdAiStart();
      if (ret == Success) { setState(Running); }

      return ret;
   }

   ErrorCode Stop()
   {
      if (IsIdle()) { return Success; }

      setState(Stopped);
      return ModuleRef().BfdAiStop();
   }

   void Release()
   {
      // Fix bug#19706: Call driver 'Release' function anyway to release resource, if any.
      // It is harmless operation even if the device is in 'FAI IDLE' state.
      // if (IsIdle()) { return; }
      // Call driver 'Release' only if we had called the method 'Prepare'
      if (m_dataPtr) {
         ModuleRef().BfdAiRelease();
      }

      setState(Idle);
      m_dataPtr    = NULL;
      m_dataBufLen = 0;
   }

   ErrorCode GetData(int32 dataType, int32 count, void *buffer, int32 timeout, 
      int32 *returned, double *startTime, int32 *markCount, DataMark *markBuf)
   {
      if (dataType != m_dataSize && dataType != sizeof(double)) { 
         if (returned) { *returned = 0; }
         return ErrorFuncNotSpted; // Not supported data type
      }
      
      ErrorCode  ret = Success;
      EventStatus es = {sizeof(es)};

      uint32   dtLen;
      uint8   *dtPos = (uint8*)buffer;

      uint32    mkReq = markCount ? *markCount : 0; 
      uint32    mkLen;
      DataMark *mkPos  = markBuf;
      
      AutoLock<MutexLock> lock(BuffLock());

      do {
         // Get running state before we read data
         ModEvent().GetStatus(EvtBufferedAiStopped, es);

         if (m_waitforStopped 
            && !es.AiDataReady.Stopped 
            && m_recordCount >= es.AiDataReady.RecordCount) {
            if (!timeout) { break; }

            timeout = WaitEventTimeout(m_dataArrived, timeout); 
            continue;
         }

         // Get data
         dtLen = count;
         mkLen = mkReq;
         if (dataType == sizeof(double)) {
            ret = ModuleRef().BfdAiGetData(dtLen, NULL, (double*)dtPos, startTime, &mkLen, mkPos);
         } else {
            ret = ModuleRef().BfdAiGetData(dtLen, dtPos, NULL, startTime, &mkLen, mkPos);
         }

         if (BioFailed(ret)) { break; } // Error occurred!

         if (dtLen) { startTime = NULL; } // save the first time-stamp and ignore the others in this loop.

         count -= dtLen;
         dtPos += dtLen * dataType;        

         mkReq -= mkLen;
         mkPos += mkLen;

         // Update the 'Read Position'
         m_dataReadyCleared = true;
         ModEvent().ClearFlag(EvtBufferedAiDataReady, dtLen, 0);

         if (ret == WarningRecordEnd) { // clear conversion stop after we had read a record
            AtomicIncrement((int32*)&m_recordCount);
         }

         // Wait data continuously if it is necessary
         if (!count || es.AiDataReady.Stopped || ret != Success) { break; }

         // Wait data ready and update the timeout value
         if (timeout) { timeout = WaitEventTimeout(m_dataArrived, timeout); }
      } while (timeout);

      if (returned)  { *returned  = uint32(dtPos - (uint8*)buffer) / dataType; }
      if (markCount) { *markCount = uint32(mkPos - markBuf); } 

      // Is something wrong in CSCL ?
      if (count && ret == Success) { 
         ret = es.AiDataReady.Stopped ? WarningFuncStopped : WarningFuncTimeout;
      }

      return ret;
   }

   bool GetBufferStatus(int32 & count, int32 & offset)
   {
      EventStatus es = {sizeof(es)};
      ModEvent().GetStatus(EvtBufferedAiDataReady, es);
      count  = (es.AiDataReady.DataCount / m_scanCHCount) * m_scanCHCount;
      offset = es.AiDataReady.DataOffset;

      return false;
   }

   // property
   void           * getBuffer()        { return m_dataPtr;    }
   int32            getBufferCapacity(){ return m_dataBufLen; }
   ConversionImpl * getConversion()    { return &m_conversion;}
   RecordImpl     * getRecord()        { return &m_record;    }

   // trigger 0, 1, ... , x
   TriggerImpl * getTrigger(int32 trigger)
   { 
      TriggerImpl *x = NULL;

      // Bug#14478
      //if ((uint32)trigger < (uint32)getFeatures()->getTriggerCount()) {
      if ((uint32)trigger < ARRAY_SIZE(m_triggers)) {
         x = &m_triggers[trigger];
      }
      return x;
   }

public: // ctor & dtor
   WaveformAiCtrlImpl()
   {
      m_dataPtr           = NULL;
      m_dataBufLen        = 0;
      m_dataReadyCleared  = false;

      m_conversion.setOwner(this);
      m_record.setOwner(this);
      
      for (int i = 0; i < MAX_TRIG_COUNT; ++i) {
         m_triggers[i].setOwner(this, i);
      }

      m_dataArrived = CreateEvent(NULL, false, false, NULL);
      m_dn4EventRegistered = false;
      m_waitforStopped = false;
      m_recordCount = 0;
   }

   virtual ~WaveformAiCtrlImpl()
   {
      // clear all user event delegate
      for (unsigned i = 0; i < ARRAY_SIZE(wav_ai_events); ++i) {
         UserEventManager::Remove(getOuter(), wav_ai_events[i]);
      }

      if (m_dataArrived) {
         CloseHandle(m_dataArrived);
      }
   }

protected: 
   //------------------------------------------------------------------------
   // override the method in base class
   //------------------------------------------------------------------------
   virtual std::vector<AccessMode> WantedModes()
   {
      AccessMode spted[] = { ModeWrite, ModeWriteWithReset };
      return std::vector<AccessMode>(&spted[0], &spted[2]);
   }

   virtual bool IsSuitable(DaqModule &module)
   {
      if (AiCtrlBaseImpl::IsSuitable(module)) {
         return DaqProperty(module).GetLength(CFG_FeatureConvertClockSources) != 0;
      }
      return false;
   }

   // update the control state
   virtual void RefreshState(ControlState &curr)
   {
      if (curr == Running) {
         EventStatus es = {sizeof(es)};
         ModEvent().GetStatus(EvtBufferedAiStopped, es);
         if (es.AiDataReady.Stopped) {
            setState(Stopped);
         }
      }
   }

   // initialize/deinitialize the control 
   virtual ErrorCode InitializeControl(DeviceInformation const &x)
   {
      // let's the base do initialization at first.
      ErrorCode ret = AiCtrlBaseImpl::InitializeControl(x);
      if (ret != Success) { return ret; }

      // buffered AI not supported
      if (!getFeatures()->getBufferedAiSupported()) {
         return ErrorFuncNotSpted;
      }

      LoadProperties();

      // activate the events       
      for (unsigned i = 0; i < ARRAY_SIZE(wav_ai_events); ++i) {
         if (UserEventManager::IsExist(getOuter(), wav_ai_events[i])) {
            RegisterKernEvent(wav_ai_events[i]);
         }
      }

      return Success;
   }

   virtual void DeinitializeControl()
   {
      // release buffered AI
      Release();

      // deactivate the events        
      for (unsigned i = 0; i < ARRAY_SIZE(wav_ai_events); ++i) {
         UnregisterKernEvent(wav_ai_events[i]);
      }

      // for blocked 'GetData'
      UnregisterKernEvent(EvtBufferedAiDataReady,   DN4_ProcessKernEvent);
      UnregisterKernEvent(EvtBufferedAiStopped,     DN4_ProcessKernEvent);
      m_dn4EventRegistered = false;
      m_waitforStopped = false;

      // reset properties to its default value
      m_dataSize = 0;

      for (int i = 0; i < getFeatures()->getTriggerCount(); ++i) {
         m_triggers[i].adjustDelayCount(0);
      }

      // let's the base do de-initialization at last.
      AiCtrlBaseImpl::DeinitializeControl();
   }

   //
   // Reload the properties' value from device
   virtual void ReloadProperties()
   {
      AiCtrlBaseImpl::ReloadProperties();
      LoadProperties();
   }

   void LoadProperties()
   {
      // cache the raw data size for the performance of method 'GetData'
      m_dataSize = getFeatures()->getDataSize();

      // get the property default value from device
      m_scanCHCount = m_conversion.getChannelCount();
      assert(m_scanCHCount);

      // trigger delay count
      int32 multiplier = 1;
      if (getFeatures()->getSamplingMethod() == EqualTimeSwitch) {
         multiplier = m_scanCHCount;
      }
      for (int i = 0; i < getFeatures()->getTriggerCount(); ++i) {
         int32 delay = 0;
         ModProp().Get(prop_id_trigDelay[i], delay);
         m_triggers[i].adjustDelayCount(delay / multiplier);
      }
   }

   // -----------------------------------------------------------------
   // Check whether the event id is supported by a counter control or not
   // -----------------------------------------------------------------
   virtual bool IsValidEvent(EventId id)
   {
      return ARRAY_INDEX_OF(wav_ai_events, id) != -1;
   }

   // -----------------------------------------------------------------
   // driver event receiver for buffered AI
   // -----------------------------------------------------------------
   virtual void OnKernEvent(EventId id)
   {
      // Discard the un-expected events
      if (IsIdle()) { return; }

      EventStatus    es   = {sizeof(es)};
      BfdAiEventArgs args = { id };

      switch(id)
      {
      default:
         break;
      case EvtBufferedAiDataReady:  /* data ready */
         m_dataReadyCleared = false;

         ModEvent().GetStatus(id, es);
         args.Count  = (es.AiDataReady.DataCount / m_scanCHCount) * m_scanCHCount;
         args.Offset = es.AiDataReady.DataOffset;
         args.MarkCount = es.AiDataReady.MarkCount;
         if (args.Count) { 
            UserEventManager::Fire(getOuter(), id, &args); 
         }

         // clear the event flag if the used hadn't clear it by calling 'GetData'.
         // using a count '0' to clear only the 'had sent' flag, but not effect the 'read pos'.
         if (!m_dataReadyCleared) { ModEvent().ClearFlag(id); }
         break;
      case EvtBufferedAiOverrun:  /* overrun */
         ModEvent().GetStatus(id, es);
         args.Count  = es.AiOverrun.DataCount;
         args.Offset = es.AiOverrun.DataOffset;
         UserEventManager::Fire(getOuter(), id, &args); 
         ModEvent().ClearFlag(id);            // clear the event flag after the user had processed it.
         break;
      case EvtBufferedAiStopped:    /* stopped */ 
         ModEvent().GetStatus(id, es);
         args.Count  = es.AiDataReady.DataCount;
         args.Offset = es.AiDataReady.DataOffset; 
         args.MarkCount = es.AiDataReady.MarkCount;
         UserEventManager::Fire(getOuter(), id, &args); 
         // need not clear the event flag
         break;
      case EvtBufferedAiCacheOverflow:  /* cache overflow */
         UserEventManager::Fire(getOuter(), id, &args); 
         ModEvent().ClearFlag(id);            // clear the event flag after the user had processed it.
         break;
      case EvtBufferedAiBurnOut:                /*burn out*/
      case EvtBufferedAiTimeStampOverrun:       /*time-stamp buffer overrun*/
      case EvtBufferedAiTimeStampCacheOverflow: /*time-stamp FIFO overflow*/
      case EvtBufferedAiMarkOverrun:            /*data marker buffer overrun*/
         UserEventManager::Fire(getOuter(), id, &args); 
         break;// need not clear the event flag
      }
   };

   //------------------------------------------------------------------------
   // internal methods
   //------------------------------------------------------------------------
   // DN4 event receiver for blocked 'GetData'
   static void BDAQCALL DN4_ProcessKernEvent(EventId id, void *param)
   {
      WaveformAiCtrlImpl *_this = (WaveformAiCtrlImpl *)param;
      SetEvent(_this->m_dataArrived);
   }

   static uint32 WaitEventTimeout(HANDLE ev, uint32 timeout)
   {
      uint32 tick = GetTickCount();
      switch (WaitForMultipleObjects(1, &ev, false, timeout)) {
      case WAIT_TIMEOUT:
      case WAIT_FAILED:
         timeout = 0;
         break;
      default:
         if (timeout != (uint32)INFINITE) { 
            uint32 elapsed = GetTickCount() - tick;
            timeout = timeout > elapsed ? timeout - elapsed : 0;
         }
         break;
      }
            
      return timeout;
   }

private:
   void *   m_dataPtr;
   int32    m_dataBufLen;
   bool     m_dataReadyCleared; // whether the data ready event has been cleared in 'GetData', used by DataReadyHandler.
   int32    m_dataSize;         // for performance of 'GetData' method, cache this value, it will be initialized in 'InitializeControl'
   int32    m_scanCHCount;      // for performance of 'GetData' method, cache this value, it will be initialized in 'Prepare'
   HANDLE   m_dataArrived;      // event for blocked 'GetData' method
   bool     m_dn4EventRegistered; // flag indicating whether the event for blocked 'GetData' method is registered or not.
   bool     m_waitforStopped;     // whether we should wait for the AI stopped before 'GetData'
   uint32   m_recordCount;
   
   ConversionImpl m_conversion;
   RecordImpl     m_record;
   TriggerImpl    m_triggers[MAX_TRIG_COUNT];
};

#endif
