//
// AO function implementation
//
#include <std_decl.h>

#include "ao_features.h"
#include "ao_channel.h"
#include "ao_ctrl_base.h"

//------------------------------------------------------------------
// Help MACRO used to get/set property and check error
//------------------------------------------------------------------
#undef  MOD_PROP
#define MOD_PROP (m_owner->ModProp())

// -----------------------------------------------------------------------
// functions of AoFeaturesImpl
//------------------------------------------------------------------------

// ADC features
int32 AoFeaturesImpl::getResolution()
{
   int32 res = 0;
   MOD_PROP.Get(CFG_FeatureResolutionInBit, res);
   return res;
}
int32 AoFeaturesImpl::getDataSize()
{
   int32 size = 0;
   MOD_PROP.Get(CFG_FeatureDataSize, size);
   return size;
}
int32 AoFeaturesImpl::getDataMask()
{
   int32 mask = 0;
   MOD_PROP.Get(CFG_FeatureDataMask, mask);
   return mask;
}

// channel features
int32 AoFeaturesImpl::getChannelCountMax()
{
   int32 count = 0;
   if (MOD_PROP.Get(CFG_FeatureChannelNumberMax, count) == Success) {
      count += 1; // convert from channel number to channel count
   }
   return count;
}
ArrayImpl<ValueRange>* AoFeaturesImpl::getValueRanges()
{
   return MOD_PROP.GetValueArrayI32<ValueRange>(CFG_FeatureVrgTypes);
}
bool AoFeaturesImpl::getExternalRefAntiPolar()
{
   int32 isAnti = 0;
   MOD_PROP.Get(CFG_FeatureExtRefAntiPolar, isAnti);
   return isAnti != 0;
}
void AoFeaturesImpl::getExternalRefRange(MathInterval &x)
{
   if (MOD_PROP.Get(CFG_FeatureExtRefRange, x) != Success) {
      ResetMathInterval(x);
   }
}

// buffered AO->basic features
bool AoFeaturesImpl::getBufferedAoSupported()
{
   return MOD_PROP.GetLength(CFG_FeatureConvertClockSources) != 0;
}
SamplingMethod AoFeaturesImpl::getSamplingMethod()
{
   int32 ret = (int32)EqualTimeSwitch;
   MOD_PROP.Get(CFG_FeatureChannelScanMethod, ret);
   return (SamplingMethod)ret;
}
int32 AoFeaturesImpl::getChannelStartBase()
{
   int32 startBase = 1;
   MOD_PROP.Get(CFG_FeatureScanChannelStartBase, startBase);
   return startBase;
}
int32 AoFeaturesImpl::getChannelCountBase()
{
   int32 countBase = 1;
   MOD_PROP.Get(CFG_FeatureScanChannelCountBase, countBase);
   return countBase;
}

// buffered AO->conversion clock features
ArrayImpl<SignalDrop> * AoFeaturesImpl::getConvertClockSources()
{
   return MOD_PROP.GetValueArrayI32<SignalDrop>(CFG_FeatureConvertClockSources);
}
void AoFeaturesImpl::getConvertClockRange(MathInterval &x)
{
   if (MOD_PROP.Get(CFG_FeatureConvertClockRateRange, x) != Success) {
      ResetMathInterval(x);
   }
}

// buffered AO->trigger features
int32 AoFeaturesImpl::getTriggerCount()
{
   int32 count = 0;
   MOD_PROP.Get(CFG_FeatureTriggersCount, count);
   return count;
}
bool AoFeaturesImpl::getRetriggerable()
{
   int32 retriggerable = 0;
   MOD_PROP.Get(CFG_FeatureRetriggerable, retriggerable);
   return retriggerable != 0;
}
ArrayImpl<SignalDrop> * AoFeaturesImpl::getTriggerSources(int32 trigger)
{
   PropertyId id = trigger < MAX_TRIG_COUNT ? feat_id_trigSrc[trigger] : PropertyId(-1);
   return MOD_PROP.GetValueArrayI32<SignalDrop>(id);
}
ArrayImpl<TriggerAction> * AoFeaturesImpl::getTriggerActions(int32 trigger)
{
   PropertyId id = trigger < MAX_TRIG_COUNT ? feat_id_trigAction[trigger] : PropertyId(-1);
   return MOD_PROP.GetValueArrayI32<TriggerAction>(id);
}
void AoFeaturesImpl::getTriggerDelayRange(int32 trigger, MathInterval &x)
{
   PropertyId id = trigger < MAX_TRIG_COUNT ? feat_id_trigDlyRng[trigger] : PropertyId(-1);
   if (MOD_PROP.Get(id, x) != Success) { ResetMathInterval(x); }
}

// -----------------------------------------------------------------------
// functions of AoChannelImpl
//------------------------------------------------------------------------
int32 AoChannelImpl::getChannel()
{
   return (int32)m_tag;
}
ValueRange AoChannelImpl::getValueRange()
{
   assert(unsigned(m_tag) < MAX_AO_CH_COUNT);

   int32 value = 0;
   MOD_PROP.GetValueNth(CFG_VrgTypeOfChannels, m_tag, value);
   return ValueRange(value);
}
ErrorCode AoChannelImpl::setValueRange(ValueRange value)
{
   assert(unsigned(m_tag) < MAX_AO_CH_COUNT);

   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }
   if (m_owner->IsRunning()) { return ErrorFuncBusy; }

   return MOD_PROP.SetValueNth(CFG_VrgTypeOfChannels, m_tag, (int32)value);
}

double AoChannelImpl::getExtRefBipolar()
{
   assert(unsigned(m_tag) < MAX_AO_CH_COUNT);

   double xbuf[128];
   uint32 size = sizeof(xbuf);
   ErrorCode ret = MOD_PROP.Read(CFG_ExtRefValueForBipolar, &size, xbuf);
   if (ret != Success) { return 0; }

   return (unsigned(m_tag) < size / sizeof(double)) ? xbuf[m_tag] : xbuf[0];
}
ErrorCode AoChannelImpl::setExtRefBipolar(double value)
{
   assert(unsigned(m_tag) < MAX_AO_CH_COUNT);

   double xbuf[128];
   uint32 size = sizeof(xbuf);
   ErrorCode ret = MOD_PROP.Read(CFG_ExtRefValueForBipolar, &size, xbuf);
   if (ret != Success) { return ret; }

   if (unsigned(m_tag) < size / sizeof(double)) {
      xbuf[m_tag] = value;
   } else {
      xbuf[0] = value;
   }

   return MOD_PROP.Write(CFG_ExtRefValueForBipolar, size, xbuf);
}
double AoChannelImpl::getExtRefUnipolar()
{
   assert(unsigned(m_tag) < MAX_AO_CH_COUNT);

   double xbuf[128];
   uint32 size = sizeof(xbuf);
   ErrorCode ret = MOD_PROP.Read(CFG_ExtRefValueForUnipolar, &size, xbuf);
   if (ret != Success) { return 0; }

   return (unsigned(m_tag) < size / sizeof(double)) ? xbuf[m_tag] : xbuf[0];
}
ErrorCode AoChannelImpl::setExtRefUnipolar(double value)
{
   assert(unsigned(m_tag) < MAX_AO_CH_COUNT);

   double xbuf[128];
   uint32 size = sizeof(xbuf);
   ErrorCode ret = MOD_PROP.Read(CFG_ExtRefValueForUnipolar, &size, xbuf);
   if (ret != Success) { return ret; }

   if (unsigned(m_tag) < size / sizeof(double)) {
      xbuf[m_tag] = value;
   } else {
      xbuf[0] = value;
   }

   return MOD_PROP.Write(CFG_ExtRefValueForUnipolar, size, xbuf);
}

// new: scale table
ErrorCode AoChannelImpl::getScaleTable(int32 &size, MapFuncPiece table[])
{
   assert(unsigned(m_tag) < MAX_AO_CH_COUNT);

   return m_owner->ModuleRef().GetScaleTable(m_tag, (uint32&)size, table);
}
ErrorCode AoChannelImpl::setScaleTable(int32  size, MapFuncPiece table[])
{
   assert(unsigned(m_tag) < MAX_AO_CH_COUNT);

   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }
   if (m_owner->IsRunning()) { return ErrorFuncBusy; }

   return m_owner->ModuleRef().SetScaleTable(m_tag, size, table);
}