#ifndef _DAQNAVI_BUFFERED_AO_CTRL_IMPL_H
#define _DAQNAVI_BUFFERED_AO_CTRL_IMPL_H

class BufferedAoCtrlImpl : public AoCtrlBaseImpl
{

public:
   //
   // event:
   // Use the default implementation in DaqCtrlBase

   // method
   ErrorCode Prepare()
   {
      AutoLock<MutexLock> lock(PropLock());

      ErrorCode ret;
      if (!getInitialized()) {
         ret = ErrorFuncNotInited;
      } else if (IsRunning()) {
         ret = ErrorFuncBusy;
      } else if (IsReady()){
         ret = Success;
      } else {
         // get scan channel count for the performance of method 'GetData'
         ModProp().Get(CFG_ScanChannelCount, m_scanCHCount);

         // configuration h/w. Only set the property we cached,
         // the others had been set when the user calls 'set' method of the property
         // 
         // set: sectionPerChan --> sectionDataCount
         ModProp().Set(CFG_SectionDataCount, m_scanChannel.getIntervalCount() * m_scanCHCount);

         // set: trigger delay count
         int32 multiplier = 1;
         if (getFeatures()->getSamplingMethod() == EqualTimeSwitch) {
            multiplier = m_scanCHCount;
         }
         ModProp().Set(CFG_TriggerDelayCount, m_trigger.getDelayCount() * multiplier);   // trigger 0
         ModProp().Set(CFG_Trigger1DelayCount, m_trigger1.getDelayCount() * multiplier); // trigger 1

         //Read back the adjustable parameters
         int32 trig0_delay = 0, trig1_delay = 0;
         ModProp().Get(CFG_TriggerDelayCount, trig0_delay);
         ModProp().Get(CFG_Trigger1DelayCount, trig1_delay);

         m_trigger.adjustDelayCount(trig0_delay / multiplier);
         m_trigger1.adjustDelayCount(trig1_delay / multiplier);

         // prepare buffered AO
         m_dataBufLen = m_scanChannel.getSamples() * m_scanCHCount;
         ret = ModuleRef().BfdAoPrepare(m_dataBufLen, m_dataPtr);
         setState(ret == Success ? Ready : Idle);
      }
      return ret;
   }

   // synchronous & one-shot
   ErrorCode RunOnce() { return StartBfdAo(false, false); }

   // asynchronous, one-shot or streaming, depends on the property 'Streaming'
   ErrorCode Start()   { return StartBfdAo(m_streaming, true); }

   ErrorCode Stop(int32 action)
   {
      if (IsIdle()) { return Success;  }
      // ji.dong mask because AO stop maybe not completed synchronously(depending on the 'action' parameter)
      // setState(Stopped);
      return ModuleRef().BfdAoStop(action);
   }

   void Release()
   {
      // Fix bug#19706: Call driver 'Release' function anyway to release resource, if any.
      // It is harmless operation even if the device is in 'FAI IDLE' state.
      // if (IsIdle()) { return; }
      // Call driver 'Release' only if we had called the method 'Prepare'
      if (m_dataPtr) {
         ModuleRef().BfdAoRelease();
      }

      setState(Idle);
      m_dataPtr    = NULL;
      m_dataBufLen = 0;
   }

   ErrorCode SetData(int32 dataType, int32 count, void *buffer)
   {
      ErrorCode ret;
      
      if (dataType != sizeof(double) && dataType != m_dataSize) {
         // Not supported data type
         ret = ErrorFuncNotSpted;
      }

      if (IsIdle()) { 
         ret = Prepare();
         if (ret != Success) { return ret; }
      }

      if (count > m_dataBufLen) { 
         count = m_dataBufLen; 
      }

      AutoLock<MutexLock> lock(BuffLock());

      //
      // The device is using the new BAO method: BfdAoSetData
      if (m_dataPtr == MULTI_BUF_PTR) {
         if (dataType == sizeof(double)) { 
            ret = ModuleRef().BfdAoSetData((uint32&)count, NULL, (double*)buffer);
         } else { 
            ret = ModuleRef().BfdAoSetData((uint32&)count, buffer, NULL);
         }
      } else {
         if (dataType == sizeof(double)) {
            ret = SetDataDbl(count, (double*)buffer);
         } else {
            ret = SetDataRaw(count, buffer);
         }
      }

      // Update the 'Write Position'
      m_dataTransedCleared = true;
      ModEvent().ClearFlag(EvtBufferedAoDataTransmitted, count);

      return ret;
   }

   bool GetBufferStatus(int32 & count, int32 & offset)
   {
      EventStatus es = {sizeof(es)};
      ModEvent().GetStatus(EvtBufferedAoDataTransmitted, es);
      count  = (es.General.lParam / m_scanCHCount) * m_scanCHCount;
      offset = es.General.rParam;

      return false;
   }

   // property
   void             * getBuffer()         { return m_dataPtr;     }
   int32              getBufferCapacity() { return m_dataBufLen;  }
   ScanChannelImpl  * getScanChannel()    { return &m_scanChannel;}
   ConvertClockImpl * getConvertClock()   { return &m_convClock;  } 

   // trigger 0, 1, ... , x
   TriggerImpl * getTrigger(int32 trigger)
   { 
      TriggerImpl *x = NULL;

      // Bug#14478
      //if ((uint32)trigger < (uint32)getFeatures()->getTriggerCount()) {
      switch(trigger) {
         default: x = NULL;        break;
         case 0:  x = &m_trigger;  break;
         case 1:  x = &m_trigger1; break;
      }

      return x;
   }

   bool getStreaming() 
   { 
      return m_streaming; 
   }
   ErrorCode setStreaming(bool value)
   {
      if (IsRunning()) { return ErrorFuncBusy; }
      m_streaming = value;
      return Success;
   }

public: // ctor & dtor
   BufferedAoCtrlImpl()
   {
      m_dataPtr           = NULL;
      m_dataBufLen        = 0;
      m_dataTransedCleared  = false;
      m_streaming         = false;

      m_scanChannel.setOwner(this, false);
      m_convClock.setOwner(this);
      m_trigger.setOwner(this, 0);
      m_trigger1.setOwner(this, 1);
   }

   virtual ~BufferedAoCtrlImpl()
   {
      // clear all user event delegate
      for (unsigned i = 0; i < ARRAY_SIZE(bfd_ao_events); ++i) {
         UserEventManager::Remove(getOuter(), bfd_ao_events[i]);
      }
   }

protected: 
   //------------------------------------------------------------------------
   // override the method in base class
   //------------------------------------------------------------------------
   virtual bool IsSuitable(DaqModule &module)
   {
      if (AoCtrlBaseImpl::IsSuitable(module)) {
         return DaqProperty(module).GetLength(CFG_FeatureConvertClockSources) != 0;
      }
      return false;
   }

   // update the control state
   virtual void RefreshState(ControlState &curr)
   {
      if (curr == Running) {
         EventStatus es = {sizeof(es)};
         ModEvent().GetStatus(EvtBufferedAoStopped, es);
         if (es.General.lParam) {
            setState(Stopped);
         }
      }
   }

   // initialize/deinitialize the control 
   virtual ErrorCode InitializeControl(DeviceInformation const &x)
   {
      // let's the base do initialization at first.
      ErrorCode ret = AoCtrlBaseImpl::InitializeControl(x);
      if (ret != Success) { return ret; }

      // buffered AO not supported
      if (!getFeatures()->getBufferedAoSupported()) {
         return ErrorFuncNotSpted;
      }

      LoadProperties();

      // activate the events       
      for (unsigned i = 0; i < ARRAY_SIZE(bfd_ao_events); ++i) {
         if (UserEventManager::IsExist(getOuter(), bfd_ao_events[i])) {
            RegisterKernEvent(bfd_ao_events[i]);
         }
      }

      return Success;
   }

   virtual void DeinitializeControl()
   {
      // release buffered AO
      Release();

      // deactivate the events        
      for (unsigned i = 0; i < ARRAY_SIZE(bfd_ao_events); ++i) {
         UnregisterKernEvent(bfd_ao_events[i]);
      }

      // reset properties to its default value
      m_dataSize = 0;

      m_scanChannel.adjustIntervalCount(0);
      m_scanChannel.adjustSamples(0);
      m_trigger.adjustDelayCount(0);
      m_trigger1.adjustDelayCount(0);

      // let's the base do de-initialization at last.
      AoCtrlBaseImpl::DeinitializeControl();
   }

   //
   // Reload the properties' value from device
   virtual void ReloadProperties()
   {
      AoCtrlBaseImpl::ReloadProperties();
      LoadProperties();
   }

   void LoadProperties()
   {
      // cache the raw data size for the performance of method 'GetData'
      m_dataSize = getFeatures()->getDataSize();

      // get the property default value from device
      m_scanCHCount = m_scanChannel.getChannelCount();
      assert(m_scanCHCount);

      int32 sectPerCh = 1024;
      ModProp().Get(CFG_SectionDataCount, sectPerCh);

      sectPerCh /= m_scanCHCount;
      m_scanChannel.adjustIntervalCount(sectPerCh);
      m_scanChannel.adjustSamples(sectPerCh ? sectPerCh * 2 : 1024);

      // trigger delay count
      int32 multiplier = 1;
      if (getFeatures()->getSamplingMethod() == EqualTimeSwitch) {
         multiplier = m_scanCHCount;
      }

      int32 trig0_delay = 0, trig1_delay = 0;
      ModProp().Get(CFG_TriggerDelayCount, trig0_delay);
      ModProp().Get(CFG_Trigger1DelayCount, trig1_delay);

      m_trigger.adjustDelayCount(trig0_delay / multiplier);
      m_trigger1.adjustDelayCount(trig1_delay / multiplier);
   }

   //------------------------------------------------------------------------
   // internal methods
   //------------------------------------------------------------------------
   ErrorCode StartBfdAo(bool streaming, bool async)
   {
      if (IsIdle()) { return ErrorFuncNotInited; }
      if (IsRunning()) { return ErrorFuncBusy;   }

      // start buffered AO
      ErrorCode ret;
      if (streaming) {
         async = true;
         ret = ModuleRef().BfdAoRun();
      } else {
         ret = ModuleRef().BfdAoRunOnce(async);
      }

      if (ret == Success) {
         setState(async ? Running : Stopped);
      }
      return ret;
   }

   ErrorCode SetDataRaw(int32 &count, void *rawData)
   {
      assert(!IsIdle() && count <= m_dataBufLen);

      // get current 'Write Position'
      EventStatus es = {sizeof(es)};
      ModEvent().GetStatus(EvtBufferedAoDataTransmitted, es);
      int32 offset = es.General.rParam;
      
      // copy the first segment, it's from 'offset' to 'buffer end', 'cpyLen' length
      int32 cpyLen = x_min(count, m_dataBufLen - offset);
      memcpy((uint8*)m_dataPtr + offset * m_dataSize, rawData, cpyLen * m_dataSize);

      // copy the rest from 'buffer start'
      if (count > cpyLen) {
         memcpy(m_dataPtr, (uint8*)rawData + cpyLen * m_dataSize, (count - cpyLen) * m_dataSize);
      }

      return Success;
   }

   ErrorCode SetDataDbl(int32 &count, double scaledData[])
   {
      assert(!IsIdle() && count <= m_dataBufLen);

      // get current 'Write Position'
      EventStatus es = {sizeof(es)};
      ModEvent().GetStatus(EvtBufferedAoDataTransmitted, es);
      int32 offset = es.General.rParam;

      int32 scaleLen = x_min(count, m_dataBufLen - offset);
      int32 chOffset = offset % m_scanCHCount;

      // copy the first segment, it's from 'offset' to 'buffer end', 'scaleLen' length
      void* raw = (uint8*)m_dataPtr + offset * m_dataSize;
      ModuleRef().BfdAoScaleData(&scaledData[0], raw, scaleLen, (uint32 *)&chOffset);

      // copy the rest from 'buffer start'
      if (count > scaleLen) {
         ModuleRef().BfdAoScaleData(&scaledData[scaleLen], m_dataPtr, count - scaleLen, (uint32 *)&chOffset);
      }

      return Success;
   }
   
   // -----------------------------------------------------------------
   // Check whether the event id is supported by a counter control or not
   // -----------------------------------------------------------------
   virtual bool IsValidEvent(EventId id)
   {
      return ARRAY_INDEX_OF(bfd_ao_events, id) != -1;
   }

   // -----------------------------------------------------------------
   // driver event receiver for buffered AO
   // -----------------------------------------------------------------
   virtual void OnKernEvent(EventId id)
   {
      // Discard the un-expected events
      if (IsIdle()) { return; }

      EventStatus    es   = {sizeof(es)};
      BfdAoEventArgs args = {id};

      switch(id)
      {
      default:  
         break;
      case EvtBufferedAoDataTransmitted:  /* data transmitted */
         m_dataTransedCleared = false;

         ModEvent().GetStatus(id, es);
         args.Count  = (es.General.lParam / m_scanCHCount) * m_scanCHCount;
         args.Offset = es.General.rParam;
         if (args.Count) { 
            UserEventManager::Fire(getOuter(), id, &args); 
         }

         // clear the event flag if the used hadn't clear it by calling 'SetData'.
         // using a count '0' to clear only the 'had sent' flag, but not effect the 'write pos'.
         if (!m_dataTransedCleared) { ModEvent().ClearFlag(id); }
         break;
      case EvtBufferedAoUnderrun:  /* under-run */
         ModEvent().GetStatus(id, es);
         args.Count  = es.General.lParam;
         args.Offset = es.General.rParam;
         UserEventManager::Fire(getOuter(), id, &args); 
         ModEvent().ClearFlag(id);            // clear the event flag after the user had processed it.
         break;
      case EvtBufferedAoCacheEmptied:  /* cache overflow */// no status data
         UserEventManager::Fire(getOuter(), id, &args); 
         ModEvent().ClearFlag(id);            // clear the event flag after the user had processed it.
         break;
      case EvtBufferedAoTransStopped: /*transition stopped*/// get the status of 'data transmitted' event to obtain the offset parameter
         ModEvent().GetStatus(EvtBufferedAoDataTransmitted, es);
         args.Count  = es.General.lParam;
         args.Offset = es.General.rParam;
         UserEventManager::Fire(getOuter(), id, &args); 
         // need not clear the event flag
         break;
      case EvtBufferedAoStopped:    /* stopped */// get the status of 'data transmitted' event to obtain the offset parameter
         ModEvent().GetStatus(EvtBufferedAoDataTransmitted, es);
         args.Count  = es.General.lParam;
         args.Offset = es.General.rParam;
         UserEventManager::Fire(getOuter(), id, &args); 
         // need not clear the event flag
         break;
      }
   };

private:
   void *   m_dataPtr;
   int32    m_dataBufLen;
   bool     m_dataTransedCleared; // whether the data ready event has been cleared in 'GetData', used by DataReadyHandler.

   bool     m_streaming;        // buffered AO running mode : streaming VS one-shot.
   int32    m_dataSize;         // for performance of 'GetData' method, cache this value, it will be initialized in 'InitializeControl'
   int32    m_scanCHCount;      // for performance of 'GetData' method, cache this value, it will be initialized in 'Prepare'
   
   ScanChannelImpl  m_scanChannel;
   ConvertClockImpl m_convClock;
   TriggerImpl      m_trigger;
   TriggerImpl      m_trigger1;
};

#endif
