#ifndef _DAQNAVI_EVENT_MONITOR_H
#define _DAQNAVI_EVENT_MONITOR_H

class KernEventMsgQueue
{
   typedef struct Event {
      EventId Id;
      bool    Queued;
   } Event;

   typedef std::deque<HANDLE>      MsgQueue;
   typedef std::map<HANDLE, Event> MsgFilter;

public:

   // Push the message into queue
   void Put(HANDLE handle)
   {
      // Filter out the duplicated event and un-registered event
      MsgFilter::iterator itor = m_filter.find(handle);
      if ( itor == m_filter.end() || itor->second.Queued) {
         return;
      } 

      // Put the message to queue
      {
         AutoLock<MutexLock> lock(m_lock);
         itor->second.Queued = true;
         m_queue.push_back(handle);
      }

      m_sema.Post();
   }

   // Get a message from the queue, if any. The function suspends if no message is available.
   HANDLE Get(EventId &id)
   {
      HANDLE h;
      id = EventUnknown;
      
      // Get a msg from queue
      m_sema.Wait();
      {
         AutoLock<MutexLock> lock(m_lock);

         MsgQueue::const_iterator head = m_queue.begin();
         if (head == m_queue.end()) {
            return NULL; // Invalid Msg
         }

         h = *head;
         m_queue.pop_front();
      }

      MsgFilter::iterator itor = m_filter.find(h);
      if ( itor == m_filter.end() || itor->second.Id == EventUnknown) {
         h =  NULL; // Invalid Msg
      } else {
         AutoLock<MutexLock> lock(m_lock);
         itor->second.Queued = false;
         id = itor->second.Id;
      }

      return h;
   }

   // Register a message into the queue. Only the registered message can be put/get.
   bool Register(HANDLE handle, EventId id)
   {
      // Duplicated event
      MsgFilter::iterator itor = m_filter.find(handle);
      if (itor != m_filter.end() && itor->second.Id != EventUnknown) {
         return false;
      } 

      Event e = { id, false };
      m_filter[handle] = e;
      return true;
   }

   // Unregister a message from the queue. 
   void Unregister(HANDLE handle)
   {
      m_filter[handle].Id = EventUnknown;
   }

   // Wake up the threads which are waiting for msg on the queue
   void Wakeup(int count)
   {
      while (count--) {
         m_sema.Post();
      }
   }

   // Find the event handle by ID.
   // Note: id-to-handle is a one-to-N map
   void Foreach(EventId id, bool (*fn)(HANDLE, void*), void *arg)
   {
      MsgFilter::const_iterator itor = m_filter.begin();
      for (; itor != m_filter.end(); ++itor) {
         if (itor->second.Id == id) {
            if (fn(itor->first, arg)) {
               break;
            }
         }
      }
   }

public:
   KernEventMsgQueue()
   {
   }

private:
   Semaphore  m_sema;
   MutexLock  m_lock;
   MsgQueue   m_queue;
   MsgFilter  m_filter; // Filter map for erasing duplicate message
};

class KernEventListener : public DaqThread
{
public:
   // maximum event count the thread supported(Please refer to MSDN 'WaitForMultipleObjects'), 
   static const uint32 CAPACITY = MAXIMUM_WAIT_OBJECTS; 

   KernEventListener() 
   { 
      memset(m_handles, 0, sizeof(m_handles));
      m_handles[0] = CreateEvent(NULL, false, false, NULL); 
      m_active     = 0x1;
      m_msgQueue   = NULL;
   }

   virtual ~KernEventListener()
   {
      if (m_handles[0] != NULL) {
         CloseHandle(m_handles[0]);
      }
   }

   void Connect(KernEventMsgQueue *queue) 
   {
      assert(queue);
      m_msgQueue = queue;
   }

public:
   int Find(HANDLE handle)
   {
      size_t idx = std::find(m_handles, m_handles + CAPACITY, handle) - m_handles;
      return idx < CAPACITY ? (int)idx : -1;
   }

   bool IsExist(HANDLE handle) 
   { 
      return Find(handle) != -1;
   }

   bool HasRoom() 
   { 
      return m_active != uint64(-1);    
   }

   int Count()
   {
      return soft_POPCNT(m_active);
   }

   // Add an event to monitor
   bool Register(HANDLE handle)
   {
      // There is no room for new event
      if (!HasRoom()) { return false; }

      // the event exists and need not add it once again.
      if (Find(handle) != -1) {
         return true;
      }

      // try to add the new event
      int index = 1;
      for (; index < 64; ++index) {
         // OK! Found a idle slot!
         if (AtomicBTS64((int64*)&m_active, index) == 0) {
            break;
         }
      }

      // Error! The idle slot is occupied by another ones. Try again later!
      if (uint32(index) >= CAPACITY) {
         return false;
      }

      m_handles[index] = handle;
      Wakeup();               // Notify the thread new event is added.
      return true;
   }

   // Unregister the specified event
   bool Unregister(HANDLE handle)
   {
      // the event doesn't exist
      int index = Find(handle);
      if (index == -1) { 
         return false; 
      }

      // mark the event as 'deleted'
      m_handles[index] = NULL;
      AtomicBTR64((int64*)&m_active, index);
      Wakeup();
      Wait(10); // Get a chance to the listener thread to know someone is removed

      return true;
   }

protected:
   virtual void Wakeup() 
   { 
      SetEvent(m_handles[0]); 
   }

   virtual int Run()
   {
      uint32 count = 1;
      HANDLE xhdls[CAPACITY] = {m_handles[0]};

      assert(m_msgQueue);

      // main-loop of the thread
      while (CanRun()) {
         uint32 waitRet = WaitForMultipleObjects(count, xhdls, false, INFINITE);
         
         // some event handle was closed, rebuild the handle array
         if (waitRet >= WAIT_OBJECT_0 + MAXIMUM_WAIT_OBJECTS) { 
            waitRet = WAIT_OBJECT_0; 
         }

         // management event: some events are added or removed
         if (waitRet == WAIT_OBJECT_0) {
            count = BuildActiveEventArray(xhdls);
         } else {
            // delivery the event
            m_msgQueue->Put(xhdls[waitRet - WAIT_OBJECT_0]);
         }
      }

      return 0;
   }

   uint32 BuildActiveEventArray(HANDLE xhdls[MAXIMUM_WAIT_OBJECTS])
   {
      // put all the device events needed.
      HANDLE   *src = m_handles;
      HANDLE   *dst = xhdls;
      uint64 active = m_active;

      for (; active; active >>= 1, ++src) {
         if ((active & 1) && *src) {
            *dst++ = *src;
         }
      }

      return uint32(dst - xhdls);
   }

private:
   HANDLE  m_handles[MAXIMUM_WAIT_OBJECTS];
   uint64  m_active;                         // Bit-mapped field, indicates the event alive currently!
   KernEventMsgQueue *m_msgQueue;
};

class KernEventDispatcher : public DaqThread
{
public:
   struct Initiator {
      void (* Initialize)();
      void (* Cleanup   )();
   };
   static Initiator & initiator()
   {
      static Initiator initor = {NULL, NULL};
      return initor;
   }

public:
   KernEventDispatcher() : m_msgQueue(NULL)
   { }
   virtual ~KernEventDispatcher() 
   { }

   void Connect(KernEventMsgQueue *queue) 
   {
      assert(queue);
      m_msgQueue = queue;
   }

protected:
   virtual void Prerun(bool start) 
   { 
      if (start) {
         if (initiator().Initialize) { 
            initiator().Initialize(); 
         } 
      } else {
         if (initiator().Cleanup) { 
            initiator().Cleanup(); 
         }
      }
   }

   virtual void Wakeup() 
   { 
      assert(m_msgQueue);
      m_msgQueue->Wakeup(1);
   }

   virtual int Run()
   {
      HANDLE  handle = NULL;
      EventId id     = EventUnknown;

      assert(m_msgQueue);

      // main-loop of the thread
      while (CanRun()) {
         handle = m_msgQueue->Get(id);
         if (CanRun()) { KernEventManager::Fire(handle, id); }
      }

      return 0;
   }

private:
   KernEventMsgQueue *m_msgQueue;
};

class KernEventMonitor
{
   static const uint32 MAX_LISTENER = 3;
   static const uint32 MAX_DISPATCH = 1;
   
   struct Context {
      KernEventProc     Proc;
      void             *Param;
      HANDLE            result;
   };

   static bool EventId2HandleCallback(HANDLE handle, void *arg)
   {
      Context *ctx = (Context *)arg;
      if (KernEventManager::IsExist(handle, ctx->Proc, ctx->Param)) {
         ctx->result = handle;
      }
      return ctx->result != NULL;
   }
   
public: // interface

   static bool Register(int32 deviceNumber, HANDLE handle, EventId id, KernEventProc proc, void *param)
   {
      AutoLock<MutexLock> lock(s_monitorMapLock);
      KernEventMonitor *& monitor = s_monitorMap[deviceNumber];

      if (monitor == NULL) {
         monitor = new KernEventMonitor();
      }

      return monitor->Register(handle, id, proc, param);
   }

   static void Unregister(int32 deviceNumber, HANDLE handle, KernEventProc proc, void *param)
   {
      AutoLock<MutexLock> lock(s_monitorMapLock);
      KEMMap::iterator itor = s_monitorMap.find(deviceNumber);

      if (itor != s_monitorMap.end()) {
         itor->second->Unregister(handle, proc, param);
      }
   }

   static void Unregister(int32 deviceNumber, EventId id, KernEventProc proc, void *param)
   {
      AutoLock<MutexLock> lock(s_monitorMapLock);
      KEMMap::iterator itor = s_monitorMap.find(deviceNumber);

      if (itor != s_monitorMap.end()) {
         Context ctx = {proc, param, NULL };
         itor->second->m_msgQueue.Foreach(id, EventId2HandleCallback, &ctx);
         if (ctx.result != NULL) {
            itor->second->Unregister(ctx.result, proc, param);
         }
      }
   }

protected:
   
   KernEventMonitor() : m_nextListener(unsigned(-1))
   {
      // Connect to message queue
      for (unsigned i = 0; i < MAX_DISPATCH; ++i) { m_dispatchers[i].Connect(&m_msgQueue); }      
      for (unsigned i = 0; i < MAX_LISTENER; ++i) { m_listeners[i].Connect(&m_msgQueue);   }
   }

   ~KernEventMonitor() 
   {
      // Stop all dispatchers & listeners
      for (unsigned i = 0; i < MAX_DISPATCH; ++i) { m_dispatchers[i].Stop(); }      
      for (unsigned i = 0; i < MAX_LISTENER; ++i) { m_listeners[i].Stop(); }

      // Wait for all dispatchers & listeners
      for (unsigned i = 0; i < MAX_LISTENER; ++i) { m_listeners[i].Wait(200); }

      for (unsigned tc = 0; tc < 10; ++ tc) {
         bool waitAgain = false;
         for (unsigned i = 0; i < MAX_DISPATCH; ++i) { 
            if (m_dispatchers[i].IsAlive() && !m_dispatchers[i].Wait(200)) { 
               waitAgain = true; 
            }
         }    
         if (!waitAgain) { break; }
      }
   }

   // Register the specified event
   bool Register(HANDLE handle, EventId id, KernEventProc proc, void *param)
   {
      assert(handle != NULL);

      //
      // Add the event proc to event manager
      KernEventManager::Add(handle, proc, param);

      //
      // Register the event into message queue
      if(!m_msgQueue.Register(handle, id)) {
         return false; // Duplicated event
      }

      // 
      // Register the event into a listener with overload balance(approximately)! 
      // Look for the thread with the least event count.
      unsigned index = (unsigned)AtomicIncrement((int32*)&m_nextListener) % MAX_LISTENER;
      KernEventListener &listener = m_listeners[index];
      if (listener.Register(handle)) {
         listener.Start();
         StartDiapatcher();
         return true;
      } 

      // 
      // The appropriate one is full, search another one.
      for (unsigned i = 0; i < MAX_LISTENER; ++i) {
         if (i == index) { continue; }

         if (m_listeners[i].Register(handle)) {
            m_nextListener = i;
            return true;
         }
      }

      //
      // Oops! All threads are full! Throw an exception to notify the user!
      throw "All listeners are full, please contact with manufacture!";  
      return false;
   }

   // Unregister the specified event
   void Unregister(HANDLE handle, KernEventProc proc, void *param)
   {
      assert(handle != NULL);

      //
      // Remove the event proc to event manager
      KernEventManager::Remove(handle, proc, param);
      if (KernEventManager::Count(handle)) {
         return; // There are still other people there
      }

      // Unregister the event from message queue
      m_msgQueue.Unregister(handle);

      //
      // Unregister the event from the listener
      for (unsigned i = 0; i < MAX_LISTENER; ++i) {
         if (m_listeners[i].Unregister(handle)) {
            break;
         }
      }
   }

   void StartDiapatcher()
   {
      // Dispatch is running, just return;
      if (m_dispatchers[0].IsAlive()) {
         return;
      }

      // Start all dispatchers
      for (unsigned i = 0; i < MAX_DISPATCH; ++i) {
         m_dispatchers[i].Start();
      }
   }

private:
   // kernel event monitor map: collects all the kernel event monitor.
   // Each kernel event monitor managers a specified device's events.
   typedef std::map<int32, KernEventMonitor*> KEMMap;

   static KEMMap       s_monitorMap;
   static MutexLock    s_monitorMapLock;

private:
   uint32              m_nextListener;             // Index of listener thread to add a new event to.
   KernEventListener   m_listeners[MAX_LISTENER];
   KernEventDispatcher m_dispatchers[MAX_DISPATCH];
   KernEventMsgQueue   m_msgQueue;
};

#endif
