//
// Miscellaneous function implementation
//
#include <std_decl.h>
#include <cntr/cntr_features.h>

#include "convert_clock.h"
#include "scan_channel.h"
#include "scan_clock.h"
#include "trigger.h"
#include "bfd_cntr_record.h"
#include "bfd_cntr_trigger.h"
#include "nos_flt_channel.h"
#include "cntr/cntr_ctrl_base.h"

#include "dio/dio_port.h"
#include "dio/dio_features.h"
#include "dio/dio_ctrl_base.h"

//------------------------------------------------------------------
// Help MACRO used to set property and check error
//------------------------------------------------------------------
#undef  MOD_PROP
#undef  PROP_COND_SET_VAL
#undef  PROP_COND_SET_ARRAY

#define MOD_PROP (m_owner->ModProp())

#define PROP_COND_SET_VAL(id, val)\
   {\
   if (m_owner->IsRunning()) return ErrorFuncBusy;\
   ErrorCode ret = m_owner->ModProp().Set(id, val);\
   if (!BioFailed(ret)) m_owner->OnPropChanged(id); \
   return ret; \
   }

#define PROP_COND_SET_ARRAY(id, count, val)\
   {\
   if (m_owner->IsRunning()) return ErrorFuncBusy;\
   ErrorCode ret = m_owner->ModProp().Set(id, count, val);\
   if (!BioFailed(ret)) m_owner->OnPropChanged(id); \
   return ret; \
   }

//------------------------------------------------------------------
// function of : DeviceCtrlImpl
//------------------------------------------------------------------
int32 DeviceCtrlImpl::getSupportedScenarios()
{
   int32 sc = 0;
   DaqModule *m = NULL;

   // AI
   if (m_device->GetModule(DaqAi, 0, m) == Success) {
      sc |= SceInstantAi;
      if (DaqProperty(m).GetLength(CFG_FeatureConvertClockSources) != 0) {
         sc |= SceWaveformAi;
      }
   }

   // AO
   if (m_device->GetModule(DaqAo, 0, m) == Success) {
      sc |= SceInstantAo;
      if (DaqProperty(m).GetLength(CFG_FeatureConvertClockSources) != 0) {
         sc |= SceBufferedAo;
      }
   }

   // DIO
   if (m_device->GetModule(DaqDio, 0, m) == Success) {
      uint8 portType[MAX_DIO_PORT_COUNT];
      uint32 size = sizeof(portType);
      uint32 i;

      DaqProperty(m).Read(CFG_FeaturePortsType, &size, portType);
      for(i = 0; i < size / sizeof(uint8); ++i) {
         if (portType[i] != PortDo) { sc |= SceInstantDi; }
         if (portType[i] != PortDi) { sc |= SceInstantDo; }
      }
      if (DaqProperty(m).GetLength(CFG_FeatureDiConvertClockSources) != 0) {
         sc |= SceBufferedDi;
      }
      if (DaqProperty(m).GetLength(CFG_FeatureDoConvertClockSources) != 0) {
         sc |= SceBufferedDo;
      }
   }

   // CNTR
   for (int ctrIdx = 0; ctrIdx < 2; ++ctrIdx) {
      if (m_device->GetModule(DaqCounter, ctrIdx, m) == Success) {
         CounterIndexerImpl<CounterCapability> indexer;
         indexer.setOwner((DaqCtModule*)m, CFG_FeatureCapabilitiesOfCounter0);

         int chCount = indexer.getLength();
         for(int ch = 0; ch < chCount; ++ch) {
            ArrayImpl<CounterCapability> *caps = indexer.getItem(ch);
            if (caps != NULL) {
               sc |= CounterCapsToScenarios(*caps);
               caps->Dispose();
            }
         }
      }
   }
   return sc;
}

// -----------------------------------------------------------------
// driver event receiver for device event
// Move the implement into .cpp due to we want to use the method of
// class DaqCtrlBaseImpl
// -----------------------------------------------------------------
void DeviceCtrlImpl::OnKernEvent(EventId id, DeviceCtrlImpl *ctrl)
{
   DeviceEventArgs args = {id};

   // Notify the underlying device module and upper device's ctrl about device is removed & reconnected
   if (id == EvtDeviceRemoved) {
      // Update the state of underlying device
      ctrl->m_device->OnDeviceStateChanged(DevState::Removed);

      // Update the state of upper device's ctrl
      if (ctrl->m_devPnpState != DevState::Removed) {
         ctrl->m_devPnpState = DevState::Removed;
         if (ctrl->m_owner) {
            ctrl->m_owner->OnDeviceStateChanged(DevState::Removed);
         }
      }      
   } else if (id == EvtDeviceReconnected) {
      // Update the state of underlying device
      ctrl->m_device->OnDeviceStateChanged(DevState::Arrived);

      // Update the state of upper device's ctrl
      if (ctrl->m_devPnpState != DevState::Arrived) {
         ctrl->m_devPnpState = DevState::Arrived;
         if (ctrl->m_owner) {
            ctrl->m_owner->OnDeviceStateChanged(DevState::Arrived);
         }
      }      
   }

   UserEventManager::Fire(ctrl, id, &args);
}

// -----------------------------------------------------------------
// helper : enumerate specified devices
// -----------------------------------------------------------------
struct DeviceEnumContext {
   DaqCtrlBaseImpl           *Ctrl;
   ArrayImpl<DeviceTreeNode> *Nodes;
};

//
// callback function for enumerating available devices
static bool BDAQCALL DeviceEnumCallBack(DeviceEnumArgs *args, void *userParam)
{
   DeviceEnumContext *ctx = (DeviceEnumContext *)userParam;
   DeviceTreeNode node;
   node.DeviceNumber = args->DeviceNumber;
#if defined(_WIN32) || defined(WIN32)
   wcscpy_s(node.Description, args->Description);
#else
   wcscpy(node.Description, args->Description);
#endif
   memset(node.ModulesIndex, -1, sizeof(node.ModulesIndex));

   // Enumerate all devices.
   if (!ctx->Ctrl) {
      ctx->Nodes->Add(node); 
      return true; // continue to enumerate the next device
   }

   // Enumerate specified devices.
   DaqModule *module;
   int count = 0;
   for (int i = 0; count < 8; ++i) {
      if (args->Device->GetModule(ctx->Ctrl->ModType(), i, module) != Success){
         break;
      }
      if (ctx->Ctrl->ModSuitable(*module)) {
         node.ModulesIndex[count++] = i;
      }
   }
   if (count > 0) {
      ctx->Nodes->Add(node);
   }

   return true; // continue to enumerate the next device
}

ArrayImpl<DeviceTreeNode>* EnumerateDeviceHelper(ModuleType typeWanted, void *daqCtrl)
{
   ArrayImpl<DeviceTreeNode> * nodes = new ArrayImpl<DeviceTreeNode>();
   DeviceEnumContext ctx = {(DaqCtrlBaseImpl*)daqCtrl, nodes};
   XDevEnumerator::EnumerateDevice(typeWanted, DeviceEnumCallBack, &ctx);

   return nodes;
}

//------------------------------------------------------------------
// function of : ScanChannelImpl
//------------------------------------------------------------------
int32 ScanChannelImpl::getChannelStart()
{
   int32 ch = 0;
   MOD_PROP.Get(CFG_ScanChannelStart, ch);
   return ch;
}
ErrorCode ScanChannelImpl::setChannelStart(int32 value)
{
   PROP_COND_SET_VAL(CFG_ScanChannelStart, value)
}
int32 ScanChannelImpl::getChannelCount()
{
   int32 count = 0;
   MOD_PROP.Get(CFG_ScanChannelCount, count);
   return count;
}
ErrorCode ScanChannelImpl::setChannelCount(int32 value)
{
   PROP_COND_SET_VAL(CFG_ScanChannelCount, value)
}
int32 ScanChannelImpl::getIntervalCount()
{
   // use the cached value here because this property is not mapped directly to a driver define property.
   return m_sectionPerChan;
}
ErrorCode ScanChannelImpl::setIntervalCount(int32 value)
{
   // use the cached value here because this property is not mapped directly to a driver define property.
   if (m_owner->IsRunning()) { return ErrorFuncBusy; }

   m_sectionPerChan = value;
   m_owner->OnPropChanged(CFG_SectionDataCount); // must re-initialize
   return Success;
}
int32 ScanChannelImpl::getSamples()
{
   // use the cached value here because this property is not mapped directly to a driver define property.
   return m_samplesPerChan;
}
ErrorCode ScanChannelImpl::setSamples(int32 value)
{
   // use the cached value here because this property is not mapped directly to a driver define property.
   if (m_owner->IsRunning()) { return ErrorFuncBusy; }

   m_samplesPerChan = value;
   m_owner->OnPropChanged(CFG_SectionDataCount); // must re-initialize, use the interval count id here. 'samples' has not own a property id in driver.
   return Success;
}

//------------------------------------------------------------------
// function of : ConvertClockImpl
//------------------------------------------------------------------
SignalDrop ConvertClockImpl::getSource()
{
   int32 src = 0;
   PropertyId propID = CFG_ConvertClockSource;

   if (m_isDio){
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? CFG_DiConvertClockSource : CFG_DoConvertClockSource;
   }

   MOD_PROP.Get(propID, src);
   return (SignalDrop)src;
}
ErrorCode ConvertClockImpl::setSource(SignalDrop value)
{
   PropertyId propID = CFG_ConvertClockSource; 

   if (m_isDio){
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? CFG_DiConvertClockSource : CFG_DoConvertClockSource;
   }

   PROP_COND_SET_VAL(propID, (int32)value);
}
double ConvertClockImpl::getRate()
{
   double rate = 0;
   PropertyId propID = CFG_ConvertClockRatePerChannel;
   if (m_isDio){
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? CFG_DiConvertClockRatePerPort : CFG_DoConvertClockRatePerPort;
   }

   MOD_PROP.Get(propID, rate);
   return rate;
}
ErrorCode ConvertClockImpl::setRate(double value)
{
   PropertyId propID = CFG_ConvertClockRatePerChannel; 

   if (m_isDio){
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? CFG_DiConvertClockRatePerPort : CFG_DoConvertClockRatePerPort;
   }

   PROP_COND_SET_VAL(propID, value)
}
//------------------------------------------------------------------
// function of : ScanClockImpl
//------------------------------------------------------------------
SignalDrop ScanClockImpl::getSource()
{
   int32 src = 0;
   MOD_PROP.Get(CFG_ScanClockSource, src);
   return (SignalDrop)src;
}
ErrorCode ScanClockImpl::setSource(SignalDrop value)
{
   PROP_COND_SET_VAL(CFG_ScanClockSource, (int32)value)
}
double ScanClockImpl::getRate()
{
   double rate = 0;
   MOD_PROP.Get(CFG_ScanClockRate, rate);
   return rate;
}
ErrorCode ScanClockImpl::setRate(double value)
{
   PROP_COND_SET_VAL(CFG_ScanClockRate, value)
}
int32 ScanClockImpl::getScanCount()
{
   int32 count = 0;
   MOD_PROP.Get(CFG_ScanCount, count);
   return count;
}
ErrorCode ScanClockImpl::setScanCount(int32 value)
{
   PROP_COND_SET_VAL(CFG_ScanCount, (int32)value)
}
// -----------------------------------------------------------------
//function of : TriggerImpl
// -----------------------------------------------------------------
SignalDrop TriggerImpl::getSource()
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   int32 src = 0;
   PropertyId propID = prop_id_trigSrc[m_tag];

   if (m_isDio){
      assert((uint32)m_tag < MAX_DIO_TRIG_COUNT);
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? prop_id_di_trigSrc[m_tag] : prop_id_do_trigSrc[m_tag];
   }

   MOD_PROP.Get(propID, src);
   return (SignalDrop)src;
}
ErrorCode TriggerImpl::setSource(SignalDrop value)
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   PropertyId propID = prop_id_trigSrc[m_tag];

   if (m_isDio){
      assert((uint32)m_tag < MAX_DIO_TRIG_COUNT);
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? prop_id_di_trigSrc[m_tag] : prop_id_do_trigSrc[m_tag];
   }

   PROP_COND_SET_VAL(propID, (int32)value)
}
ActiveSignal TriggerImpl::getEdge()
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   int32 src = 0;
   PropertyId propID = prop_id_trigEdge[m_tag];

   if (m_isDio){
      assert((uint32)m_tag < MAX_DIO_TRIG_COUNT);
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? prop_id_di_trigEdge[m_tag] : prop_id_do_trigEdge[m_tag];
   }

   MOD_PROP.Get(propID, src);
   return (ActiveSignal)src;
}
ErrorCode TriggerImpl::setEdge(ActiveSignal value)
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   PropertyId propID = prop_id_trigEdge[m_tag];
   
   if (m_isDio){
      assert((uint32)m_tag < MAX_DIO_TRIG_COUNT);
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? prop_id_di_trigEdge[m_tag] : prop_id_do_trigEdge[m_tag];
   }

   PROP_COND_SET_VAL(propID, (int32)value)
}
double TriggerImpl::getLevel()
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   double level = 0;
   PropertyId propID = prop_id_trigLevel[m_tag];

   if (m_isDio){
      assert((uint32)m_tag < MAX_DIO_TRIG_COUNT);
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? prop_id_di_trigLevel[m_tag] : prop_id_do_trigLevel[m_tag];
   }

   MOD_PROP.Get(propID, level);
   return level;
}
ErrorCode TriggerImpl::setLevel(double value)
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   PropertyId propID = prop_id_trigLevel[m_tag];

   if (m_isDio){
      assert((uint32)m_tag < MAX_DIO_TRIG_COUNT);
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? prop_id_di_trigLevel[m_tag] : prop_id_do_trigLevel[m_tag];
   }   

   PROP_COND_SET_VAL(propID, value)
}
TriggerAction TriggerImpl::getAction()
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   int32 action = 0;
   PropertyId propID = prop_id_trigAction[m_tag];

   if (m_isDio){
      assert((uint32)m_tag < MAX_DIO_TRIG_COUNT);
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? prop_id_di_trigAction[m_tag] : prop_id_do_trigAction[m_tag];
   }

   MOD_PROP.Get(propID, action);
   return (TriggerAction)action;
}
ErrorCode TriggerImpl::setAction(TriggerAction value)
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   PropertyId propID = prop_id_trigAction[m_tag];

   if (m_isDio){
      assert((uint32)m_tag < MAX_DIO_TRIG_COUNT);
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? prop_id_di_trigAction[m_tag] : prop_id_do_trigAction[m_tag];
   }

   PROP_COND_SET_VAL(propID, (int32)value)
}
int32 TriggerImpl::getDelayCount()
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);

   // use the cached value here because this property is not mapped directly to a driver define property.
   return m_trigDelayPerChan; 
}
ErrorCode TriggerImpl::setDelayCount(int32 value)
{
   // use the cached value here because this property is not mapped directly to a driver define property.
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   if (m_owner->IsRunning()) {
      return ErrorFuncBusy;
   }
   PropertyId propID = prop_id_trigDelay[m_tag];
   PropertyId featID = feat_id_trigDlyRng[m_tag];

   if (m_isDio){
      assert((uint32)m_tag < MAX_DIO_TRIG_COUNT);
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? prop_id_di_trigDelay[m_tag]  : prop_id_do_trigDelay[m_tag];
      featID = (dir == Input) ? feat_id_di_trigDlyRng[m_tag] : feat_id_do_trigDlyRng[m_tag];
   }

   m_owner->OnPropChanged(propID);

   MathInterval range = {0};
   MOD_PROP.Get(featID, range);

   m_trigDelayPerChan = value;
   if ((uint32)value < range.Min || (uint32)value > range.Max) {
      return WarningPropValueOutOfRange;
   }
   return Success;
}
double TriggerImpl::getHysteresisIndex()
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   double index = 0;
   PropertyId propID = prop_id_trigHystIdx[m_tag];

   if (m_isDio){
      assert((uint32)m_tag < MAX_DIO_TRIG_COUNT);
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? prop_id_di_trigHystIdx[m_tag] : prop_id_do_trigHystIdx[m_tag];
   }

   MOD_PROP.Get(propID, index);
   return index;
}
ErrorCode TriggerImpl::setHysteresisIndex(double value)
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   PropertyId propID = prop_id_trigHystIdx[m_tag];

   if (m_isDio){
      assert((uint32)m_tag < MAX_DIO_TRIG_COUNT);
      DioPortDir dir = ((DioCtrlBaseImpl*)m_owner)->WantedPort();
      propID = (dir == Input) ? prop_id_di_trigHystIdx[m_tag] : prop_id_do_trigHystIdx[m_tag];
   }

   ErrorCode ret = m_owner->ModProp().Set(propID, value);
   if (!BioFailed(ret)) {
      m_owner->OnPropChanged(propID); 
   }

   return ret; 
}

FilterType TriggerImpl::getFilterType()
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   int32 filter = FilterNone;
   MOD_PROP.Get(prop_id_trigFltType[m_tag], filter);
   return (FilterType)filter;
}
ErrorCode  TriggerImpl::setFilterType(FilterType value)
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   PROP_COND_SET_VAL(prop_id_trigFltType[m_tag], (int32)value)
}

double TriggerImpl::getFilterCutoffFreq()
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   double freq = 0;
   MOD_PROP.Get(prop_id_trigFltCutoffFreq[m_tag], freq);
   return freq;
}
ErrorCode TriggerImpl::setFilterCutoffFreq(double value)
{
   assert((uint32)m_tag < MAX_TRIG_COUNT);
   PROP_COND_SET_VAL(prop_id_trigFltCutoffFreq[m_tag], value)
}

// -----------------------------------------------------------------
//function of : NosFltChannelImpl
// -----------------------------------------------------------------
int32 NosFltChannelImpl::getChannel()
{
   return (int32)m_tag;
}
bool NosFltChannelImpl::getEnabled()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT * 8);

   uint8 value = 0;
   MOD_PROP.GetValueNth(CFG_NoiseFilterEnabledChannels, m_tag, value);
   return value != 0;
}
ErrorCode NosFltChannelImpl::setEnabled(bool value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT * 8);

   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }

   return MOD_PROP.SetValueNth(CFG_NoiseFilterEnabledChannels, m_tag, (uint8)value);
}

// -----------------------------------------------------------------
//function of : ConversionImpl
// -----------------------------------------------------------------
//
// clock
SignalDrop ConversionImpl::getClockSource()
{
   int32 src = 0;
   MOD_PROP.Get(CFG_ConvertClockSource, src);
   return (SignalDrop)src;
}
ErrorCode ConversionImpl::setClockSource(SignalDrop value)
{
   PROP_COND_SET_VAL(CFG_ConvertClockSource, (int32)value);
}
double ConversionImpl::getClockRate()
{
   double rate = 0;
   MOD_PROP.Get(CFG_ConvertClockRatePerChannel, rate);
   return rate;
}
ErrorCode ConversionImpl::setClockRate(double value)
{
   PROP_COND_SET_VAL(CFG_ConvertClockRatePerChannel, value)
}
//
// channel
int32 ConversionImpl::getChannelStart()
{
   int32 ch = 0;
   MOD_PROP.Get(CFG_ScanChannelStart, ch);
   return ch;
}
ErrorCode ConversionImpl::setChannelStart(int32 value)
{
   PROP_COND_SET_VAL(CFG_ScanChannelStart, value)
}
int32 ConversionImpl::getChannelCount()
{
   int32 count = 0;
   MOD_PROP.Get(CFG_ScanChannelCount, count);
   return count;
}
ErrorCode ConversionImpl::setChannelCount(int32 value)
{
   PROP_COND_SET_VAL(CFG_ScanChannelCount, value)
}
void ConversionImpl::getChannelMap(int32 count, int8 *data)
{
   MOD_PROP.Read(CFG_ScanEnabledChannels, count, data);
}
ErrorCode ConversionImpl::setChannelMap(int32 count, int8 *data)
{
   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }
   if (m_owner->IsRunning()) { return ErrorFuncBusy; }

   return MOD_PROP.Set(CFG_ScanEnabledChannels, count, (uint8 *)data);
}

// -----------------------------------------------------------------
//function of : RecordImpl
// -----------------------------------------------------------------
int32 RecordImpl::getSectionLength()
{
   int32 length = 0;
   MOD_PROP.Get(CFG_RecordSectionLength, length);
   return length; // Not implemented
}
ErrorCode RecordImpl::setSectionLength(int32 value)
{
   PROP_COND_SET_VAL(CFG_RecordSectionLength, value);
}
int32 RecordImpl::getSectionCount()
{
   int32 count;
   MOD_PROP.Get(CFG_RecordSectionCount, count);
   return count;
}
ErrorCode RecordImpl::setSectionCount(int32 value)
{
   PROP_COND_SET_VAL(CFG_RecordSectionCount, value);
}
int32 RecordImpl::getCycles()
{
   int32 count = 1;
   MOD_PROP.Get(CFG_RecordCycles, count);
   return count;
}
ErrorCode RecordImpl::setCycles(int32 value)
{
   PROP_COND_SET_VAL(CFG_RecordCycles, value);
}

// -----------------------------------------------------------------
//function of : BfdCntrRecordImpl
// -----------------------------------------------------------------
int32 BfdCntrRecordImpl::getSectionLength()
{
   return m_sectionLength;
}

ErrorCode BfdCntrRecordImpl::setSectionLength(int32 value)
{   
   if (m_owner->IsRunning()) {
      return ErrorFuncBusy;
   }

   m_sectionLength = value;

   return Success;
}

int32 BfdCntrRecordImpl::getSectionCount()
{
   return m_sectionCount;
}

ErrorCode BfdCntrRecordImpl::setSectionCount(int32 value)
{
   if (m_owner->IsRunning()) {
      return ErrorFuncBusy;
   }

   m_sectionCount = value;

   return Success;
}

int32 BfdCntrRecordImpl::getCycles()
{
   return 1;
}

ErrorCode BfdCntrRecordImpl::setCycles(int32 value)
{
   if (m_owner->IsRunning()) {
      return ErrorFuncBusy;
   }

   if(1 == value ) { return Success; }

   return WarningPropValueNotSpted;
}

// -----------------------------------------------------------------
//function of : BfdCntrTriggerImpl
// -----------------------------------------------------------------

SignalDrop BfdCntrTriggerImpl::getSource()
{
   MOD_PROP.GetValueNth<SignalDrop>(CFG_TriggerSources, m_owner->getChannelStart(), m_src);

   return m_src;
}

ErrorCode BfdCntrTriggerImpl::setSource(SignalDrop value)
{
   if(m_owner->IsRunning()) return ErrorFuncBusy;

   ErrorCode ret = m_owner->ModProp().SetValueNth(CFG_TriggerSources, m_owner->getChannelStart(), value);
   if (!BioFailed(ret)) {
      m_owner->OnPropChanged(CFG_TriggerSources);

      MOD_PROP.GetValueNth<SignalDrop>(CFG_TriggerSources, m_owner->getChannelStart(), m_src);
   }

   return ret; 
}

ActiveSignal BfdCntrTriggerImpl::getEdge()
{
   MOD_PROP.GetValueNth<ActiveSignal>(CFG_TriggerSourcesEdge, m_owner->getChannelStart(), m_edge);

   return m_edge;
}

ErrorCode BfdCntrTriggerImpl::setEdge(ActiveSignal value)
{
   if(m_owner->IsRunning()) return ErrorFuncBusy;

   ErrorCode ret = m_owner->ModProp().SetValueNth(CFG_TriggerSourcesEdge, m_owner->getChannelStart(), value);
   if (!BioFailed(ret)) {
      m_owner->OnPropChanged(CFG_TriggerSourcesEdge);

      MOD_PROP.GetValueNth<ActiveSignal>(CFG_TriggerSourcesEdge, m_owner->getChannelStart(), m_edge);
   }

   return ret; 
}

double BfdCntrTriggerImpl::getLevel()
{
   MOD_PROP.GetValueNth<double>(CFG_TriggerSourcesLevel, m_owner->getChannelStart(), m_level);

   return m_level;
}

ErrorCode BfdCntrTriggerImpl::setLevel(double value)
{
   if(m_owner->IsRunning()) return ErrorFuncBusy;

   ErrorCode ret = m_owner->ModProp().SetValueNth(CFG_TriggerSourcesLevel, m_owner->getChannelStart(), value);
   if (!BioFailed(ret)) {
      m_owner->OnPropChanged(CFG_TriggerSourcesLevel);

      MOD_PROP.GetValueNth<double>(CFG_TriggerSourcesLevel, m_owner->getChannelStart(), m_level);
   }
   
   return ret; 
}

TriggerAction BfdCntrTriggerImpl::getAction()
{
   return m_action;
}

ErrorCode BfdCntrTriggerImpl::setAction(TriggerAction value)
{
   if (m_owner->IsRunning()) {
      return ErrorFuncBusy;
   }

   uint32 xbuf[16] = {0};
   uint32 size = sizeof(xbuf);
   MOD_PROP.Read(feat_id_trigAction[m_tag], &size, xbuf);

   for(unsigned i = 0; i < size / sizeof(int32); i++) {
      if(value == xbuf[i]) {
         m_action = value;
         return Success;
      }
   }

   return size ? ErrorPropValueNotSpted : ErrorPropNotSpted;
}

int32 BfdCntrTriggerImpl::getDelayCount()
{
   return m_delayCount;
}

ErrorCode BfdCntrTriggerImpl::setDelayCount(int32 value)
{
   if (m_owner->IsRunning()) {
      return ErrorFuncBusy;
   }

   m_delayCount = value;
   return Success;
}

double BfdCntrTriggerImpl::getHysteresisIndex()
{
   MOD_PROP.GetValueNth<double>(CFG_TriggerSourcesHysteresisIndex, m_owner->getChannelStart(), m_hysteresis);

   return m_hysteresis;
}

ErrorCode BfdCntrTriggerImpl::setHysteresisIndex(double value)
{
   if(m_owner->IsRunning()) return ErrorFuncBusy;

   ErrorCode ret = m_owner->ModProp().SetValueNth(CFG_TriggerSourcesHysteresisIndex, m_owner->getChannelStart(), value);
   if (!BioFailed(ret)) {
      m_owner->OnPropChanged(CFG_TriggerSourcesHysteresisIndex);

      MOD_PROP.GetValueNth<double>(CFG_TriggerSourcesHysteresisIndex, m_owner->getChannelStart(), m_hysteresis);
   }

   return ret; 
}

FilterType BfdCntrTriggerImpl::getFilterType()
{
   return m_filterType;
}

ErrorCode BfdCntrTriggerImpl::setFilterType(FilterType value)
{
   if (m_owner->IsRunning()) {
      return ErrorFuncBusy;
   }

   m_filterType = value;
   return Success;
}

double BfdCntrTriggerImpl::getFilterCutoffFreq()
{
   return m_cutoffFreq;
}

ErrorCode BfdCntrTriggerImpl::setFilterCutoffFreq(double value)
{
   if (m_owner->IsRunning()) {
      return ErrorFuncBusy;
   }

   m_cutoffFreq = value;
   return Success;
}

//------------------------------------------------------------------
// kernel event monitor map: collects all the kernel event monitor.
// Each kernel event monitor managers a specified device's events.
//------------------------------------------------------------------
KernEventMonitor::KEMMap KernEventMonitor::s_monitorMap;
MutexLock                KernEventMonitor::s_monitorMapLock;
