#ifndef _DAQNAVI_CNTR_CTRL_BASE_IMPL_H
#define _DAQNAVI_CNTR_CTRL_BASE_IMPL_H

#define CHK_RESULT(ret) {if(BioFailed(ret)) return ret;}
#define CNTR_PROP_BUF_LEN  16

class BDAQ_NO_VTABLE CntrCtrlBaseImpl : public DaqCtrlBaseImpl
{
public:
   CntrFeaturesImpl* getFeatures() 
   { 
      return &m_features; 
   }

   int32 getChannelStart() 
   { 
      return m_chStart; 
   }
   virtual ErrorCode setChannelStart(int32 ch)
   {

      if (!getInitialized()) { return ErrorFuncNotInited; }

      if (ch >= 0) {
         if (ch >= getFeatures()->getChannelCountMax()) {
            return ErrorParamOutOfRange;
         }
         if (!IsCntrFuncSpted(ch, WantedCapability())) {
            return ErrorFuncNotSpted;
         }
      }

      AutoLock<MutexLock> lock(PropLock());
      ErrorCode ret = Success;
      if (m_chStart != ch) {
         StartCounter(false);
         InitializeCounter(false);
         m_chStart = ch;

         if (IsValidChannel()) {
            InitializeCounter(true);
            ret = StartCounter(true);
         }
      }
      return ret;
   }

   int32 getChannelCount() 
   { 
      return m_chCount; 
   }
   virtual ErrorCode setChannelCount(int32 count)
   {
      if (!getInitialized()) { return ErrorFuncNotInited; }

      if (count <= 0 || count > getFeatures()->getChannelCountMax()) {
         return ErrorParamOutOfRange;
      }

      AutoLock<MutexLock> lock(PropLock());
      ErrorCode ret = Success;
      if (m_chCount != count) {
         StartCounter(false);
         InitializeCounter(false);
         m_chCount = count;

         // Bug#8568
         if (IsValidChannel()) {
            InitializeCounter(true);
            ret = StartCounter(true);
         }
      }
      return ret;
   }

   bool getEnabled() 
   { 
      return m_enabled;  
   }
   ErrorCode setEnabled(bool enabled)
   {
      if (!getInitialized()) { return ErrorFuncNotInited; }

      AutoLock<MutexLock> lock(PropLock());
      RefreshState(m_ctrlState);

      ErrorCode ret = Success;
      if (m_enabled != enabled) {
         m_enabled = enabled;

         if (m_chStart >= 0){ ret = StartCounter(enabled); }
      }
      return ret;
   }

   double getNoiseFilterBlockTime()
   {
      double val;
      ModProp().Get(CFG_NoiseFilterOverallBlockTime, val);
      return val;
   }
   ErrorCode setNoiseFilterBlockTime(double value)
   {
      return ModProp().Set(CFG_NoiseFilterOverallBlockTime, value);
   }

   // measurement timeout value
   double getMeasurementTimeout()
   {
      double val;
      ModProp().Get(CFG_MeasurementTimeout, val);
      return val;
   }
   ErrorCode setMeasurementTimeout(double value)
   {
      return ModProp().Set(CFG_MeasurementTimeout, value);
   }

   // 
   //NOTE: For compatible ONLY! DON'T USE THIS PROPERTY ANYMORE! 
   // Please use the counter channel's noise filter property instead!
   NosFltChannelImpl* getNoiseFilter()
   {
      if (!getInitialized() || !getFeatures()->getNoiseFilterSupported()) {
         return NULL;
      }

      AutoLock<MutexLock> lock(PropLock());
      m_noiseFilter.setOwner(this, m_chStart);
      return m_chStart < 0 ? NULL : &m_noiseFilter;
   }

   bool getRunning() { return IsRunning(); }

   //
   // event:
   virtual void addEventHandler(EventId id, DaqEventProc proc, void * param)
   {
      if (!IsValidEvent(id)) {  
         // Not our event, let 'base' to do default process
         DaqCtrlBaseImpl::addEventHandler(id, proc, param);
         return; 
      }

      UserEventManager::Add(getOuter(), id, (UserEventProc)proc, param);

      // The first handler is added.
      if (UserEventManager::Count(getOuter(), id) == 1 
         && getChannelStart() >= 0 
         && getChannelCount() > 0) { 

         bool succeeded = false;
         for (int i = 0; i < getChannelCount(); ++i) {
            ErrorCode ret = RegisterKernEvent(EventId(GetCntrEventId() + getChannelStart() + i));
            succeeded = succeeded || ret == Success;
         }

         if (succeeded) { 
            StartCounter(true); // restart counter if needed
         }
      }
   }

   virtual void removeEventHandler(EventId id, DaqEventProc proc, void * param)
   {
      if (!IsValidEvent(id)) { 
         // Not our event, let 'base' to do default process
         DaqCtrlBaseImpl::removeEventHandler(id, proc, param);
         return; 
      }

      UserEventManager::Remove(getOuter(), id, (UserEventProc)proc, param);
      if (getChannelStart() >= 0 
         && getChannelCount() > 0 
         && !UserEventManager::IsExist(getOuter(), id)) {

         for (int i = 0; i < getChannelCount(); ++i) {
            UnregisterKernEvent(EventId(GetCntrEventId() + getChannelStart() + i));
         }
      }
   }

public: // ctor & dtor
   CntrCtrlBaseImpl() : DaqCtrlBaseImpl(DaqCounter)
   {
      m_chStart = -1;
      m_chCount = 1;

      m_enabled = false;
      m_features.setOwner(this);
      m_noiseFilter.setOwner(this, -1);
   }

   virtual ~CntrCtrlBaseImpl() 
   {
   }

   DaqCtModule & ModuleRef() { return *(DaqCtModule*)getModule(); }

protected: 
   //------------------------------------------------------------------------
   // override the method in base class
   //------------------------------------------------------------------------
   virtual ModuleType WantedType() { return DaqCounter; }

   virtual std::vector<AccessMode> WantedModes()
   {
      AccessMode spted[] = { ModeWrite, ModeWriteWithReset };
      return std::vector<AccessMode>(&spted[0], &spted[2]);
   }

   virtual bool IsSuitable(DaqModule & module)
   {
      if (module.getType() != DaqCounter){ return false; }

      // get each counter channel's capability to check that
      // whether some channel is suitable for the class
      CounterIndexerImpl<CounterCapability> indexer;
      indexer.setOwner((DaqCtModule*)&module, CFG_FeatureCapabilitiesOfCounter0);

      int     count = indexer.getLength();
      bool suitable = false;
      for(int i = 0; i < count && !suitable; ++i) {
         ArrayImpl<CounterCapability> *caps = indexer.getItem(i);
         if (caps != NULL) {
            for (int j = 0; j < caps->getLength() && !suitable; ++j){
               suitable = caps->Item(j) == WantedCapability();
            }
            caps->Dispose();
         }
      }
      return suitable;
   }

   // initialize/de-initialize the control 
   virtual ErrorCode InitializeControl(DeviceInformation const &x)
   {
      // let's the base do initialization at first.
      ErrorCode ret = DaqCtrlBaseImpl::InitializeControl(x);
      if (ret != Success) { return ret; }

      // not counter module
      if (getFeatures()->getChannelCountMax() == 0){
         return ErrorFuncNotSpted;
      }
      getFeatures()->BuildSupportedScenarios();

      m_chStart = -1;
      m_chCount = 1;
      m_enabled = false;
      return Success;
   }

   virtual void DeinitializeControl()
   {
      // reset all configuration to null
      StartCounter(false);
      InitializeCounter(false);

      m_chStart = -1;
      m_chCount = 1;
      m_enabled = false;

      DaqCtrlBaseImpl::DeinitializeControl();
   }

   //------------------------------------------------------------------------
   // virtual functions the derived class must override
   //------------------------------------------------------------------------
   virtual CounterCapability WantedCapability() =  0;

   //Initialize/Deinitialize counter
   virtual ErrorCode InitializeCounter(bool init)
   {
      if (!IsValidChannel()) { return ErrorParamOutOfRange; }

      // initialize the channel related properties/events
      //
      // Activate / deactivate the events
      if (UserEventManager::IsExist(getOuter(), GetCntrEventId())) {
         if (init) { 
            for (int i = 0; i < getChannelCount(); ++i) {
               RegisterKernEvent(EventId(GetCntrEventId() + getChannelStart() + i)); 
            }
         } else { 
            for (int i = 0; i < getChannelCount(); ++i) {
               UnregisterKernEvent(EventId(GetCntrEventId() + getChannelStart() + i)); 
            }
         }
      }
      return Success;
   }

   // Start/stop counter
   virtual ErrorCode StartCounter(bool start) = 0;

   // Supported Counter Event ID
   virtual EventId GetCntrEventId() { return EventUnknown; }

   // Check whether the event id is supported by a counter control or not
   virtual bool IsValidEvent(EventId id) { return id != EventUnknown && id == GetCntrEventId(); } 
   
   //------------------------------------------------------------------------
   // helper functions
   //------------------------------------------------------------------------
   bool IsValidChannel()    
   { 
      int32 chMax = getFeatures()->getChannelCountMax();
      return getChannelStart() >= 0 && getChannelStart() < chMax
          && getChannelCount() > 0  && getChannelCount() <= chMax;
   }

   bool IsCntrFuncSpted(int32 ch, CounterCapability func)
   {
      ArrayImpl<CounterCapability> *caps = getFeatures()->getCapabilities()->getItem(ch);
      bool spted = false;
      if (caps != NULL) {
         for (int j = 0; j < caps->getLength() && !spted; ++j) {
            spted = caps->Item(j) == func;
         }
         caps->Dispose();
      }
      return spted;
   }

protected:
   int32  m_chStart;
   int32  m_chCount;
   bool   m_enabled;
   CntrFeaturesImpl  m_features;
   NosFltChannelImpl m_noiseFilter;
};

#endif
