#ifndef _DAQNAVI_CNTR_FEATURE_IMPL_H
#define _DAQNAVI_CNTR_FEATURE_IMPL_H

class CntrCtrlBaseImpl;

class XCntrIndexer
{
public:
   virtual void     BDAQCALL Dispose() = 0;   // destroy the instance
   virtual int32    BDAQCALL getLength() = 0;
   virtual XArray * BDAQCALL getItem(int32 channel) = 0;
};

template<class T>
class CounterIndexerImpl : public XCntrIndexer
{
   enum{OwnerNotValid, OwnerIsModule, OwnerIsCtrl};

public:
   void BDAQCALL Dispose()
   {
      if (m_disposable) { 
         delete this; 
      }
   }
   int32 BDAQCALL getLength()
   {
      int32 count = 0;
      if (OwnerProp().Get(CFG_FeatureChannelNumberMax, count) == Success) {
         count += 1; // convert from channel number to channel count
      }

      return count;
   }
   ArrayImpl<T>* BDAQCALL getItem(int32 channel)
   {
      ArrayImpl<T>* vals = NULL;
      if (channel >= 0 && channel < MAX_CNTR_CH_COUNT) {
         PropertyId id = PropertyId(m_baseId + channel);

         int32  xbuf[128];
         uint32 size = sizeof(xbuf);
         int32 *data = (int32 *)OwnerProp().ReadValueArray(id, size, xbuf);

         if (size) {
            vals = new ArrayImpl<T>();
            vals->Assign((T*)data, size/sizeof(int32));
         }

         if (data != xbuf) { free(data); }
      }
      return vals;
   }

public:
   CounterIndexerImpl(bool disposable = false)
   {
      m_disposable = disposable;
      m_ownerType  = OwnerNotValid;
      m_owner      = NULL;
      m_baseId     = PropertyId(-1);
   }
   virtual ~CounterIndexerImpl()
   {
   }  
   void setOwner(DaqCtModule *owner, PropertyId baseId)
   {
      m_owner     = owner;
      m_ownerType = OwnerIsModule;
      m_baseId    = baseId;
   }
   void setOwner(CntrCtrlBaseImpl *owner, PropertyId baseId)
   {
      m_owner     = owner;
      m_ownerType = OwnerIsCtrl;
      m_baseId    = baseId;
   }

protected:
   DaqProperty OwnerProp()
   {
      return m_ownerType == OwnerIsModule ? 
         DaqProperty((DaqModule*)m_owner) : ((DaqCtrlBaseImpl*)m_owner)->ModProp();
   }

private:
   bool  m_disposable;
   int32 m_ownerType; // 0 -- invalid, 1 -- DaqModule, 2 -- CntrCtrlBaseImpl
   void  *m_owner;
   PropertyId m_baseId;
};

typedef CounterIndexerImpl<CounterCapability> CounterCapabilityIndexerImpl;
typedef CounterIndexerImpl<SignalDrop>        CounterClockSourceIndexerImpl;
typedef CounterIndexerImpl<SignalDrop>        CounterGateSourceIndexerImpl;

class CntrFeaturesImpl
{
public:
   // channel features
   int32 getChannelCountMax();
   int32 getResolution();
   int32 getDataSize();
   CounterCapabilityIndexerImpl * getCapabilities();

   // noise filter features
   bool                           getNoiseFilterSupported();
   void                           getNoiseFilterBlockTimeRange(MathInterval &x); 
   ArrayImpl<uint8>*              getNoiseFilterOfChannels();

   // new: measurement timeout range
   void                           getMeasurementTimeoutRange(MathInterval &x);

   // event counting features
   CounterClockSourceIndexerImpl* getEcClockSources();
   ArrayImpl<SignalPolarity>*     getEcClockPolarities();
   ArrayImpl<int32>*              getEcGateControlOfChannels();
   ArrayImpl<SignalPolarity>*     getEcGatePolarities();

   // frequency measurement features
   ArrayImpl<FreqMeasureMethod>*  getFmMethods();

   // one-shot features
   CounterClockSourceIndexerImpl* getOsClockSources();
   ArrayImpl<SignalPolarity>*     getOsClockPolarities();
   CounterGateSourceIndexerImpl*  getOsGateSources();
   ArrayImpl<SignalPolarity>*     getOsGatePolarities();
   ArrayImpl<OutSignalType>*      getOsOutSignals();
   void                           getOsDelayCountRange(MathInterval &x);

   // Timer/pulse features
   ArrayImpl<int32>*              getTmrGateControlOfChannels();
   ArrayImpl<SignalPolarity>*     getTmrGatePolarities();
   ArrayImpl<OutSignalType>*      getTmrOutSignals();
   void                           getTmrFrequencyRange(MathInterval &x);

   // pulse width measurement features
   ArrayImpl<CounterCascadeGroup>* getPiCascadeGroup();

   // pulse width modulation features
   ArrayImpl<int32>*              getPoGateControlOfChannels();
   ArrayImpl<SignalPolarity>*     getPoGatePolarities();
   ArrayImpl<OutSignalType>*      getPoOutSignals();
   void                           getPoHiPeriodRange(MathInterval &x);
   void                           getPoLoPeriodRange(MathInterval &x);
   void                           getPoOutCountRange(MathInterval &x);

   // new: trigger features for pulse width modulation
   int32                          getPoTriggerCount();
   ArrayImpl<SignalDrop>*         getPoTriggerSources();
   ArrayImpl<TriggerAction>*      getPoTriggerActions();
   void                           getPoTriggerDelayRange(MathInterval &x);

   // Up-down counter features
   ArrayImpl<CountingType>*       getUdCountingTypes(); 
   ArrayImpl<int32>*              getUdInitialValues();
   ArrayImpl<int32>*              getUdSnapEventSources();
   void                           getUdValueResetTimes(MathInterval &x);

   // new features for Counter continue comparing: outputting pulse settings
   ArrayImpl<int32>*              getCcpGateControlOfChannels();
   ArrayImpl<SignalPolarity>*     getCcpGatePolarities();
   ArrayImpl<OutSignalType>*      getCcpOutSignals();
   void                           getCcpHiPeriodRange(MathInterval &x);
   void                           getCcpLoPeriodRange(MathInterval &x);
   void                           getCcpOutCountRange(MathInterval &x);

   // new: buffered counter features
   int32                          getTriggerCount();
   ArrayImpl<SignalDrop>*         getTriggerSources(int32 trigger);
   ArrayImpl<TriggerAction>*      getTriggerActions(int32 trigger);
   
   CounterClockSourceIndexerImpl* getEcSampleClkSrcs();
   CounterClockSourceIndexerImpl* getPiSampleClkSrcs();
   CounterClockSourceIndexerImpl* getPoSampleClkSrcs();
   CounterClockSourceIndexerImpl* getUdSampleClkSrcs();

public:
   CntrFeaturesImpl(CntrCtrlBaseImpl * owner = NULL)
   {
      m_owner = owner;
      m_scenarios = 0;
   }

   void setOwner(CntrCtrlBaseImpl * owner)
   {
      assert(owner);
      m_owner = owner;
      m_capsIndexer.setOwner(owner, CFG_FeatureCapabilitiesOfCounter0);
      m_ecClkSrcIndexer.setOwner(owner, CFG_FeatureEcClkSourceOfCounter0);
      m_osClkSrcIndexer.setOwner(owner, CFG_FeatureOsClkSourceOfCounter0);
      m_osGateSrcIndexer.setOwner(owner, CFG_FeatureOsGateSourceOfCounter0);
      m_ecSampleSrcIndexer.setOwner(owner, CFG_FeatureEcSampleClkSourceOfCounter0);
      m_piSampleSrcIndexer.setOwner(owner, CFG_FeaturePiSampleClkSourceOfCounter0);
      m_poSampleSrcIndexer.setOwner(owner, CFG_FeaturePoSampleClkSourceOfCounter0);
      m_udSampleSrcIndexer.setOwner(owner, CFG_FeatureUdSampleClkSourceOfCounter0);
   }

   void BuildSupportedScenarios()
   {
      // calculate capability overview
      m_scenarios = 0;
      for (int i = 0; i < getChannelCountMax(); ++i) {
         ArrayImpl<CounterCapability> * caps = m_capsIndexer.getItem(i);
         if (!caps) { continue; }

         for (int j = 0; j < caps->getLength(); ++j) {
            switch(caps->Item(j))
            {
            case InstantEventCount:   m_scenarios |= SceEventCounter;         break;
            case OneShot:             m_scenarios |= SceOneShot;              break;
            case TimerPulse:          m_scenarios |= SceTimerPulse;           break;
            case InstantFreqMeter:    m_scenarios |= SceFreqMeter;            break;
            case InstantPwmIn:        m_scenarios |= ScePwMeter;              break;
            case InstantPwmOut:       m_scenarios |= ScePwModulator;          break;
            case UpDownCount:         m_scenarios |= SceUdCounter;            break;
            case BufferedEventCount:  m_scenarios |= SceBufferedEventCounter; break;
            case BufferedPwmIn:       m_scenarios |= SceBufferedPwMeter;      break;
            case BufferedPwmOut:      m_scenarios |= SceBufferedPwModulator;  break;
            case BufferedUpDownCount: m_scenarios |= SceBufferedUdCounter;    break;
            default: 
               break;
            }
         }

         caps->Dispose();
      }
   }
private:
   CntrCtrlBaseImpl * m_owner;
   uint32             m_scenarios;
   CounterIndexerImpl<CounterCapability> m_capsIndexer;
   
   CounterIndexerImpl<SignalDrop>        m_ecClkSrcIndexer;
   CounterIndexerImpl<SignalDrop>        m_osClkSrcIndexer;
   CounterIndexerImpl<SignalDrop>        m_osGateSrcIndexer;

   CounterIndexerImpl<SignalDrop>        m_ecSampleSrcIndexer;
   CounterIndexerImpl<SignalDrop>        m_piSampleSrcIndexer;
   CounterIndexerImpl<SignalDrop>        m_poSampleSrcIndexer;
   CounterIndexerImpl<SignalDrop>        m_udSampleSrcIndexer;
};


#endif
