#ifndef _DAQNAVI_UP_DOWN_COUNTER_CTRL_IMPL_H
#define _DAQNAVI_UP_DOWN_COUNTER_CTRL_IMPL_H


class UdCounterCtrlImpl : public CntrCtrlBaseImpl
{
   typedef std::map<int32, int32 *> UdEventMap;

public:
   // Property
   ArrayImpl<UdChannelImpl>* getChannels()
   {
      if (getInitialized()) {
         if (m_channels.getLength() == 0) {
            int count = getFeatures()->getChannelCountMax();
            m_channels.setCapacity(count);

            for (int i = 0; i < count; ++i) {
               m_channels.Add(UdChannelImpl(this, i));
            }
         }
      }

      return &m_channels;
   }

   // method
   ErrorCode SnapStart(int32 srcId)
   {
      if (!getInitialized() || !IsValidChannel()) { 
         return ErrorFuncNotInited; 
      }
      if (!IsValidSnapSrc(srcId)) { 
         return ErrorParamNotSpted;  
      }

      ErrorCode ret = Success;
      if (m_events.find(srcId) == m_events.end()) {
         // register the event
         ret = RegisterKernEvent(EventId(srcId));
         if (ret != Success) { return ret; }

         // add new entry and initialize it.
         m_events[srcId] = NULL;  

         // clear the flag of un-expected event signaled 
         // during we are starting counter snap.
         ModEvent().ClearFlag(EventId(srcId), 0, 0);
      }

      if (m_events[srcId] == NULL) {
         ret = ModuleRef().CtSnapStart(EventId(srcId), 0, 
                              m_udChanMax, (uint32**)&m_events[srcId], NULL);
      }

      return ret;
   }

   ErrorCode SnapStop(int32 srcId)
   {
      if (!getInitialized() || !IsValidChannel()) {
         return ErrorFuncNotInited;
      }
      if (!IsValidSnapSrc(srcId)) {
         return ErrorParamNotSpted;
      }

      UdEventMap::iterator itor = m_events.find(srcId);
      if (itor != m_events.end()) {
         // stop snap
         ModuleRef().CtSnapStop(EventId(srcId));
         itor->second = NULL;

         // unregister the event
         if (uint32(srcId) < EvtUdIndex0 || uint32(srcId) > EvtUdIndex7) {
            UnregisterKernEvent(EventId(srcId));
            m_events.erase(itor);
         }
      }

      return Success;
   }

   ErrorCode CompareSetTable(int channel, int32 count, int32 *table)
   {
      if (!getInitialized() || !IsValidChannel()) { return ErrorFuncNotInited; }
      if (count < 0) { return ErrorParamOutOfRange; }

      return ModuleRef().CCmpSetTable(channel, count, (uint32 *)table);
   }
   ErrorCode CompareSetInterval(int channel, int32 start, int32 increment,int32 count)
   {
      if (!getInitialized() || !IsValidChannel()) { return ErrorFuncNotInited; }
      if (count < -1) { return ErrorParamOutOfRange; }

      return ModuleRef().CCmpSetInterval(channel, start, increment, count);
   }
   ErrorCode CompareClear(int channel)
   {
      if (!getInitialized() || !IsValidChannel()) { return ErrorFuncNotInited; }
      return ModuleRef().CCmpClear(channel);
   }

   ErrorCode ValueReset()
   {
      if (!getInitialized() || !IsValidChannel()) { return ErrorFuncNotInited; }
      return ModuleRef().UdResetValue(getChannelStart(), getChannelCount());
   }

   ErrorCode Read(int32 count, int32 *data)
   {
      if (count > getChannelCount()) {
         count = getChannelCount();
      }

      ErrorCode ret = ModuleRef().UdRead(getChannelStart(), count, (uint32*)data);

      // Cache the data
      if (ret == Success) {
         for (int i = 0; i < count; ++i) {
            m_values[(getChannelStart() + i) % MAX_CNTR_CH_COUNT] = data[i];
         }
      } else { // Return the last good value
         for (int i = 0; i < count; ++i) {
            data[i] = m_values[(getChannelStart() + i) % MAX_CNTR_CH_COUNT];
         }
      }
      return ret;
   }

   //
   // event:
   virtual void addEventHandler(EventId id, DaqEventProc proc, void * param)
   {
      //
      // Not our event, let 'base' to do default process
      if (!IsValidEvent(id)) {
         CntrCtrlBaseImpl::addEventHandler(id, proc, param);
         return;
      }
     
      // Add up/down counter event
      UserEventManager::Add(getOuter(), id, (UserEventProc)proc, param);

      if (UserEventManager::Count(getOuter(), id) == 1 && IsValidChannel()) {
         RegisterDefaultEvents();
         StartCounter(true);  // restart counter if needed
      }
   }

   virtual void removeEventHandler(EventId id, DaqEventProc proc, void * param)
   {
      //
      // Not our event, let 'base' to do default process
      if (!IsValidEvent(id)) {
         CntrCtrlBaseImpl::removeEventHandler(id, proc, param);
         return;
      }

      //
      // Remove up/down counter event
      UserEventManager::Remove(getOuter(), id, (UserEventProc)proc, param);

      if (IsValidChannel() && !UserEventManager::IsExist(getOuter(), id)) {
         UdEventMap::const_iterator itor;
         for (itor = m_events.begin(); itor != m_events.end(); ++itor) {
            UnregisterKernEvent((EventId)itor->first);
         }
         m_events.clear();
      }
   }

public: // ctor & dtor
   UdCounterCtrlImpl() : m_channels(false)
   { 
      memset(m_values, 0, sizeof(m_values));  
      m_udChanMax = 0;
   }
   virtual ~UdCounterCtrlImpl() 
   { 
      UserEventManager::Remove(getOuter(), this->GetCntrEventId());
   }

protected:
   //------------------------------------------------------------------------
   // override the method in base class
   //------------------------------------------------------------------------
   virtual CounterCapability WantedCapability() { return UpDownCount; }

   // Supported Counter Event ID, uses 'EvtUdIndex0' as a general event id for up/down counter 
   virtual EventId GetCntrEventId() { return EvtUdIndex0; }

   // initialize the control 
   virtual ErrorCode InitializeControl(DeviceInformation const &x)
   {
      // let's the base do initialization at first.
      ErrorCode ret = CntrCtrlBaseImpl::InitializeControl(x);
      if (ret != Success) { return ret; }

      m_channels.Clear();
      return Success;
   }

   // initialize the counter channel 
   virtual ErrorCode InitializeCounter(bool init)
   {
      ErrorCode ret = CntrCtrlBaseImpl::InitializeCounter(init);
      if (ret != Success) { return ret; }

      memset(m_values, 0, sizeof(m_values));  

      // initialize / de-initialize the channel related properties/events
      if (init) {
         // calculate the max. channel count that supports up-down counting
         m_udChanMax = 0;
         for (int i = 0; i < getFeatures()->getChannelCountMax(); ++i) {
            uint32 caps[16];
            uint32 size = sizeof(caps);
            ModProp().Read(PropertyId(CFG_FeatureCapabilitiesOfCounter0 + i), &size, caps);
            if (ArrayIndexOf(caps, size / sizeof(uint32), (uint32)UpDownCount) != -1) {
               ++m_udChanMax;
            }
         }

         // activate the counter interrupt events  
         if (UserEventManager::IsExist(getOuter(), GetCntrEventId())) {
            RegisterDefaultEvents();
         }
      } else {
         m_udChanMax = 0;

         // deactivate the events 
         UdEventMap::const_iterator itor;
         for (itor = m_events.begin(); itor != m_events.end(); ++itor) {
            UnregisterKernEvent(EventId(itor->first));
         }
         m_events.clear();
      }
      return Success;
   }

   // start/restart counter after something had been changed
   virtual ErrorCode StartCounter(bool start)
   {
      if (!IsValidChannel()) { return ErrorParamOutOfRange; }

      // stop counter at first
      if (IsRunning()) {
         ModuleRef().CtReset(getChannelStart(), getChannelCount());
         setState(Stopped);
      }

      // start action
      ErrorCode err = Success;
      if (start && getEnabled()) {
         err = ModuleRef().UdStart(getChannelStart(), getChannelCount(), true);
         if (err == Success) {
            setState(Running);
         }
      }
      return err;
   }

   void RegisterDefaultEvents()
   {
      ErrorCode ret;
      // register the native event into kernel driver
      for (int i = 0; i < getChannelCount(); ++i) {
         int32 ch = getChannelStart() + i;
         ret = RegisterKernEvent(EventId(EvtUdIndex0 + ch));
         if (ret == Success) {
            m_events[EventId(EvtUdIndex0 + ch)] = NULL;
         }

         ret = RegisterKernEvent(EventId(EvtCntPatternMatch0 + ch));
         if (ret == Success) {
            m_events[EventId(EvtCntPatternMatch0 + ch)] = NULL;
         }

         ret = RegisterKernEvent(EventId(EvtCntCompareTableEnd0 + ch));
         if (ret == Success) {
            m_events[EventId(EvtCntCompareTableEnd0 + ch)] = NULL;
         }
      }
   }

   bool IsValidSnapSrc(int32 srcId)
   {
      ArrayImpl<int32> *sptedSrcs = getFeatures()->getUdSnapEventSources();
      bool valid = false;

      if (sptedSrcs) {
         for (int i = 0; !valid && i < sptedSrcs->getLength(); ++i) {
            valid = sptedSrcs->Item(i) == srcId;
         }
         sptedSrcs->Dispose();
      }
      return valid;
   }

   // -----------------------------------------------------------------
   // driver event receiver for up/down counter
   // -----------------------------------------------------------------
   virtual void OnKernEvent(EventId id)
   {
      UdEventMap::iterator itor = m_events.find(id);
      if (itor != m_events.end()) {
         UdCntrEventArgs args;
         args.Id = id;
         if (itor->second) {
            args.Length = m_udChanMax;
            memcpy(args.Data, itor->second, m_udChanMax * sizeof(int32));
         } else {
            args.Length = 0;
         }

         // call user handler
         UserEventManager::Fire(getOuter(), GetCntrEventId(), &args);

         // automatically clear the event flag after the user had processed it.
         ModEvent().ClearFlag(id);
      }
   };

private:
   ArrayImpl<UdChannelImpl> m_channels;
   UdEventMap               m_events;
   int32                    m_udChanMax; // max. channel count that supports Up-down counting
   int32                    m_values[MAX_CNTR_CH_COUNT]; // cached counter value for device broken down unexpected.
};

#endif
