#ifndef _DAQNAVI_BUFFERED_DO_CTRL_IMPL_H
#define _DAQNAVI_BUFFERED_DO_CTRL_IMPL_H

class BufferedDoCtrlImpl : public DioCtrlBaseImpl
{

public:
   //
   // event:
   // Use the default implementation in DaqCtrlBase

   // method
   ErrorCode Prepare()
   {
      AutoLock<MutexLock> lock(PropLock());

      ErrorCode ret;
      if (!getInitialized()) {
         ret = ErrorFuncNotInited;
      } else if (IsRunning()) {
         ret = ErrorFuncBusy;
      } else if (IsReady()){
         ret = Success;
      } else {
         // get scan port count for the performance of method 'GetData'
         ModProp().Get(CFG_DoScanPortCount,    m_scanPortCount);

         //get section count
         ModProp().Get(CFG_DoPortSectionCount, m_sectionCount);

         // configuration h/w. Only set the property we cached,
         // the others had been set when the user calls 'set' method of the property
         // 
         // set: trigger delay count
         int32 multiplier = 1;
         if (getFeatures()->getDoSamplingMethod() == EqualTimeSwitch) {
            multiplier = m_scanPortCount;
         }
         ModProp().Set(CFG_DoTriggerDelayCount, m_trigger.getDelayCount() * multiplier);   // trigger 0
         ModProp().Set(CFG_DoTrigger1DelayCount, m_trigger1.getDelayCount() * multiplier); // trigger 1

         //Read back the adjustable parameters
         int32 trig0_delay = 0, trig1_delay = 0;
         ModProp().Get(CFG_DoTriggerDelayCount, trig0_delay);
         ModProp().Get(CFG_DoTrigger1DelayCount, trig1_delay);

         m_trigger.adjustDelayCount(trig0_delay / multiplier);
         m_trigger1.adjustDelayCount(trig1_delay / multiplier);

         // prepare buffered DO         
         m_dataBufLen = 0;
         ret = ModuleRef().BfdDoPrepare((uint32 &)m_dataBufLen, m_dataPtr);
         setState(ret == Success ? Ready : Idle);
      }
      return ret;
   }

   // synchronous & one-shot
   ErrorCode RunOnce() {
      //
      // We cannot run with synchronous mode if record cycles is zero 
      // and/or section count is zero(unlimited record).
      int32 rec_cycles = 1;
      ModProp().Get(CFG_DoPortCycles, rec_cycles);

      if (!rec_cycles || !m_sectionCount) {
         return ErrorPropValueConflict;
      }

      return StartBfdDo(false, false); 
   }

   // asynchronous, one-shot or streaming, depends on the property 'Streaming'
   ErrorCode Start() { 
      return StartBfdDo(m_sectionCount == 0, true); 
   }

   ErrorCode Stop(int32 action) {
      if (IsIdle()) { return Success;  }
      // ji.dong mask because DO stop maybe not completed synchronously
      // (depending on the 'action' parameter)
      // setState(Stopped);
      return ModuleRef().BfdDoStop(action);
   }

   void Release()
   {
      // Fix bug#19706: Call driver 'Release' function anyway to release resource, if any.
      // It is harmless operation even if the device is in 'FAI IDLE' state.
      // if (IsIdle()) { return; }
      // Call driver 'Release' only if we had called the method 'Prepare'
      if (m_dataPtr) {
         ModuleRef().BfdDoRelease();
      }

      setState(Idle);
      m_dataPtr    = NULL;
      m_dataBufLen = 0;
   }

   ErrorCode SetData(int32 count, void *buffer)
   {
      ErrorCode ret;
      if (IsIdle()) { 
         ret = Prepare();
         if (ret != Success) { return ret; }
      }

      if (count > m_dataBufLen) { 
         count = m_dataBufLen; 
      }

      AutoLock<MutexLock> lock(BuffLock());

      //
      // The device is using the new BAO method: BfdDoSetData
      if (m_dataPtr == MULTI_BUF_PTR) {
         ret = ModuleRef().BfdDoSetData((uint32&)count, buffer);
      } else {
         ret = SetDataRaw(count, buffer);
      }

      // Update the 'Write Position'
      m_dataTransedCleared = true;
      ModEvent().ClearFlag(EvtBufferedDoDataTransmitted, count);

      return ret;
   }

   bool GetBufferStatus(int32 & count, int32 & offset)
   {
      EventStatus es = {sizeof(es)};
      ModEvent().GetStatus(EvtBufferedDoDataTransmitted, es);
      count  = (es.General.lParam / m_scanPortCount) * m_scanPortCount;
      offset = es.General.rParam;

      return false;
   }

   // property
   void *                  getBuffer()         { return m_dataPtr;     }
   int32                   getBufferCapacity() { return m_dataBufLen;  }
   ScanPortImpl *          getScanPort()       { return &m_scanPort;   }
   ConvertClockImpl *      getConvertClock()   { return &m_convClock;  } 

   // trigger 0, 1, ... , x
   TriggerImpl * getTrigger(int32 trigger)
   { 
      TriggerImpl *x = NULL;

      // Bug#14478
      //if ((uint32)trigger < (uint32)getFeatures()->getTriggerCount()) {
      switch(trigger) {
         default: x = NULL;        break;
         case 0:  x = &m_trigger;  break;
         case 1:  x = &m_trigger1; break;
      }

      return x;
   }

public: // ctor & dtor
   BufferedDoCtrlImpl()
   {
      m_dataPtr            = NULL;
      m_dataBufLen         = 0;
      m_dataTransedCleared = false;

      m_scanPort.setOwner(this);
      m_convClock.setOwner(this, true);
      m_trigger.setOwner(this, 0, true);
      m_trigger1.setOwner(this, 1, true);
   }

   virtual ~BufferedDoCtrlImpl()
   {
      // clear all user event delegate
      for (unsigned i = 0; i < ARRAY_SIZE(bfd_do_events); ++i) {
         UserEventManager::Remove(getOuter(), bfd_do_events[i]);
      }
   }

protected: 
   //------------------------------------------------------------------------
   // override the method in base class
   //------------------------------------------------------------------------
   virtual DioPortDir WantedPort() { return Output; }

   virtual std::vector<AccessMode> WantedModes()
   {
      AccessMode spted[] = { ModeWrite, ModeWriteWithReset };
      return std::vector<AccessMode>(&spted[0], &spted[2]);
   }
   
   virtual bool IsSuitable(DaqModule &module)
   {
      if (DioCtrlBaseImpl::IsSuitable(module)) {
         return DaqProperty(module).GetLength(CFG_FeatureDoConvertClockSources) != 0;
      }
      return false;
   }

   // update the control state
   virtual void RefreshState(ControlState &curr)
   {
      if (curr == Running) {
         EventStatus es = {sizeof(es)};
         ModEvent().GetStatus(EvtBufferedDoStopped, es);
         if (es.General.lParam) {
            setState(Stopped);
         }
      }
   }

   // initialize/deinitialize the control 
   virtual ErrorCode InitializeControl(DeviceInformation const &x)
   {
      // let's the base do initialization at first.
      ErrorCode ret = DioCtrlBaseImpl::InitializeControl(x);
      if (ret != Success) { return ret; }

      // buffered DO not supported
      if (!getFeatures()->getBufferedDoSupported()) {
         return ErrorFuncNotSpted;
      }

      LoadProperties();

      // activate the events       
      for (unsigned i = 0; i < ARRAY_SIZE(bfd_do_events); ++i) {
         if (UserEventManager::IsExist(getOuter(), bfd_do_events[i])) {
            RegisterKernEvent(bfd_do_events[i]);
         }
      }

      return Success;
   }

   virtual void DeinitializeControl()
   {
      // release buffered DO
      Release();

      // deactivate the events        
      for (unsigned i = 0; i < ARRAY_SIZE(bfd_do_events); ++i) {
         UnregisterKernEvent(bfd_do_events[i]);
      }

      // reset properties to its default value
      m_dataSize = 0;

      m_trigger.adjustDelayCount(0);
      m_trigger1.adjustDelayCount(0);

      // let's the base do de-initialization at last.
      DioCtrlBaseImpl::DeinitializeControl();
   }

   //
   // Reload the properties' value from device
   virtual void ReloadProperties()
   {
      DioCtrlBaseImpl::ReloadProperties();
      LoadProperties();
   }

   void LoadProperties()
   {
      // cache the raw data size for the performance of method 'GetData'
      m_dataSize = getFeatures()->getDataSize();

      // get the property default value from device
      m_scanPortCount = m_scanPort.getPortCount();
      assert(m_scanPortCount);

      // trigger delay count
      int32 multiplier = 1;
      if (getFeatures()->getDoSamplingMethod() == EqualTimeSwitch) {
         multiplier = m_scanPortCount;
      }

      int32 trig0_delay = 0, trig1_delay = 0;
      ModProp().Get(CFG_DoTriggerDelayCount, trig0_delay);
      ModProp().Get(CFG_DoTrigger1DelayCount, trig1_delay);

      m_trigger.adjustDelayCount(trig0_delay / multiplier);
      m_trigger1.adjustDelayCount(trig1_delay / multiplier);
   }

   //------------------------------------------------------------------------
   // internal methods
   //------------------------------------------------------------------------
   ErrorCode StartBfdDo(bool streaming, bool async)
   {
      if (IsIdle()) { return ErrorFuncNotInited; }
      if (IsRunning()) { return ErrorFuncBusy;   }

      // start buffered AO
      ErrorCode ret;
      if (streaming) {
         async = true;
         ret = ModuleRef().BfdDoRun();
      } else {
         ret = ModuleRef().BfdDoRunOnce(async);
      }

      if (ret == Success) {
         setState(async ? Running : Stopped);
      }
      return ret;
   }

   ErrorCode SetDataRaw(int32 &count, void *rawData)
   {
      assert(!IsIdle() && count <= m_dataBufLen);

      // get current 'Write Position'
      EventStatus es = {sizeof(es)};
      ModEvent().GetStatus(EvtBufferedDoDataTransmitted, es);
      int32 offset = es.General.rParam;
      
      // copy the first segment, it's from 'offset' to 'buffer end', 'cpyLen' length
      int32 cpyLen = x_min(count, m_dataBufLen - offset);
      memcpy((uint8*)m_dataPtr + offset * m_dataSize, rawData, cpyLen * m_dataSize);

      // copy the rest from 'buffer start'
      if (count > cpyLen) {
         memcpy(m_dataPtr, (uint8*)rawData + cpyLen * m_dataSize, (count - cpyLen) * m_dataSize);
      }

      return Success;
   }
   
   // -----------------------------------------------------------------
   // Check whether the event id is supported by a counter control or not
   // -----------------------------------------------------------------
   virtual bool IsValidEvent(EventId id)
   {
      return ARRAY_INDEX_OF(bfd_do_events, id) != -1;
   }

   // -----------------------------------------------------------------
   // driver event receiver for buffered AO
   // -----------------------------------------------------------------
   virtual void OnKernEvent(EventId id)
   {
      // Discard the un-expected events
      if (IsIdle()) { return; }

      EventStatus    es   = {sizeof(es)};
      BfdAoEventArgs args = {id};

      switch(id)
      {
      default:  
         break;
      case EvtBufferedDoDataTransmitted:  /* data transmitted */
         m_dataTransedCleared = false;

         ModEvent().GetStatus(id, es);
         args.Count  = (es.General.lParam / m_scanPortCount) * m_scanPortCount;
         args.Offset = es.General.rParam;
         if (args.Count) { 
            UserEventManager::Fire(getOuter(), id, &args); 
         }

         // clear the event flag if the used hadn't clear it by calling 'SetData'.
         // using a count '0' to clear only the 'had sent' flag, but not effect the 'write pos'.
         if (!m_dataTransedCleared) { ModEvent().ClearFlag(id); }
         break;
      case EvtBufferedDoUnderrun:  /* under-run */
         ModEvent().GetStatus(id, es);
         args.Count  = es.General.lParam;
         args.Offset = es.General.rParam;
         UserEventManager::Fire(getOuter(), id, &args); 
         ModEvent().ClearFlag(id);            // clear the event flag after the user had processed it.
         break;
      case EvtBufferedDoCacheEmptied:  /* cache overflow */// no status data
         UserEventManager::Fire(getOuter(), id, &args); 
         ModEvent().ClearFlag(id);            // clear the event flag after the user had processed it.
         break;
      case EvtBufferedDoTransStopped: /*transition stopped*/// get the status of 'data transmitted' event to obtain the offset parameter
         ModEvent().GetStatus(EvtBufferedDoDataTransmitted, es);
         args.Count  = es.General.lParam;
         args.Offset = es.General.rParam;
         UserEventManager::Fire(getOuter(), id, &args); 
         // need not clear the event flag
         break;
      case EvtBufferedDoStopped:    /* stopped */// get the status of 'data transmitted' event to obtain the offset parameter
         ModEvent().GetStatus(EvtBufferedDoDataTransmitted, es);
         args.Count  = es.General.lParam;
         args.Offset = es.General.rParam;
         UserEventManager::Fire(getOuter(), id, &args); 
         // need not clear the event flag
         break;
      }
   };

private:
   void *   m_dataPtr;
   int32    m_dataBufLen;
   bool     m_dataTransedCleared; // whether the data ready event has been cleared in 'GetData', used by DataReadyHandler.

   int32    m_dataSize;         // for performance of 'GetData' method, cache this value, it will be initialized in 'InitializeControl'
   int32    m_scanPortCount;    // for performance of 'GetData' method, cache this value, it will be initialized in 'Prepare'
   int32    m_sectionCount;     //  

   ScanPortImpl         m_scanPort;
   ConvertClockImpl     m_convClock;
   TriggerImpl          m_trigger;
   TriggerImpl          m_trigger1;
};

#endif
