//
// DIO function implementation
//
#include <std_decl.h>

#include "dio_features.h"
#include "dio_port.h"
#include "scan_port.h"
#include "dio_trigger.h"
#include "dio_convert_clock.h"
#include "di_int_channel.h"
#include "di_cosint_port.h"
#include "di_pmint_port.h"
#include "dio_ctrl_base.h"

//------------------------------------------------------------------
// Help MACRO used to get/set property and check error
//------------------------------------------------------------------
#undef  MOD_PROP
#define MOD_PROP (m_owner->ModProp())

#define PROP_COND_SET_VAL(id, val)\
   {\
   if (m_owner->IsRunning()) return ErrorFuncBusy;\
   ErrorCode ret = m_owner->ModProp().Set(id, val);\
   if (!BioFailed(ret)) m_owner->OnPropChanged(id); \
   return ret; \
   }

// -----------------------------------------------------------------------
// functions of DioFeaturesImpl
//------------------------------------------------------------------------
// COMMON
bool DioFeaturesImpl::getPortProgrammable()
{
   uint8  xbuf[128] = {0};
   uint32 size = sizeof(xbuf);
   MOD_PROP.Read(CFG_FeaturePortsType, &size, xbuf);

   bool spted = false;
   for (unsigned i = 0; !spted && i < size; ++i) {
       spted = (xbuf[i] == Port8255A 
             || xbuf[i] == Port8255C 
             || xbuf[i] == PortIndvdlDio);
   }
   return spted;
}
int32 DioFeaturesImpl::getChannelCountMax()
{
   int32 count = 0;
   if (getPortProgrammable()) {
      if (MOD_PROP.Get(CFG_FeatureChannelNumberMax, count) == Success) {
         count += 1;
      }
      return count;
   }

   //
   // Fix bug: 14442,  5850: non-symmetry DI/DO,
   // We cannot query the DI/DO channel count directly from driver 
   // because the driver SPEC have not defined these two properties individually,
   // We have to calculate it according to the data mask here.
   // Here we cannot calculate the channel count from port count due to some one's
   // port owns less 8 channels.
   uint8  xbuf[128] = {0};
   uint32 size = sizeof(xbuf);
   PropertyId  id = m_owner->WantedPort() == Input ? CFG_DiDataMaskOfPorts : CFG_DoDataMaskOfPorts;

   MOD_PROP.Read(id, &size, xbuf);
   for (unsigned i = 0; i < size; ++i) {
      for (unsigned mask = xbuf[i]; mask; mask >>= 1) { 
         count += mask & 0x1; 
      }
   }

   return count;
}
int32 DioFeaturesImpl::getPortCount()
{
   int32 count = 0;

   //
   // Fix bug: 8654,  PCM-27D24DI: non-symmetry DI/DO,
   // We cannot query the DI/DO port count directly from driver 
   // because the driver SPEC have not defined these two properties individually,
   // We have to calculate it according to the port type here.
   uint8  xbuf[128] = {0};
   uint32 size = sizeof(xbuf);
   uint8  ignored = m_owner->WantedPort() == Input ? PortDo : PortDi;

   MOD_PROP.Read(CFG_FeaturePortsType, &size, xbuf);
   for (unsigned i = 0; i < size; ++i) {
      if (xbuf[i] != ignored) { 
         ++count; 
      }
   }

   return count;
}
ArrayImpl<uint8> * DioFeaturesImpl::getPortsType()
{
   return MOD_PROP.GetValueArrayI8(CFG_FeaturePortsType);
}
bool DioFeaturesImpl::getDiSupported()
{
   uint8  xbuf[128] = {0};
   uint32 size = sizeof(xbuf);

   MOD_PROP.Read(CFG_FeaturePortsType, &size, xbuf);
   return size && xbuf[0] != PortDo;
}
bool DioFeaturesImpl::getDoSupported()
{
   uint8  xbuf[128] = {0};
   uint32 size = sizeof(xbuf);

   MOD_PROP.Read(CFG_FeaturePortsType, &size, xbuf);
   return size && xbuf[0] != PortDi;
}

//Buffered DI / Buffered DO 
bool DioFeaturesImpl::getBufferedDiSupported()
{
   return MOD_PROP.GetLength(CFG_FeatureDiConvertClockSources) != 0;
}

//buffered do->basic feature
bool DioFeaturesImpl::getBufferedDoSupported()
{
   return MOD_PROP.GetLength(CFG_FeatureDoConvertClockSources) != 0;
}

int32 DioFeaturesImpl::getDataSize()
{
   int32 size = 0;
   MOD_PROP.Get(CFG_FeatureDataSize, size);
   return size;
}

int32 DioFeaturesImpl::getResolution()
{
   int32 size = 0;
   MOD_PROP.Get(CFG_FeatureResolutionInBit, size);
   return size;
}

//------------------------------------------------------------------------
// DI
ArrayImpl<uint8> * DioFeaturesImpl::getDiDataMask()
{
   return MOD_PROP.GetValueArrayI8(CFG_DiDataMaskOfPorts);
}

// DI noise filter features
bool DioFeaturesImpl::getDiNoiseFilterSupported()
{
   return MOD_PROP.GetLength(CFG_FeatureNoiseFilterOfChannels) != 0;
}
ArrayImpl<uint8>* DioFeaturesImpl::getDiNoiseFilterOfChannels()
{
   return MOD_PROP.GetValueArrayI8(CFG_FeatureNoiseFilterOfChannels);
}
void DioFeaturesImpl::getDiNoiseFilterBlockTimeRange(MathInterval &x)
{
   if (MOD_PROP.Get(CFG_FeatureNoiseFilterBlockTimeRange, x) != Success) {
      ResetMathInterval(x);
   }
}

// DI interrupt features
bool DioFeaturesImpl::getDiintSupported()
{
   return MOD_PROP.GetLength(CFG_FeatureDiintOfChannels) != 0;
}
bool DioFeaturesImpl::getDiintGateSupported()
{
   return MOD_PROP.GetLength(CFG_FeatureDiintGateOfChannels) != 0;
}
bool DioFeaturesImpl::getDiCosintSupported()
{
   return MOD_PROP.GetLength(CFG_FeatureDiCosintOfChannels) != 0;
}
bool DioFeaturesImpl::getDiPmintSupported()
{
   return MOD_PROP.GetLength(CFG_FeatureDiPmintOfChannels) != 0;
}
ArrayImpl<ActiveSignal>* DioFeaturesImpl::getDiintTriggerEdges()
{
   return MOD_PROP.GetValueArrayI8<ActiveSignal>(CFG_FeatureDiintTriggerEdges);
}
ArrayImpl<uint8>* DioFeaturesImpl::getDiintOfChannels()
{
   return MOD_PROP.GetValueArrayI8(CFG_FeatureDiintOfChannels);
}
ArrayImpl<uint8>* DioFeaturesImpl::getDiintGateOfChannels()
{
   return MOD_PROP.GetValueArrayI8(CFG_FeatureDiintGateOfChannels);
}
ArrayImpl<uint8>* DioFeaturesImpl::getDiCosintOfPorts()
{
   ArrayImpl<uint8> * ports = NULL;
   uint8  xbuf[128] = {0};
   uint32 size = sizeof(xbuf);

   MOD_PROP.Read(CFG_FeatureDiCosintOfChannels, &size, xbuf);
   if (size) {
      // convert the information from per channel to per ports and add into the collection
      ports = new ArrayImpl<uint8>();
      ports->setCapacity((size/sizeof(uint8)) / 8);
      for (unsigned i = 0; i < (size/sizeof(uint8)) / 8; ++i) {
         uint8 value = 0;
         for (int j = 0; j < 8; ++j) { 
            value |= xbuf[i * 8 + j] << j; 
         }
         ports->Add((uint8)value);
      }
   }
   return ports;
}
ArrayImpl<uint8>* DioFeaturesImpl::getDiPmintOfPorts()
{
   ArrayImpl<uint8> * ports = NULL;
   uint8  xbuf[128] = {0};
   uint32 size = sizeof(xbuf);

   MOD_PROP.Read(CFG_FeatureDiPmintOfChannels, &size, xbuf);
   if (size) {
      // convert the information from per channel to per ports and add into the collection
      ports = new ArrayImpl<uint8>();
      ports->setCapacity((size/sizeof(uint8)) / 8);
      for (unsigned i = 0; i < (size/sizeof(uint8)) / 8; ++i) {
         uint8 value = 0;
         for (int j = 0; j < 8; ++j) { 
            value |= xbuf[i * 8 + j] << j; 
         }
         ports->Add((uint8)value);
      }
   }
   return ports;
}
ArrayImpl<int32>* DioFeaturesImpl::getDiSnapEventSources()
{
   return MOD_PROP.GetValueArrayI32<int32>(CFG_FeatureDiSnapEventSources);
}

//-----------------------------------------------------------------------------------
// DO
ArrayImpl<uint8> * DioFeaturesImpl::getDoDataMask()
{
   return MOD_PROP.GetValueArrayI8(CFG_DoDataMaskOfPorts);
}

// DO freeze features
ArrayImpl<SignalDrop> * DioFeaturesImpl::getDoFreezeSignalSources()
{
   return MOD_PROP.GetValueArrayI32<SignalDrop>(CFG_FeatureDoFreezeSignalSources);
}

// DO reflect watchdog features
void DioFeaturesImpl::getDoReflectWdtFeedIntervalRange(MathInterval &x)
{
   if (MOD_PROP.Get(CFG_FeatureDoReflectWdtFeedIntervalRange, x) != Success) {
      ResetMathInterval(x);
   }
}

// DO preset value depository
Depository DioFeaturesImpl::getDoPresetValueDepository()
{
   int32 dep = 0;
   MOD_PROP.Get(CFG_FeatureDoInitialStateDepository, dep);
   return Depository(dep);
}

// DO circuit selectable types
ArrayImpl<DoCircuitType>* DioFeaturesImpl::getDoCircuitSelectableTypes()
{
   return MOD_PROP.GetValueArrayI8<DoCircuitType>(CFG_FeatureDoCircuitSelectableTypes);
}

// For compatible
ArrayImpl<uint8>* DioFeaturesImpl::getDataMask()
{
   return m_owner->WantedPort() == Input ? getDiDataMask() : getDoDataMask();
}

//------------------------------------------------------------------
// function of : DioPortImpl
//------------------------------------------------------------------
int32 DioPortImpl::getPort()
{
   return (int32)m_tag;
}
DioPortDir DioPortImpl::getDirection()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   uint8 value = m_wanted;
   MOD_PROP.GetValueNth(CFG_DirectionOfPorts, m_tag, value);

   if (!m_owner->getFeatures()->getPortProgrammable()) {
      PropertyId id = m_wanted == Input ? CFG_DiDataMaskOfPorts : CFG_DoDataMaskOfPorts;
      uint8    mask = 0;
      MOD_PROP.GetValueNth(id, m_tag, mask);
      value &= mask;
   }

   return DioPortDir(value);
}
ErrorCode DioPortImpl::setDirection(DioPortDir value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }
   if (!m_owner->getFeatures()->getPortProgrammable()) { return ErrorPropReadOnly; }

   return MOD_PROP.SetValueNth(CFG_DirectionOfPorts, m_tag, (uint8)value);
}
// DI port property
uint8 DioPortImpl::getDiInversePort()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   if (m_wanted != Input) { return 0; }

   uint8 value = 0;
   MOD_PROP.GetValueNth(CFG_DiInversePorts, m_tag, value);
   return value;
}
ErrorCode DioPortImpl::setDiInversePort(uint8 value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   if (m_wanted != Input) { return ErrorFuncNotSpted;  }

   return MOD_PROP.SetValueNth(CFG_DiInversePorts, m_tag, (uint8)value);
}
uint8 DioPortImpl::getDiOpenState()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   if (m_wanted != Input) { return 0; }

   uint8 value = 0;
   MOD_PROP.GetValueNth(CFG_DiOpenStatesOfPorts, m_tag, value);
   return value;
}
ErrorCode DioPortImpl::setDiOpenState(uint8 value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   if (m_wanted != Input) { return ErrorFuncNotSpted;  }

   return MOD_PROP.SetValueNth(CFG_DiOpenStatesOfPorts, m_tag, (uint8)value);
}
// DO port property
uint8 DioPortImpl::getDoPresetValue()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   uint8 value = 0;
   MOD_PROP.GetValueNth(CFG_DoInitialStateOfPorts, m_tag, value);
   return value;
}
ErrorCode DioPortImpl::setDoPresetValue(uint8 value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   return MOD_PROP.SetValueNth(CFG_DoInitialStateOfPorts, m_tag, (uint8)value);
}
DoCircuitType DioPortImpl::getDoCircuitType()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   uint8 value = 0;
   MOD_PROP.GetValueNth(CFG_DoCircuitTypeOfPorts, m_tag, value);
   return DoCircuitType(value);
}
ErrorCode DioPortImpl::setDoCircuitType(DoCircuitType value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   return MOD_PROP.SetValueNth(CFG_DoCircuitTypeOfPorts, m_tag, (uint8)value);
}

//------------------------------------------------------------------
// function of : DiintChannelImpl
//------------------------------------------------------------------
int32 DiintChannelImpl::getChannel()
{
   return (int32)m_tag;
}
bool DiintChannelImpl::getEnabled()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT * 8);

   return m_intEnabled != 0;
}
ErrorCode DiintChannelImpl::setEnabled(bool value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT * 8);
   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }

   m_intEnabled = value;
   return Success;
}
bool DiintChannelImpl::getGated()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT * 8);

   uint8 value = 0;
   MOD_PROP.GetValueNth(CFG_DiintGateEnabledChannels, m_tag, value);
   return value != 0;
}
ErrorCode DiintChannelImpl::setGated(bool value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT * 8);
   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }

   return MOD_PROP.SetValueNth(CFG_DiintGateEnabledChannels, m_tag, (uint8)value);
}
ActiveSignal DiintChannelImpl::getTrigEdge()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT * 8);

   uint8 value = 0;
   MOD_PROP.GetValueNth(CFG_DiintTriggerEdgeOfChannels, m_tag, value);
   return ActiveSignal(value);
}
ErrorCode DiintChannelImpl::setTrigEdge(ActiveSignal value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT * 8);
   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }

   return MOD_PROP.SetValueNth(CFG_DiintTriggerEdgeOfChannels, m_tag, (uint8)value);
}
//------------------------------------------------------------------
// function of : DiCosintPortImpl
//------------------------------------------------------------------
int32 DiCosintPortImpl::getPort()
{
   return (int32)m_tag;
}
uint8 DiCosintPortImpl::getMask()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   uint8 xbuf[MAX_DIO_PORT_COUNT * 8] = {0};
   MOD_PROP.Get(CFG_DiCosintEnabledChannels, ARRAY_SIZE(xbuf), xbuf);

   uint8 enabledCHs = 0;
   for (int i = 0; i < 8; ++i) {
      if (xbuf[m_tag * 8 + i] != 0) {
         enabledCHs |= 0x1 << i;
      }
   }
   return enabledCHs;
}
ErrorCode DiCosintPortImpl::setMask(uint8 value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);
   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }

   uint8 xbuf[MAX_DIO_PORT_COUNT * 8];
   MOD_PROP.Get(CFG_DiCosintEnabledChannels, ARRAY_SIZE(xbuf), xbuf);

   for (int i = 0; i < 8; ++i) {
      xbuf[m_tag * 8 + i] = (value >> i) & 0x1;
   }

   // Mask out the DO channels
   ArrayImpl<uint8> *port_mask = m_owner->getFeatures()->getDiDataMask();
   xbuf[m_tag] &= port_mask->Item(m_tag);
   port_mask->Dispose();

   return MOD_PROP.Set(CFG_DiCosintEnabledChannels, ARRAY_SIZE(xbuf), xbuf);
}
//------------------------------------------------------------------
// function of : DiPmintPortImpl
//------------------------------------------------------------------
int32 DiPmintPortImpl::getPort()
{
   return (int32)m_tag;
}
uint8 DiPmintPortImpl::getMask()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   uint8 xbuf[MAX_DIO_PORT_COUNT * 8] = {0};
   MOD_PROP.Get(CFG_DiPmintEnabledChannels, ARRAY_SIZE(xbuf), xbuf);

   uint8 enabledCHs = 0;
   for (int i = 0; i < 8; ++i) {
      if (xbuf[m_tag * 8 + i] != 0) {
         enabledCHs |= 0x1 << i;
      }
   }
   return enabledCHs;
}
ErrorCode DiPmintPortImpl::setMask(uint8 value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);
   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }

   uint8 xbuf[MAX_DIO_PORT_COUNT * 8];
   MOD_PROP.Get(CFG_DiPmintEnabledChannels, ARRAY_SIZE(xbuf), xbuf);

   // Mask out the DO channels
   ArrayImpl<uint8> *port_mask = m_owner->getFeatures()->getDiDataMask();
   uint8 di_mask = port_mask->Item(m_tag);
   port_mask->Dispose();

   for (int i = 0; i < 8; ++i) {
      uint8 ch_mask = 1 << i;
      xbuf[m_tag * 8 + i] = (value & ch_mask) && (di_mask & ch_mask);
   }

   return MOD_PROP.Set(CFG_DiPmintEnabledChannels, ARRAY_SIZE(xbuf), xbuf);
}
uint8 DiPmintPortImpl::getPattern()
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);

   uint8 value = 0;
   MOD_PROP.GetValueNth(CFG_DiPmintValueOfPorts, m_tag, value);
   return value;
}
ErrorCode DiPmintPortImpl::setPattern(uint8 value)
{
   assert((unsigned)m_tag < MAX_DIO_PORT_COUNT);
   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }

   return MOD_PROP.SetValueNth(CFG_DiPmintValueOfPorts, m_tag, (uint8)value);
}

// Buffered Di->trigger features
bool DioFeaturesImpl::getDiRetriggerable()
{
   int32 retriggerable = 0;
   MOD_PROP.Get(CFG_FeatureRetriggerable, retriggerable);
   return retriggerable != 0;
}
int32 DioFeaturesImpl::getDiTriggerCount()
{
   int32 count = 0;
   MOD_PROP.Get(CFG_FeatureDiTriggersCount, count);
   return count;
}

ArrayImpl<TriggerAction> * DioFeaturesImpl::getDiTriggerActions(int32 trigger)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_di_trigAction[trigger] : PropertyId(-1);
   return MOD_PROP.GetValueArrayI32<TriggerAction>(id);
}

ArrayImpl<SignalDrop> * DioFeaturesImpl::getDiTriggerSources(int32 trigger)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_di_trigSrc[trigger] : PropertyId(-1);
   return MOD_PROP.GetValueArrayI32<SignalDrop>(id);
}

void DioFeaturesImpl::getDiTriggerDelayRange(int32 trigger, MathInterval &x)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_di_trigDlyRng[trigger] : PropertyId(-1);
   if (MOD_PROP.Get(id, x) != Success) { ResetMathInterval(x); }
}

ValueRange DioFeaturesImpl::getDiTriggerSourceVrg(int32 trigger)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_trigSrcVrg[trigger] : PropertyId(-1);

   int32 vrg = 0;
   MOD_PROP.Get(id, vrg);
   return ValueRange(vrg);
}

double DioFeaturesImpl::getDiTriggerHysteresisIndexMax(int32 trigger)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_di_trigHystIdxMax[trigger] : PropertyId(-1);

   double indexMax = 0;
   MOD_PROP.Get(id, indexMax);
   return indexMax;
}
int32 DioFeaturesImpl::getDiTriggerHysteresisIndexStep(int32 trigger)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_di_trigHystStep[trigger] : PropertyId(-1);

   int32 step = 0;
   MOD_PROP.Get(id, step);
   return step;
}

//Buffered DI Scan Method 
SamplingMethod DioFeaturesImpl::getDiSamplingMethod()
{
   int32 ret = (int32)EqualTimeSwitch;
   MOD_PROP.Get(CFG_FeatureDiPortScanMethod, ret);
   return (SamplingMethod)ret;
}

// buffered di->conversion clock features
ArrayImpl<SignalDrop> * DioFeaturesImpl::getDiConvertClockSources()
{
   return MOD_PROP.GetValueArrayI32<SignalDrop>(CFG_FeatureDiConvertClockSources);
}
void DioFeaturesImpl::getDiConvertClockRange(MathInterval &x)
{
   if (MOD_PROP.Get(CFG_FeatureDiConvertClockRateRange, x) != Success) {
      ResetMathInterval(x);
   }
}

//Buffered DO Scan Method 
SamplingMethod DioFeaturesImpl::getDoSamplingMethod()
{
   int32 ret = (int32)EqualTimeSwitch;
   MOD_PROP.Get(CFG_FeatureDoPortScanMethod, ret);
   return (SamplingMethod)ret;
}

// buffered do->conversion clock features
ArrayImpl<SignalDrop> * DioFeaturesImpl::getDoConvertClockSources()
{
   return MOD_PROP.GetValueArrayI32<SignalDrop>(CFG_FeatureDoConvertClockSources);
}
void DioFeaturesImpl::getDoConvertClockRange(MathInterval &x)
{
   if (MOD_PROP.Get(CFG_FeatureDoConvertClockRateRange, x) != Success) {
      ResetMathInterval(x);
   }
}

// Buffered Do->trigger features
bool DioFeaturesImpl::getDoRetriggerable()
{
   int32 retriggerable = 0;
   MOD_PROP.Get(CFG_FeatureRetriggerable, retriggerable);
   return retriggerable != 0;
}
int32 DioFeaturesImpl::getDoTriggerCount()
{
   int32 count = 0;
   MOD_PROP.Get(CFG_FeatureDoTriggersCount, count);
   return count;
}

ArrayImpl<TriggerAction> * DioFeaturesImpl::getDoTriggerActions(int32 trigger)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_do_trigAction[trigger] : PropertyId(-1);
   return MOD_PROP.GetValueArrayI32<TriggerAction>(id);
}

ArrayImpl<SignalDrop> * DioFeaturesImpl::getDoTriggerSources(int32 trigger)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_do_trigSrc[trigger] : PropertyId(-1);
   return MOD_PROP.GetValueArrayI32<SignalDrop>(id);
}

void DioFeaturesImpl::getDoTriggerDelayRange(int32 trigger, MathInterval &x)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_do_trigDlyRng[trigger] : PropertyId(-1);
   if (MOD_PROP.Get(id, x) != Success) { ResetMathInterval(x); }
}

ValueRange DioFeaturesImpl::getDoTriggerSourceVrg(int32 trigger)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_trigSrcVrg[trigger] : PropertyId(-1);

   int32 vrg = 0;
   MOD_PROP.Get(id, vrg);
   return ValueRange(vrg);
}

double DioFeaturesImpl::getDoTriggerHysteresisIndexMax(int32 trigger)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_do_trigHystIdxMax[trigger] : PropertyId(-1);

   double indexMax = 0;
   MOD_PROP.Get(id, indexMax);
   return indexMax;
}
int32 DioFeaturesImpl::getDoTriggerHysteresisIndexStep(int32 trigger)
{
   PropertyId id = trigger < MAX_DIO_TRIG_COUNT ? feat_id_do_trigHystStep[trigger] : PropertyId(-1);

   int32 step = 0;
   MOD_PROP.Get(id, step);
   return step;
}

//------------------------------------------------------------------
// function of : ScanPortImpl
//------------------------------------------------------------------
int32 ScanPortImpl::getPortStart()
{
   int32 port = 0;
   DioPortDir dir = m_owner->WantedPort();

   if (dir == Input){
      MOD_PROP.Get(CFG_DiScanPortStart, port);        
   } else {
      MOD_PROP.Get(CFG_DoScanPortStart, port);        
   }

   return port;
}

ErrorCode ScanPortImpl::setPortStart(int32 value)
{
   DioPortDir dir = m_owner->WantedPort();

   if (dir == Input){
      PROP_COND_SET_VAL(CFG_DiScanPortStart, value);
   } else {
      PROP_COND_SET_VAL(CFG_DoScanPortStart, value);   
   }

   return Success;
}
int32 ScanPortImpl::getPortCount()
{
   int32 count = 0;
   DioPortDir dir = m_owner->WantedPort();

   if (dir == Input){
      MOD_PROP.Get(CFG_DiScanPortCount, count);       
   } else {
      MOD_PROP.Get(CFG_DoScanPortCount, count);     
   }
   return count;
}

ErrorCode ScanPortImpl::setPortCount(int32 value)
{
   DioPortDir dir = m_owner->WantedPort();

   if (dir == Input){
      PROP_COND_SET_VAL(CFG_DiScanPortCount, value);
   } else {
      PROP_COND_SET_VAL(CFG_DoScanPortCount, value);
   }

   return Success;
}

void ScanPortImpl::getPortMap(int32 count, int8 *data)
{
   DioPortDir dir = m_owner->WantedPort(); 
   if (dir == Input){
      MOD_PROP.Read(CFG_DiScanEnabledPort, count, data);
   } else {
      MOD_PROP.Read(CFG_DoScanEnabledPort, count, data);
   }
}

ErrorCode ScanPortImpl::setPortMap(int32 count, int8 *data)
{
   if (!m_owner->CanEditProperty()) { return ErrorPrivilegeNotHeld; }
   if (m_owner->IsRunning()) { return ErrorFuncBusy; }
   
   DioPortDir dir = m_owner->WantedPort(); 
   if (dir == Input){
      return MOD_PROP.Set(CFG_DiScanEnabledPort, count, (uint8 *)data);
   } else {         
      return MOD_PROP.Set(CFG_DoScanEnabledPort, count, (uint8 *)data);
   }

   return Success;
}

int32 ScanPortImpl::getSectionLength()
{
   int32 length = 0;
   DioPortDir dir = m_owner->WantedPort();

   if (dir == Input){
      MOD_PROP.Get(CFG_DiPortSectionLength, length);       
   } else {
      MOD_PROP.Get(CFG_DoPortSectionLength, length);     
   }

   return length; 
}

ErrorCode ScanPortImpl::setSectionLength(int32 value)
{
   DioPortDir dir = m_owner->WantedPort();

   if (dir == Input){
      PROP_COND_SET_VAL(CFG_DiPortSectionLength, value);  
   } else {
      PROP_COND_SET_VAL(CFG_DoPortSectionLength, value);  
   }

   return Success;
}

int32 ScanPortImpl::getSectionCount()
{
   int32 count;
   DioPortDir dir = m_owner->WantedPort();

   if (dir == Input){
      MOD_PROP.Get(CFG_DiPortSectionCount, count);       
   } else {
      MOD_PROP.Get(CFG_DoPortSectionCount, count);     
   }
   return count;
}

ErrorCode ScanPortImpl::setSectionCount(int32 value)
{
   DioPortDir dir = m_owner->WantedPort();

   if (dir == Input){
      PROP_COND_SET_VAL(CFG_DiPortSectionCount, value);
   } else {             
      PROP_COND_SET_VAL(CFG_DoPortSectionCount, value);   
   }

   return Success;
}

int32 ScanPortImpl::getCycles()
{
   int32 count = 1;
   DioPortDir dir = m_owner->WantedPort();
   if (dir == Input){
      MOD_PROP.Get(CFG_DiPortCycles, count);      
   } else {
      MOD_PROP.Get(CFG_DoPortCycles, count);      
   }
   
   return count;
}
ErrorCode ScanPortImpl::setCycles(int32 value)
{
   DioPortDir dir = m_owner->WantedPort();

   if (dir == Input){
      PROP_COND_SET_VAL(CFG_DiPortCycles, value);
   } else {
      PROP_COND_SET_VAL(CFG_DoPortCycles, value);
   }
}

