#ifndef _DAQNAVI_INSTANT_DI_CTRL_IMPL_H
#define _DAQNAVI_INSTANT_DI_CTRL_IMPL_H


class InstantDiCtrlImpl : public DioCtrlBaseImpl
{
   // -----------------------------------------------------------------
   // nested types : user event info for DI snap event
   // -----------------------------------------------------------------
   struct DiEventInfo {
      int32   Src;
      uint8 * Buffer;
   };
   typedef std::map<EventId, DiEventInfo> DiEventMap;

public:
   //
   // event:
   virtual void addEventHandler(EventId id, DaqEventProc proc, void * param)
   {
      if (!IsValidEvent(id)) { 
         // Not our event, let 'base' to do default process
         DioCtrlBaseImpl::addEventHandler(id, proc, param);
         return; 
      }

      UserEventManager::Add(getOuter(), id, (UserEventProc)proc, param);
      // Defer registering kernel event till the user start snap.
   }

   virtual void removeEventHandler(EventId id, DaqEventProc proc, void * param)
   {
      if (!IsValidEvent(id)) { 
         // Not our event, let 'base' to do default process
         DioCtrlBaseImpl::removeEventHandler(id, proc, param);
         return; 
      }

      UserEventManager::Remove(getOuter(), id, (UserEventProc)proc, param);
      // Defer unregistering kernel event till the user stop snap.
   }

   // method
   ErrorCode ReadAny(int32 portStart, int32 portCount, uint8 data[])
   {
      return ModuleRef().DiReadPort(portStart, portCount, data);
   }

   ErrorCode ReadBit(int32 port, int32 bit, uint8 *data)
   {
      ErrorCode warning = Success;
      if (uint32(bit) > 7) {
         bit &= 7; 
         warning = WarningParamOutOfRange;
      }

      ErrorCode ret = ModuleRef().DiReadBit(port, bit, *data);
      return ret != Success ? ret : warning;
   }

   ErrorCode SnapStart()
   {
      if (  !getFeatures()->getDiintSupported() 
         && !getFeatures()->getDiCosintSupported() 
         && !getFeatures()->getDiPmintSupported()) {
         return ErrorFuncNotSpted;
      }

      //
      // don't firing event before we returned
      m_canFire = false;

      // start each enabled DI event
      ErrorCode ret = Success;
      ErrorCode err;

      // Start DiInt
      if (getFeatures()->getDiintSupported()) {
         ArrayImpl<DiintChannelImpl> * chs = getDiintChannels();
         for (int i = 0; i < chs->getLength(); ++i) {
            DiintChannelImpl & x = chs->Item(i);
            if (x.getEnabled()) {
               err = EnableKernEvent(EvtDiintChannel000 + x.getChannel(), x.getChannel());
               if (err != Success) {
                  ret = err;
               }
            }
         }
      }

      // Start CosInt
      if (getFeatures()->getDiCosintSupported()) {
         ArrayImpl<DiCosintPortImpl> *ports = getDiCosintPorts();
         for (int i = 0; i < ports->getLength(); ++i) {
            DiCosintPortImpl &x = ports->Item(i);
            if (x.getMask() != 0) {
               err = EnableKernEvent(EvtDiCosintPort000 + x.getPort(), x.getPort());
               if (err != Success) {
                  ret = err;
               }
            }
         }
      }

      // Start PmInt
      if (getFeatures()->getDiPmintSupported()) {
		   ArrayImpl<DiPmintPortImpl> *ports = getDiPmintPorts();
         for (int i = 0; i < ports->getLength(); ++i) {
            DiPmintPortImpl &x = ports->Item(i);
            if (x.getMask()) {
               err = EnableKernEvent(EvtDiPmintPort000 + x.getPort(), x.getPort());
               if (err != Success) {
                  ret = err;
               }
            }
         }
      }

      //
      // Enable firing event now
      m_canFire = true;

      return ret;
   }

   ErrorCode SnapStop()
   {
      ErrorCode ret = Success;
      ErrorCode err;

      //Masked for bug: 20064, imported when fix bug 19522
      //Dead-lock when snap stop while the event callback is executing.
      //AutoLock<MutexLock> lock(EventLock());

      // disable firing event to user
      m_canFire = false;

      // stop all 
      DiEventMap::iterator itor = m_events.begin();
      for (; itor != m_events.end(); ++itor) {
         UnregisterKernEvent(itor->first);
         err = ModuleRef().DiSnapStop(itor->first);
         if (err != Success) {
            ret = err;
         }
      }

      m_events.clear();

      return ret;
   }

   // property
   double getNoiseFilterBlockTime()
   {
      double val;
      ModProp().Get(CFG_NoiseFilterOverallBlockTime, val);
      return val;
   }
   ErrorCode setNoiseFilterBlockTime(double value)
   {
      return ModProp().Set(CFG_NoiseFilterOverallBlockTime, value);
   }

   ArrayImpl<NosFltChannelImpl> * getNoiseFilter()
   {
      if (!getInitialized() || !getFeatures()->getDiNoiseFilterSupported()) {
         return NULL;
      }

      AutoLock<MutexLock> lock(PropLock());
      if (m_noiseFilterCh.getLength() == 0) {
         uint8 nosFlt[MAX_DIO_PORT_COUNT * 8] = {0};
         ModProp().Get(CFG_FeatureNoiseFilterOfChannels, ARRAY_SIZE(nosFlt), nosFlt);

         for (int i = 0; i < m_portCount * 8; ++i) {
            if (nosFlt[i] != 0){
               m_noiseFilterCh.Add(NosFltChannelImpl(this, i));
            }
         }
      }

      return &m_noiseFilterCh;
   }

   ArrayImpl<DiintChannelImpl> * getDiintChannels()
   {
      if (!getInitialized() || !getFeatures()->getDiintSupported()) {
         return NULL;
      }

      AutoLock<MutexLock> lock(PropLock());
      if (m_diintChannels.getLength() == 0) {
         uint8 chcap[MAX_DIO_PORT_COUNT * 8] = {0};
         ModProp().Get(CFG_FeatureDiintOfChannels, ARRAY_SIZE(chcap), chcap);

         for (int i = 0; i < m_portCount * 8; ++i) {
            if (chcap[i] != 0){
               m_diintChannels.Add(DiintChannelImpl(this, i));
            }
         }
      }

      return &m_diintChannels;
   }

   ArrayImpl<DiCosintPortImpl>* getDiCosintPorts()
   {
      if (!getInitialized() || !getFeatures()->getDiCosintSupported()) {
         return NULL;
      }

      AutoLock<MutexLock> lock(PropLock());
      if (m_cosintPorts.getLength() == 0) {
         uint8 chcap[MAX_DIO_PORT_COUNT * 8] = {0};
         ModProp().Get(CFG_FeatureDiCosintOfChannels, ARRAY_SIZE(chcap), chcap);

         for (int i = 0; i < m_portCount; ++i) {
            for (int j = 0; j < 8; ++j) {
               if (chcap[i * 8 + j] != 0) {
                  m_cosintPorts.Add(DiCosintPortImpl(this, i));
                  break;
               }
            }
         }
      }

      return &m_cosintPorts;
   }

   ArrayImpl<DiPmintPortImpl>* getDiPmintPorts()
   {
      if (!getInitialized() || !getFeatures()->getDiPmintSupported()) {
         return NULL;
      }

      AutoLock<MutexLock> lock(PropLock());
      if (m_pmintPorts.getLength() == 0) {
         uint8 chcap[MAX_DIO_PORT_COUNT * 8] = {0};
         ModProp().Get(CFG_FeatureDiPmintOfChannels, ARRAY_SIZE(chcap), chcap);

         for (int i = 0; i < m_portCount; ++i) {
            for (int j = 0; j < 8; ++j) {
               if (chcap[i * 8 + j] != 0) {
                  m_pmintPorts.Add(DiPmintPortImpl(this, i));
                  break;
               }
            }
         }
      }

      return &m_pmintPorts;
   }

public: // ctor & dtor
   InstantDiCtrlImpl() : m_noiseFilterCh(false), m_diintChannels(false), m_cosintPorts(false), m_pmintPorts(false)
   {
      m_canFire = false;
   }
   virtual ~InstantDiCtrlImpl()
   {
      // clear all user event delegate
      for (unsigned i = 0; i < ARRAY_SIZE(inst_di_events); ++i) {
         UserEventManager::Remove(getOuter(), inst_di_events[i]);
      }
   }

protected: 
   //------------------------------------------------------------------------
   // override the method in base class
   //------------------------------------------------------------------------
   virtual DioPortDir WantedPort() { return Input; }

   virtual bool IsSuitable(DaqModule & module)
   {
      if (DioCtrlBaseImpl::IsSuitable(module)) {
         uint8 portType[MAX_DIO_PORT_COUNT];
         uint32 size = sizeof(portType);

         DaqProperty(module).Read(CFG_FeaturePortsType, &size, portType);
         for(unsigned i = 0; i < size / sizeof(uint8); ++i) {
            if (portType[i] != PortDo) {
               return true;
            }
         }
      }
      return false;
   }

   // initialize/de-initialize the control 
   virtual ErrorCode InitializeControl(DeviceInformation const &x)
   {
      // let's the base do initialization at first.
      ErrorCode ret = DioCtrlBaseImpl::InitializeControl(x);
      if (ret != Success) {
         return ret;
      }
      if (!getFeatures()->getDiSupported()) {
         return ErrorFuncNotSpted;
      }

      m_portCount = (int)getPortCount();
      return Success;
   }

   virtual void DeinitializeControl()
   {
      // Disable DI snap
      SnapStop();

      // reset all configuration to null
      m_diintChannels.Clear();
      m_cosintPorts.Clear();
      m_pmintPorts.Clear();
      m_noiseFilterCh.Clear();

      DioCtrlBaseImpl::DeinitializeControl();
   }

   //
   // Check whether the event id is supported by a counter control or not
   virtual bool IsValidEvent(EventId id)
   {
      return ARRAY_INDEX_OF(inst_di_events, id) != -1;
   }

   //
   // Process the kernel event
   virtual void OnKernEvent(EventId id)
   {
      DiEventMap::iterator itor = m_events.find(id);
      if (itor != m_events.end()) {
         // call user handler
         if (m_canFire) {
            DiSnapEventArgs args;
            args.Id     = id;
            args.SrcNum = itor->second.Src;
            args.Length = m_portCount;
            memcpy(args.PortData, itor->second.Buffer, args.Length);

            // General Event ID, such as EvtDiInterrupt
            EventId genId = EventId(id - args.SrcNum);
            UserEventManager::Fire(getOuter(), genId, &args);
        	}

         // automatically clear the event flag after the user had processed it.
         ModEvent().ClearFlag(id);
      }
   }

   // -----------------------------------------------------------------------
   // helpers
   //------------------------------------------------------------------------
   ErrorCode EnableKernEvent(int eventId, int eventSrc)
   {
      EventId       id = (EventId)eventId;
      DiEventInfo info = {eventSrc, NULL};
      ErrorCode   ret;

      AutoLock<MutexLock> lock(PropLock());

      // the event had been in snapped event list
      DiEventMap::iterator itor = m_events.find(id);
      if (itor != m_events.end()) { return Success; }

      // try to start the snapping at first
      ret = ModuleRef().DiSnapStart(id, 0, m_portCount, &info.Buffer);
      if (ret != Success) { return ret; }

      // try to register the event into driver after we had started the snapping
      ret = RegisterKernEvent(id);
      if (ret != Success) { return ret; }

      // save event info into snapped event list
      m_events[id] = info;

      // clear the flag of un-expected event signaled during we are starting DI snap.
      ModEvent().ClearFlag(id);

      return ret;
   }

private:
   ArrayImpl<NosFltChannelImpl> m_noiseFilterCh;
   ArrayImpl<DiintChannelImpl>  m_diintChannels;
   ArrayImpl<DiCosintPortImpl>  m_cosintPorts;
   ArrayImpl<DiPmintPortImpl>   m_pmintPorts;

   DiEventMap  m_events;
   bool        m_canFire;
   int         m_portCount;
};

#endif
