#include <std_decl.h>
#include "daqnavi3_api.h"
//-------------------------------------------------------------------------
// Helper macros used to check the parameters/conditions
//-------------------------------------------------------------------------
#define CHK_HANDLE_TYPE(hdl, type)     CHK_EXPR(hdl && ((DaqModule*)hdl)->getType() == type, ErrorHandleNotValid)

ErrorCode DeviceAddrToSpace(DaqDevModule *device, uint32 startAddr, uint32 &space, uint32 &offset)
{
   union {
      uint64 i64[6];
      uint32 i32[6];
   } xbuf;

   uint32 size = sizeof(xbuf);
   DaqProperty(device).Read(CFG_DeviceBaseAddresses, &size, &xbuf);
   if (size == 0) { return ErrorFuncNotSpted; }

   // Note: 
   // The driver returned the device's addresses in two formats:
   //    1. 'int64' for memory-mapped IO address, or
   //    2. 'int32' for IO-mapped IO address and return only one address.
   // Here we had to determine which format the driver used.
   if (size > sizeof(uint32)) { 
      for (unsigned i = 0; i < size / sizeof(int64); ++i) {
         uint32 baseAddr = (uint32)xbuf.i64[i];
         if (baseAddr <= startAddr && startAddr < baseAddr + 4096) {
            space = i;
            offset = startAddr - baseAddr;
            return Success;
         }
      }
   } else { 
      for (unsigned i = 0; i < size / sizeof(uint32); ++i) {
         uint32 baseAddr = xbuf.i32[i];
         if (baseAddr <= startAddr && startAddr < baseAddr + 256) {
            space = i;
            offset = startAddr - baseAddr;
            return Success;
         }
      }
   }
   return ErrorParamOutOfRange;
}

// ---------------------------------------------------------------
// Global Helper
// ---------------------------------------------------------------
ErrorCode BDAQCALL AdxGetValueRangeInformation(
   IN           uint32        vrg,
   IN           uint32        sizeofDesc,
   OUT OPTIONAL wchar_t       *description,
   OUT OPTIONAL PMathInterval range,
   OUT OPTIONAL uint32        *unit)
{
   return QueryVrgInfoDB(vrg, unit, range, sizeofDesc, description);
}

ErrorCode BDAQCALL AdxGetSignalConnectionInformation(
   IN           uint32  connectorType,
   IN           uint32  descriptionSize,
   OUT OPTIONAL wchar_t *description,
   OUT OPTIONAL uint32  *position)
{
   if (description != NULL){
      EnumToString(L"SignalDrop", connectorType, descriptionSize / sizeof(wchar_t), description);
   }

   if (position != NULL){
      if (connectorType <= SigDiStatusChange ) {
         *position = 0;
      } else if (connectorType >= SigAmsiPin0 && connectorType <= SigAmsiPin19 ) {
         *position = 2;
      } else {
         *position = 1;
      }
   }

   return Success;
}

ErrorCode BDAQCALL AdxEnumToString(
   IN wchar_t const *enumTypeName,
   IN uint32        enumValue,
   IN uint32        enumStringLength,
   OUT wchar_t      *enumString)
{
   // For compatible with DN3
   if (enumTypeName && wcscmp(enumTypeName, L"SignalCountingType") == 0) {
      enumTypeName = L"CountingType";
   }
   return EnumToString(enumTypeName, enumValue, enumStringLength, enumString);
}

ErrorCode BDAQCALL AdxStringToEnum(
   IN wchar_t const *enumTypeName,
   IN wchar_t const *enumString,
   OUT uint32       *enumValue)
{
   // For compatible with DN3
   if (enumTypeName && wcscmp(enumTypeName, L"SignalCountingType") == 0) {
      enumTypeName = L"CountingType";
   }
   return StringToEnum(enumTypeName, enumString, enumValue);
}

double BDAQCALL AdxTranslateTemperatureScale(
   IN uint32 degreeType,
   IN double degreeCelsius )
{
   double temperature = degreeCelsius;
   switch (degreeType)
   {
   case Fahrenheit: temperature = (9 * degreeCelsius + 160) / 5;   break;
   case Rankine:    temperature = (degreeCelsius + 273.15) * 1.8;  break;
   case Kelvin:     temperature = degreeCelsius + 273.15;          break;
   case Celsius:
   default:
      break;
   }

   return temperature;
}

void BDAQCALL AdxUnloadIdleDrivers()
{
   XDevFactory::UnloadIdleDriver();
}

// ---------------------------------------------------------------
// Event Function
// ---------------------------------------------------------------
ErrorCode BDAQCALL AdxEventGetHandle(
   IN  HANDLE handle,
   IN  uint32 eventType,
   OUT HANDLE *eventHandle)
{
   CHK_HANDLE(handle);
   CHK_USER_BUF(eventHandle);

   return ((DaqModule*)handle)->EventGetHandle((EventId)eventType, *eventHandle);
}

ErrorCode BDAQCALL AdxEventGetLastStatus(
   IN           HANDLE handle,
   IN           uint32 eventType,
   OUT OPTIONAL uint32 *statusLParam,
   OUT OPTIONAL uint32 *statusRParam)
{
   CHK_HANDLE(handle);
   uint32 dummy;
   uint32 *lptr = statusLParam ? statusLParam : &dummy;
   uint32 *rptr = statusRParam ? statusRParam : &dummy;

   EventStatus es = {sizeof(es)};
   ErrorCode ret = ((DaqModule*)handle)->EventGetStatus((EventId)eventType, es);

   switch(eventType) 
   {
   case EvtBufferedAiDataReady:
      *lptr = es.AiDataReady.DataCount;
      *rptr = es.AiDataReady.DataOffset;
      break;
   case EvtBufferedAiStopped:
      *lptr = es.AiDataReady.Stopped;
      *rptr = es.AiDataReady.DataCount;
      break;
   default:
      *lptr = es.General.lParam;
      *rptr = es.General.rParam;
      break;
   }
   return ret;
}

ErrorCode BDAQCALL AdxEventClearFlag(
   IN HANDLE handle,
   IN uint32 eventType,
   IN uint32 flagLParam,
   IN uint32 flagRParam)
{
   CHK_HANDLE(handle);

   EventStatus flags;
   flags.Size = sizeof(flags);
   flags.General.lParam = flagLParam;
   flags.General.rParam = flagRParam;
   return ((DaqModule*)handle)->EventClearFlag((EventId)eventType, flags);
}

// ---------------------------------------------------------------
// Property Access Function
// ---------------------------------------------------------------
ErrorCode BDAQCALL AdxPropertyRead(
   IN           HANDLE handle,
   IN           uint32 propertyId,
   IN           uint32 bufferSize,
   OUT OPTIONAL void   *buffer,
   OUT OPTIONAL uint32 *dataLength,
   OUT OPTIONAL uint32 *attribute)
{
   CHK_HANDLE(handle);

   DaqProperty prop((DaqModule*)handle);

   ErrorCode  ret = prop.Read((PropertyId)propertyId, &bufferSize, buffer);
   if (dataLength) {
      *dataLength = ret == ErrorPropNotSpted ? 0 : bufferSize; 
   }

   if (attribute) {
      *attribute = prop.GetAttribute((PropertyId)propertyId);
   }

   return ret;
}

ErrorCode BDAQCALL AdxPropertyWrite(
   IN HANDLE handle,
   IN uint32 propertyId,
   IN uint32 dataLength,
   IN void   *buffer,
   IN uint32 notifyNow)
{
   CHK_HANDLE(handle);
   CHK_USER_BUF_LEN(buffer, dataLength);

#define DN3_UPDHW 0x100
   uint32 flag = (notifyNow & DN3_UPDHW) ? PropFlag::UpdHW : 0;
   if (notifyNow & (~DN3_UPDHW)) {
      flag |= PropFlag::NtfNow;
   }

   DaqProperty prop((DaqModule*)handle);
   return prop.Write((PropertyId)propertyId, dataLength, buffer, flag);
}

ErrorCode BDAQCALL AdxPropertyReadFromRegistry(
   IN           HANDLE handle,
   IN           uint32 propertyId,
   IN           uint32 bufferSize,
   OUT OPTIONAL void   *buffer,
   OUT OPTIONAL uint32 *dataLength)
{
   CHK_HANDLE(handle);

   DaqProperty prop((DaqModule*)handle);

   ErrorCode ret = prop.Read((PropertyId)propertyId, &bufferSize, buffer, PropFlag::OpGet | PropFlag::SysDB);
   if (dataLength) {
      *dataLength = ret == ErrorPropNotSpted ? 0 : bufferSize;
   }
   return ret;
}

ErrorCode BDAQCALL AdxPropertyWriteToRegistry(
   IN HANDLE handle,
   IN uint32 propertyId,
   IN uint32 dataLength,
   IN void   *buffer)
{
   CHK_HANDLE(handle);
   CHK_USER_BUF_LEN(buffer, dataLength);

   DaqProperty prop((DaqModule*)handle);

   return prop.Write((PropertyId)propertyId, dataLength, buffer, PropFlag::SysDB);
}

// ---------------------------------------------------------------
// Device Function
// ---------------------------------------------------------------
ErrorCode BDAQCALL AdxDeviceGetLinkageInfo(
   IN           int32   deviceParent,
   IN           uint32  index,
   OUT          int32   *deviceNumber,
   OUT OPTIONAL wchar_t *deviceDescription,
   OUT OPTIONAL uint32  *subDeviceCount)
{
   CHK_USER_BUF(deviceNumber);
   *deviceNumber = -1;
   if (deviceDescription != NULL){
      *deviceDescription = 0;
   }
   if (subDeviceCount != NULL){
      *subDeviceCount = 0;
   }
   int32 active;
   return XDevFactory::EnumDeviceInfo(index, active, *deviceNumber, deviceDescription);
}

ErrorCode BDAQCALL AdxDeviceOpen(
   IN int32   deviceNumber,
   IN uint32  accessMode,
   OUT HANDLE *deviceHandle)
{
   CHK_USER_BUF(deviceHandle);

   ErrorCode ret = XDevManager::OpenDevice(deviceNumber, (AccessMode)accessMode, *(DaqDevModule **)deviceHandle);
   if (ret == Success && accessMode == ModeWriteWithReset) {
      ((DaqDevModule *)*deviceHandle)->Reset(1);
   }
   return ret;
}

ErrorCode BDAQCALL AdxDeviceClose(
   IN HANDLE deviceHandle)
{
   CHK_HANDLE_TYPE(deviceHandle, DaqDevice);

   XDevManager::CloseDevice((DaqDevModule *)deviceHandle);
   return Success;
}

ErrorCode BDAQCALL AdxDeviceReset(
   IN HANDLE deviceHandle,
   IN uint32 state)
{
   CHK_HANDLE_TYPE(deviceHandle, DaqDevice);

   return ((DaqDevModule*)deviceHandle)->Reset(state);
}

ErrorCode BDAQCALL AdxDeviceGetModuleHandle(
   IN HANDLE  deviceHandle,
   IN uint32  type,
   IN uint32  index,
   OUT HANDLE *moduleHandle )
{
   CHK_HANDLE_TYPE(deviceHandle, DaqDevice);

   return ((DaqDevModule*)deviceHandle)->GetModule((ModuleType)type, index, *(DaqModule**)moduleHandle);
}

ErrorCode BDAQCALL AdxDeviceRefreshProperties(
   IN HANDLE deviceHandle)
{
   CHK_HANDLE_TYPE(deviceHandle, DaqDevice);

   return ((DaqDevModule*)deviceHandle)->Refresh();
}

ErrorCode BDAQCALL AdxDeviceShowConfigDialogBox(
   IN HANDLE  deviceHandle,
   IN HANDLE  parentWindow,
   IN void    *region,
   IN uint32  dataSource,
   IN int32   modal,
   OUT HANDLE *dialogWindow)
{
   CHK_HANDLE_TYPE(deviceHandle, DaqDevice);
#if defined(_WIN32) || defined(WIN32)
   DaqDevModule *device = (DaqDevModule*)deviceHandle;
   if (device->getIntfType() == IntfType::DN3_API) {
      DN3Api::DN3Device * dn3Device = (DN3Api::DN3Device *)device;
      return dn3Device->ShowConfigDialogBox((HWND)parentWindow, (LPRECT)region, dataSource, modal, (HWND *)dialogWindow);
   }
#endif
   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxDeviceReadPorts(
   IN  HANDLE deviceHandle,
   IN  uint32 startAddress,
   IN  uint32 length,
   OUT void   *data)
{
   CHK_HANDLE_TYPE(deviceHandle, DaqDevice);

   uint32 space, offset;
   ErrorCode ret = DeviceAddrToSpace((DaqDevModule*)deviceHandle, startAddress, space, offset);
   return ret != Success ? ret : ((DaqDevModule*)deviceHandle)->ReadRegister(space, offset, length, data);
}

ErrorCode BDAQCALL AdxDeviceWritePorts(
   IN HANDLE deviceHandle,
   IN uint32 startAddress,
   IN uint32 length,
   IN void   *data)
{
   CHK_HANDLE_TYPE(deviceHandle, DaqDevice);

   uint32 space, offset;
   ErrorCode ret = DeviceAddrToSpace((DaqDevModule*)deviceHandle, startAddress, space, offset);
   return ret != Success ? ret : ((DaqDevModule*)deviceHandle)->WriteRegister(space, offset, length, data);
}

ErrorCode BDAQCALL AdxDevicePrivateRegionRead(
   IN  HANDLE deviceHandle,
   IN  uint32 signature,
   IN  uint32 dataLen,
   OUT uint8  *dataBuf)
{
   CHK_HANDLE_TYPE(deviceHandle, DaqDevice);

   return ((DaqDevModule*)deviceHandle)->ReadPrivRgn(signature, dataLen, dataBuf);
}

ErrorCode BDAQCALL AdxDevicePrivateRegionWrite(
   IN HANDLE deviceHandle,
   IN uint32 signature,
   IN uint32 dataLen,
   IN uint8  *dataBuf)
{
   CHK_HANDLE_TYPE(deviceHandle, DaqDevice);

   return ((DaqDevModule*)deviceHandle)->WritePrivRgn(signature, dataLen, dataBuf);
}

// ---------------------------------------------------------------
// AI Function
// ---------------------------------------------------------------
ErrorCode BDAQCALL AdxAiReadSamples(
   IN           HANDLE aiHandle,
   IN           uint32 channelStart,
   IN           uint32 channelCount,
   OUT OPTIONAL void   *rawData,
   OUT OPTIONAL double *scaledData)
{
   CHK_HANDLE_TYPE(aiHandle, DaqAi);
   CHK_USER_BUF(rawData || scaledData);

   return ((DaqAiModule*)aiHandle)->ReadSamples(channelStart, channelCount, rawData, scaledData);
}

ErrorCode BDAQCALL AdxBufferedAiPrepare(
   IN HANDLE aiHandle,
   IN uint32 dataCount,
   OUT void  **dataBuffer)
{
   CHK_HANDLE_TYPE(aiHandle, DaqAi);
   CHK_DATA_PTR(dataBuffer);
   *dataBuffer = NULL;

   return ((DaqAiModule*)aiHandle)->BfdAiPrepare(dataCount, *dataBuffer);
}

ErrorCode BDAQCALL AdxBufferedAiRunOnce(
   IN HANDLE aiHandle,
   IN int32  asynchronous)
{
   CHK_HANDLE_TYPE(aiHandle, DaqAi);

   ErrorCode ret;
   switch(((DaqAiModule*)aiHandle)->getIntfType())
   {
   case IntfType::DN3_API: ret = ((DN3Api::DN3Ai*)aiHandle)->BfdAiRunOnce(!!asynchronous); break;
   case IntfType::DN3_OBJ: ret = ((DN3Obj::DN3Ai*)aiHandle)->BfdAiRunOnce(!!asynchronous); break; // int32 --> bool
   default:                ret = ErrorFuncNotSpted; break;
   }
   return ret;
}

ErrorCode BDAQCALL AdxBufferedAiRun(
   IN HANDLE aiHandle)
{
   CHK_HANDLE_TYPE(aiHandle, DaqAi);

   return ((DaqAiModule*)aiHandle)->BfdAiStart();
}

ErrorCode BDAQCALL AdxBufferedAiStop(
   IN HANDLE aiHandle)
{
   CHK_HANDLE_TYPE(aiHandle, DaqAi);

   return ((DaqAiModule*)aiHandle)->BfdAiStop();
}

ErrorCode BDAQCALL AdxBufferedAiRelease(
   IN HANDLE aiHandle)
{
   CHK_HANDLE_TYPE(aiHandle, DaqAi);

   return ((DaqAiModule*)aiHandle)->BfdAiRelease();
}

ErrorCode BDAQCALL AdxBufferedAiScaleData(
   IN              HANDLE aiHandle,
   IN              void   *rawData,
   OUT             double *scaledData,
   IN              uint32 dataCount,
   IN OUT OPTIONAL uint32 *channelOffset)
{
   // ***************************************************************************************
   // The capacity of the data buffer dataRaw and dataScaled should be larger than dataCount,
   // or else the BioDAQ.Dll or the Device driver will not be responsible for any exception
   // caused by memory overflow
   // ***************************************************************************************
   CHK_HANDLE_TYPE(aiHandle, DaqAi);
   CHK_USER_BUF(rawData && scaledData);

   ErrorCode ret;
   switch(((DaqAiModule*)aiHandle)->getIntfType())
   {
   case IntfType::DN4_OBJ: ret = ((DaqAiModule*)aiHandle)->BfdAiGetData(dataCount, NULL, scaledData, NULL, NULL, NULL);     break; 
   case IntfType::DN3_API: ret = ((DN3Api::DN3Ai*)aiHandle)->BfdAiScaleData(rawData, scaledData, dataCount, channelOffset); break;
   case IntfType::DN3_OBJ: ret = ((DN3Obj::DN3Ai*)aiHandle)->BfdAiScaleData(rawData, scaledData, dataCount, channelOffset); break; 
   default:                 ret = ErrorFuncNotSpted; break;
   }
   return ret;
}

ErrorCode BDAQCALL AdxAiGetScaleTable(
   IN           HANDLE       aiHandle,
   IN           uint32       valueRangeType,
   IN OUT       uint32       *size,
   OUT OPTIONAL MapFuncPiece *tables)
{
   CHK_HANDLE_TYPE(aiHandle, DaqAi);
   CHK_USER_BUF(size);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxAiSetScaleTable(
   IN HANDLE       aiHandle,
   IN uint32       valueRangeType,
   IN uint32       size,
   IN MapFuncPiece *tables)
{
   CHK_HANDLE_TYPE(aiHandle, DaqAi);
   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxAiGenerateNewValueRange(
   IN          HANDLE       aiHandle,
   IN          uint32       baseType,
   IN          uint32       size,
   IN OPTIONAL MapFuncPiece *tables,
   OUT         uint32       *newType)
{
   CHK_HANDLE_TYPE(aiHandle, DaqAi);
   CHK_USER_BUF(newType);
   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxAiGetCjcScaleTable(
   IN           HANDLE       aiHandle,
   IN OUT       uint32       *size,
   OUT OPTIONAL MapFuncPiece *tables)
{
   CHK_HANDLE_TYPE(aiHandle, DaqAi);
   CHK_USER_BUF(size);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxAiSetCjcScaleTable(
   IN HANDLE       aiHandle,
   IN uint32       size,
   IN MapFuncPiece *tables)
{
   CHK_HANDLE_TYPE(aiHandle, DaqAi);
   return ErrorFuncNotSpted;
}

// ---------------------------------------------------------------
// AO Function
// ---------------------------------------------------------------
ErrorCode BDAQCALL AdxAoWriteSamples(
   IN          HANDLE aoHandle,
   IN          uint32 channelStart,
   IN          uint32 channelCount,
   IN OPTIONAL void   *rawData,
   IN OPTIONAL double *scaledData)
{
   CHK_HANDLE_TYPE(aoHandle, DaqAo);
   CHK_USER_BUF(rawData || scaledData);

   return ((DaqAoModule*)aoHandle)->WriteSamples(channelStart, channelCount, rawData, scaledData);
}

ErrorCode BDAQCALL AdxBufferedAoPrepare(
   IN HANDLE aoHandle,
   IN uint32 dataCount,
   OUT void  **dataBuffer)
{
   CHK_HANDLE_TYPE(aoHandle, DaqAo);
   CHK_DATA_PTR(dataBuffer);
   *dataBuffer = NULL;

   return ((DaqAoModule*)aoHandle)->BfdAoPrepare(dataCount, *dataBuffer);
}

ErrorCode BDAQCALL AdxBufferedAoRunOnce(
   IN HANDLE aoHandle,
   IN int32  asynchronous)
{
   CHK_HANDLE_TYPE(aoHandle, DaqAo);

   return ((DaqAoModule*)aoHandle)->BfdAoRunOnce(!!asynchronous);
}

ErrorCode BDAQCALL AdxBufferedAoRun(
   IN HANDLE aoHandle)
{
   CHK_HANDLE_TYPE(aoHandle, DaqAo);

   return ((DaqAoModule*)aoHandle)->BfdAoRun();
}

ErrorCode BDAQCALL AdxBufferedAoStop(
   IN HANDLE aoHandle,
   IN uint32 action)
{
   CHK_HANDLE_TYPE(aoHandle, DaqAo);

   return ((DaqAoModule*)aoHandle)->BfdAoStop(action);
}

ErrorCode BDAQCALL AdxBufferedAoRelease(
   IN HANDLE aoHandle)
{
   CHK_HANDLE_TYPE(aoHandle, DaqAo);

   return ((DaqAoModule*)aoHandle)->BfdAoRelease();
}

ErrorCode BDAQCALL AdxBufferedAoScaleData(
   IN              HANDLE aoHandle,
   IN              double *scaledData,
   OUT             void   *rawData,
   IN              uint32 dataCount,
   IN OUT OPTIONAL uint32 *channelOffset)
{
   // ***************************************************************************************
   // The capacity of the data buffer dataRaw and dataScaled should be larger than dataCount,
   // or else the BioDAQ.Dll or the Device driver will not be responsible for any exception
   // caused by memory overflow
   // ***************************************************************************************
   CHK_HANDLE_TYPE(aoHandle, DaqAo);
   CHK_USER_BUF(rawData && scaledData);

   return ((DaqAoModule*)aoHandle)->BfdAoScaleData(scaledData, rawData, dataCount, channelOffset);
}

ErrorCode BDAQCALL AdxAoGetScaleTable(
   IN           HANDLE       aoHandle,
   IN           uint32       valueRangeType,
   IN OUT       uint32       *size,
   OUT OPTIONAL MapFuncPiece *tables)
{
   CHK_HANDLE_TYPE(aoHandle, DaqAo);
   CHK_USER_BUF(size);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxAoSetScaleTable(
   IN HANDLE       aoHandle,
   IN uint32       valueRangeType,
   IN uint32       size,
   IN MapFuncPiece *tables)
{
   CHK_HANDLE_TYPE(aoHandle, DaqAo);
   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxAoGenerateNewValueRange(
   IN          HANDLE       aoHandle,
   IN          uint32       baseType,
   IN          uint32       size,
   IN OPTIONAL MapFuncPiece *tables,
   OUT         uint32       *newType)
{
   CHK_HANDLE_TYPE(aoHandle, DaqAo);
   CHK_USER_BUF(newType);

   return ErrorFuncNotSpted;
}

// ---------------------------------------------------------------
// DI/O Function
// ---------------------------------------------------------------
ErrorCode BDAQCALL AdxDiReadPorts(
   IN  HANDLE dioHandle,
   IN  uint32 portStart,
   IN  uint32 portCount,
   OUT uint8  *dataBuffer)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);
   CHK_USER_BUF(dataBuffer);

   return ((DaqDioModule*)dioHandle)->DiReadPort(portStart, portCount, dataBuffer);
}

ErrorCode BDAQCALL AdxDoWritePorts(
   IN HANDLE dioHandle,
   IN uint32 portStart,
   IN uint32 portCount,
   IN uint8  *dataBuffer)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);
   CHK_USER_BUF(dataBuffer);

   return ((DaqDioModule*)dioHandle)->DoWritePort(portStart, portCount, dataBuffer);
}

ErrorCode BDAQCALL AdxDoReadBackPorts(
   IN HANDLE dioHandle,
   IN uint32 portStart,
   IN uint32 portCount,
   OUT uint8 *dataBuffer)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);
   CHK_USER_BUF(dataBuffer);

   return ((DaqDioModule*)dioHandle)->DoReadPort(portStart, portCount, dataBuffer);
}

ErrorCode BDAQCALL AdxDiReadBit(
   IN  HANDLE dioHandle,
   IN  uint32 port,
   IN  uint32 bit,
   OUT uint8  *dataBuffer)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);
   CHK_USER_BUF(dataBuffer);

   return ((DaqDioModule*)dioHandle)->DiReadBit(port, bit, *dataBuffer);
}

ErrorCode BDAQCALL AdxDoWriteBit(
   IN  HANDLE dioHandle,
   IN  uint32 port,
   IN  uint32 bit,
   IN  uint8  data)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);

   return ((DaqDioModule*)dioHandle)->DoWriteBit(port, bit, data);
}

ErrorCode BDAQCALL AdxDoReadBackBit(
   IN  HANDLE dioHandle,
   IN  uint32 port,
   IN  uint32 bit,
   OUT uint8  *dataBuffer)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);
   CHK_USER_BUF(dataBuffer);

   return ((DaqDioModule*)dioHandle)->DoReadBit(port, bit, *dataBuffer);
}

ErrorCode BDAQCALL AdxBufferedDiPrepare(
   IN HANDLE dioHandle,
   IN uint32 dataCount,
   OUT void  **dataBuffer)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);
   CHK_DATA_PTR(dataBuffer);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxBufferedDiRunOnce(
   IN HANDLE dioHandle,
   IN int32  asynchronous)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxBufferedDiRun(
   IN HANDLE dioHandle)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxBufferedDiStop(
   IN HANDLE dioHandle)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxBufferedDiRelease(
   IN HANDLE dioHandle)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxBufferedDoPrepare(
   IN HANDLE dioHandle,
   IN uint32 dataCount,
   OUT void  **dataBuffer)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);
   CHK_DATA_PTR(dataBuffer);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxBufferedDoRunOnce(
   IN HANDLE dioHandle,
   IN int32  asynchronous)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxBufferedDoRun(
   IN HANDLE dioHandle)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxBufferedDoStop(
   IN HANDLE dioHandle,
   IN uint32 action)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxBufferedDoRelease(
   IN HANDLE dioHandle)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);

   return ErrorFuncNotSpted;
}

ErrorCode BDAQCALL AdxDiSnapStart(
   IN  HANDLE dioHandle,
   IN  uint32 eventType,
   IN  uint32 portStart,
   IN  uint32 portCount,
   OUT uint8  **dataBuffer)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);

   return ((DaqDioModule*)dioHandle)->DiSnapStart((EventId)eventType, portStart, portCount, dataBuffer);
}

ErrorCode BDAQCALL AdxDiSnapStop(
   IN HANDLE dioHandle,
   IN uint32 eventType)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);

   return ((DaqDioModule*)dioHandle)->DiSnapStop((EventId)eventType);
}

ErrorCode BDAQCALL AdxDoReflectWdtLockStart(
   IN HANDLE dioHandle)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);
   return ((DaqDioModule*)dioHandle)->WdtExec(DaqDioModule::WdtStart);
}

ErrorCode BDAQCALL AdxDoReflectWdtFeed(
   IN HANDLE dioHandle)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);
   return ((DaqDioModule*)dioHandle)->WdtExec(DaqDioModule::WdtFeed);
}

ErrorCode BDAQCALL AdxDoReflectWdtLockStop(
   IN HANDLE dioHandle)
{
   CHK_HANDLE_TYPE(dioHandle, DaqDio);
   return ((DaqDioModule*)dioHandle)->WdtExec(DaqDioModule::WdtStop);
}

// ---------------------------------------------------------------
// Counter Function
// ---------------------------------------------------------------
ErrorCode BDAQCALL AdxCounterReset(
   IN HANDLE cntrHandle,
   IN uint32 counterStart,
   IN uint32 counterCount)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->CtReset(counterStart, counterCount);
}

ErrorCode BDAQCALL AdxCounterStart(
   IN HANDLE  cntrHandle,
   IN uint32  counterStart,
   IN uint32  counterCount)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->CtStart(counterStart, counterCount);
}

ErrorCode BDAQCALL AdxCounterRead(
   IN  HANDLE cntrHandle,
   IN  uint32 counterStart,
   IN  uint32 counterCount,
   OUT uint32 *loadValue,
   OUT uint32 *holdValue)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);
   CHK_USER_BUF(loadValue || holdValue);

   return ((DaqCtModule*)cntrHandle)->CtRead(counterStart, counterCount, loadValue, holdValue);
}

ErrorCode BDAQCALL AdxEventCountStart(
   IN HANDLE cntrHandle,
   IN uint32 counterStart,
   IN uint32 counterCount)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->EcStart(counterStart, counterCount);
}

ErrorCode BDAQCALL AdxEventCountRead(
   IN  HANDLE cntrHandle,
   IN  uint32 counterStart,
   IN  uint32 counterCount,
   OUT uint32 *eventCount )
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);
   CHK_USER_BUF(eventCount);

   return ((DaqCtModule*)cntrHandle)->EcRead(counterStart, counterCount, eventCount);
}

ErrorCode BDAQCALL AdxOneShotStart(
   IN HANDLE cntrHandle,
   IN uint32 counterStart,
   IN uint32 counterCount)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->OsStart(counterStart, counterCount);
}

ErrorCode BDAQCALL AdxTimerPulseStart(
   IN HANDLE cntrHandle,
   IN uint32 counterStart,
   IN uint32 counterCount)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->TmrStart(counterStart, counterCount);
}

ErrorCode BDAQCALL AdxFrequencyMeasureStart(
   IN HANDLE cntrHandle,
   IN uint32 counterStart,
   IN uint32 counterCount)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->FmStart(counterStart, counterCount);
}

ErrorCode BDAQCALL AdxFrequencyMeasureRead(
   IN  HANDLE cntrHandle,
   IN  uint32 counterStart,
   IN  uint32 counterCount,
   OUT double *frequency)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);
   CHK_USER_BUF(frequency);

   return ((DaqCtModule*)cntrHandle)->FmRead(counterStart, counterCount, frequency);
}

ErrorCode BDAQCALL AdxPwmInStart(
   IN HANDLE cntrHandle,
   IN uint32 counterStart,
   IN uint32 groupCount)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->PiStart(counterStart, groupCount);
}

ErrorCode BDAQCALL AdxPwmInRead(
   IN  HANDLE cntrHandle,
   IN  uint32 counterStart,
   IN  uint32 groupCount,
   OUT double *hiPeriod,
   OUT double *lowPeriod)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);
   CHK_USER_BUF(hiPeriod && lowPeriod);

   return ((DaqCtModule*)cntrHandle)->PiRead(counterStart, groupCount, hiPeriod, lowPeriod);
}

ErrorCode BDAQCALL AdxPwmOutStart(
   IN HANDLE cntrHandle,
   IN uint32 counterStart,
   IN uint32 counterCount)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->PoStart(counterStart, counterCount);
}

ErrorCode BDAQCALL AdxUpDownCountStart(
   IN HANDLE cntrHandle,
   IN uint32 counterStart,
   IN uint32 counterCount,
   IN uint32 resetCounter)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->UdStart(counterStart, counterCount, resetCounter);
}

ErrorCode BDAQCALL AdxUpDownCountValueReset(
   IN HANDLE cntrHandle,
   IN uint32 counterStart,
   IN uint32 counterCount)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->UdResetValue(counterStart, counterCount);
}

ErrorCode BDAQCALL AdxUpDownCountRead(
   IN HANDLE cntrHandle,
   IN uint32 counterStart,
   IN uint32 counterCount,
   IN uint32 *values)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->UdRead(counterStart, counterCount, values);
}

ErrorCode BDAQCALL AdxCounterSnapStart(
   IN HANDLE cntrHandle,
   IN uint32  eventType,
   IN uint32  counterStart,
   IN uint32  counterCount,
   OUT OPTIONAL uint32 **loadValues,
   OUT OPTIONAL uint32 **holdValues)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->CtSnapStart((EventId)eventType, counterStart, counterCount, loadValues, holdValues);
}

ErrorCode BDAQCALL AdxCounterSnapStop(
   IN HANDLE cntrHandle,
   IN uint32 eventType)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->CtSnapStop((EventId)eventType);
}

ErrorCode BDAQCALL AdxCounterCompareSetTable(
   IN HANDLE cntrHandle,
   IN uint32 counter,
   IN uint32 dataCount,
   IN uint32 *buffer)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->CCmpSetTable(counter, dataCount, buffer);
}

ErrorCode BDAQCALL AdxCounterCompareSetInterval(
   IN HANDLE cntrHandle,
   IN uint32 counter,
   IN uint32 firstValue,
   IN int32  increment,
   IN uint32 count)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->CCmpSetInterval(counter, firstValue, increment, count);
}

ErrorCode BDAQCALL AdxCounterCompareClear(
   IN HANDLE cntrHandle,
   IN uint32 counter)
{
   CHK_HANDLE_TYPE(cntrHandle, DaqCounter);

   return ((DaqCtModule*)cntrHandle)->CCmpClear(counter);
}
