//
// DAQNAVI4 fusion device management routine
// Note:
//      The APIs is mainly used by DAQNAVI navigator to manage the fusion devices. 
//
#define  _CRT_SECURE_NO_WARNINGS
#include <std_decl.h>
#include <ioctl_helper.hpp>

typedef struct _FUSION_DRIVER {
   ProductId   fusionType;
   char const *driverName;
} FUSION_DRIVER;

static FUSION_DRIVER const s_spted_fusion_drv[] = {
   {BD_IDAQ1CHASSYNC, "bioidaq"},
};

#define NAME_FusionDeviceInfoGet     "BDaqFusionDeviceInfoGet" 
#define NAME_DeviceFusionableCheck   "BDaqDeviceFusionableCheck"
#define NAME_DeviceFuse              "BDaqDeviceFuse"  
#define NAME_DeviceDefuse            "BDaqDeviceDefuse" 
#define NAME_FusionDeviceChildrenGet "BDaqFusionDeviceChildrenGet"

typedef ErrorCode (BDAQCALL * FN_FusionDeviceInfoGet)    (DEVICE_HWINFO_EX *hwInfo);
typedef ErrorCode (BDAQCALL * FN_DeviceFusionableCheck)  (int32 fusedDevice, int32 count, int32 children[]);
typedef ErrorCode (BDAQCALL * FN_DeviceFuse)             (int32 fusedDevice, int32 count, int32 children[]);
typedef ErrorCode (BDAQCALL * FN_DeviceDefuse)           (int32 fusedDevice, int32 count, int32 children[], int32 *rest);
typedef ErrorCode (BDAQCALL * FN_FusionDeviceChildrenGet)(int32 fusedDevice, int32 *count, int32 children[]);

//--------------------------------------------------------------------------------------------------
// Helpers
//--------------------------------------------------------------------------------------------------


//--------------------------------------------------------------------------------------------------
// DAQNAVI4 APIs for fusion devices
//--------------------------------------------------------------------------------------------------
extern "C" {
   ErrorCode BDAQCALL AdxFusionDeviceCreate(int32 count, int32 children[], int32 *fusionNumber)
   {
      if (!fusionNumber) {
         return ErrorBufferIsNull;
      }

      *fusionNumber = -1;

      if (!count) {
         return ErrorBufferTooSmall;
      }

      //
      // Get the fusion type from the first child
      int32 prod_id = -1;
      ErrorCode ret = QueryDeviceFieldI32DB(children[0], DFN_PRODID, &prod_id);
      if (ret != Success) {
         return ret;
      }

      uint32 fusionTypes[16];
      uint32 size = sizeof(fusionTypes);
      ret = PropReadValueDB(prod_id, catofModule(DaqDevice, 0), CFG_DeviceSupportedFusionTypes, &size, fusionTypes);
      if (ret != Success) {
         return ret;
      }

      //
      // Looking for the driver information
      char const  *drvName = NULL;
      for (unsigned i = 0; i < ARRAY_SIZE(s_spted_fusion_drv); ++i) {
         if (s_spted_fusion_drv[i].fusionType == fusionTypes[0]) {
            drvName = s_spted_fusion_drv[i].driverName;
            break;
         }
      }

      if (drvName == NULL) {
         return ErrorDeviceUnrecognized;
      }

      //
      // Check whether the children can be fused
      FN_DeviceFusionableCheck fn_fusionable_check = (FN_DeviceFusionableCheck)XDevFactory::GetDriverFunc(drvName, NAME_DeviceFusionableCheck);
      if (fn_fusionable_check == NULL) {
         return ErrorFuncNotSpted;
      }

      ret = fn_fusionable_check(-1/* No parent*/, count, children);
      if (ret != Success) {
         return ret;
      }

      //
      // Get fusion device information
      FN_FusionDeviceInfoGet fn_info_get = (FN_FusionDeviceInfoGet)XDevFactory::GetDriverFunc(drvName, NAME_FusionDeviceInfoGet);
      if (fn_info_get == NULL) {
         return ErrorFuncNotSpted;
      }

      DEVICE_HWINFO_EX info = {0};
      fn_info_get(&info);

      //
      // Allocate device number for the device if we had not done.
#if defined(_WIN32) || defined(WIN32)
      wchar_t const *devPath = L"\\\\.\\FusionDevice";
#else
      char const    *devPath = "FusionDevice";
#endif

      ret = AllocDeviceNumberDB(info, devPath,  (int32*)&info.DeviceNumber);
      if (ret != Success) { 
         return ret; 
      }

      //
      // Fuse the children
      FN_DeviceFuse fn_dev_fuse = (FN_DeviceFuse)XDevFactory::GetDriverFunc(drvName, NAME_DeviceFuse);
      if (fn_dev_fuse == NULL) { 
         return ErrorFuncNotSpted; 
      }

      ret = fn_dev_fuse(info.DeviceNumber, count, children);
      if (ret != Success) {
         FreeDeviceNumberDB(info.DeviceNumber);
         return ret;
      }

      *fusionNumber = info.DeviceNumber;
      return ret;
   }

   ErrorCode BDAQCALL AdxFusionDeviceAddChildren(int32 fusionNumber, int32 count, int32 children[])
   {
      //
      // Get driver name
      std::string drvName;
      ErrorCode ret = XDevFactory::GetDriverName(fusionNumber, drvName);
      if (ret != Success) {
         return ret;
      }

      //
      // Fuse the children
      FN_DeviceFuse fn_dev_fuse = (FN_DeviceFuse)XDevFactory::GetDriverFunc(drvName, NAME_DeviceFuse);
      if (fn_dev_fuse == NULL) { 
         return ErrorFuncNotSpted; 
      }

      ret = fn_dev_fuse(fusionNumber, count, children);

      return ret;
   }

   ErrorCode BDAQCALL AdxFusionDeviceDeleteChildren(int32 fusionNumber, int32 count, int32 children[])
   {
      //
      // Get driver name
      std::string drvName;
      ErrorCode ret = XDevFactory::GetDriverName(fusionNumber, drvName);
      if (ret != Success) {
         return ret;
      }

      //
      // Delete the children
      FN_DeviceDefuse fn_dev_defuse = (FN_DeviceDefuse)XDevFactory::GetDriverFunc(drvName, NAME_DeviceDefuse);
      if (fn_dev_defuse == NULL) { 
         return ErrorFuncNotSpted; 
      }

      int32 rest = -1;
      ret = fn_dev_defuse(fusionNumber, count, children, &rest);

      if (rest == 0) {
         FreeDeviceNumberDB(fusionNumber);
      }

      return ret;
   }

   ErrorCode BDAQCALL AdxFusionDeviceGetChildren(int32 fusionNumber, int32 *count, int32 children[])
   {
      //
      // Get driver name
      std::string drvName;
      ErrorCode ret = XDevFactory::GetDriverName(fusionNumber, drvName);
      if (ret != Success) {
         return ret;
      }

      //
      // Delete the children
      FN_FusionDeviceChildrenGet fn_children_get = (FN_FusionDeviceChildrenGet)XDevFactory::GetDriverFunc(drvName, NAME_FusionDeviceChildrenGet);
      if (fn_children_get == NULL) { 
         return ErrorFuncNotSpted; 
      }

      ret = fn_children_get(fusionNumber, count, children);

      return ret;
   }

   ErrorCode BDAQCALL AdxDeviceFusionableCheck(int32 fusionNumber, int32 count, int32 children[])
   {
      //
      // Get driver name
      std::string drvName;
      ErrorCode ret = XDevFactory::GetDriverName(fusionNumber, drvName);
      if (ret != Success) {
         return ret;
      }

      //
      // Fuse the children
      FN_DeviceFusionableCheck fn_fusion_chk = (FN_DeviceFusionableCheck)XDevFactory::GetDriverFunc(drvName, NAME_DeviceFusionableCheck);
      if (fn_fusion_chk == NULL) { 
         return ErrorFuncNotSpted; 
      }

      ret = fn_fusion_chk(fusionNumber, count, children);

      return ret;
   }

   ErrorCode BDAQCALL AdxIsFusionDevice(int32 fusionNumber)
   {
      //
      // Get the fusion type from the first child
      int32 prod_id = -1;
      ErrorCode ret = QueryDeviceFieldI32DB(fusionNumber, DFN_PRODID, &prod_id);
      if (ret != Success) {
         return ret;
      }

      for (unsigned i = 0; i < ARRAY_SIZE(s_spted_fusion_drv); ++i) {
         if (s_spted_fusion_drv[i].fusionType == prod_id) {
            return Success;
         }
      }

      return ErrorDeviceNotFusionable;
   }
};
