#ifndef _BDAQNAVI_PROPERTY_ACCESSOR_H
#define _BDAQNAVI_PROPERTY_ACCESSOR_H

// **********************************************************
// helper class to access device property
// **********************************************************
class DaqProperty
{
public:
   DaqProperty(DaqModule *owner)
   {
      m_owner = owner;
   }
   DaqProperty(DaqModule &owner)
   {
      m_owner = &owner;
   }
   // ------------------------------------------------------------------------
   // Helpers for property access
   // ------------------------------------------------------------------------
   ErrorCode Read(PropertyId id, uint32 size, void *buffer, uint32 flag = PropFlag::OpGet)
   {
      return Read(id, &size, buffer, flag);
   }
   ErrorCode Read(PropertyId id, uint32 *size, void *buffer, uint32 flag = PropFlag::OpGet)
   {
      extern ErrorCode PropReadHelper(DaqModule &, PropertyId, uint32 &, void *, uint32);
      return PropReadHelper(*m_owner, id, *size, buffer, flag);
   }

   ErrorCode Write(PropertyId id, uint32 size, void *buffer, uint32 flag = 0)
   {
      if (m_owner->getMode() == ModeRead) {
         return ErrorPrivilegeNotHeld;
      }

      extern ErrorCode PropWriteHelper(DaqModule &, PropertyId, uint32, void *, uint32);
      return PropWriteHelper(*m_owner, id, size, buffer, flag);
   }

   PropertyAttribute GetAttribute(PropertyId id)
   {
      extern ErrorCode PropAttributeHelper(DaqModule &, PropertyId, PropertyAttribute &);
      PropertyAttribute attr = ReadOnly;
      PropAttributeHelper(*m_owner, id, attr);
      return attr;
   }

   uint32 GetLength(PropertyId id)
   {
      uint32 bufLen = 0;
      Read(id, &bufLen, NULL);
      return bufLen;
   }

   // Get single-value property
   ErrorCode Get(PropertyId id, int32 &data)
   {
      return Read(id, sizeof(data), &data);
   }
   ErrorCode Get(PropertyId id, double &data)
   {
      return Read(id, sizeof(data), &data);
   }
   ErrorCode Get(PropertyId id, MathInterval& data)
   {
      return Read(id, sizeof(data), &data);
   }

   // Get array type property
   ErrorCode Get(PropertyId id, uint32 count, uint8 data[])
   {
      return Read(id, count * sizeof(*data), data);
   }
   ErrorCode Get(PropertyId id, uint32 count, int32 data[])
   {
      return Read(id, count * sizeof(*data), data);
   }
   ErrorCode Get(PropertyId id, uint32 count, double data[])
   {
      return Read(id, count * sizeof(*data), data);
   }
   ErrorCode Get(PropertyId id, uint32 count, wchar_t data[])
   {
      return Read(id, count * sizeof(*data), data);
   }

   // Set single-value property
   ErrorCode Set(PropertyId id, int32 data)
   {
      return Write(id, sizeof(data), &data);
   }
   ErrorCode Set(PropertyId id, double data)
   {
      return Write(id, sizeof(data), &data);
   }
   ErrorCode Set(PropertyId id, MathInterval data)
   {
      return Write(id, sizeof(data), &data);
   }

   // Set array type property
   ErrorCode Set(PropertyId id, uint32 count, uint8 data[])
   {
      return Write(id, count * sizeof(*data), data);
   }
   ErrorCode Set(PropertyId id, uint32 count, int32 data[])
   {
      return Write(id, count * sizeof(*data), data);
   }
   ErrorCode Set(PropertyId id, uint32 count, double data[])
   {
      return Write(id, count * sizeof(*data), data);
   }
   ErrorCode Set(PropertyId id, uint32 count, wchar_t data[])
   {
      return Write(id, count * sizeof(*data), data);
   }

public:
   // helpers used to access the array type property value
   template<class T>
   ErrorCode GetValueNth(PropertyId propId, int32 nth, T &propValue)
   {
      T  xBuf[128];
      T  *data = xBuf;
      uint32 size = sizeof(xBuf);

      ErrorCode ret = Read(propId, &size, data);
      if (ret == ErrorPropNotSpted) { return ret; }
      if (unsigned(nth) >= size/sizeof(T)) { return ErrorParamOutOfRange; }

      if (ret == ErrorBufferTooSmall) {
         data = (T *)_malloca(size);
         ret = Read(propId, &size, data);
      }

      propValue = data[nth];

      if (data != xBuf) { _freea(data); }
      return ret;
   }

   template<class T>
   ErrorCode SetValueNth(PropertyId propId, int32 nth, T propValue, bool isForce = false)
   {
      T xbuf[128];
      T *data = xbuf;
      uint32 size = sizeof(xbuf);

      ErrorCode ret = Read(propId, &size, data);
      if (ret == ErrorPropNotSpted) { return ret;  }
      if (unsigned(nth) >= size/sizeof(T)) { return ErrorParamOutOfRange; }

      if (ret == ErrorBufferTooSmall) {
         data = (T *)_malloca(size);
         ret = Read(propId, &size, data);
      }

      if (data[nth] != propValue || isForce) {
         data[nth] = propValue;
         ret = Write(propId, size, data);
      }

      if (data != xbuf) { _freea(data); }
      return ret;
   }

   ErrorCode GetValueNth(PropertyId propId, int32 nth, int32 &length, wchar_t value[])
   {
      wchar_t  xBuf[1024];
      wchar_t  *data = xBuf;
      uint32   size = sizeof(xBuf);

      ErrorCode ret = Read(propId, &size, data);
      if (ret == ErrorPropNotSpted)  { 
         return ret; 
      }

      do {
         if (ret == ErrorBufferTooSmall){
            data = (wchar_t *)_malloca(size);
            ret = Read(propId, &size, data);
         }
         
         if (ret != Success) {
            break;
         }

         // Forward to the Nth string value
         wchar_t *start = wcschr(data, L'{');
         for (; start && nth; --nth) {
            start = wcschr(start + 1, L'{');
         }

         if (!start){
            ret = ErrorParamOutOfRange;
            break;
         } else {
            ++start; // skip the open-brace '{'
         }

         // Find the close-brace
         wchar_t *end =  wcschr(start, L'}');
         if (!end) {
            ret = ErrorParamFmtUnexpted;
            break;
         } else {
            *end = 0; // erase the close-brace '}'
         }
         
         // Copy data
         length *= sizeof(wchar_t);
         ret = DefCopyPtr(start, uint32((end - start + 1) * sizeof(wchar_t)), value, (uint32 &)length);
         length /= sizeof(wchar_t);

      } while (0);
     
      if (data != xBuf) { _freea(data); }
      return ret;
   }

   ErrorCode SetValueNth(PropertyId propId, int32 nth, int32 length, wchar_t value[], bool isForce = false)
   {
      wchar_t  xBuf[1024];
      wchar_t  *data = xBuf;
      uint32   size = sizeof(xBuf);

      ErrorCode ret = Read(propId, &size, data);
      if (ret == ErrorPropNotSpted)  { 
         return ret; 
      }

      do {
         if (ret == ErrorBufferTooSmall){
            data = (wchar_t *)_malloca(size);
            ret = Read(propId, &size, data);
         }

         if (ret != Success) {
            break;
         }

         // Forward to the Nth string value
         wchar_t *start = wcschr(data, L'{');
         for (; start && nth; --nth) {
            start = wcschr(start + 1, L'{');
         }

         if (!start){
            ret = ErrorParamOutOfRange;
            break;
         } else {
            ++start; // skip the open-brace '{'
         }

         // Find the close-brace
         wchar_t *end =  wcschr(start, L'}');
         if (!end) {
            ret = ErrorParamFmtUnexpted;
            break;
         } 

         // Replace the data in the {...}
         if (wcslen(value) != size_t(end - start) || wcsncmp(start, value, end - start) || isForce) {
            std::wstring xval = data;
            xval.replace(start - data, end - start, value);
            ret = Write(propId, uint32((xval.size() + 1) * sizeof(wchar_t)), (void*)xval.c_str());
         }

      } while (0);

      if (data != xBuf) { _freea(data); }
      return ret;
   }

   template<class T>
   ArrayImpl<T>* GetValueArrayI32(PropertyId id)
   {
      int32  xbuf[128];
      uint32 size = sizeof(xbuf);
      int32 *data = (int32 *)ReadValueArray(id, size, xbuf);

      ArrayImpl<T>* vals = new ArrayImpl<T>();
      if (size) {
         vals->setCapacity(size/sizeof(T));
         for (unsigned i = 0; i < size/sizeof(int32); ++i) {
            vals->Add((T)data[i]);
         }
      }

      if (data != xbuf) { free(data); }
      return vals;
   }

   template<class T>
   ArrayImpl<T>* GetValueArrayI8(PropertyId id)
   {
      uint8  xbuf[128];
      uint32 size = sizeof(xbuf);
      uint8 *data = (uint8 *)ReadValueArray(id, size, xbuf);

      ArrayImpl<T>* vals = new ArrayImpl<T>();
      if (size) {
         vals->setCapacity(size/sizeof(T));
         for (unsigned i = 0; i < size/sizeof(uint8); ++i) {
            vals->Add((T)data[i]);
         }
      }

      if (data != xbuf) { free(data); }
      return vals;
   }

   ArrayImpl<uint8>* GetValueArrayI8(PropertyId id)
   {
      uint8   xbuf[128];
      uint32  size = sizeof(xbuf);
      uint8  *data = (uint8 *)ReadValueArray(id, size, xbuf);

      ArrayImpl<uint8>* vals = new ArrayImpl<uint8>();
      if (size) {
         vals->Assign(data, size/sizeof(uint8));
      }

      if (data != xbuf) { free(data); }
      return vals;
   }

   void* ReadValueArray(PropertyId id, uint32 &size, void *buffer)
   {
      ErrorCode ret = Read(id, &size, buffer);
      if (ret == ErrorBufferTooSmall || (ret == Success && buffer == NULL)) {
         buffer = malloc(size);
         ret = Read(id, &size, buffer);
         assert(ret == Success);
      } 

      return buffer;
   }

protected:
   DaqModule * m_owner;
};


#endif
