#if defined(_WIN32) || defined(WIN32) 
#  define  _CRT_SECURE_NO_WARNINGS
#  include <Windows.h>
#else
#  include <stdio.h>
#  include <sys/ioctl.h>
#  include <sys/mman.h>
#endif

#include <ctype.h>
#include <wchar.h>
#include <assert.h>
#include <vector>
#include <property_id.h>
#include <bdaq_util.h>
#include <device_adaptor.h>

#include "legacy/daqnavi3_adaptor.h"
//
// 
static ErrorCode DevicePropReadHelper(DaqModule &device, PropertyId id, uint32 &size, void *buffer, uint32 flag);
static ErrorCode DevicePropWriteHelper(DaqModule &device, PropertyId id, uint32 size, void *buffer, uint32 flag);

static ErrorCode AiPropReadHelper(DaqModule &ai, PropertyId id, uint32 &size, void *buffer, uint32 flag);
static ErrorCode AiPropWriteHelper(DaqModule &ai, PropertyId id, uint32 size, void *buffer, uint32 flag);

static ErrorCode AoPropReadHelper(DaqModule &ao, PropertyId id, uint32 &size, void *buffer, uint32 flag);
static ErrorCode AoPropWriteHelper(DaqModule &ao, PropertyId id, uint32 size, void *buffer, uint32 flag);

static ErrorCode DioPropReadHelper(DaqModule &dio, PropertyId id, uint32 &size, void *buffer, uint32 flag);
static ErrorCode DioPropWriteHelper(DaqModule &dio, PropertyId id, uint32 size, void *buffer, uint32 flag);

static ErrorCode CntrPropReadHelper(DaqModule &cntr, PropertyId id, uint32 &size, void *buffer, uint32 flag);
static ErrorCode CntrPropWriteHelper(DaqModule &cntr, PropertyId id, uint32 size, void *buffer, uint32 flag);

//
// Property default value implement
//

ErrorCode PropReadHelper(DaqModule &module, PropertyId id, uint32 &size, void *buffer, uint32 flag)
{
   ErrorCode (* readFunc)(DaqModule &, PropertyId, uint32 &, void *, uint32);
   switch(module.getType())
   {
   case DaqDevice:  readFunc = DevicePropReadHelper; break;
   case DaqAi:      readFunc = AiPropReadHelper;     break;
   case DaqAo:      readFunc = AoPropReadHelper;     break;
   case DaqDio:     readFunc = DioPropReadHelper;    break;
   case DaqCounter: readFunc = CntrPropReadHelper;   break;
   default: 
      size = 0;
      return ErrorFuncNotSpted;
   }

   return readFunc(module, id, size, buffer, flag);

}

ErrorCode PropWriteHelper(DaqModule &module, PropertyId id, uint32 size, void *buffer, uint32 flag)
{
   ErrorCode (* writeFunc)(DaqModule &, PropertyId, uint32, void *, uint32);
   switch(module.getType())
   {
   case DaqDevice:  writeFunc = DevicePropWriteHelper; break;
   case DaqAi:      writeFunc = AiPropWriteHelper;     break;
   case DaqAo:      writeFunc = AoPropWriteHelper;     break;
   case DaqDio:     writeFunc = DioPropWriteHelper;    break;
   case DaqCounter: writeFunc = CntrPropWriteHelper;   break;
   default: 
      size = 0;
      return ErrorFuncNotSpted;
   }

   return writeFunc(module, id, size, buffer, flag);
}

ErrorCode PropAttributeHelper(DaqModule &module, PropertyId id, PropertyAttribute &attr)
{
   ErrorCode ret = module.PropAttribute(id, attr);
   if (ret == Success) { return ret;}

   // Get attribute from database
   uint32 attribute = ReadOnly | Modal;
   ret = PropIdToAttrI32DB(id, PROPATTR_ATTR, &attribute);
   attr = PropertyAttribute(attribute);
   return ret;
}


/*-------------------------------------------------------------------
* default value of string properties of bionic components 
*-------------------------------------------------------------------*/
#define GroupDefaultDesc   L"Unknown group"
#define DeviceDefaultDesc  L"Unknown device" 
#define AiDefaultDesc      L"AI" 
#define AoDefaultDesc      L"AO" 
#define DioDefaultDesc     L"DIO"
#define CntrDefaultDesc    L"Counter"
#define BoardVerDefault    L"-1,-1,-1,-1"
#define DrvVersionDefault  L"0,0,0,0"

// helper function used to get the root property ID of the specified property
static 
PropertyId GetRootPropId(PropertyId child)
{
   switch (child)
   {
   // AI/AO
   case CFG_TriggerSourceEdge:
   case CFG_TriggerSourceLevel:  
   case CFG_TriggerDelayCount:
   case CFG_TriggerAction:
      {
         return CFG_FeatureTriggerSources;
      }
   case CFG_Trigger1SourceEdge:
   case CFG_Trigger1SourceLevel: 
   case CFG_Trigger1DelayCount:
   case CFG_Trigger1Action:
      {
         return CFG_FeatureTrigger1Sources;
      }

   // DI
   case CFG_FeatureDiPortScanMethod:
   case CFG_DiConvertClockSource:
      {
         return CFG_FeatureDiConvertClockSources;
      }
   case CFG_DiScanClockRate:
      {
         return CFG_DiConvertClockRatePerPort;
      }
   case CFG_DiTriggerSourceEdge:
   case CFG_DiTriggerSourceLevel:
   case CFG_DiTriggerDelayCount:
   case CFG_DiTriggerAction:
      {
         return CFG_FeatureDiTriggerSources;
      }
   case CFG_DiTrigger1SourceEdge:
   case CFG_DiTrigger1SourceLevel:
   case CFG_DiTrigger1DelayCount:
   case CFG_DiTrigger1Action:
      {
         return CFG_FeatureDiTrigger1Sources;
      }

   // DO
   case CFG_FeatureDoPortScanMethod:
   case CFG_DoConvertClockSource:
      {
         return CFG_FeatureDoConvertClockSources;
      }
   case CFG_DoScanClockRate:
      {
         return CFG_DoConvertClockRatePerPort;
      }
   case CFG_DoTriggerSourceEdge:
   case CFG_DoTriggerSourceLevel:
   case CFG_DoTriggerDelayCount:
   case CFG_DoTriggerAction:
      {
         return CFG_FeatureDoTriggerSources;
      }
   case CFG_DoTrigger1SourceEdge:
   case CFG_DoTrigger1SourceLevel:
   case CFG_DoTrigger1DelayCount:
   case CFG_DoTrigger1Action:
      {
         return CFG_FeatureDoTrigger1Sources;
      }

   default:
      break;
    }

   return PropertyId(-1);
}

// helper function used to get channel max count, using by GetPropertyDefaultValue.
static 
uint32 GetChannelCountMax(DaqModule &module)
{
   uint32 count = 0;
   uint32 size = sizeof(count);

   if (module.PropAccess(CFG_FeatureChannelNumberMax,  size, &count, PropFlag::OpGet) == Success) {
      count += 1;
   }
   return count;
}

// helper function used to fill the memory using type 'T' with count 'count'
template< class T >
ErrorCode FillBuffer(T val, uint32 count, void *dst, uint32& dstLen)
{
   uint32 orgLen = dstLen;
   dstLen = sizeof(T) * count;

   if (dst) {
      T * ptr    = (T*)dst;
      T * endPtr = ptr + x_min(count, orgLen / sizeof(T));

      while (ptr < endPtr) { *ptr++ = val; }
      if (orgLen < dstLen) { return ErrorBufferTooSmall; }
   } 
   return Success;
}

template< class T >
ErrorCode FillBufferInc(T base, T inc, uint32 count, void *dst, uint32& dstLen)
{
   uint32 orgLen = dstLen;
   dstLen = sizeof(T) * count;
   
   if (dst) {
      T * ptr    = (T*)dst;
      T * endPtr = ptr + x_min(count, orgLen / sizeof(T));
      while (ptr < endPtr) {
         *ptr++ = base;
         base  += inc;
      }

      if (orgLen < dstLen) {
         return ErrorBufferTooSmall;
      }
   } 

   return Success;
}

// helper function used to calculate the DI/DO data mask
static 
void ParseDiDataMask(uint32 portCount, uint32 chCount, uint8 portType[], uint8 portDir[], uint8 dataMask[])
{
   for (uint32 i = 0; i < portCount; ++i) {
      switch(portType[i])
      {
      case PortDi:
      case PortDio:
         if (chCount - (i * 8) >= 8) { // Is the channel count of this port less than 8 ?
            dataMask[i] = 0xFF;        // all channels is valid
         } else {                      
            dataMask[i] = (0x1 << (chCount - (i * 8))) - 1; // some channels is valid
         }
         break;
      case PortDo:
         dataMask[i] = 0;
         break;
      case Port8255A:
      case Port8255C:
      case PortIndvdlDio:
         dataMask[i] = ~portDir[i];
         break;
      }
   }
}

static 
void ParseDoDataMask(uint32 portCount, uint32 chCount, uint8 portType[], uint8 portDir[], uint8 dataMask[])
{
   for (uint32 i = 0; i < portCount; ++i) {
      switch(portType[i])
      {
      case PortDi:
         dataMask[i] = 0;
         break;
      case PortDo:
      case PortDio:
         if (chCount - (i * 8) >= 8) {  // Is the channel count of this port less than 8 ?
            dataMask[i] = 0xFF;         // all channels is valid
         } else {
            dataMask[i] = (0x1 << (chCount - (i * 8))) - 1; // some channels is valid
         }
         break;
      case Port8255A:
      case Port8255C:
      case PortIndvdlDio:
         dataMask[i] = portDir[i];
         break;
      }
   }
}

static 
bool IsAoExtRefSupported(DaqModule &ao, bool bipolar)
{
   uint32 vrgs[64];
   uint32 count = sizeof(vrgs);
   
   AoPropReadHelper(ao, CFG_FeatureVrgTypes, count, vrgs, PropFlag::OpGet);
   count /= sizeof(uint32); // convert from bytes to element count

   uint32 targetVrg = bipolar ? V_ExternalRefBipolar : V_ExternalRefUnipolar;
   for (uint32 i = 0; i < count; ++i) {
      if (vrgs[i] == targetVrg) { return true; }
   }
   return false;
}

static 
ErrorCode DevicePropReadHelper(DaqModule &device, PropertyId id, uint32 &size, void *buffer, uint32 flag)
{
   assert(device.getType() == DaqDevice);

   uint32 sizeBkup = size;
   ErrorCode ret = device.PropAccess(id, size, buffer, flag);
   if (ret != ErrorPropNotSpted) { return ret; }

   size = sizeBkup;
   // The underlying driver returned 'not supported', try to get a default value for the property.
   switch(id)
   {
   case CFG_Number:
      {
         ret = DefCopyT((uint32)device.getNumber(), buffer, size);
      }
      break;
   case CFG_ComponentType:
      {
         ret  = DefCopyT((uint32)DaqDevice, buffer, size);
      }
      break;
   case CFG_Description:
      {
         DefCopyPtr(DeviceDefaultDesc, sizeof(DeviceDefaultDesc), buffer, size); 
      }
      break;
   case CFG_Parent:
      {
         DefCopyT((uint32)-1, buffer, size);
      }
      break;
   case CFG_ChildList:               // default is NULL : no child
   case CFG_DeviceLocation:          // default is NULL : no available location info
   case CFG_DeviceBaseAddresses:     // default is NULL : not I/O based device
   case CFG_DeviceInterrupts:        // default is NULL : no interrupt function
   case CFG_DeviceSupportedEvents:   // default is NULL : no event
      {
         size = 0;
      }
      break;
   case CFG_DeviceGroupNumber:                  // default is 0 : root, belongs to none of group
   case CFG_DeviceProductID:                    // default is 0 : unknown device
   case CFG_DeviceBoardID:                      // default is 0 : factory value
   case CFG_DeviceSupportedTerminalBoardTypes:  // default is 0 : WiringBoard
   case CFG_DeviceHotResetPreventable:          // default is 0 : not support Hot reset prevent.
      {
         ret  = DefCopyT((uint32)0, buffer, size); 
      }
      break;
   case CFG_DeviceBoardVersion:  // default is: (-1, -1, -1, -1)
   case CFG_DeviceDriverVersion: // default is: (0, 0, 0, 0)
   case CFG_DeviceDllVersion:    // default is: (0, 0, 0, 0)
      {
         wchar_t const * ver = DrvVersionDefault;
         uint32  len = sizeof(DrvVersionDefault);
         if (id == CFG_DeviceBoardVersion) {
            len = sizeof(BoardVerDefault);
            ver = BoardVerDefault;
         }
         DefCopyPtr(ver, len, buffer, size); 
      }
      break;
   case CFG_DeviceTerminalBoardType: // default is : CFG_DeviceSupportedTerminalBoardTypes[0]
      {
         uint32 sptedTypes = WiringBoard;
         uint32 xsize = sizeof(uint32);
         DevicePropReadHelper(device, CFG_DeviceSupportedTerminalBoardTypes, xsize, &sptedTypes, flag);
         ret = DefCopyT(sptedTypes, buffer, size); 
      }
      break;
   case CFG_DeviceLoadingTimeInit: // depending on CFG_DeviceHotResetPreventable, 1 (do initialize) if hot reset prevent not supported.
      {
         uint32 hotResetPrevent = 0;
         uint32 xsize = sizeof(uint32);
         DevicePropReadHelper(device, CFG_DeviceHotResetPreventable, xsize, &hotResetPrevent, flag);
         ret = DefCopyT((uint32)(hotResetPrevent == 0), buffer, size); 
      }
      break;
   case CFG_DeviceWaitingForReconnect: // default is 0ms : no wait
   case CFG_DeviceWaitingForSleep:     // default is 0ms : no wait
      {
         DefCopyT((double)0.0, buffer, size); 
      }
      break;
   case CFG_DevicePrivateRegionLength:
      {
         DefCopyT((uint32)0, buffer, size); // default is 0 : no private region
      }
      break;
   default: // Not supported
      if (buffer) { memset(buffer, 0, size); }
      size = 0;
      break;
   }
 
   if (ret == ErrorPropNotSpted) { size = 0; }
   return ret;
}

static 
ErrorCode DevicePropWriteHelper(DaqModule &device, PropertyId id, uint32 size, void *buffer, uint32 flag)
{
   assert(device.getType() == DaqDevice);
   if (device.getMode() == ModeRead) { return ErrorPrivilegeNotHeld; }

   ErrorCode ret = device.PropAccess(id, size, buffer, flag);
   if (ret != ErrorPropNotSpted && ret != ErrorPropReadOnly) { return ret; }

   // Does the property be implemented in integrated layer ? 
   // Or is the property implemented as read-only by underlying driver ?
   uint32 xsize = 0;
   DevicePropReadHelper(device, id, xsize, NULL, PropFlag::OpGet);
   if (xsize == 0)   { return ErrorPropNotSpted; }
   if (size < xsize) { return ErrorBufferTooSmall; }

   // Do further check to get a proper error code.
   void *xbuf = _malloca(xsize);
   DevicePropReadHelper(device, id, xsize, xbuf, PropFlag::OpGet);

   if(memcmp(xbuf,  buffer, xsize) == 0) {
      ret = Success; // return success if the property value not changed.
   } else {
      switch (id)
      {
      case CFG_DeviceTerminalBoardType: 
      case CFG_DeviceLoadingTimeInit:
         ret = ErrorPropValueNotSpted;
         break;
      case CFG_DeviceWaitingForReconnect:
      case CFG_DeviceWaitingForSleep:    
         ret = ErrorPropValueOutOfRange;
         break;
      default:
         ret = ErrorPropReadOnly;
         break;
      }
   }
   _freea(xbuf);

   return ret;
}

//
// Helper: read property with const buffer size parameter 'size'.
static inline
ErrorCode AiPropReadHelperConst(DaqModule &ai, PropertyId id, uint32 size, void *buffer, uint32 flag)
{
   return AiPropReadHelper(ai, id, size, buffer, flag);
}

static 
ErrorCode AiPropReadHelper(DaqModule &ai, PropertyId id, uint32 &size, void *buffer, uint32 flag)
{
   assert(ai.getType() == DaqAi);

   uint32 sizeBkup = size;
   ErrorCode ret = ai.PropAccess(id, size, buffer, flag);
   if (ret != ErrorPropNotSpted) { return ret; }

   size = sizeBkup;
   // The underlying driver returned 'not supported', try to get a default value for the property.
   ErrorCode bufErr = Success;
   switch(id)
   {
   case CFG_ComponentType: // Must be DaqAi
      {
         ret  = DefCopyT((uint32)DaqAi, buffer, size); 
      }
      break;
   case CFG_Description: // default is "AI"
      {
         ret  = DefCopyPtr(AiDefaultDesc, sizeof(AiDefaultDesc), buffer, size); 
      }
      break;
   case CFG_Number: // component index
      {
         ret  = DefCopyT(ai.getNumber(), buffer, size);
      }
      break;
   case CFG_FeatureResolutionInBit: // default: 0, not support Ai; return: 'not supported'
      {
         DefCopyT((uint32)0, buffer, size);  
      }
      break;
   case CFG_FeatureDataSize:
   case CFG_FeatureDataMask: // default: calculated from resolution; return: just as 'CFG_FeatureResolutionInBit'
      {
         uint32 resolution = 0;
         ret = AiPropReadHelperConst(ai, CFG_FeatureResolutionInBit, sizeof(uint32), &resolution, flag);

         uint32 propVal = (0x1 << resolution) - 1; // data mask
         if (id == CFG_FeatureDataSize) {
            propVal = (resolution + 7) / 8;
            if (propVal > 4) {
               propVal = 8; // quad-word
            } else if (propVal > 2) {
               propVal = 4; // dword
            }
         }

         bufErr = DefCopyT(propVal, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_FeatureChannelNumberMax: // default: -1, not support; return: 'not supported'
      {
         DefCopyT((uint32)-1, buffer, size);  
      }
      break;
   case CFG_FeatureChannelConnectionType:  // default is AllSingleEnded, return: just as 'CFG_FeatureChannelNumberMax'
      {
         uint32 chCount = 0;
         ret = AiPropReadHelperConst(ai, CFG_FeatureChannelNumberMax, sizeof(uint32), &chCount, flag);
         bufErr = DefCopyT((uint32)AllSingleEnded, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_FeatureBurnDetectedReturnTypes:// default is ReturnCurrent; return: 'not support'
      {
         DefCopyT((uint32)Current, buffer, size);
      }
      break;
   case CFG_FeatureBurnoutDetectionChannels: // default is all zero(not support); return: 'not support'
      {
         FillBuffer((uint8)0, GetChannelCountMax(ai), buffer, size);
      }
      break;
   case CFG_FeatureOverallVrgType: // default is 0 : each channel can be set individually; return : just as the 'CFG_FeaturevrgTypes'
      {
         ret = DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_FeatureVrgTypes: // default : NULL, return : 'not support'
      {
         size = 0;
      }
      break;
   case CFG_FeatureConnectionTypes:
      {
         uint32 connType = AllSingleEnded;
         AiPropReadHelperConst(ai, CFG_FeatureChannelConnectionType, sizeof(uint32), &connType, flag);

         uint32 types[2] = {SingleEnded};
         uint32 count    = 1;
         switch(connType)
         {
         case AllSingleEnded:  types[0] = SingleEnded; break;
         case AllDifferential: types[0] = Differential; break;
         default:
            types[0] = SingleEnded;
            types[1] = Differential;
            count    = 2;
         }
         ret = DefCopyPtr(types, count * sizeof(uint32), buffer, size);
      }
      break;
   case CFG_FeatureOverallConnection:
      {
         uint32 connType = AllSingleEnded;
         AiPropReadHelperConst(ai, CFG_FeatureChannelConnectionType, sizeof(uint32), &connType, flag);

         ret = DefCopyT(uint32(connType != MixedSeDiffAdj), buffer, size);
      }
      break;
   case CFG_ConnectionTypeOfChannels: // default: SE, return; just as 'CFG_FeatureChanCnntType'
      {
         uint32 types[2] = {SingleEnded};
         AiPropReadHelperConst(ai, CFG_FeatureConnectionTypes, sizeof(types), &types, flag);
        
         ret = FillBuffer(types[0], GetChannelCountMax(ai), buffer, size);
      }
      break;
   case CFG_ChannelCount: // default is channel number max + 1, all single-ended
      {
         ret  = DefCopyT(GetChannelCountMax(ai), buffer, size);
      }
      break;
   case CFG_VrgTypeOfChannels: // default is CFG_FeatureVrgTypes[0] for each channel.
      {
         uint32 sptedVrg = V_OMIT;
         ret  = AiPropReadHelperConst(ai, CFG_FeatureVrgTypes, sizeof(uint32), &sptedVrg, flag);

         bufErr = FillBuffer(sptedVrg, GetChannelCountMax(ai), buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_BurnDetectedReturnTypeOfChannels:// default is CFG_FeatureBurnDetectedReturnTypes[0] for each channel.
      {
         uint32 sptedType = Current;
         ret = AiPropReadHelperConst(ai, CFG_FeatureBurnDetectedReturnTypes, sizeof(uint32), &sptedType, flag);

         bufErr = FillBuffer(sptedType, GetChannelCountMax(ai), buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_BurnoutReturnValueOfChannels: 
      {
         FillBuffer((double)0.0, GetChannelCountMax(ai), buffer, size);
      }
      break;
   case CFG_FeatureCjcChannels:     // default is NULL, not support CJC.
      {
         size = 0;
      }
      break;
   case CFG_CjcChannel:   
      {
         uint32 sptedCH = 0;
         ret = AiPropReadHelperConst(ai, CFG_FeatureCjcChannels, sizeof(uint32), &sptedCH, flag);
         if (ret == ErrorPropNotSpted) { sptedCH = GetChannelCountMax(ai); }

         bufErr = DefCopyT(sptedCH, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_CjcUpdateFrequency: // default is 2Hz, return: 'not support'
      {
         DefCopyT((double)2.0, buffer, size);
      }
      break;
   case CFG_CjcValue:
      {
         DefCopyT((double)25.0, buffer, size);
      }
      break;
   case CFG_FeatureChannelScanMethod:  // default is 0 : equal time switch
      {
         uint32 dummy;
         ret  = AiPropReadHelperConst(ai, CFG_FeatureConvertClockSources, sizeof(uint32), &dummy, flag);

         bufErr = DefCopyT((uint32)EqualTimeSwitch, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_FeatureScanChannelStartBase:
   case CFG_FeatureScanChannelCountBase:
      {
         uint32 scanMethod = EqualTimeSwitch;
         ret = AiPropReadHelperConst(ai, CFG_FeatureChannelScanMethod, sizeof(uint32), &scanMethod, flag);

         bufErr = DefCopyT((uint32)(scanMethod == EqualTimeSwitch ? 1 : 2), buffer, size); 
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_FeatureConvertClockSources: // default: SignalNone, return: 'not support'
   case CFG_FeatureScanClockSources:    // default: SignalNone, return: 'not support'
      {
         DefCopyT((uint32)SignalNone, buffer, size);
      }
      break;
   case CFG_FeatureConvertClockRateRange: // default: (0,0), return: 'not support'
   case CFG_FeatureScanClockRateRange:    // default: (0,0), return: 'not support'
   case CFG_FeatureTriggerDelayCountRange: // default: (0,0), return: 'not support'
   case CFG_FeatureTrigger1DelayCountRange: // default: (0,0), return: 'not support'
   case CFG_FeatureTriggerFilterCutoffFreqRange: // default: (0,0), return: 'not support'
      {
         MathInterval range = { 0 };
         DefCopyT(range, buffer, size);
      }
      break;
   case CFG_FeatureScanCountMax: // default 0, return: 'not support'
   case CFG_SectionDataCount:    // default 0, return: 'not support'
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_ConvertClockSource: // default : CFG_FeatureConvertClockSources[0]
      {
         uint32 sptedClk = SigInternalClock;
         ret = AiPropReadHelperConst(ai, CFG_FeatureConvertClockSources, sizeof(uint32), &sptedClk, flag);

         bufErr = DefCopyT(sptedClk, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_ConvertClockRatePerChannel: // default is 0Hz, return : 'not support'
      {
         DefCopyT((double)0.0, buffer, size);
      }
      break;
   case CFG_ScanChannelStart: // default is 1, return : 'not support'
   case CFG_ScanChannelCount: // default is 1, return : 'not support'
      {
         DefCopyT((uint32)1, buffer, size);
      }
      break;
   case CFG_ScanClockSource: // default is signal none, return : 'not support'
      {
         DefCopyT((uint32)SignalNone, buffer, size);
      }
      break;
   case CFG_ScanClockRate: // default is CFG_ConvertClockRatePerChannel, return: 'not support'
      {
         double cvtClkRate = 0;
         AiPropReadHelperConst(ai, CFG_ConvertClockRatePerChannel, sizeof(double), &cvtClkRate, flag);
         DefCopyT(cvtClkRate, buffer, size);
      }
      break;
   case CFG_ScanCount: // default is 0, return : 'not support'
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_FeatureTriggersCount:      // default is 0 : not support trigger, return : 'not support'
   case CFG_FeatureTriggerFilterTypes:     // default is 0 : FilterNone, return : 'not support'
   case CFG_FeatureTriggerSources:     // default is 0 : SignalNone, return : 'not support'
   case CFG_FeatureTriggerActions:     // default is 0 : ActionNone, return : 'not support'
   case CFG_FeatureTrigger1Sources:    // default is 0 : SignalNone, return : 'not support'
   case CFG_FeatureTrigger1Actions:    // default is 0 : ActionNone, return : 'not support'
   case CFG_TriggerSource:             // default is 0 : SignalNone, return : 'not support'
   case CFG_Trigger1Source:            // default is 0 : SignalNone, return : 'not support'
   case CFG_TriggerFilterType:   // default is 0 : SignalNone, return : 'not support'
   case CFG_Trigger1FilterType:  // default is 0 : SignalNone, return : 'not support'
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_TriggerSourceEdge:      // default is 1 : rising edge
   case CFG_Trigger1SourceEdge:     // default is 1 : rising edge
      {
         ret = AiPropReadHelperConst(ai, GetRootPropId(id), 0, NULL, flag);

         bufErr = DefCopyT((uint32)RisingEdge, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_TriggerSourceLevel:              // default is 0 : no means.
   case CFG_TriggerFilterCutoffFreq:   // default is 0 : no means.
   case CFG_Trigger1SourceLevel:             // default is 0 : no means.
   case CFG_Trigger1FilterCutoffFreq:  // default is 0 : no means.
      {
         ret = AiPropReadHelperConst(ai, GetRootPropId(id), 0, NULL, flag);

         bufErr = DefCopyT((double)0.0, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_TriggerDelayCount:
   case CFG_Trigger1DelayCount:
      {
         ret = AiPropReadHelperConst(ai, GetRootPropId(id), 0, NULL, flag);

         bufErr = DefCopyT((uint32)0, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_TriggerAction:
   case CFG_Trigger1Action:
      {
         ret = AiPropReadHelperConst(ai, GetRootPropId(id), 0, NULL, flag);

         bufErr = DefCopyT((uint32)ActionNone, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_ParentSignalConnectionChannel:
   case CFG_ParentCjcConnectionChannel:
   case CFG_ParentControlPort:
      {
         DefCopyT((uint32)-1, buffer, size);
      }
      break;
   case CFG_OrderOfChannels:
      {
         if (size == 0 || buffer == NULL) {
            AiPropReadHelperConst(ai, CFG_FeatureChannelNumberMax, sizeof(uint32), &size, flag);
            if (size != 0) {
               size += 1;
               size *= sizeof(uint32);
            }
            ret = Success;
         } else{
            if (((DaqAiModule &)ai).getIntfType() == IntfType::DN3_API) {
               ret  = ((DN3Api::DN3Ai &)ai).GetChanMapping(size/sizeof(uint32),(uint32*)buffer);
            } else {
               ret = ErrorFuncNotSpted;
            }
            if (ret == ErrorFuncNotSpted) {
               size = 0;
            }
         }
      }
      break;
   case CFG_RecordCycles:
      {
         ret = DefCopyT(uint32(1), buffer, size);
      }
      break;
   case CFG_FeatureTimestampResolution:
      {
         DefCopyT(0.0, buffer, size);
      }
      break;
   default: // Not supported
      if (buffer) { memset(buffer, 0, size); }
      size = 0;
      break;
   }

   if (ret == ErrorPropNotSpted) {
      size = 0; 
   }

   return ret;
}

static
ErrorCode AiPropWriteHelper(DaqModule &ai, PropertyId id, uint32 size, void *buffer, uint32 flag)
{
   assert(ai.getType() == DaqAi);
   if (ai.getMode() == ModeRead) { return ErrorPrivilegeNotHeld; }

   ErrorCode ret = ai.PropAccess(id, size, buffer, flag);
   if (ret != ErrorPropNotSpted && ret != ErrorPropReadOnly) { return ret; }

   // Does the property be implemented in integrated layer ? 
   // Or is the property be implemented as read-only by underlying driver ?
   uint32 xsize = 0;
   AiPropReadHelper(ai, id, xsize, NULL, PropFlag::OpGet);
   if (xsize == 0)   { return ErrorPropNotSpted; }
   if (size < xsize) { return ErrorBufferTooSmall; }

   // Do further check to get a proper error code.
   void *xbuf = _malloca(xsize);
   AiPropReadHelper(ai, id, xsize, xbuf, PropFlag::OpGet);
   if(memcmp(xbuf,  buffer, xsize) == 0) {
      ret = Success; // return success if the property value not changed.
   } else {
      switch (id)
      {
      case CFG_ConnectionTypeOfChannels: 
      case CFG_VrgTypeOfChannels:
      case CFG_BurnDetectedReturnTypeOfChannels:
      case CFG_CjcChannel:
      case CFG_ConvertClockSource:
      case CFG_ScanClockSource:
      case CFG_TriggerSource:
      case CFG_TriggerSourceEdge:
      case CFG_TriggerAction:
         ret = ErrorPropValueNotSpted;
         break;
      default: 
         ret = ErrorPropReadOnly; 
         break;
      }
   }
   _freea(xbuf);

   return ret;
}

//
// Helper: read property with const buffer size parameter 'size'.
static inline
ErrorCode AoPropReadHelperConst(DaqModule &ao, PropertyId id, uint32 size, void *buffer, uint32 flag)
{
   return AoPropReadHelper(ao, id, size, buffer, flag);
}

static
ErrorCode AoPropReadHelper(DaqModule &ao, PropertyId id, uint32 &size, void *buffer, uint32 flag)
{
   assert(ao.getType() == DaqAo);

   uint32 sizeBkup = size;
   ErrorCode ret = ao.PropAccess(id, size, buffer, flag);
   if (ret != ErrorPropNotSpted) { return ret; }

   size = sizeBkup;
   // The underlying driver returned 'not supported', try to get a default value
   // for the property.
   ErrorCode bufErr;
   switch(id)
   {
   case CFG_ComponentType: // Must be DaqAo
      {
         ret  = DefCopyT((uint32)DaqAo, buffer, size); 
      }
      break;
   case CFG_Description: // default is "AO"
      {
         ret  = DefCopyPtr(AoDefaultDesc, sizeof(AoDefaultDesc), buffer, size); 
      }
      break;
   case CFG_Number: // component index
      {
         ret  = DefCopyT(ao.getNumber(), buffer, size);
      }
      break;
   case CFG_FeatureResolutionInBit: // default: 0, not support Ao; return: 'not supported'
      {
         DefCopyT((uint32)0, buffer, size);  
      }
      break;
   case CFG_FeatureDataSize:
   case CFG_FeatureDataMask: // default: calculated from resolution; return: just as 'CFG_FeatureResolutionInBit'
      {
         uint32 resolution = 0;
         ret = AoPropReadHelperConst(ao, CFG_FeatureResolutionInBit, sizeof(resolution), &resolution, flag);

         uint32 propVal = (0x1 << resolution) - 1; // data mask
         if (id == CFG_FeatureDataSize) {
            propVal = (resolution + 7) / 8;
            if (propVal > 4) { // quad-word
               propVal = 8; 
            } else if (propVal > 2) { // dword
               propVal = 4; 
            }
         }

         bufErr = DefCopyT(propVal, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_FeatureChannelNumberMax: // default: -1, not support; return: 'not supported'
      {
         DefCopyT((uint32)-1, buffer, size);  
      }
      break;
   case CFG_FeatureVrgTypes: // default : NULL, return : 'not support'
      {
         size = 0;
      }
      break;
   case CFG_FeatureExtRefRange:
      {
         MathInterval range = { 0 };
         DefCopyT(range, buffer, size);
      }
      break;
   case CFG_FeatureExtRefAntiPolar:
      {
         ret = AoPropReadHelperConst(ao, CFG_FeatureExtRefRange, 0, NULL, flag);
         bufErr = DefCopyT((uint32)0, buffer, size);
         if (ret == Success){ ret = bufErr; }
      }
      break;
   case CFG_VrgTypeOfChannels: // default is CFG_FeatureVrgTypes[0] for each channel.
      {
         uint32 sptedVrg = V_OMIT;
         ret  = AoPropReadHelperConst(ao, CFG_FeatureVrgTypes, sizeof(sptedVrg), &sptedVrg, flag);
         
         bufErr = FillBuffer(sptedVrg, GetChannelCountMax(ao), buffer, size);
         if (ret != ErrorPropNotSpted){ ret = bufErr; }
      }
      break;
   case CFG_ExtRefValueForUnipolar:
   case CFG_ExtRefValueForBipolar:
      {
         MathInterval range = {0};
         if (IsAoExtRefSupported(ao, id == CFG_ExtRefValueForBipolar)) {
            ret = AoPropReadHelperConst(ao, CFG_FeatureExtRefRange, sizeof(range), &range, flag);
         }

         bufErr = DefCopyT(range.Max, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_FeatureChannelScanMethod:  // default is 0 : equal time switch
      {
         ret  = AoPropReadHelperConst(ao, CFG_FeatureConvertClockSources, 0, NULL, flag);
         bufErr = DefCopyT((uint32)Simultaneous, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_FeatureScanChannelStartBase:
   case CFG_FeatureScanChannelCountBase:
      {
         uint32 scanMethod = EqualTimeSwitch;
         ret = AoPropReadHelperConst(ao, CFG_FeatureChannelScanMethod, sizeof(scanMethod), &scanMethod, flag);
         bufErr = DefCopyT((uint32)(scanMethod == EqualTimeSwitch ? 1 : 2), buffer, size); 
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_FeatureConvertClockSources: // default: SignalNone, return: 'not support'
      {
         DefCopyT((uint32)SignalNone, buffer, size);
      }
      break;
   case CFG_FeatureConvertClockRateRange: // default: (0,0), return: 'not support'
   case CFG_FeatureTriggerDelayCountRange: // default: (0,0), return: 'not support'
   case CFG_FeatureTrigger1DelayCountRange: // default: (0,0), return: 'not support'
      {
         MathInterval range = { 0 };
         DefCopyT(range, buffer, size);
      }
      break;
   case CFG_SectionDataCount:    // default 0, return: 'not support'
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_ConvertClockSource: // default : CFG_FeatureConvertClockSources[0]
      {
         uint32 sptedClk = SigInternalClock;
         ret = AoPropReadHelperConst(ao, CFG_FeatureConvertClockSources, sizeof(sptedClk), &sptedClk, flag);
         bufErr = DefCopyT(sptedClk, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_ConvertClockRatePerChannel: // default is 0Hz, return : 'not support'
      {
         DefCopyT((double)0.0, buffer, size);
      }
      break;
   case CFG_ScanChannelStart: // default is 0, return : 'not support'
   case CFG_ScanChannelCount: // default is 0, return : 'not support'
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_FeatureTriggersCount:      // default is 0 : not support trigger, return : 'not support'
   case CFG_FeatureTriggerSources:     // default is 0 : SignalNone, return : 'not support'
   case CFG_FeatureTriggerActions:     // default is 0 : ActionNone, return : 'not support'
   case CFG_FeatureTrigger1Sources:    // default is 0 : SignalNone, return : 'not support'
   case CFG_FeatureTrigger1Actions:    // default is 0 : ActionNone, return : 'not support'
   case CFG_TriggerSource:             // default is 0 : SignalNone, return : 'not support'
   case CFG_Trigger1Source:            // default is 0 : SignalNone, return : 'not support'
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_TriggerSourceEdge:      // default is 1 : rising edge
   case CFG_Trigger1SourceEdge:     // default is 1 : rising edge
      {
         ret = AoPropReadHelperConst(ao, GetRootPropId(id), 0, NULL, flag);
         bufErr = DefCopyT((uint32)RisingEdge, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_TriggerSourceLevel:  // default is 0 : no means.
   case CFG_Trigger1SourceLevel: // default is 0 : no means.
      {
         ret = AoPropReadHelperConst(ao, GetRootPropId(id), 0, NULL, flag);
         bufErr = DefCopyT((double)0.0, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_TriggerDelayCount:
   case CFG_Trigger1DelayCount:
      {
         ret = AoPropReadHelperConst(ao, GetRootPropId(id), 0, NULL, flag);
         bufErr = DefCopyT((uint32)0, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_TriggerAction:
   case CFG_Trigger1Action:
      {
         ret = AoPropReadHelperConst(ao, GetRootPropId(id), 0, NULL, flag);
         bufErr = DefCopyT((uint32)ActionNone, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   default: // Not supported
      if (buffer) { memset(buffer, 0, size); }
      size = 0;
      break;
   }

   if (ret == ErrorPropNotSpted) { size = 0; }
   return ret;
}
static
ErrorCode AoPropWriteHelper(DaqModule &ao, PropertyId id, uint32 size, void *buffer, uint32 flag)
{
   assert(ao.getType() == DaqAo);
   if (ao.getMode() == ModeRead) { return ErrorPrivilegeNotHeld; }

   ErrorCode ret = ao.PropAccess(id, size, buffer, flag);
   if (ret != ErrorPropNotSpted && ret != ErrorPropReadOnly) { return ret; }

   // Does the property be implemented in integrated layer ? 
   // Or is the property be implemented as read-only by underlying driver ?
   uint32 xsize = 0;
   AoPropReadHelper(ao, id, xsize, NULL, PropFlag::OpGet);
   if (xsize == 0)   { return ErrorPropNotSpted; }
   if (size < xsize) { return ErrorBufferTooSmall; }

   // Do further check to get a proper error code.
   void *xbuf = _malloca(xsize);
   AoPropReadHelper(ao, id, xsize, xbuf, PropFlag::OpGet);

   if(memcmp(xbuf,  buffer, xsize) == 0) {
      ret = Success; // return success if the property value not changed.
   } else {
      switch (id)
      {
      case CFG_VrgTypeOfChannels:
      case CFG_ConvertClockSource:
      case CFG_TriggerSource:
      case CFG_TriggerSourceEdge:
      case CFG_TriggerAction:
         ret = ErrorPropValueNotSpted;
         break;
      default:
         ret = ErrorPropReadOnly;
         break;
      }
   }
   _freea(xbuf);

   return ret;
}

//
// Helper: read property with const buffer size parameter 'size'.
static inline
ErrorCode DioPropReadHelperConst(DaqModule &dio, PropertyId id, uint32 size, void *buffer, uint32 flag)
{
   return DioPropReadHelper(dio, id, size, buffer, flag);
}
static
ErrorCode DioPropReadHelper(DaqModule &dio, PropertyId id, uint32 &size, void *buffer, uint32 flag)
{
   assert(dio.getType() == DaqDio);

   uint32 sizeBkup = size;
   ErrorCode ret = dio.PropAccess(id, size, buffer, flag);
   if (ret != ErrorPropNotSpted) { return ret; }

   size = sizeBkup;
   // The underlying driver returned 'not supported', try to get a default value
   // for the property.
   ErrorCode bufErr;
   switch(id)
   {
   case CFG_ComponentType: // Must be DaqDio
      {
         ret  = DefCopyT((uint32)DaqDio, buffer, size); 
      }
      break;
   case CFG_Description: // default is "DIO"
      {
         ret  = DefCopyPtr(DioDefaultDesc, sizeof(DioDefaultDesc), buffer, size); 
      }
      break;
   case CFG_Number: // component index
      {
         ret  = DefCopyT(dio.getNumber(), buffer, size);
      }
      break;
   case CFG_FeaturePortsCount: // default: 0, not support DIO; return: 'not supported'
      {
         uint32 chCount = 0;
         ret = DioPropReadHelperConst(dio, CFG_FeatureChannelNumberMax, sizeof(chCount), &chCount, flag);
         if (ret == Success) { chCount += 1; }

         bufErr = DefCopyT((chCount + 7) / 8, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_FeatureChannelNumberMax: // default: -1, not support; return: 'not supported'
      {
         DefCopyT((uint32)-1, buffer, size);  
      }
      break;
   case CFG_FeaturePortsType: // default: DIO
      {
         uint32 portCount = 0;
         ret = DioPropReadHelperConst(dio, CFG_FeaturePortsCount, sizeof(uint32), &portCount, flag);
         
         bufErr = FillBuffer((uint8)PortDio, portCount, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break; 
   case CFG_DirectionOfPorts:
      {
         uint32 portCount = 0;
         DioPropReadHelperConst(dio, CFG_FeaturePortsCount, sizeof(uint32), &portCount, flag);
         FillBuffer((uint8)Input, portCount, buffer, size);
      }
      break;
   case CFG_DiDataMaskOfPorts:
   case CFG_DoDataMaskOfPorts:
      {
         uint32 portCount = 0;
         ret = DioPropReadHelperConst(dio, CFG_FeaturePortsCount, sizeof(uint32), &portCount, flag);
         if (ret != Success) { break; }

         uint8 *portType = (uint8 *)_malloca(portCount * 2); // port type + port dir
         uint8 *portDir  = portType + portCount;
         uint8 *portMask = portType;
         if (portType == NULL) { 
            ret = ErrorMemoryNotEnough; 
            break;
         }

         DioPropReadHelperConst(dio, CFG_FeaturePortsType, portCount, portType, flag);
         DioPropReadHelperConst(dio, CFG_DirectionOfPorts, portCount, portDir, flag);
         uint32 chCount = GetChannelCountMax(dio);
         if (id == CFG_DiDataMaskOfPorts) {
            ParseDiDataMask(portCount, chCount, portType, portDir, portMask);
         } else {
            ParseDoDataMask(portCount, chCount, portType, portDir, portMask);
         }
         ret = DefCopyPtr(portMask, portCount, buffer, size);
         _freea(portType);
      }
      break;
   case CFG_FeatureNoiseFilterOfChannels:
   case CFG_NoiseFilterEnabledChannels:
      {
         FillBuffer((uint8)0, GetChannelCountMax(dio), buffer, size);
      }
      break;
   case CFG_FeatureNoiseFilterBlockTimeRange:
      {
         MathInterval range = { 0 };
         DefCopyT(range, buffer, size);
      }
      break;
   case CFG_NoiseFilterOverallBlockTime:
      {
         DefCopyT((double)0.0, buffer, size);
      }
      break;
   case CFG_FeatureDiintTriggerEdges:
      {
         DefCopyT((uint8)0, buffer, size);
      }
      break;
   case CFG_FeatureDiintOfChannels:
   case CFG_FeatureDiintGateOfChannels:
   case CFG_FeatureDiCosintOfChannels:
   case CFG_FeatureDiPmintOfChannels:
   case CFG_DiintGateEnabledChannels:
   case CFG_DiCosintEnabledChannels:
   case CFG_DiPmintEnabledChannels:
      {
         FillBuffer((uint8)0, GetChannelCountMax(dio), buffer, size);
      }
      break;
   case CFG_DiintTriggerEdgeOfChannels:
      {
         uint32 chCount = 0;
         ret = DioPropReadHelper(dio, CFG_FeatureDiintOfChannels, chCount, NULL, flag);
         if (ret != Success) { break; }

         assert(chCount);
         uint8 *diintCH = (uint8 *)_malloca(chCount);
         if (diintCH == NULL) {
            ret = ErrorMemoryNotEnough;
            break;
         }

         DioPropReadHelper(dio, CFG_FeatureDiintOfChannels, chCount, diintCH, flag);
         for (uint32 i = 0; i < chCount; ++i) {
            if (diintCH[i]) {
               diintCH[i] = RisingEdge;
            }
         }
         ret = DefCopyPtr(diintCH, chCount, buffer, size);
         _freea(diintCH);
      }
      break;
   case CFG_DiPmintValueOfPorts:
   case CFG_DoInitialStateOfPorts:
      {
         uint32 chCount = GetChannelCountMax(dio);
         FillBuffer((uint8)0, (chCount + 7) / 8, buffer, size);
      }
      break;
   case CFG_FeatureDiSnapEventSources:
      {
         DefCopyT((uint32)-1, buffer, size);
      }
      break;
   case CFG_FeatureDoInitialStateDepository:
      {
         DefCopyT((uint32)DepositoryNone, buffer, size);
      }
      break;
   case CFG_FeatureDoFreezeSignalSources:
   case CFG_DoFreezeEnabled:
   case CFG_DoFreezeSignalState:
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_FeatureDoReflectWdtFeedIntervalRange:
      {
         MathInterval range = { 0 };
         DefCopyT(range, buffer, size);
      }
      break;
   case CFG_DoReflectWdtFeedInterval:
      {
         MathInterval range = {0};
         ret = DioPropReadHelperConst(dio, CFG_FeatureDoReflectWdtFeedIntervalRange, sizeof(range), &range, flag);
         bufErr = DefCopyT(range.Max, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_DoReflectWdtLockValue:
      {
         uint32 doPortCount = 0;
         DioPropReadHelper(dio, CFG_FeaturePortsCount, doPortCount, NULL, flag);
         FillBuffer((uint8)0, doPortCount, buffer, size);
      }
      break;
   case CFG_FeatureDiPortScanMethod:
   case CFG_FeatureDoPortScanMethod:
      {
         ret = DioPropReadHelperConst(dio, GetRootPropId(id), 0, NULL, flag);
         bufErr = DefCopyT((uint32)Simultaneous, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_FeatureDiConvertClockSources: // default: SignalNone, return: 'not support'
   case CFG_FeatureDiScanClockSources:    // default: SignalNone, return: 'not support'
   case CFG_FeatureDoConvertClockSources: // default: SignalNone, return: 'not support'
   case CFG_FeatureDoScanClockSources:    // default: SignalNone, return: 'not support'
      {
         DefCopyT((uint32)SignalNone, buffer, size);
      }
      break;
   case CFG_FeatureDiConvertClockRateRange: // default: (0,0), return: 'not support'
   case CFG_FeatureDiScanClockRateRange:    // default: (0,0), return: 'not support'
   case CFG_FeatureDiTriggerDelayCountRange: // default: (0,0), return: 'not support'
   case CFG_FeatureDiTrigger1DelayCountRange: // default: (0,0), return: 'not support'
   case CFG_FeatureDoConvertClockRateRange: // default: (0,0), return: 'not support'
   case CFG_FeatureDoScanClockRateRange:    // default: (0,0), return: 'not support'
   case CFG_FeatureDoTriggerDelayCountRange: // default: (0,0), return: 'not support'
   case CFG_FeatureDoTrigger1DelayCountRange: // default: (0,0), return: 'not support'
      {
         MathInterval range = { 0 };
         DefCopyT(range, buffer, size);
      }
      break;
   case CFG_FeatureDiScanCountMax: // default 0, return: 'not support'
   case CFG_FeatureDoScanCountMax: // default 0, return: 'not support'
   case CFG_DiSectionDataCount:    // default 0, return: 'not support'
   case CFG_DoSectionDataCount:    // default 0, return: 'not support'
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_DiConvertClockSource: // default : CFG_FeatureConvertClockSources[0]
   case CFG_DoConvertClockSource:  // default : CFG_FeatureConvertClockSources[0]
      {
         uint32 sptedClk = SigInternalClock;
         ret = DioPropReadHelperConst(dio, GetRootPropId(id), sizeof(sptedClk), &sptedClk, flag);
         bufErr = DefCopyT(sptedClk, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_DiConvertClockRatePerPort: // default is 0Hz, return : 'not support'
   case CFG_DoConvertClockRatePerPort: // default is 0Hz, return : 'not support'
      {
         DefCopyT((double)0.0, buffer, size);
      }
      break;
   case CFG_DiScanPortStart: // default is 0, return : 'not support'
   case CFG_DiScanPortCount: // default is 0, return : 'not support'
   case CFG_DoScanPortStart: // default is 0, return : 'not support'
   case CFG_DoScanPortCount: // default is 0, return : 'not support'
     {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_DiScanClockSource: // default is signal none, return : 'not support'
   case CFG_DoScanClockSource: // default is signal none, return : 'not support'
      {
         DefCopyT((uint32)SignalNone, buffer, size);
      }
      break;
   case CFG_DiScanClockRate: // default is CFG_ConvertClockRatePerChannel, return: 'not support'
   case CFG_DoScanClockRate: // default is CFG_ConvertClockRatePerChannel, return: 'not support'
      {
         double cvtClkRate = 0;
         DioPropReadHelperConst(dio, GetRootPropId(id), sizeof(double), &cvtClkRate, flag);
         DefCopyT(cvtClkRate, buffer, size);
      }
      break;
   case CFG_DiScanCount: // default is 0, return : 'not support'
   case CFG_DoScanCount: // default is 0, return : 'not support'
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_FeatureDiTriggersCount:      // default is 0 : not support trigger, return : 'not support'
   case CFG_FeatureDiTriggerSources:     // default is 0 : SignalNone, return : 'not support'
   case CFG_FeatureDiTriggerActions:     // default is 0 : ActionNone, return : 'not support'
   case CFG_FeatureDiTrigger1Sources:    // default is 0 : SignalNone, return : 'not support'
   case CFG_FeatureDiTrigger1Actions:    // default is 0 : ActionNone, return : 'not support'
   case CFG_DiTriggerSource:             // default is 0 : SignalNone, return : 'not support'
   case CFG_DiTrigger1Source:            // default is 0 : SignalNone, return : 'not support'
   case CFG_FeatureDoTriggersCount:      // default is 0 : not support trigger, return : 'not support'
   case CFG_FeatureDoTriggerSources:     // default is 0 : SignalNone, return : 'not support'
   case CFG_FeatureDoTriggerActions:     // default is 0 : ActionNone, return : 'not support'
   case CFG_FeatureDoTrigger1Sources:    // default is 0 : SignalNone, return : 'not support'
   case CFG_FeatureDoTrigger1Actions:    // default is 0 : ActionNone, return : 'not support'
   case CFG_DoTriggerSource:             // default is 0 : SignalNone, return : 'not support'
   case CFG_DoTrigger1Source:            // default is 0 : SignalNone, return : 'not support'
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_DiTriggerSourceEdge:      // default is 1 : rising edge
   case CFG_DiTrigger1SourceEdge:     // default is 1 : rising edge
   case CFG_DoTriggerSourceEdge:      // default is 1 : rising edge
   case CFG_DoTrigger1SourceEdge:     // default is 1 : rising edge
      {
         ret = DioPropReadHelperConst(dio, GetRootPropId(id), 0, NULL, flag);
         bufErr = DefCopyT((uint32)RisingEdge, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_DiTriggerSourceLevel:  // default is 0 : no means.
   case CFG_DiTrigger1SourceLevel: // default is 0 : no means.
   case CFG_DoTriggerSourceLevel:  // default is 0 : no means.
   case CFG_DoTrigger1SourceLevel: // default is 0 : no means.
      {
         ret = DioPropReadHelperConst(dio, GetRootPropId(id), 0, NULL, flag);
         bufErr = DefCopyT((double)0.0, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_DiTriggerDelayCount:
   case CFG_DiTrigger1DelayCount:
   case CFG_DoTriggerDelayCount:
   case CFG_DoTrigger1DelayCount:
      {
         ret = DioPropReadHelperConst(dio, GetRootPropId(id), 0, NULL, flag);
         bufErr = DefCopyT((uint32)0, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_DiTriggerAction:
   case CFG_DiTrigger1Action:
   case CFG_DoTriggerAction:
   case CFG_DoTrigger1Action:
     {
         ret = DioPropReadHelperConst(dio, GetRootPropId(id), 0, NULL, flag);
         bufErr = DefCopyT((uint32)ActionNone, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   default: // Not supported
      if (buffer) { memset(buffer, 0, size); }
      size = 0;
      break;
   }

   if (ret == ErrorPropNotSpted) { size = 0; }
   return ret;
}

static
ErrorCode DioPropWriteHelper(DaqModule &dio, PropertyId id, uint32 size, void *buffer, uint32 flag)
{
   assert(dio.getType() == DaqDio);
   if (dio.getMode() == ModeRead) { return ErrorPrivilegeNotHeld; }

   ErrorCode ret = dio.PropAccess(id, size, buffer, flag);
   if (ret != ErrorPropNotSpted && ret != ErrorPropReadOnly) { return ret; }

   // Does the property be implemented in integrated layer ? 
   // Or is the property be implemented as read-only by underlying driver ?
   uint32 xsize = 0;
   ret = DioPropReadHelper(dio, id, xsize, NULL, PropFlag::OpGet);
   if (xsize == 0)   { return ErrorPropNotSpted; }
   if (size < xsize) { return ErrorBufferTooSmall; }

   // Do further check to get a proper error code.
   void *xbuf = _malloca(xsize);
   DioPropReadHelper(dio, id, xsize, xbuf, PropFlag::OpGet);

   if(memcmp(xbuf,  buffer, xsize) == 0) {
      ret = Success; // return success if the property value not changed.
   } else {
      switch (id)
      {
      case CFG_DiintTriggerEdgeOfChannels: 
         ret = ErrorPropValueNotSpted; 
         break;
      default:  
         ret = ErrorPropReadOnly; 
         break;
      }
   }
   _freea(xbuf);

   return ret;
}


//
// Helper: read property with const buffer size parameter 'size'.
static inline
ErrorCode CntrPropReadHelperConst(DaqModule &cntr, PropertyId id, uint32 size, void *buffer, uint32 flag)
{
   return CntrPropReadHelper(cntr, id, size, buffer, flag);
}

static
ErrorCode CntrPropReadHelper(DaqModule &cntr, PropertyId id, uint32 &size, void *buffer, uint32 flag)
{
   assert(cntr.getType() == DaqCounter);

   uint32 sizeBkup = size;
   ErrorCode ret = cntr.PropAccess(id, size, buffer, flag);
   if (ret != ErrorPropNotSpted) { return ret; }

   size = sizeBkup;
   // The underlying driver returned 'not supported', try to get a default value
   // for the property.
   uint32 chCount = GetChannelCountMax(cntr);
   ErrorCode bufErr;
   switch(id)
   {
//-----------------------------------------------------
// common
//-----------------------------------------------------
   case CFG_ComponentType: // Must be DaqCntr
      {
         ret  = DefCopyT((uint32)DaqCounter, buffer, size); 
      }
      break;
   case CFG_Description: // default is "Counter"
      {
         ret  = DefCopyPtr(CntrDefaultDesc, sizeof(CntrDefaultDesc), buffer, size); 
      }
      break;
   case CFG_Number: // component index
      {
         ret  = DefCopyT(cntr.getNumber(), buffer, size);
      }
      break;
   case CFG_FeatureChannelNumberMax: // default: -1, not support; return: 'not supported'
      {
         DefCopyT((uint32)-1, buffer, size);  
      }
      break;
   case CFG_FeatureResolutionInBit: // default: 0, not support; return: 'not supported'
      {
         DefCopyT((uint32)0, buffer, size);  
      }
      break;
   case CFG_FeatureDataSize:
      {
         uint32 resolution = 0;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureResolutionInBit, sizeof(uint32), &resolution, flag);

         uint32 propVal = (0x1 << resolution) - 1; // data mask
         if (id == CFG_FeatureDataSize) {
            propVal = (resolution + 7) / 8;
            if (propVal > 4) {
               propVal = 8; // quad-word
            } else if (propVal > 2) {
               propVal = 4; // dword
            }
         }
         bufErr = DefCopyT(propVal, buffer, size);
         if (ret == Success) { ret = bufErr; }
      }
      break;
   case CFG_FeatureCapabilitiesOfCounter0:
   case CFG_FeatureCapabilitiesOfCounter1:
   case CFG_FeatureCapabilitiesOfCounter2:
   case CFG_FeatureCapabilitiesOfCounter3:
   case CFG_FeatureCapabilitiesOfCounter4:
   case CFG_FeatureCapabilitiesOfCounter5:
   case CFG_FeatureCapabilitiesOfCounter6:
   case CFG_FeatureCapabilitiesOfCounter7:
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_FeatureNoiseFilterOfChannels:
   case CFG_NoiseFilterEnabledChannels:
      {
         FillBuffer((uint8)0, chCount, buffer, size);
      }
      break;
   case CFG_FeatureNoiseFilterBlockTimeRange:
      {
         MathInterval range = { 0 };
         DefCopyT(range, buffer, size);
      }
      break;
   case CFG_NoiseFilterOverallBlockTime:
      {
         DefCopyT((double)0.0, buffer, size);
      }
      break;
//-----------------------------------------------------
// primal counter
//-----------------------------------------------------
   case CFG_FeatureChipClkSourceOfCounter0:
   case CFG_FeatureChipClkSourceOfCounter1:
   case CFG_FeatureChipClkSourceOfCounter2:
   case CFG_FeatureChipClkSourceOfCounter3:
   case CFG_FeatureChipClkSourceOfCounter4:
   case CFG_FeatureChipClkSourceOfCounter5:
   case CFG_FeatureChipClkSourceOfCounter6:
   case CFG_FeatureChipClkSourceOfCounter7:
      {
         DefCopyT((uint32)SignalNone, buffer, size);
      }
      break;
   case CFG_FeatureChipGateSourceOfCounter0:
   case CFG_FeatureChipGateSourceOfCounter1:
   case CFG_FeatureChipGateSourceOfCounter2:
   case CFG_FeatureChipGateSourceOfCounter3:
   case CFG_FeatureChipGateSourceOfCounter4:
   case CFG_FeatureChipGateSourceOfCounter5:
   case CFG_FeatureChipGateSourceOfCounter6:
   case CFG_FeatureChipGateSourceOfCounter7:
      {
         DefCopyT((uint32)(SigCntGate0 + id - CFG_FeatureChipGateSourceOfCounter0), buffer, size);
      }
      break;
   case CFG_FeatureChipValueRegisters:
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_FeatureChipOperationModes:
      {
         DefCopyT((uint32)-1, buffer, size);
      }
      break;
   case CFG_FeatureChipSignalCountingTypes:
      {
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureChipOperationModes, 0, NULL, flag);
         bufErr = DefCopyT((uint32)DownCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_ChipClkSourceOfCounters:
      {
         uint32 *clkSrc = (uint32 *)_malloca(chCount * sizeof(uint32));
         if (clkSrc == NULL) {
            ret = ErrorMemoryNotEnough;
            break;
         }
         for (uint32 i = 0; i < chCount; ++i) {
            clkSrc[i] = SignalNone;
            CntrPropReadHelperConst(cntr, PropertyId(CFG_FeatureChipClkSourceOfCounter0 + i), sizeof(uint32), &clkSrc[i], flag);
         }
         ret = DefCopyPtr(clkSrc, chCount * sizeof(uint32), buffer, size);
         _freea(clkSrc);
      }
      break;
   case CFG_ChipGateSourceOfCounters:
      {
         uint32 *gateSrc = (uint32 *)_malloca(chCount * sizeof(uint32));
         if (gateSrc == NULL) {
            ret = ErrorMemoryNotEnough;
            break;
         }
         uint32 gateSpted = 0;
         for (uint32 i = 0; i < chCount; ++i) {
            gateSrc[i] = SignalNone;
            ret = CntrPropReadHelperConst(cntr, PropertyId(CFG_FeatureChipGateSourceOfCounter0 + i), sizeof(uint32), &gateSrc[i], flag);
            if (ret != ErrorPropNotSpted) { ++gateSpted; }
         }
         bufErr = DefCopyPtr(gateSrc, chCount * sizeof(uint32), buffer, size);
         if (gateSpted) { ret = bufErr; }
         _freea(gateSrc);
      }
      break;
   case CFG_ChipOperationModeOfCounters:
      {
         FillBuffer((uint32)-1, chCount, buffer, size);
      }
      break;
   case CFG_ChipLoadValueOfCounters:
      {
         uint32 resolution = 0;
         CntrPropReadHelperConst(cntr, CFG_FeatureResolutionInBit, sizeof(uint32), &resolution, flag);
         FillBuffer(((uint32)0x1 << resolution) - 1, chCount, buffer, size);
      }
      break;
   case CFG_ChipHoldValueOfCounters:
   case CFG_ChipOverCompareValueOfCounters:
   case CFG_ChipUnderCompareValueOfCounters:
      {
         FillBuffer((uint32)0, chCount, buffer, size);
      }
      break;
   case CFG_ChipOverCompareEnabledCounters:
   case CFG_ChipUnderCompareEnabledCounters:
      {
         FillBuffer((uint8)0, chCount, buffer, size);
      }
      break;
   case CFG_ChipSignalCountingTypeOfCounters:
      {
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureChipOperationModes, 0, NULL, flag);
         
         uint32 sptedType = DownCount;
         CntrPropReadHelperConst(cntr, CFG_FeatureChipSignalCountingTypes, sizeof(uint32), &sptedType, flag);
         bufErr = FillBuffer(sptedType, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   // -------------------------------------------------------
   // V1.4 added
   // -------------------------------------------------------
   case CFG_FeatureChipClkPolarities:
   case CFG_FeatureChipGatePolarities:
      {
         ret  = DefCopyT((uint32)Positive, buffer, size);
      }
      break;
   case CFG_FeatureChipOutSignals:
      {
         ret  = FillBuffer((uint32)PositivePulse, chCount, buffer, size);
      }
      break;
   case CFG_ChipClkPolarityOfCounters:
      {
         uint32 sptedPolarity = Positive;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureChipClkPolarities, sizeof(uint32), &sptedPolarity, flag);
         bufErr = FillBuffer(sptedPolarity, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_ChipGatePolarityOfCounters:
      {
         uint32 sptedPolarity = Positive;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureChipGatePolarities, sizeof(uint32), &sptedPolarity, flag);
         bufErr = FillBuffer(sptedPolarity, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_ChipOutSignalOfCounters:
      {
         uint32 sig = PositivePulse;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureChipOutSignals, sizeof(uint32), &sig, flag);
         bufErr = FillBuffer(sig, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;

//-----------------------------------------------------
// Event counting
//-----------------------------------------------------
   // -------------------------------------------------------
   // V1.3 added
   // -------------------------------------------------------
   case CFG_FeatureEcClkPolarities:
   case CFG_FeatureEcGatePolarities:
      {
         ret  = DefCopyT((uint32)Positive, buffer, size);
      }
      break;
   case CFG_FeatureEcGateControlOfCounters:
      {
         ret  = FillBuffer((uint32)0, chCount, buffer, size);
      }
      break;
   case CFG_EcClkPolarityOfCounters:
      {
         uint32 sptedPolarity = Positive;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureEcClkPolarities, sizeof(uint32), &sptedPolarity, flag);
         bufErr = FillBuffer(sptedPolarity, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_EcGatePolarityOfCounters:
      {
         uint32 sptedPolarity = Positive;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureEcGatePolarities, sizeof(uint32), &sptedPolarity, flag);
         bufErr = FillBuffer(sptedPolarity, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_EcGateEnabledOfCounters:
      {
         ret  = FillBuffer((uint32)1, chCount, buffer, size);
      }
      break;
   // -------------------------------------------------------
   // V3.3 added
   // -------------------------------------------------------
   case CFG_FeatureEcClkSourceOfCounter0:
   case CFG_FeatureEcClkSourceOfCounter1:
   case CFG_FeatureEcClkSourceOfCounter2:
   case CFG_FeatureEcClkSourceOfCounter3:
   case CFG_FeatureEcClkSourceOfCounter4:
   case CFG_FeatureEcClkSourceOfCounter5:
   case CFG_FeatureEcClkSourceOfCounter6:
   case CFG_FeatureEcClkSourceOfCounter7:
      {
         DefCopyT((uint32)(SigCntClk0 + id - CFG_FeatureEcClkSourceOfCounter0), buffer, size);
      }
      break;
   case CFG_EcClkSourceOfCounters:
      {
         uint32 *clkSrc = (uint32 *)_malloca(chCount * sizeof(uint32));
         if (clkSrc == NULL) {
            ret = ErrorMemoryNotEnough;
            break;
         }

         for (uint32 i = 0; i < chCount; ++i) {
            clkSrc[i] = SigCntClk0 + i;
            CntrPropReadHelperConst(cntr, PropertyId(CFG_FeatureEcClkSourceOfCounter0 + i), sizeof(uint32), &clkSrc[i], flag);
         }
         ret = DefCopyPtr(clkSrc, chCount * sizeof(uint32), buffer, size);
         _freea(clkSrc);
      }
      break;
//-----------------------------------------------------
// One-shot
//-----------------------------------------------------
   case CFG_FeatureOsClkSourceOfCounter0:
   case CFG_FeatureOsClkSourceOfCounter1:
   case CFG_FeatureOsClkSourceOfCounter2:
   case CFG_FeatureOsClkSourceOfCounter3:
   case CFG_FeatureOsClkSourceOfCounter4:
   case CFG_FeatureOsClkSourceOfCounter5:
   case CFG_FeatureOsClkSourceOfCounter6:
   case CFG_FeatureOsClkSourceOfCounter7:
      {
         DefCopyT((uint32)SignalNone, buffer, size);
      }
      break;
   case CFG_FeatureOsGateSourceOfCounter0:
   case CFG_FeatureOsGateSourceOfCounter1:
   case CFG_FeatureOsGateSourceOfCounter2:
   case CFG_FeatureOsGateSourceOfCounter3:
   case CFG_FeatureOsGateSourceOfCounter4:
   case CFG_FeatureOsGateSourceOfCounter5:
   case CFG_FeatureOsGateSourceOfCounter6:
   case CFG_FeatureOsGateSourceOfCounter7:
      {
         DefCopyT((uint32)(SigCntGate0 + id - CFG_FeatureOsGateSourceOfCounter0), buffer, size);
      }
      break;
   case CFG_OsClkSourceOfCounters:
      {
         uint32 *clkSrc = (uint32 *)_malloca(chCount * sizeof(uint32));
         if (clkSrc == NULL) {
            ret = ErrorMemoryNotEnough;
            break;
         }

         for (uint32 i = 0; i < chCount; ++i) {
            clkSrc[i] = SignalNone;
            CntrPropReadHelperConst(cntr, PropertyId(CFG_FeatureOsClkSourceOfCounter0 + i), sizeof(uint32), &clkSrc[i], flag);
         }
         ret = DefCopyPtr(clkSrc, chCount * sizeof(uint32), buffer, size);
         _freea(clkSrc);
      }
      break;
   case CFG_OsGateSourceOfCounters:
      {
         uint32 *gateSrc = (uint32 *)_malloca(chCount * sizeof(uint32));
         if (gateSrc == NULL) {
            ret = ErrorMemoryNotEnough;
            break;
         }

         uint32 gateSpted = 0;
         for (uint32 i = 0; i < chCount; ++i) {
            gateSrc[i] = SignalNone;
            ret = CntrPropReadHelperConst(cntr, PropertyId(CFG_FeatureOsGateSourceOfCounter0 + i), sizeof(uint32), &gateSrc[i], flag);
            if (ret != ErrorPropNotSpted) { ++gateSpted; }
         }
         bufErr = DefCopyPtr(gateSrc, chCount * sizeof(uint32), buffer, size);
         if (gateSpted) { ret = bufErr; }
         _freea(gateSrc);
      }
      break;
   case CFG_OsDelayCountOfCounters: // the driver must supply this property value if the one-shot is supported
      {
         FillBuffer((uint32)0, chCount, buffer, size);
      }
      break;
   // -------------------------------------------------------
   // V1.3 added
   // -------------------------------------------------------
   case CFG_FeatureOsClkPolarities:
   case CFG_FeatureOsGatePolarities:
      {
         ret  = DefCopyT((uint32)Positive, buffer, size);
      }
      break;
   case CFG_FeatureOsOutSignals:
      {
         ret  = DefCopyT((uint32)PositivePulse, buffer, size);
      }
      break;
   case CFG_OsClkPolarityOfCounters:
      {
         uint32 sptedPolarity = Positive;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureOsClkPolarities, sizeof(uint32), &sptedPolarity, flag);
         bufErr = FillBuffer(sptedPolarity, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_OsGatePolarityOfCounters:
      {
         uint32 sptedPolarity = Positive;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureOsGatePolarities, sizeof(uint32), &sptedPolarity, flag);
         bufErr = FillBuffer(sptedPolarity, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_OsOutSignalOfCounters:
      {
         uint32 sig = PositivePulse;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureOsOutSignals, sizeof(uint32), &sig, flag);
         bufErr = FillBuffer(sig, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
      // -------------------------------------------------------
      // V1.5 added
      // -------------------------------------------------------
   case CFG_FeatureOsDelayCountRange:
      {
         MathInterval range = { 0 };
         uint32 resolution = 0;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureResolutionInBit, sizeof(uint32), &resolution, flag);
         if (ret == Success) {
            uint32 chipMode = -1;
            CntrPropReadHelperConst(cntr, CFG_FeatureChipOperationModes, sizeof(chipMode), &chipMode, flag);
            range.Min = chipMode > C8254_M5 ? 1 : 2;
            range.Max = (uint32)((1 << resolution) - 1);
            range.Type= LCBRCB;
         }

         bufErr = DefCopyT(range, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
//-----------------------------------------------------
// Timer/Pulse
//-----------------------------------------------------
   case CFG_FeatureTmrCascadeGroups:
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_TmrFrequencyOfCounters: // the driver must supply this property value if the one-shot is supported
      {
         FillBuffer((double)0.0, chCount, buffer, size);
      }
      break;
   // -------------------------------------------------------
   // V1.3 added
   // -------------------------------------------------------
   case CFG_FeatureTmrGateControlOfCounters:
      {
         ret  = FillBuffer((uint32)0, chCount, buffer, size);
      }
      break;
   case CFG_FeatureTmrGatePolarities:
      {
         ret  = DefCopyT((uint32)Positive, buffer, size);
      }
      break;
   case CFG_FeatureTmrOutSignals:
      {
         ret  = DefCopyT((uint32)PositivePulse, buffer, size);
      }
      break;
   case CFG_FeatureTmrFrequencyRange:
      {
         MathInterval range = { 0 };
         DefCopyT(range, buffer, size);
      }
      break;
   case CFG_TmrGateEnabledOfCounters:
      {
         ret  = FillBuffer((uint32)1, chCount, buffer, size);
      }
      break;
   case CFG_TmrGatePolarityOfCounters:
      {
         uint32 sptedPolarity = Positive;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureTmrGatePolarities, sizeof(uint32), &sptedPolarity, flag);
         bufErr = FillBuffer(sptedPolarity, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_TmrOutSignalOfCounters:
      {
         uint32 sig = PositivePulse;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureTmrOutSignals, sizeof(uint32), &sig, flag);
         bufErr = FillBuffer(sig, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;

//-----------------------------------------------------
// Frequency measurement
//-----------------------------------------------------
   case CFG_FeatureFmMethods:
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_FmMethodOfCounters:
      {
         uint32 fmMethod = AutoAdaptive;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureFmMethods, sizeof(uint32), &fmMethod, flag);
         bufErr = FillBuffer(fmMethod, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_FmCollectionPeriodOfCounters:
      {
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureFmMethods, 0, NULL, flag);
         bufErr = FillBuffer((double)0.0, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
//-----------------------------------------------------
// Pulse width measurement
//-----------------------------------------------------
   case CFG_FeaturePiCascadeGroups:
      {
         DefCopyT((uint32)0, buffer, size);
      }
      break;
//-----------------------------------------------------
// Pulse width modulation
//-----------------------------------------------------
   case CFG_PoHiPeriodOfCounters: // the driver must supply this property value if the one-shot is supported
      {
         FillBuffer((double)0.08, chCount, buffer, size);
      }
      break;
   case CFG_PoLoPeriodOfCounters: // the driver must supply this property value if the one-shot is supported
      {
         FillBuffer((double)0.02, chCount, buffer, size);
      }
      break;

   // -------------------------------------------------------
   // V1.3 added
   // -------------------------------------------------------
   case CFG_FeaturePoGateControlOfCounters:
      {
         ret  = FillBuffer((uint32)0, chCount, buffer, size);
      }
      break;
   case CFG_FeaturePoGatePolarities:
      {
         ret  = DefCopyT((uint32)Positive, buffer, size);
      }
      break;
   case CFG_FeaturePoHiPeriodRange:
   case CFG_FeaturePoLoPeriodRange:
   case CFG_FeaturePoOutCountRange:
      {
         MathInterval range = { 0 };
         DefCopyT(range, buffer, size);
      }
      break;
   case CFG_PoGateEnabledOfCounters:
      {
         ret  = FillBuffer((uint32)1, chCount, buffer, size);
      }
      break;
   case CFG_PoGatePolarityOfCounters:
      {
         uint32 sptedPolarity = Positive;
         ret = CntrPropReadHelperConst(cntr, CFG_FeaturePoGatePolarities, sizeof(uint32), &sptedPolarity, flag);
         bufErr = FillBuffer(sptedPolarity, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_PoOutCountOfCounters:
      {
         ret  = FillBuffer((uint32)0, chCount, buffer, size);
      }
      break;
//-----------------------------------------------------
// Up-down count, V1.6 added
//-----------------------------------------------------
   case CFG_FeatureSnapEventSources:
      {
         DefCopyT((uint32)-1, buffer, size);
         size = 0;
      }
      break;
   case CFG_FeatureUdCountingTypes:
      {
         DefCopyT((uint32)CountingNone, buffer, size);
         size = 0;
      }
      break;
   case CFG_FeatureUdInitialValues:
      {
         ret = DefCopyT((uint32)0, buffer, size);
      }
      break;
   case CFG_UdCountingTypeOfCounters:
      {
         uint32 cntType = CountingNone;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureUdCountingTypes, sizeof(uint32), &cntType, flag);
         bufErr = FillBuffer(cntType, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_UdInitialValueOfCounters:
      {
         uint32 initVal = CountingNone;
         ret = CntrPropReadHelperConst(cntr, CFG_FeatureUdInitialValues, sizeof(uint32), &initVal, flag);
         bufErr = FillBuffer(initVal, chCount, buffer, size);
         if (ret != ErrorPropNotSpted) { ret = bufErr; }
      }
      break;
   case CFG_UdCountValueResetTimesByIndexs:
      {
         ret = FillBuffer((uint32)0, chCount, buffer, size);
      }
      break;

   // -------------------------------------------------------
   // others
   // -------------------------------------------------------
   default: // Not supported
      if (buffer) { memset(buffer, 0, size); }
      size = 0;
      break;
   }

   if (ret == ErrorPropNotSpted) { size = 0; }
   return ret;
}

static
ErrorCode CntrPropWriteHelper(DaqModule &cntr, PropertyId id, uint32 size, void *buffer, uint32 flag)
{
   assert(cntr.getType() == DaqCounter);
   if (cntr.getMode() == ModeRead) { return ErrorPrivilegeNotHeld; }

   ErrorCode ret = cntr.PropAccess(id, size, buffer, flag);
   if (ret != ErrorPropNotSpted && ret != ErrorPropReadOnly) { return ret; }

   // Does the property be implemented in integrated layer ? 
   // Or is the property be implemented as read-only by underlying driver ?
   uint32 xsize = 0;
   ret = CntrPropReadHelper(cntr, id, xsize, NULL, PropFlag::OpGet);
   if (xsize == 0)   { return ErrorPropNotSpted; }
   if (size < xsize) { return ErrorBufferTooSmall; }

   // Do further check to get a proper error code.
   void *xbuf = _malloca(xsize);
   CntrPropReadHelper(cntr, id, xsize, xbuf, PropFlag::OpGet);

   if(memcmp(xbuf,  buffer, xsize) == 0) {
      ret = Success; // return success if the property value not changed.
   } else {
      switch (id)
      {
      case CFG_ChipClkPolarityOfCounters:
      case CFG_ChipGatePolarityOfCounters:
      case CFG_ChipOutSignalOfCounters:
      case CFG_ChipClkSourceOfCounters:
      case CFG_ChipGateSourceOfCounters:
      case CFG_ChipOperationModeOfCounters:
      case CFG_ChipOverCompareEnabledCounters:
      case CFG_ChipUnderCompareEnabledCounters:
      case CFG_ChipSignalCountingTypeOfCounters:
      case CFG_EcClkSourceOfCounters:
      case CFG_EcClkPolarityOfCounters:
      case CFG_EcGatePolarityOfCounters:
      case CFG_EcGateEnabledOfCounters:
      case CFG_OsClkSourceOfCounters:
      case CFG_OsClkPolarityOfCounters:
      case CFG_OsGateSourceOfCounters:
      case CFG_OsGatePolarityOfCounters:
      case CFG_OsOutSignalOfCounters:
      case CFG_TmrGateEnabledOfCounters:
      case CFG_TmrGatePolarityOfCounters:
      case CFG_TmrOutSignalOfCounters:
      case CFG_PoGateEnabledOfCounters:
      case CFG_PoGatePolarityOfCounters:
      case CFG_PoOutCountOfCounters:
      case CFG_FmMethodOfCounters:
      case CFG_UdCountingTypeOfCounters:
      case CFG_UdInitialValueOfCounters:
         ret = ErrorPropValueNotSpted;
         break;
      case CFG_FmCollectionPeriodOfCounters:
         ret = ErrorPropValueOutOfRange;
         break;
      default:
         ret = ErrorPropReadOnly;
         break;
      }
   }
   _freea(xbuf);

   return ret;
}
