#if defined(_WIN32) || defined(WIN32) 
#  define  _CRT_SECURE_NO_WARNINGS
#  include <Windows.h>
#  include <win32/registry.h>
#  include <win32/reg_key.h>
#else
#  include <stdio.h>
#  include <dlfcn.h>
#  include <dirent.h>
#  include <sys/ioctl.h>
#  include <sys/mman.h>
#endif

#include <vector>
#include <map>
#include <string>
#include <assert.h>

#include <bdaq_util.h>
#include <property_id.h>
#include <device_adaptor.h>
#include <oal.h>
#include <lock.h>

#include "legacy/daqnavi3_adaptor.h"
#include "xdev_factory.h"

#if defined(_WIN32) || defined(WIN32)
#  define IsDeviceEntryKey(KeyName)\
   (_wcsnicmp(KeyName, DEVX_KEY_NAME_PREFIX, DEVX_KEY_NAME_PREFIXLEN) == 0)
#endif

ErrorCode XDevFactory::OpenDevice(int32 deviceNumber, AccessMode mode, DaqDevModule *& device)
{
   device = NULL;

   // 
   // Get driver name.
   std::string drvName;
   ErrorCode ret = GetDriverName(deviceNumber, drvName);
   if (ret != Success) { return ret; }
   assert(drvName.length());

   // 
   // Load driver 
   XDevFactory &_this = getInstance();
   DriverEntry *entry = NULL;

   ret = LoadDriver(drvName, entry);
   if (ret != Success) {
      return ret; 
   }

   //
   // Open device 
   switch(entry->IntfType)
   {
   case IntfType::DN3_API:
      {
         DN3Api::DN3Device *dev = new DN3Api::DN3Device();
         ret = dev->Initialize(deviceNumber, mode, entry->DrvApis);
         if (ret != Success) {
            dev->Close();
         } else {
            device = dev;
         }
      }
      break;
   case IntfType::DN3_OBJ:
      {
         DN3Obj::DN3Device *dev = new DN3Obj::DN3Device();
         ret = dev->Initialize(deviceNumber, mode, entry->ObjCreate);
         if (ret != Success) {
            dev->Close();
         } else {
            device = dev;
         }
      }
      break;
   case IntfType::DN4_OBJ:
      {
         assert(entry->DN4ObjCreate);
         ret = entry->DN4ObjCreate(deviceNumber, mode, &device);
         if (ret == Success && mode == ModeWriteWithReset) {
            device->Reset(1);
         }
      }
      break;
   default: 
      ret = ErrorDeviceUnrecognized; 
      break;
   }

   // Update driver entry 
   if (ret == Success) {
      AutoLock<MutexLock> lock(_this.m_lock);
      
      ++_this.m_drivers[drvName]->RefCount; 
   }

   return ret;
}

void XDevFactory::CloseDevice(DaqDevModule * device)
{
   assert(device);

   // Get driver name
   std::string drvName;
   GetDriverName(device->getNumber(), drvName);

   // Close the underlying device
   device->Close();

   // For performance reason, here we just decrement the 
   // reference count of the driver, but not unload it immediately.
   // To unload all idle driver (reference count is zero),
   // Please call 'UnloadIdleDriver'
   XDevFactory &_this = getInstance();
   {
      AutoLock<MutexLock> lock(_this.m_lock);

      DriverTable::iterator itor = _this.m_drivers.find(drvName);
      if (itor != _this.m_drivers.end()) {
         --itor->second->RefCount;
      }
   }
}

void XDevFactory::UnloadIdleDriver()
{
   XDevFactory &_this = getInstance();
   AutoLock<MutexLock> lock(_this.m_lock);

   DriverTable::iterator itor = _this.m_drivers.begin();
   while (itor != _this.m_drivers.end()) {
      DriverTable::iterator curr = itor++; 
      DriverEntry * entry = curr->second;
      if (!entry->RefCount) {
         UnloadDriver(entry);
         _this.m_drivers.erase(curr);
      } 
   }
}

XDevFactory & XDevFactory::getInstance()
{
   static XDevFactory instance;
   return instance;
}

ErrorCode XDevFactory::LoadDriver(std::string const &driverName, DriverEntry * &entry)
{
   XDevFactory &_this = getInstance();
   entry = NULL;

   //
   // If the driver is loaded, just return it.
   {
      AutoLock<MutexLock> lock(_this.m_lock);

      DriverTable::const_iterator itor = _this.m_drivers.find(driverName); 

      // OK, we found it.
      if (itor != _this.m_drivers.end()) {
         entry = itor->second;
         return Success; 
      }
   }
   
   //
   // Load the new driver now!
   //

#if defined(_WIN32) || defined(WIN32)
   
   HMODULE lib = LoadLibraryA(driverName.c_str());
   if (lib == NULL) {
      return ErrorDriverNotFound;
   }

   // DN4: Get driver entry point
   //
   DN4FNOBJECTCREATE objectCreate = (DN4FNOBJECTCREATE)GetProcAddress(lib, "BDaqDeviceObjectCreate");
   if (objectCreate != NULL) {
      // allocate memory for DriverEntry 
      entry = (DriverEntry *)malloc(sizeof(DriverEntry));
      if (entry == NULL) {
         FreeLibrary(lib);
         return ErrorMemoryNotEnough;
      }

      memset(entry, 0, sizeof(DriverEntry));
      entry->IntfType  = IntfType::DN4_OBJ;
      entry->RefCount  = 0;
      entry->LibHandle = lib;
      entry->DN4ObjCreate = objectCreate;
   }
   else
   {
      // DN3: Get driver entry point
      DN3Api::FNDRIVERENTRY drvEntry = (DN3Api::FNDRIVERENTRY)GetProcAddress(lib, DN3Api::nameofFnDriverEntry());
      if (drvEntry == NULL) {
         FreeLibrary(lib);
         return ErrorDriverVerMismatch;
      }

      // allocate memory for DriverEntry 
      entry = (DriverEntry *)malloc(sizeof(DriverEntry) + sizeof(DN3Api::DriverApis));
      if (entry == NULL) {
         FreeLibrary(lib);
         return ErrorMemoryNotEnough;
      }

      memset(entry, 0, sizeof(DriverEntry) + sizeof(DN3Api::DriverApis));
      entry->IntfType  = IntfType::DN3_API;
      entry->RefCount  = 0;
      entry->LibHandle = lib;
      entry->DrvApis   = (DN3Api::DriverApis *)(entry + 1);

      // set default api table: all api return 'NotSpted'
      extern DN3Api::DriverApis const g_dn3apiStub;
      *entry->DrvApis = g_dn3apiStub;

      // get driver api table
      if (drvEntry(sizeof(DN3Api::DriverApis), entry->DrvApis) != Success) {
         free(entry);
         FreeLibrary(lib);
         return ErrorDriverVerMismatch;
      }
   }

#else
   
   void * lib = dlopen(driverName.c_str(), RTLD_LAZY);
   if (lib == NULL) {
      return ErrorDriverNotFound;
   }

   DN4FNOBJECTCREATE objectCreate = (DN4FNOBJECTCREATE)dlsym(lib, "BDaqDeviceObjectCreate");
   if (objectCreate != NULL) {
      // allocate memory for DriverEntry 
      entry = (DriverEntry *)malloc(sizeof(DriverEntry));
      if (entry == NULL) {
         dlclose(lib);
         return ErrorMemoryNotEnough;
      }

      memset(entry, 0, sizeof(DriverEntry));
      entry->IntfType  = IntfType::DN4_OBJ;
      entry->RefCount  = 0;
      entry->LibHandle = lib;
      entry->DN4ObjCreate = objectCreate;
   } else {
      DN3Obj::FNOBJECTCREATE dn3ObjCreate = (DN3Obj::FNOBJECTCREATE)dlsym(lib, DN3Obj::nameofFnObjectCreate());
      if (dn3ObjCreate == NULL) {
         dlclose(lib);
         return ErrorDriverVerMismatch;
      }
      // allocate memory for DriverEntry 
      entry = (DriverEntry *)malloc(sizeof(DriverEntry));
      if (entry == NULL) {
         dlclose(lib);
         return ErrorMemoryNotEnough;
      }

      memset(entry, 0, sizeof(DriverEntry));
      entry->IntfType  = IntfType::DN3_OBJ;
      entry->RefCount  = 0;
      entry->LibHandle = lib;
      entry->ObjCreate = dn3ObjCreate;
   }

#endif

   // Cache the driver into driver table
   { 
      AutoLock<MutexLock> lock(_this.m_lock);
      _this.m_drivers[driverName] = entry; 
   }

   return Success;
}

void XDevFactory::UnloadDriver(DriverEntry * &entry)
{
   if (entry->LibHandle) {
#if defined(_WIN32) || defined(WIN32)
      FreeLibrary((HMODULE)entry->LibHandle);
#else
      dlclose(entry->LibHandle);
#endif
   }

   free(entry);
}

ErrorCode XDevFactory::GetDriverName(int32 deviceNumber, std::string &driverName)
{
   char ansiBuf[260] = {0};
   uint32 len = 260;
   
   // Is it a DaqNavi4 device (DN4)
   ErrorCode ret = QueryDeviceFieldADB(deviceNumber, DFN_DRVNAME, &len, ansiBuf);
   if (ret == Success) {
#if defined(_WIN32) || defined(WIN32)
      driverName  = ansiBuf;
      driverName += ".dll";
#else
      driverName  = "lib";
      driverName += ansiBuf;
      driverName += ".so";
#endif
   }

   return ret;
}

void * XDevFactory::GetDriverFunc(std::string const &driverName, std::string const funcName)
{
   DriverEntry *entry = NULL;

   //
   // Load driver at first
   ErrorCode ret = LoadDriver(driverName, entry);
   if (ret != Success) {
      return NULL;
   }

   //
   // find the function address
#if defined(_WIN32) || defined(WIN32)
   void *fn = (void *)GetProcAddress(HMODULE(entry->LibHandle), funcName.c_str());
#else
   void *fn = (void *)dlsym(entry->LibHandle, funcName.c_str());
#endif

   return fn;
}

// Helpers
ErrorCode XDevFactory::EnumDeviceInfo(int32 deviceIndex, int32 &active, int32 & deviceNumber, wchar_t desc[MAX_DEVICE_DESC_LEN])
{
   char xbuf[260] = {0};

   ErrorCode ret = EnumDeviceInfoDB(deviceIndex, active, deviceNumber, desc ? xbuf : NULL);
   if (ret != Success) {
      return ret;
   }

   if (desc) { 
      mbstowcs(desc, xbuf, MAX_DEVICE_DESC_LEN); 
   }
   return Success;
}
