#ifndef _DAQNAVI_DEVICE_MANAGER_H
#define _DAQNAVI_DEVICE_MANAGER_H

class XDevManager 
{
   struct DeviceEntry {
      int32        RefCount;
      DaqDevModule *Device;
   };
   typedef std::map<uint32, DeviceEntry> DeviceTable; // indexed by device number & access mode

public:   
   static ErrorCode OpenDevice(int32 deviceNumber, AccessMode mode, DaqDevModule * &device)
   {
      // 'Singleton': all devices which own same device number and access mode
      //  will reference to the same object.
      ErrorCode ret = Success;
      device = NULL;
           
      uint32 key = GetDeviceKey(deviceNumber, mode);
      XDevManager &_this = getInstance();

      {
         AutoLock<MutexLock> lock(_this.m_lock);
         DeviceTable::iterator itor = _this.m_devices.find(key);
         if (itor != _this.m_devices.end())  {
            ++itor->second.RefCount;
            device = itor->second.Device;
         } else{
            ret = XDevFactory::OpenDevice(deviceNumber, mode, device);
            // Cache device into device table
            if (ret == Success) {
               DeviceEntry entry = { 1, device};
               _this.m_devices[key] = entry;
            }
         }
      }

      return ret;
   }

   static ErrorCode OpenDevice(wchar_t const deviceDescription[], AccessMode mode, DaqDevModule * &device)
   {
      device = NULL;
      if (deviceDescription == NULL || deviceDescription[0] == 0) { 
         return ErrorBufferIsNull; 
      }

      wchar_t   curDevDesc[MAX_DEVICE_DESC_LEN];
      int32     deviceNumber;
      int32     deviceIndex = 0;
      int32     active      = 0;
      ErrorCode ret         = ErrorDeviceNotExist;

      do {
         XDevFactory::EnumDeviceInfo(deviceIndex, active, deviceNumber, curDevDesc);

         // Found the device 
         if (  active != 0
            && wcscmp(curDevDesc, deviceDescription) == 0
            && (ret = OpenDevice(deviceNumber, mode, device)) == Success) { 
            return Success;
         }

         ++deviceIndex;
      } while (deviceNumber != -1);

      return ret;
   }

   static void CloseDevice(DaqDevModule * device)
   {
      if (device == NULL){
         return;
      }

      uint32 key = GetDeviceKey(device->getNumber(), device->getMode());
      XDevManager &_this = getInstance();

      {
         AutoLock<MutexLock> lock(_this.m_lock);
         DeviceTable::iterator itor = _this.m_devices.find(key);
         if (itor != _this.m_devices.end())  {
            --itor->second.RefCount;
            if (itor->second.RefCount == 0){
               XDevFactory::CloseDevice(itor->second.Device);
               _this.m_devices.erase(itor);
            }
         }
      }
   }

   static bool IsValidDevicePtr(DaqDevModule *ptr)
   {
      XDevManager &_this = getInstance();
      
      {
         AutoLock<MutexLock> lock(_this.m_lock);
         DeviceTable::iterator itor = _this.m_devices.begin();
         while (itor != _this.m_devices.end())  {
            if (itor->second.Device == ptr) {
               return true;
            }
         }
      }

      return false;
   }

   static ErrorCode SynchronizeDeviceWithDB(int32 deviceNumber, uint32 productId, ModuleType const mtype[])
   {
      //
      // Open device
      DaqDevModule *devObj = NULL;
      ErrorCode ret = OpenDevice(deviceNumber, ModeWrite, devObj);
      if (ret != Success) { 
         return ret; 
      }

      // 
      // configure: DEVICE/AI/AO/DIO/CNTR...
      for (; *mtype != DaqAny; ++mtype) {      
         if (*mtype== DaqDevice) { 
            // configure device module
            PropEnumerateDB(productId, catofModule(DaqDevice, 0), DaqNaviPropEnumerateCallBack, (DaqModule*)devObj);

            // terminal board need somehow special treatment
            int32 board = WiringBoard;
            QueryDeviceFieldI32DB(deviceNumber, DFN_TERMBD, &board);
            DaqProperty(devObj).Set(CFG_DeviceTerminalBoardType, board);
         } else {  
            // configure each module of type 'mtype[x]'
            DaqModule *m = NULL;
            for (unsigned j = 0; devObj->GetModule(*mtype, j, m) == Success; ++j) {
               PropEnumerateDB(productId, catofModule(*mtype, j), DaqNaviPropEnumerateCallBack, m);
            }
         }
      }

      CloseDevice(devObj);
      return ret;
   }

private:
   XDevManager()          { }
   virtual ~XDevManager() { }

   static uint32 GetDeviceKey(int32 deviceNumber, AccessMode mode)
   {
      if (mode != ModeRead) { mode = ModeWrite; }
      return (uint32)deviceNumber | ((uint32)mode << 16);
   }

   static XDevManager & getInstance()
   {
      static XDevManager instance;
      return instance;
   }

   //
   // Configure the device using the property value stored in config DB
   static 
   void DaqNaviPropEnumerateCallBack(uint32 category, uint32 id, uint32 access,  uint32 size, void *buffer, void *userParam)
   {
      assert(userParam);
      if (access == PROPACCESS_READWRITE) {
         DaqProperty((DaqModule *)userParam).Write((PropertyId)id, size, buffer);
      }
   }

private:   
   MutexLock    m_lock;
   DeviceTable  m_devices;
};

#endif
