/*
 * misc.c
 *
 *  Created on: 2011-9-8
 *      Author: rocky
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/vmalloc.h>
#include <asm/io.h>

#include <bdaq_def.h>
#include <bdaq_kern.h>
#include <plx905x.h>

#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#endif

unsigned daq_ai_calc_log_chan_count(unsigned char *chType, unsigned count)
{
   unsigned logCount = count;
   unsigned i;

   for (i = 0; i < count; i += 2) {
      if (chType[i] == Differential) {
         --logCount;
      }
   }

   return logCount;
}

unsigned daq_ai_calc_phy_chan_range(
   unsigned char *chType,  unsigned count,
   unsigned      phyStart, unsigned logCount)
{
   unsigned nextChan;

   BUG_ON(phyStart >= count || logCount > count);

   if (chType[phyStart] == Differential) {
      phyStart &= ~0x1;
   }

   nextChan = phyStart;
   while (logCount--) {
      if (chType[nextChan] == Differential) {
         nextChan += 2;
      } else {
         ++nextChan;
      }
      nextChan %= count;
   }

   if (!nextChan) {
      nextChan = count - 1;
   } else {
      --nextChan;
   }

   return (nextChan << 8) | phyStart;
}

unsigned daq_plx905x_calc_sgl_length(unsigned data_len, unsigned sect_len)
{
   unsigned count = 0;

   if (sect_len >= PAGE_SIZE){
      count = data_len / PAGE_SIZE + ((data_len % PAGE_SIZE) != 0);
      if (sect_len % PAGE_SIZE) {
         count += data_len / sect_len;
      }
   } else {
      unsigned i, remains = 0;
      if (PAGE_SIZE % sect_len) {
         for (i = 0; i < data_len / PAGE_SIZE; i++) {
            count   += remains != 0;
            count   += (PAGE_SIZE - remains) / sect_len;
            remains  = (PAGE_SIZE - remains) % sect_len ;
            count   += remains != 0;
            remains  = sect_len - remains;
         }
      } else {
         count = (data_len / PAGE_SIZE) * (PAGE_SIZE / sect_len);
      }

      count += (remains != 0);
      data_len %= PAGE_SIZE;
      if (data_len > remains) {
         data_len -= remains;
         remains   = data_len % sect_len;
         count    += data_len / sect_len + (remains != 0);
      }
   }

   return count * sizeof(PLX_SGL_ENTRY);
}

int daq_plx905x_alloc_sgl_mem(unsigned sgl_len, daq_plx_sgl_t *sgl_mem)
{
#ifndef GFP_DMA32
#  define GFP_DMA32 0u
#endif

   struct page **pages;
   int           nr_pages;
   int           i;

   BUG_ON(!sgl_mem);

   if (!sgl_len || sgl_mem->pages) {
      return -EINVAL;
   }

   memset(sgl_mem, 0, sizeof(*sgl_mem));

   nr_pages = (sgl_len + ~PAGE_MASK) >> PAGE_SHIFT;
   pages    = kmalloc(nr_pages * sizeof(void *), GFP_KERNEL);
   if (pages == NULL) {
      return -ENOMEM;
   }

   for (i = 0; i < nr_pages; ++i) {
      pages[i] = alloc_page(GFP_KERNEL | GFP_DMA32);
      if (pages[i] == 0) {
         break;
      }
   }

   if (i == nr_pages) {
      sgl_mem->start = vmap(pages, nr_pages, VM_MAP, PAGE_KERNEL);
      if (sgl_mem->start != NULL) {
         sgl_mem->nr_pages = nr_pages;
         sgl_mem->pages    = pages;
         sgl_mem->count    = sgl_len / sizeof(PLX_SGL_ENTRY);
         sgl_mem->startPA  = page_to_phys(pages[0]);
         return 0;
      }      
   }

   while (--i >= 0) {
      __free_page(pages[i]);
   }

   kfree(pages);
   return -ENOMEM;
}

void daq_plx905x_free_sgl_mem(daq_plx_sgl_t *sgl_mem)
{
   int i;

   BUG_ON(!sgl_mem || (sgl_mem->nr_pages && !sgl_mem->pages));

   if (sgl_mem->nr_pages) {
      vunmap(sgl_mem->start);

      for (i = 0; i < sgl_mem->nr_pages; ++i){
         __free_page(sgl_mem->pages[i]);
      }

      kfree(sgl_mem->pages);
      memset(sgl_mem, 0, sizeof(*sgl_mem));
   }
}

static
PLX_SGL_ENTRY * daq_plx905x_alloc_entry(daq_plx_sgl_t *sgl_mem, unsigned *index)
{
   PLX_SGL_ENTRY *entry = NULL;
   unsigned pg_num;
   unsigned pg_off;

   if (*index < sgl_mem->count) {
      entry   = &sgl_mem->start[*index];
      *index += 1;

      pg_num = (*index * sizeof(PLX_SGL_ENTRY)) >> PAGE_SHIFT;
      pg_off = (*index * sizeof(PLX_SGL_ENTRY)) &  ~PAGE_MASK;
      if (pg_num >= sgl_mem->nr_pages) {
         pg_num = pg_off = 0;
      }

      *(__u32*)&entry->DescPtr = (__u32)page_to_phys(sgl_mem->pages[pg_num]) + pg_off;
   }

   return entry;
}

int daq_plx905x_build_sgl(
   daq_user_mem_t *data_mem, unsigned data_len, unsigned  sect_len,
   unsigned dev_addr, unsigned from_device, daq_plx_sgl_t *sgl_mem)
{
   struct page  **data_pages = data_mem->pages;
   PLX_SGL_ENTRY *entry      = NULL;
   unsigned       index      = 0;
   unsigned       bp         = sect_len;
   unsigned       desc_flag;
   unsigned       curLen;
   dma_addr_t     curPA;
   
#  define DUMP_SGL_ENTRY(entry)   \
   daq_trace((KERN_INFO "entry(%lx): 0:0x%x, %d, 0x%x\n", \
      (unsigned long)entry, entry->PciAddrLow, entry->DataLength, *(__u32*)&entry->DescPtr));

   BUG_ON(!sgl_mem  || !sgl_mem->nr_pages  || !sgl_mem->pages);
   BUG_ON(!data_mem || !data_mem->nr_pages || !data_mem->pages);

   if (from_device){
      desc_flag = PLX_DESC_IN_PCI | (PLX_DESC_DIR_FROM_DEV << 3);
   } else {
      desc_flag = PLX_DESC_IN_PCI | (PLX_DESC_DIR_TO_DEV << 3);
   }

   while (data_len) {
      curPA     = page_to_phys(*data_pages++);
      curLen    = x_min((unsigned)PAGE_SIZE, data_len);
      data_len -= curLen;

      while (curLen) {
         entry = daq_plx905x_alloc_entry(sgl_mem, &index);
         if (!entry) {
            return -EINVAL;
         }

         *(__u32*)&entry->DescPtr |= desc_flag;
         entry->LocAddress = dev_addr;
         PLX_SET_PCI_ADDR(entry, curPA);

         if (curLen >= bp) {
            entry->DescPtr.TermCountInt = 1;
            entry->DataLength = bp;
            curPA += bp;
            bp     = sect_len;
         } else {
            entry->DataLength  = curLen;
            bp                -= curLen;
         }

         curLen -= entry->DataLength;
         DUMP_SGL_ENTRY(entry);
      }
   }

   entry->DescPtr.TermCountInt = 1;
   entry->DescPtr.Address      = PLX_DESC_PTR_ADDR(sgl_mem->startPA);
   DUMP_SGL_ENTRY(entry); 

   sgl_mem->end = entry;

   return 0;
}

static unsigned char __dummy_buffer[128];
int daq_plx9054_flush_fifo(unsigned bridge_base, int dma_chan)
{
   AdxIoOutD(bridge_base, BR_PLX_INTCSR, 0);

   {
      PLX_DMA_MODE dmamode = {0};
      dmamode.LocBusWidth   = 1;
      dmamode.IntlWaitState = 3;
      dmamode.LocBurstEn    = 1;
      dmamode.LocAddrMode   = 1;
      AdxIoOutD(bridge_base, dma_chan ? BR_PLX_DMAMODE1 : BR_PLX_DMAMODE0, dmamode.Value);
   }

   AdxIoOutD(bridge_base, dma_chan ? BR_PLX_DMAPADR1 : BR_PLX_DMAPADR0, virt_to_phys(__dummy_buffer));

   AdxIoOutD(bridge_base, dma_chan ? BR_PLX_DMALADR1 : BR_PLX_DMALADR0, 0);

   AdxIoOutD(bridge_base, dma_chan ? BR_PLX_DMASIZ1 :BR_PLX_DMASIZ0, 128);

   {
      unsigned reg_offset = dma_chan ? BR_PLX_DMACSR1 : BR_PLX_DMACSR0;
      unsigned try_count = 0; 
      PLX_DMA_CSR dmacsr  = {0};
      dmacsr.Enable   = 1;
      dmacsr.ClearInt = 1;
      AdxIoOutB(bridge_base, reg_offset, dmacsr.Value);

      dmacsr.Start = 1;
      AdxIoOutW(bridge_base, reg_offset, dmacsr.Value);

      dmacsr.Value = AdxIoInB(bridge_base, reg_offset);
      while (!dmacsr.Done && try_count < 1000) {
         dmacsr.Value = AdxIoInB(bridge_base, reg_offset);
         ++try_count;
      }
      if (try_count >= 1000)
      {
         printk("Warning, clear FIFO out of time!\n");
      }
   }

   return 0;
}

unsigned daq_ver_string_to_integer(char * ver_str)
{
   unsigned ver = 0;
   unsigned   x = 0;

   BUG_ON(!ver_str);

   do {
      for (x = 0; *ver_str; ++ver_str) {
         if (*ver_str == '.') {
            ++ver_str;
            break;
         }
         x = x * 10 + (*ver_str - '0');
      }
      ver = (ver << 8) | (x & 0xff);
   } while (*ver_str);

   return ver;
}

/************************************************************************
Routine : daq_dev_schedule_work
************************************************************************/
typedef struct _daq_work {
   struct work_struct  work;
   daq_work_func_t     func;
   void               *device;
   unsigned long       arg1;
   unsigned long       arg2;
} daq_work_t;

static void daq_work_func(struct work_struct *work)
{
   daq_work_t *wk = container_of(work, daq_work_t, work);
   BUG_ON(!wk || !wk->func);

   wk->func(wk->device, wk->arg1, wk->arg2);
   
   kfree(wk);
}

int daq_dev_schedule_work(void *device, daq_work_func_t func, unsigned long arg1, unsigned long arg2)
{
   daq_work_t *wk;

   BUG_ON(!func);
   
   //
   // Queue a work item so that the callback function
   // can be executed in the context of a system thread.
   wk = kmalloc(sizeof(daq_work_t), GFP_KERNEL);
   if  (!wk) {
      return -ENOMEM;
   }

   DAQ_INIT_WORK(&wk->work, daq_work_func);
   wk->func   = func;
   wk->device = device;
   wk->arg1   = arg1;
   wk->arg2   = arg2;

   // Execute this work item.   
   schedule_work(&wk->work);

   return 0;
}


EXPORT_SYMBOL_GPL(daq_ai_calc_log_chan_count);
EXPORT_SYMBOL_GPL(daq_ai_calc_phy_chan_range);
EXPORT_SYMBOL_GPL(daq_plx905x_calc_sgl_length);
EXPORT_SYMBOL_GPL(daq_plx905x_alloc_sgl_mem);
EXPORT_SYMBOL_GPL(daq_plx905x_free_sgl_mem);
EXPORT_SYMBOL_GPL(daq_plx905x_build_sgl);
EXPORT_SYMBOL_GPL(daq_plx9054_flush_fifo);
EXPORT_SYMBOL_GPL(daq_ver_string_to_integer);
EXPORT_SYMBOL_GPL(daq_dev_schedule_work);
