
#include "std_decl.h"
#include <string>
#include <vector>
#include <property_id.h>
#include <sstream>
#include <iomanip>
#include <clocale> 

#include <sqlite3.h>
#include <sqlite_wrapper.h>

#if defined(_WIN32) || defined(WIN32)
#  include <io.h>
#  include <win32/registry.h>
#  include <win32/reg_key.h>
#endif

#if defined(_WIN32) || defined(WIN32)
#  define x_strncpy    strncpy
#  define x_stricmp    _stricmp
#  define x_wcsicmp    _wcsicmp
#  define x_wcstoui64  _wcstoui64
#  define x_strtoui64  _strtoui64
#  define x_vsnprintf  _vsnprintf
#  define x_snprintf   _snprintf
//
#  define KEY_ADVAN_PUBLIC          _T("SOFTWARE\\Advantech\\Public")
//
#  define IsDeviceEntryKey(KeyName) (_wcsnicmp(KeyName, DEVX_KEY_NAME_PREFIX, DEVX_KEY_NAME_PREFIXLEN) == 0)
#else
#  define x_strncpy    strncpy
#  define x_stricmp    strcasecmp
#  define x_wcsicmp    wcscasecmp
#  define x_wcstoui64  wcstoull
#  define x_strtoui64  strtoull
#  define x_vsnprintf  vsnprintf
#  define x_snprintf   snprintf
#endif


#define ELEM_DELIMITA    ','
#define ELEM_S_DELIMITA  ';'

inline char const * GetDBFilePath()
{
#if defined(_WIN32) || defined(WIN32)
   // win32 db path(default): c:/advantech/daqnavi/db/daqnavi.config.db 
   static char dbFile[MAX_PATH] = {0};
   RegKey key(HKEY_LOCAL_MACHINE, KEY_ADVAN_PUBLIC, KEY_READ | KEY_WOW64_64KEY);
   if (key.IsOpened()) {
      DWORD size = sizeof(dbFile);
      if (RegQueryValueExA(key.GetKey(), "RootPath", NULL, NULL, (LPBYTE)dbFile, &size) == ERROR_SUCCESS) {
         strcat(dbFile, "\\daqnavi\\db\\daqnavi.config.db");
      }
   } 
  
   // check file existence
   if (_access(dbFile, 0) != 0) {
      return NULL;
   }

   return dbFile;
#else // linux db path:
   char const *dbFile = "/var/lib/daq/daqnavi.config.db";
   if (access(dbFile, F_OK)) {
      return NULL;
   }
   return dbFile;
#endif
}

// singleton
static SQLiteWrapper & DaqNaviConfigDB(bool is_open = true)
{
   static SQLiteWrapper db;

   db.Lock();

   if (is_open && !db.IsOpened()) {
      char const *dbFile = GetDBFilePath();
      if (dbFile != NULL) {
         db.Open(dbFile);
      }
   }
   
   return db;
}

BDAQ_API
void BDAQCALL CloseDaqNaviConfigDB()
{
   SQLiteWrapper &db = DaqNaviConfigDB(false);
   db.Close();
   db.Unlock();
}

//--------------------------------------------------------------------------
//Helper method
//--------------------------------------------------------------------------
void FloatStringTrim(char *str)
{
   size_t last_nonezero = strlen(str) - 1;

   // Chop trailing zeros
   while (last_nonezero > 0 && str[last_nonezero] == '0') {
      --last_nonezero;
   }
   
   // Chop trailing dot
   if (str[last_nonezero] == '.') { 
      --last_nonezero;
   } 

   str[last_nonezero + 1] = 0;
};

char const * NextElement(char const * str, char delimit)
{
   char const *next = strchr(str, delimit);
   return next != NULL ? next + 1 : next;
}

bool IsHexString(char const *str)
{
   if (*str++ != '0') { 
      return false; 
   } 

   return *str == 'x' || *str == 'X';
}

uint32 CountOfElement(char const * str, char delimit)
{
   uint32 count = 0;
   char   const *p = str;

   while (p && *p) {
      ++count;
      p = NextElement(p, delimit);
   }

   return count;
}

uint32 SizeOfElement(uint32 type)
{
   uint32 size = 0;
   switch(type & 0xffff) 
   {
   case PT_I8:   size = sizeof(uint8);        break;
   case PT_I16:  size = sizeof(uint16);       break;
   case PT_I32:  size = sizeof(uint32);       break;
   case PT_I64:  size = sizeof(uint64);       break;
   case PT_F32:  size = sizeof(float);        break;
   case PT_F64:  size = sizeof(double);       break;
   case PT_WSTR: size = sizeof(wchar_t);      break;
   case PT_S_MI: size = sizeof(MathInterval); break;
   }

   return size;
}

void FillDeviceRecord(DEVICE_RECORD *rec, size_t size, SQLiteTable &table, int row)
{
   assert(rec != NULL);

   rec->DeviceNumber = table.Cell(row, DFI_DEVNUM, int(0));
   rec->ProductId    = table.Cell(row, DFI_PRODID, int(0));
   rec->OnSystem     = table.Cell(row, DFI_ACTIVE, int(0));
   rec->DriverSPEC   = table.Cell(row, DFI_DRVSPEC, int(0));

   x_strncpy(rec->DriverName, table.Cell(row, DFI_DRVNAME), 32);
   x_strncpy(rec->DevicePath, table.Cell(row, DFI_DEVPATH), 256);
   x_strncpy(rec->DeviceDesc, table.Cell(row, DFI_DESC),    64);

   if (size >= sizeof(DEVICE_RECORD_EX)) {
      ((DEVICE_RECORD_EX *)rec)->BoardId = table.Cell(row, DFI_BOARDID, int(0));
      ((DEVICE_RECORD_EX *)rec)->BusSlot = table.Cell(row, DFI_BUSSLOT, int(0));
   }
}

ErrorCode BDAQCALL PropValueToStringImpl(
   uint32      propId, 
   uint32      valSize, 
   void        *valBuf,
   std::string &str)
{
   // Get property data type
   uint32 type = 0;
   if (PropIdToAttrI32DB(propId, PROPATTR_TYPE, &type) != Success) {
      return ErrorPropNotSpted;
   }

   std::ostringstream oss;
   oss << std::setprecision(12); // Sets the precision for floating-point values.

   switch(type & 0xffff) 
   {
   case PT_I8:
      {
         uint8 *p = (uint8*)valBuf;
         for(; p < ((uint8*)valBuf) + valSize; ++p) {
            oss << (uint16)*p << ELEM_DELIMITA; // The ostringstream leaks the operator for 'BYTE', uses 'unsigned short' instead. 
         }
      }
      break;
   case PT_I16:
      {
         uint16 *p = (uint16*)valBuf;
         for(; (uint8*)p < ((uint8*)valBuf) + valSize; ++p) {
            oss << *p << ELEM_DELIMITA;
         }
      }
      break;
   case PT_I32:
      {
         uint32 *p = (uint32*)valBuf;
         for(; (uint8*)p < ((uint8*)valBuf) + valSize; ++p) {
            oss << *p << ELEM_DELIMITA;
         }
      }
      break;
   case PT_I64:
      {
         uint64 *p = (uint64*)valBuf;
         for(; (uint8*)p < ((uint8*)valBuf) + valSize; ++p) {
            oss << *p << ELEM_DELIMITA;
         }
      }
      break;
   case PT_F32:
      {
         float *p = (float*)valBuf;
         for(; (uint8*)p < ((uint8*)valBuf) + valSize; ++p) {
            oss << *p << ELEM_DELIMITA;
         }
      }
      break;
   case PT_F64:
      {
         double *p = (double*)valBuf;
         for(; (uint8*)p < ((uint8*)valBuf) + valSize; ++p) {
            oss << *p << ELEM_DELIMITA;
         }
      }
      break;
   case PT_WSTR: 
      {
         // The ostringstream leaks the operator for 'wchar_t', uses 'write' method instead. 
         XBuffer txtBuf;
         uint32 len = (uint32)wcslen((wchar_t *)valBuf) * 2 + 1;
         wcstombs(txtBuf.Buffer(len), (wchar_t *)valBuf, len);
         oss.write(txtBuf.Buffer(), len);
      }
      break;
   case PT_S_MI:
      {
         MathInterval *p = (MathInterval*)valBuf;
         for(; (uint8*)p < ((uint8*)valBuf) + valSize; ++p) {
            // {Type, Min, Max}
            oss << p->Type << ELEM_DELIMITA\
               << p->Min << ELEM_DELIMITA\
               << p->Max << ELEM_S_DELIMITA;
         }
      }
      break;
   default:
      assert(false);
      return ErrorPropNotSpted;
   }

   str = oss.str();
   if (!str.empty()) {
      if (*str.rbegin() == ELEM_DELIMITA || *str.rbegin() == ELEM_S_DELIMITA) {
         str.erase(str.length() - 1);
      }
   }
   
   return Success;
}

ErrorCode BDAQCALL PropStringToValueImpl(
   uint32     propId, 
   char const *strBuf,
   uint32     *valSize, 
   void       *valBuf)
{
   if (!strBuf || strBuf[0] == 0) {
      return ErrorParamNotSpted;
   }

   // Get property data type
   uint32 type = 0;
   if (PropIdToAttrI32DB(propId, PROPATTR_TYPE, &type) != Success) {
      return ErrorPropNotSpted;
   }

   // Calculate elements count & size
   uint32 size = SizeOfElement(type);
   uint32 count = 0;
   if (type == PT_WSTR) {
      count = (uint32)strlen(strBuf) + 1;
   } else if (type == PT_S_MI) {
      count = CountOfElement(strBuf, ELEM_S_DELIMITA);
   } else {
      count = CountOfElement(strBuf, ELEM_DELIMITA);
   }

   assert(valSize);
   if (valBuf == NULL || *valSize < count * size) {
      *valSize = count * size;
      return valBuf == NULL ? Success : ErrorBufferTooSmall; 
   }

   // Convert string
   switch(type & 0xffff) 
   {
   case PT_I8:
      {
         uint8 *p = (uint8*)valBuf;
         for(; strBuf && *strBuf; strBuf = NextElement(strBuf, ELEM_DELIMITA)) {
            *p++ = (uint8)strtoul(strBuf, NULL, IsHexString(strBuf) ? 16 : 10);
         }
      }
      break;
   case PT_I16:
      {
         uint16 *p = (uint16*)valBuf;
         for(; strBuf && *strBuf; strBuf = NextElement(strBuf, ELEM_DELIMITA)) {
            *p++ = (uint16)strtoul(strBuf, NULL, IsHexString(strBuf) ? 16 : 10);
         }
      }
      break;
   case PT_I32:
      {
         uint32 *p = (uint32*)valBuf;
         for(; strBuf && *strBuf; strBuf = NextElement(strBuf, ELEM_DELIMITA)) {
            *p++ = strtoul(strBuf, NULL, IsHexString(strBuf) ? 16 : 10);
         }
      }
      break;
   case PT_I64:
      {
         uint64 *p = (uint64*)valBuf;
         for(; strBuf && *strBuf; strBuf = NextElement(strBuf, ELEM_DELIMITA)) {
            *p++ = x_strtoui64(strBuf, NULL, IsHexString(strBuf) ? 16 : 10);
         }
      }
      break;
   case PT_F32:
      {
         float *p = (float*)valBuf;
         for(; strBuf && *strBuf; strBuf = NextElement(strBuf, ELEM_DELIMITA)) {
            *p++ = (float)strtod(strBuf, NULL);
         }
      }
      break;
   case PT_F64:
      {
         double *p = (double*)valBuf;
         for(; strBuf && *strBuf; strBuf = NextElement(strBuf, ELEM_DELIMITA)) {
            *p++ = strtod(strBuf, NULL);
         }
      }
      break;
   case PT_WSTR: 
      {
         mbstowcs((wchar_t *)valBuf, strBuf, count);
      }
      break;
   case PT_S_MI:
      {
         MathInterval *p = (MathInterval*)valBuf;
         for(; strBuf && *strBuf; strBuf = NextElement(strBuf, ELEM_S_DELIMITA)) {
            // Type
            p->Type = strtoul(strBuf, NULL, 10);
            strBuf = NextElement(strBuf, ELEM_DELIMITA);

            // Min
            p->Min = strtod(strBuf, NULL);
            strBuf = NextElement(strBuf, ELEM_DELIMITA);

            // Max
            p->Max = strtod(strBuf, NULL);
         }
      }
      break;
   default:
      assert(false);
      return ErrorPropNotSpted;
   }

   *valSize = count * size;

   return Success;
}

//--------------------------------------------------------------------------
// APIs
//--------------------------------------------------------------------------
BDAQ_API
ErrorCode BDAQCALL EnumToString(
   wchar_t const *enumType,
   uint32         enumValue,
   uint32         stringLen,
   wchar_t       *stringBuf)
{
   // init string buffer
   if (stringBuf) { 
      stringBuf[0] = 0; 
   }

   // Open db
   SQLiteScopedPtr db(DaqNaviConfigDB());

   // prepare query string
   if (x_wcsicmp(enumType, L"PropertyId") == 0) { // PropertyId --> Description   
      db->Prepare("SELECT Description FROM prop_attr WHERE ID = %d;", enumValue);    
   } else { // Other Enum
      db->Prepare("SELECT Text FROM enum_string WHERE Category = \"%ls\" AND ID = %d;", enumType, enumValue);
   }

   // query
   if (db->Step() != SQLITE_ROW) {
      return ErrorParamNotSpted;
   } 

   char const *text = db->Column(0);
   uint32 len = (uint32)strlen(text) + 1;
   if (stringLen < len) {
      return ErrorBufferTooSmall;
   } else {
      mbstowcs(stringBuf, text, len);
   }

   return Success;
}

BDAQ_API
ErrorCode BDAQCALL StringToEnum(
   wchar_t const *enumType,
   wchar_t const *enumString,
   uint32        *enumValue)
{
   // Open db
   SQLiteScopedPtr db(DaqNaviConfigDB());

   // prepare query string
   if (x_wcsicmp(enumType, L"PropertyId") == 0) { // PropertyId --> Description
      db->Prepare("SELECT ID FROM prop_attr WHERE Description = \"%ls\";", enumString);      
   } else { // Other Enum
      db->Prepare("SELECT ID FROM enum_string WHERE Category = \"%ls\" AND Text = \"%ls\";", enumType, enumString);
   }

   // query
   if (db->Step() != SQLITE_ROW) {
      return ErrorParamNotSpted;
   } 

   *enumValue = db->Column(0, int(0));
   
   return Success;
}

BDAQ_API
ErrorCode BDAQCALL QueryVrgInfoDB(
   uint32       vrg, 
   uint32       *unit, 
   MathInterval *range, 
   uint32       sizeofDesc, 
   wchar_t      *desc)
{
   // Open db
   SQLiteScopedPtr db(DaqNaviConfigDB());

   // prepare query string
   db->Prepare("SELECT * FROM vrg_info WHERE ID = %d;", vrg);      

   // query
   if (db->Step() != SQLITE_ROW) {
      return ErrorParamNotSpted;
   } 
  
   // retrieve the values, note: skip the first column(ID)
   if (unit)  { *unit = db->Column(1, int(0));  }

   if (range) { 
      range->Type = LCBROS;   // LeftClosedBoundary | RightOpenBoundary
      range->Min  = db->Column(2, double(0));
      range->Max  = db->Column(3, double(0));
   }

   if (desc) {
      char const *text = db->Column(4);
      size_t       len = strlen(text) + 1;
      if (sizeofDesc / sizeof(wchar_t) >= len) {
         mbstowcs(desc, text, len);
      }
   }
   
   return Success;
}

BDAQ_API
ErrorCode BDAQCALL PropIdToAttrDB(uint32 propId, char const *field, uint32 *strLen, char *strBuf)
{
   assert(field && strLen && strBuf);

   // Open db
   SQLiteScopedPtr db(DaqNaviConfigDB());

   // prepare query string
   db->Prepare("SELECT \"%s\" FROM prop_attr WHERE ID = %d;", field, propId);    

   // query
   if (db->Step() != SQLITE_ROW) {
      return ErrorPropNotSpted;
   }
   
   char const *text = db->Column(0);
   uint32 len = (uint32)strlen(text);
   if (!text || !len) {
      return ErrorConfigDataLost;
   }

   if (*strLen < len + 1) {
      *strLen = len + 1;
      return ErrorBufferTooSmall;
   }

   *strLen = len + 1;
   x_strncpy(strBuf, text, len + 1);
   return Success;
}

BDAQ_API
ErrorCode BDAQCALL PropIdToAttrI32DB(uint32 propId, char const *field, uint32 *value)
{
   assert(field && value);

   // Open db
   SQLiteScopedPtr db(DaqNaviConfigDB());

   // prepare query string
   db->Prepare("SELECT \"%s\" FROM prop_attr WHERE ID = %d;", field, propId);    

   // query
   if (db->Step() != SQLITE_ROW) {
      return ErrorPropNotSpted;
   }

   *value = db->Column(0, int(0));
   return Success;
}
BDAQ_API
ErrorCode BDAQCALL PropSymToAttrDB(
   char const *symbol, 
   char const *field, 
   uint32     *strLen, 
   char       *strBuf)
{
   assert(symbol && field && strLen && strBuf);

   // Open db
   SQLiteScopedPtr db(DaqNaviConfigDB());

   // prepare query string
   db->Prepare("SELECT \"%s\" FROM prop_attr WHERE Symbol = \"%s\";", field, symbol);    

   // query
   if (db->Step() != SQLITE_ROW) {
      return ErrorPropNotSpted;
   }

   char const *text = db->Column(0);
   uint32 len = (uint32)strlen(text);
   if (!text || !len) {
      return ErrorConfigDataLost;
   }

   if (*strLen < len + 1) {
      *strLen = len + 1;
      return ErrorBufferTooSmall;
   }

   *strLen = len + 1;
   x_strncpy(strBuf, text, len + 1);
   return Success;
}

BDAQ_API
ErrorCode BDAQCALL PropSymToAttrI32DB(char const *symbol, char const *field, uint32 *value)
{
   assert(symbol && field && value);

   // Open db
   SQLiteScopedPtr db(DaqNaviConfigDB());

   // prepare query string
   db->Prepare("SELECT \"%s\" FROM prop_attr WHERE Symbol = \"%s\";", field, symbol);    

   // query
   if (db->Step() != SQLITE_ROW) {
      return ErrorPropNotSpted;
   }

   *value = db->Column(0, int(0));
   return Success;
}

BDAQ_API
ErrorCode BDAQCALL PropEnumerateDB(
   uint32           productId,
   int32            category,
   PropEnumCallback enumCallBack,
   void             *userParam)
{
   if (!enumCallBack) {
      return ErrorParamNotSpted;
   }

   SQLiteScopedPtr db(DaqNaviConfigDB());
   SQLiteTable table(*db);

   if(-1 == category) {
      table.Load("SELECT * FROM prop_dev_0x%x;", productId, category);
   } else {
      table.Load("SELECT * FROM prop_dev_0x%x WHERE Category = %d;", productId, category);
   }
   
   XBuffer valBuf, txtBuf;
   uint32     propId;
   uint32     size;
   uint32     access;
   char const *text;

   for (int i = 0; i < table.Rows(); ++i) {
      propId = table.Cell(i, 1, int(0));
      access = table.Cell(i, 2, int(0));
      size   = table.Cell(i, 3, int(0));

      // Copy the 'value' into local buffer due to it will 
      // be destroyed when the next sqlite3 command executes.
      text = table.Cell(i, 4);
      strcpy(txtBuf.Buffer((uint32)strlen(text) + 1), text);

      if (size == 0) { // for 'wchar_t' property
         size = (uint32)(strlen(text) + 1) * sizeof(wchar_t);
      }

      if (PropStringToValueA(propId, txtBuf.Buffer(), &size, valBuf.Buffer(size)) == Success) {
         enumCallBack(category, propId, access, size, valBuf.Buffer(), userParam);
      }
   }

   return Success;
}

BDAQ_API
ErrorCode BDAQCALL PropReadValueDB(
   uint32 productId, 
   uint32 category, 
   uint32 propId, 
   uint32 *bufSize, 
   void   *buffer)
{
   // Open DB
   SQLiteScopedPtr db(DaqNaviConfigDB());

   // try to query the property value from device's property table
   db->Prepare("SELECT Size, Value FROM prop_dev_0x%x WHERE Category = %d AND ID = %d;", productId, category, propId);    
   if (db->Step() != SQLITE_ROW) {
      return ErrorPropNotSpted;
   }

   int         size = db->Column(0, int(0)); 
   char const *text = db->Column(1);

   if (size == 0) { // for 'wchar_t' property
      size = (int)(strlen(text) + 1) * sizeof(wchar_t);
   }

   assert(bufSize);
   if (!buffer || *bufSize < unsigned(size)) {
      *bufSize = size;
      return buffer ? ErrorBufferTooSmall : Success;
   }

   //
   // Note:
   //    We must copy 'value' into local buffer because 
   //    it points to the buffer allocated by sqlite3 and  
   //    will be reused when the next sqlite3 command executes.
   XBuffer txtlBuf;
   strcpy(txtlBuf.Buffer((uint32)strlen(text) + 1), text);
   return PropStringToValueA(propId, txtlBuf.Buffer(), bufSize, buffer);
}

BDAQ_API
ErrorCode BDAQCALL PropWriteValueDB(
   uint32 productId, 
   uint32 category, 
   uint32 propId, 
   uint32 bufSize, 
   void   *buffer)
{
   if (!buffer) {
      return ErrorBufferIsNull;
   }

   // Open DB
   SQLiteScopedPtr db(DaqNaviConfigDB());

   // The Property is writable ?
   uint32 attr = 0;
   PropIdToAttrI32DB(propId, PROPATTR_ATTR, &attr);
   if ((attr & Writable) == 0) {
      return ErrorPropReadOnly;
   }
   
   // Query value length from device's property table
   db->Prepare("SELECT Size FROM prop_dev_0x%x WHERE Category = %d AND ID = %d;", productId, category, propId);    
   if (db->Step() != SQLITE_ROW) {
      return ErrorFuncNotSpted;
   }

   int size = db->Column(0, int(0)); 
   if (bufSize < unsigned(size)) {
      return ErrorBufferTooSmall;
   }

   // Limit the user data size to 'Size' of DB 
   // because the user may provide a bigger buffer than required with garbage value filled.
   if (size) {
      bufSize = size; 
   }

   // Convert the property value to string
   XBuffer txtBuf;
   uint32  strLen = txtBuf.Size();

   // Convert the value with a default length text buffer
   ErrorCode ret = PropValueToStringA(propId, bufSize, buffer, &strLen, txtBuf.Buffer());
   if (ret != Success) { // Convert again with required text buffer length
      ret = PropValueToStringA(propId, bufSize, buffer, &strLen, txtBuf.Buffer(strLen + 1)); 
   }
   if (ret != Success) {
      return ret;
   }

   // Update device property table
   db->Prepare("UPDATE prop_dev_0x%x SET Value = \"%s\" WHERE Category = %d AND ID = %d;", 
      productId, txtBuf.Buffer(), category, propId);    

   return db->Step() == SQLITE_DONE ? Success : ErrorUndefined;
}

BDAQ_API
ErrorCode BDAQCALL PropValueToStringA(
   uint32  propId, 
   uint32  valSize, 
   void   *valBuf,
   uint32 *strLen,
   char   *strBuf)
{
   assert(strLen);

   if (!valBuf || !valSize) {
      return ErrorParamNotSpted;
   }


   std::string str;
   PropValueToStringImpl(propId, valSize, valBuf, str);

   uint32 orgLen = *strLen;
   *strLen = (uint32)str.length() + 1;

   if (strBuf != NULL && orgLen >= str.length() + 1) {
      strcpy(strBuf, str.c_str());
      return Success;
   }

   return strBuf != NULL ? ErrorBufferTooSmall : Success;
}

BDAQ_API
ErrorCode BDAQCALL PropValueToString(
   uint32   propId, 
   uint32   valSize, 
   void    *valBuf,
   uint32  *strLen,
   wchar_t *strBuf)
{
   assert(strLen);
   
   if (!valBuf || !valSize) {
      return ErrorParamNotSpted;
   }

   std::string str;
   PropValueToStringImpl(propId, valSize, valBuf, str);

   uint32 orgLen = *strLen;
   *strLen = (uint32)str.length() + 1;

   if (strBuf != NULL && orgLen >= str.length() + 1) {
      mbstowcs(strBuf, str.c_str(), str.length() + 1);
      return Success;
   }

   return strBuf != NULL ? ErrorBufferTooSmall : Success;
}

BDAQ_API
ErrorCode BDAQCALL PropStringToValueA(
   uint32     propId, 
   char const *strBuf,
   uint32     *valSize, 
   void       *valBuf)
{
   return PropStringToValueImpl(propId, strBuf, valSize, valBuf);
}

BDAQ_API
ErrorCode BDAQCALL PropStringToValue(
   uint32        propId, 
   wchar_t const *strBuf,
   uint32        *valSize, 
   void          *valBuf)
{
   uint32 len = (uint32)wcslen(strBuf) * 2 + 1;
   XBuffer txtBuf;
   wcstombs(txtBuf.Buffer(len), strBuf, len);

   return PropStringToValueImpl(propId, txtBuf.Buffer(), valSize, valBuf);
}

//---------------------------------------------------------------------
// device information DB field access
//
BDAQ_API
ErrorCode BDAQCALL EnumDeviceInfoDB(
   int32 &deviceIndex, 
   int32 &active,
   int32 &deviceNumber, 
   char  desc[MAX_DEVICE_DESC_LEN])
{
   active = 0;
   deviceNumber = -1;
   if (desc) { desc[0] = 0; }

   // Open db
   SQLiteScopedPtr db(DaqNaviConfigDB());
   SQLiteTable     table(*db);

   // query device number
   table.Load("SELECT DeviceNumber, Description, Active FROM device_map;");      
   for (int i = 0; i < table.Rows(); ++i) {
      if (!deviceIndex) {
         deviceNumber = table.Cell(i, 0, int(0));
         if (desc) { strncpy(desc, table.Cell(i, 1), MAX_DEVICE_DESC_LEN); }
         active = table.Cell(i, 2, int(0));

         break;
      } 
      --deviceIndex;
   } 

   return deviceNumber != -1 ? Success : ErrorDeviceNotExist;
}

//
// RETURN:
//   ErrorFixNumberConflict: 
//      the fix number conflict with other device's, the parameter 'deviceNumber' 
//      contains the device number with which we conflict.
//   ErrorDeviceNotExist:
//      the device number is not found with the specified information
//   Success: 
//      we found the device number allocated previously to the device;
BDAQ_API
ErrorCode BDAQCALL FindDeviceNumberDB(
   DEVICE_HWINFO_EX const &hw, 
   int32                  *deviceNumber)
{
   assert(deviceNumber);
   *deviceNumber = -1;

   SQLiteScopedPtr db(DaqNaviConfigDB());
   SQLiteTable     table(*db);
   int             row = 0;


   // try to find the device number using 'Fix number' mechanism
   if (hw.BoardId) {
      table.Load("SELECT DeviceNumber, Active, DriverSPEC FROM device_map WHERE ProductId = %d AND BoardID = %d;", hw.ProductId, hw.BoardId);      

      if (table.Rows()) {
         if (table.Cell(row, 1, int(0)) != 0) { // Fix number conflict
            return ErrorFixNumberConflict;
         }
         *deviceNumber = table.Cell(row, 0, int(0));
      }
   } 

   // try to find the device number using 'HW address' mechanism
   if (*deviceNumber == -1) { 
      table.Load("SELECT DeviceNumber, Active, DriverSPEC FROM device_map WHERE ProductId = %d AND BusSlot = %d;", hw.ProductId, hw.BusSlot);      

      for (row = 0; row < table.Rows(); ++row) {
         if (table.Cell(row, 1, int(0)) == 0) {
            *deviceNumber = table.Cell(row, 0, int(0));
            break;
         }
      } 
   }

   // the device number does not exist.
   if(*deviceNumber == -1) {
      return ErrorDeviceNotExist;
   }

   // If the driver SPEC changed, update it now.
   if (hw.DriverSpec != (unsigned)table.Cell(row, 2, int(0))) {
      // Update registry
#if defined(_WIN32) || defined(WIN32)
      // Create a mirrored registry key to compatible with DN3
      TCHAR xbuf[256];
      _sntprintf_s(xbuf, _TRUNCATE, REG_PATH_BIODAQ_UM _T("\\") DEVX_KEY_NAME_FMT, *deviceNumber);

      RegKey devKey;
      devKey.Open(HKEY_LOCAL_MACHINE, xbuf, KEY_WRITE);
      if (devKey.IsOpened()) {
         // Mark the key as a mirror of DB item
         DWORD mirror = hw.DriverSpec == DRVSPEC_DN4;
         devKey.SetValue(ValMirror, mirror);

         // product id & board id
         devKey.SetValue(ValProductID, hw.ProductId);
         devKey.SetValue(ValBoardID, hw.BoardId);

         // bus & slot
         uint32 x = hw.BusSlot & 0xff;
         devKey.SetValue(ValBusNumber, x);

         x = hw.BusSlot >> 8;
         devKey.SetValue(ValSlotNumber, x);
      }
#endif

      // Update DB
      db->Exec("UPDATE device_map SET DriverSPEC = %d WHERE DeviceNumber = %d;", hw.DriverSpec, *deviceNumber);
   }
   
   return Success;
}

//
// Allocate a device number for new device
//
BDAQ_API
ErrorCode BDAQCALL AllocDeviceNumberDB(
   DEVICE_HWINFO_EX const &hw, 
#if defined(_WIN32) || defined(WIN32)
   wchar_t const *devicePath,
#else 
   char const    *devicePath,
#endif
   int32         *deviceNumber)
{
   assert(deviceNumber);
   *deviceNumber = -1;

   // Max. device number count : 4096
   int const countofDevNumber = 4096;
   uint32 devNumberInUse[countofDevNumber/32] = {0};

   #define MARK_DEVNUMBER(num)  devNumberInUse[num / 32] |= 1 << (num & 31)
   #define TEST_DEVNUMBER(num)  (devNumberInUse[num / 32] & (1 << (num & 31)))

   // scanning DB for free device number
   SQLiteScopedPtr db(DaqNaviConfigDB());
   SQLiteTable     table(*db);

   table.Load("SELECT DeviceNumber FROM device_map;");      
   for (int i = 0; i < table.Rows(); ++i) {
      int32 devNumber = table.Cell(i, 0, int(0));
      if (devNumber < countofDevNumber) {
         MARK_DEVNUMBER(devNumber);
      }
   } 

   // Search the free device number
   for (int i = 0; i < countofDevNumber; ++i) {
      if (!TEST_DEVNUMBER(i)) {
         *deviceNumber = i;
         break;
      }
   }

   if (*deviceNumber == -1) {
      return ErrorDriverCountExceedLimit; // It should be 'ErrorDeviceCountExceedLimit'
   }

   // Prepare the device description.
   // Using the kernel driver supplied description, if there is one.
#if defined(_WIN32) || defined(WIN32)
#  define DEV_STR_FMT "%ls" 
#else
#  define DEV_STR_FMT "%s" 
#endif

   char desc[MAX_DEVICE_DESC_LEN];
   if (hw.Description[0] != 0) {
      x_snprintf(desc, MAX_DEVICE_DESC_LEN, DEV_STR_FMT, hw.Description);
   } else {
      x_snprintf(desc, MAX_DEVICE_DESC_LEN, DEV_STR_FMT ",BID#%d", hw.DeviceName, hw.BoardId);
   }

   // Update table 'device_map', create item for the new device number
   desc[MAX_DEVICE_DESC_LEN - 1] = 0;

   db->Prepare("INSERT INTO device_map("
      DFN_DEVNUM "," DFN_PRODID  "," DFN_BOARDID "," DFN_BUSSLOT "," DFN_ACTIVE  "," 
      DFN_TERMBD "," DFN_DRVSPEC "," DFN_DRVNAME "," DFN_DEVPATH "," DFN_DESC    
      ") VALUES(%d, %d, %d, %d, %d, %d, %d,"
#if defined(_WIN32) || defined(WIN32)
      " \"%ls\", \"%ls\", \"%s\");",
#else
      " \"%s\", \"%s\", \"%s\");",
#endif
      *deviceNumber, hw.ProductId, hw.BoardId, hw.BusSlot, 1, WiringBoard, hw.DriverSpec, 
      hw.DriverName, devicePath, desc);
   
   if (db->Step() != SQLITE_DONE) {
      return ErrorConfigDataLost; // configure DB access error!
   }

#if defined(_WIN32) || defined(WIN32)
   // Create a mirrored registry key to compatible with DN3
   TCHAR xbuf[256];
   _sntprintf_s(xbuf, _TRUNCATE, REG_PATH_BIODAQ_UM _T("\\") DEVX_KEY_NAME_FMT, *deviceNumber);
   
   RegKey devKey;
   devKey.Create(HKEY_LOCAL_MACHINE, xbuf, KEY_WRITE);
   if (devKey.IsOpened()) {
      // Mark the key as a mirror of DB item
      DWORD mirror = 1;
      devKey.SetValue(ValMirror, mirror);

      // product id & board id
      devKey.SetValue(ValProductID, hw.ProductId);
      devKey.SetValue(ValBoardID, hw.BoardId);

      // bus & slot
      uint32 x = hw.BusSlot & 0xff;
      devKey.SetValue(ValBusNumber, x);

      x = hw.BusSlot >> 8;
      devKey.SetValue(ValSlotNumber, x);

      // description
      mbstowcs(xbuf, desc, MAX_DEVICE_DESC_LEN);
      devKey.SetValue(ValDeviceDesc, xbuf);

      // kernel key
      _tcscpy(xbuf, hw.DriverName);
      TCHAR *dot = _tcsrchr(xbuf, _T('.'));
      if (dot) { 
         *dot++ = _T('S'); 
         *dot   = 0;
      } else {
         _tcscat(xbuf, _T("S"));
      }
      devKey.SetValue(ValKernelKey, xbuf);
   }
#endif

   return Success;
}

//
// Allocate a device number for audio endpoint device
//
BDAQ_API
ErrorCode BDAQCALL AllocDeviceNumberDBForAudio(
   DEVICE_HWINFO_EX const &hw, 
#if defined(_WIN32) || defined(WIN32)
   wchar_t const *devicePath,
   wchar_t const *deviceDesc,
#else 
   char const    *devicePath,
   char const    *deviceDesc,
#endif
   int32         *deviceNumber)
{
   assert(deviceNumber);
   *deviceNumber = -1;

   // Max. device number count : 4096
   int const countofDevNumber = 4096;
   uint32 devNumberInUse[countofDevNumber/32] = {0};

   #define MARK_DEVNUMBER(num)  devNumberInUse[num / 32] |= 1 << (num & 31)
   #define TEST_DEVNUMBER(num)  (devNumberInUse[num / 32] & (1 << (num & 31)))

   // scanning DB for free device number
   SQLiteScopedPtr db(DaqNaviConfigDB());
   SQLiteTable     table(*db);

   table.Load("SELECT DeviceNumber FROM device_map;");      
   for (int i = 0; i < table.Rows(); ++i) {
      int32 devNumber = table.Cell(i, 0, int(0));
      if (devNumber < countofDevNumber) {
         MARK_DEVNUMBER(devNumber);
      }
   } 

   // Search the free device number
   for (int i = 0; i < countofDevNumber; ++i) {
      if (!TEST_DEVNUMBER(i)) {
         *deviceNumber = i;
         break;
      }
   }

   if (*deviceNumber == -1) {
      return ErrorDriverCountExceedLimit; // It should be 'ErrorDeviceCountExceedLimit'
   }

   // Set locale information for string conversion
   std::setlocale(LC_CTYPE, "");          

   // Update table 'device_map', create item for the new device number
   db->Prepare("INSERT INTO device_map("
      DFN_DEVNUM "," DFN_PRODID  "," DFN_BOARDID "," DFN_BUSSLOT "," DFN_ACTIVE  "," 
      DFN_TERMBD "," DFN_DRVSPEC "," DFN_DRVNAME "," DFN_DEVPATH "," DFN_DESC    
      ") VALUES(%d, %d, %d, %d, %d, %d, %d,"
#if defined(_WIN32) || defined(WIN32)
      " \"%ls\", \"%ls\", \"%ls, %ls\");",
#else
      " \"%s\", \"%s\", \"%s, %s\");",
#endif
      *deviceNumber, hw.ProductId, hw.BoardId, hw.BusSlot, 1, WiringBoard, hw.DriverSpec, 
      hw.DriverName, devicePath, hw.DeviceName, deviceDesc);
   
   if (db->Step() != SQLITE_DONE) {
      return ErrorConfigDataLost; // configure DB access error!
   }

#if defined(_WIN32) || defined(WIN32)
   // Create a mirrored registry key to compatible with DN3
   TCHAR xbuf[256];
   _sntprintf_s(xbuf, _TRUNCATE, REG_PATH_BIODAQ_UM _T("\\") DEVX_KEY_NAME_FMT, *deviceNumber);
   
   RegKey devKey;
   devKey.Create(HKEY_LOCAL_MACHINE, xbuf, KEY_WRITE);
   if (devKey.IsOpened()) {
      // Mark the key as a mirror of DB item
      DWORD mirror = 1;
      devKey.SetValue(ValMirror, mirror);

      // product id & board id
      devKey.SetValue(ValProductID, hw.ProductId);
      devKey.SetValue(ValBoardID, hw.BoardId);

      // bus & slot
      uint32 x = hw.BusSlot & 0xff;
      devKey.SetValue(ValBusNumber, x);

      x = hw.BusSlot >> 8;
      devKey.SetValue(ValSlotNumber, x);

      // description
      _sntprintf_s(xbuf, _TRUNCATE, _T("%s, %s"), hw.DeviceName, deviceDesc);
      devKey.SetValue(ValDeviceDesc, xbuf);

      // kernel key
      _tcscpy(xbuf, hw.DriverName);
      TCHAR *dot = _tcsrchr(xbuf, _T('.'));
      if (dot) { 
         *dot++ = _T('S'); 
         *dot   = 0;
      } else {
         _tcscat(xbuf, _T("S"));
      }
      devKey.SetValue(ValKernelKey, xbuf);
   }
#endif

   return Success;
}

//
// Free a device number 
//
BDAQ_API
ErrorCode BDAQCALL FreeDeviceNumberDB(int32 devNumber)
{
   //
   // Delete the device number information from DB
   SQLiteScopedPtr db(DaqNaviConfigDB());

   if (!db->IsOpened()) {
      return ErrorConfigDataLost;
   }
   db->Exec("DELETE FROM device_map WHERE %s = %d;", DFN_DEVNUM, devNumber);

#if defined(_WIN32) || defined(WIN32) 
   //
   // delete the device registry key

   // open the DAQ key
   RegKey daqKey(HKEY_LOCAL_MACHINE, _T("SYSTEM\\CurrentControlSet\\Services\\BIODAQ"), KEY_WRITE);
   if (!daqKey.IsOpened()) {
      return ErrorConfigDataLost;
   }

   // does the device key is a mirror of DB
   wchar_t keyName[64];
   swprintf(keyName, L"Device%03d", devNumber);
   daqKey.DeleteSubKey(keyName);
#endif

   return Success;
}

//
// Re-synchronize the device_map with installed DN3 devices.
//
BDAQ_API
void BDAQCALL ResyncDeviceInfoDB()
{
   SQLiteScopedPtr db(DaqNaviConfigDB());

   db->Begin();

   // Remove all DN3 drivers
   // Fix bug: 19773, change the WHERE condition from '= 3' to '!= 4' to erase the DN3 and the wrong items
   char const *db_cmd = "DELETE FROM device_map WHERE " DFN_DRVSPEC " != 4;";
   db->Exec(db_cmd);

   //
   // Enumerate all DB3 device and insert into device_map table
#if defined(_WIN32) || defined(WIN32)
   RegKey daqKey(HKEY_LOCAL_MACHINE, REG_PATH_BIODAQ_UM, KEY_READ);
   if (daqKey.IsOpened()) { 
      uint32 index = 0;
      wchar_t buffer[256];

      RegKey mapKey = daqKey.SubKey(MAPENTRY_KEY_NAME, KEY_READ);

      // Set locale information for string conversion
      std::setlocale(LC_CTYPE, "");          

      // Enumerate the DAQKEY('XXXX\BIODAQ') to traversal all device keys('DeviceNNN')
      while (daqKey.EnumKey(index++, buffer, 256) == ERROR_SUCCESS) {

         // Not a 'deviceXXX' key, ignore it!
         if (!IsDeviceEntryKey(buffer)) { continue; }

         RegKey devKey = daqKey.SubKey(buffer, KEY_READ);

         // The key is a mirror of the DB item, ignore it!
         DWORD mirror = 0;
         devKey.QueryValue(ValMirror, mirror);
         if (mirror) { continue; }

         // A true device key, get device information
         int32 deviceNumber = wcstol(&buffer[DEVX_KEY_NAME_PREFIXLEN], NULL, 10);
         DWORD productId = -1, boardId = 0, bus = 0, slot = 0, termBD = 0, active = 0;

         WCHAR description[64];
         WCHAR driverName[64];
         WCHAR devicePath[256];
         DWORD nchars;

         devKey.QueryValue(ValProductID, productId);
         devKey.QueryValue(ValBoardID, boardId);
         devKey.QueryValue(ValBusNumber, bus);
         devKey.QueryValue(ValSlotNumber, slot);
         devKey.QueryValue(ValTerminalBoardType, termBD);

         nchars = 64;
         description[0] = 0;
         devKey.QueryValue(ValDeviceDesc, description, &nchars);
         BioTrace(("ResyncDeviceInfoDB, device desc:%ls\n", description));
         nchars = 64;
         driverName[0] = 0;
         devKey.QueryValue(ValDriverName, driverName, &nchars);
         WCHAR *dot = _tcsrchr(driverName, _T('.'));
         if (dot) { 
            *dot = 0; 
         }
         if (x_wcsicmp(driverName, L"susiGPIO") == 0) {
            active = 1;
         }

         WCHAR mapValName[10];
         _itot(deviceNumber, mapValName, 10);
         nchars = 256;
         devicePath[0] = 0;
         mapKey.QueryValue(mapValName, devicePath, &nchars);

         db->Exec("INSERT INTO device_map("
            DFN_DEVNUM "," DFN_PRODID  "," DFN_BOARDID "," DFN_BUSSLOT "," DFN_ACTIVE  "," 
            DFN_TERMBD "," DFN_DRVSPEC "," DFN_DRVNAME "," DFN_DEVPATH "," DFN_DESC    
            ") VALUES(%d, %d, %d, %d, %d, %d, %d, \"%ls\", \"%ls\", \"%ls\");",
            deviceNumber, productId, boardId, bus | (slot << 8), active, termBD, 3, driverName, devicePath, description);
      }
   }
#endif 

   db->Commit();
}

// ----------------------------------------------------------------------------------------
// Query the device_map for device information
// Parameter:
//   deviceNumber : device number of the item to be queried.
//
BDAQ_API
ErrorCode BDAQCALL QueryDeviceInfoDB(
   QueryType     type,
   wchar_t const *key,
   uint32        *count,
   DEVICE_RECORD *buffer)
{
   XBuffer xbuf;
   uint32 len = (uint32)wcslen(key) * 2 + 1;
   wcstombs(xbuf.Buffer(len), key, xbuf.Size());

   return QueryDeviceInfoADB(type, xbuf.Buffer(), count, buffer);
}

BDAQ_API
ErrorCode BDAQCALL QueryDeviceInfoADB(
   QueryType     type,
   char const    *key,
   uint32        *count,
   DEVICE_RECORD *buffer)
{
   assert(count && *count && buffer);

   DEVICE_RECORD *rec = buffer;
   DEVICE_RECORD *end = buffer + *count;
   *count = 0;

   // Open db
   SQLiteScopedPtr db(DaqNaviConfigDB());
   SQLiteTable table(*db);

   // List all device information
   if (type == ListAll) {
      table.Load("SELECT * FROM device_map;");      
      for (int i = 0; i < table.Rows(); ++i) {
         if (rec >= end) {
            *count = uint32(rec - buffer);
            return ErrorBufferTooSmall;
         }         

         FillDeviceRecord(rec, sizeof(*rec), table, i);
         ++rec;
      }

      *count = uint32(rec - buffer);
      return Success;
   }

   if (type == ByDevNumber) {
      table.Load("SELECT * FROM device_map WHERE DeviceNumber = %s;", key);
      if (!table.Rows()) {
         return ErrorDeviceNotExist;
      }

      FillDeviceRecord(rec, sizeof(*rec), table, 0);
      *count = 1;
      return Success;
   }

   //
   // Due the case sensitive reason, here we had manual compare the key with db.
   if (type == ByDevDesc) {
      table.Load("SELECT * FROM device_map;");      
      for (int i = 0; i < table.Rows(); ++i) {
         if (x_stricmp(table.Cell(i, DFI_DESC), key) == 0) {
            if (rec >= end) {
               *count = uint32(rec - buffer);
               return ErrorBufferTooSmall;
            } 

            FillDeviceRecord(rec, sizeof(*rec), table, i);
            ++rec;
         }
      }

      *count = uint32(rec - buffer);
      return Success;
   }

   if (type == ByDevPath) {
      table.Load("SELECT * FROM device_map;");      
      for (int i = 0; i < table.Rows(); ++i) {
#if defined(_WIN32) || defined(WIN32)
         // To compatible with DN3, skip the prefix "\\."
         if (x_stricmp(table.Cell(i, DFI_DEVPATH) + 3, key + 3) == 0)
#else
         if (x_stricmp(table.Cell(i, DFI_DEVPATH), key) == 0)
#endif
         {
            if (rec >= end) {
               *count = uint32(rec - buffer);
               return ErrorBufferTooSmall;
            } 

            FillDeviceRecord(rec, sizeof(*rec), table, i);
            ++rec;
         }
      }

      *count = uint32(rec - buffer);
      return Success;
   }

   return ErrorParamNotSpted;
}

BDAQ_API
ErrorCode BDAQCALL QueryDeviceInfoDBEx(
   QueryType        type,
   wchar_t const    *key,
   uint32           *count,
   DEVICE_RECORD_EX *buffer)
{
   XBuffer xbuf;
   uint32 len = (uint32)wcslen(key) * 2 + 1;
   wcstombs(xbuf.Buffer(len), key, xbuf.Size());

   return QueryDeviceInfoADBEx(type, xbuf.Buffer(), count, buffer);
}

BDAQ_API
ErrorCode BDAQCALL QueryDeviceInfoADBEx(
   QueryType        type,
   char const       *key,
   uint32           *count,
   DEVICE_RECORD_EX *buffer)
{
   assert(count && *count && buffer);

   DEVICE_RECORD_EX *rec = buffer;
   DEVICE_RECORD_EX *end = buffer + *count;
   *count = 0;

   // Open db
   SQLiteScopedPtr db(DaqNaviConfigDB());
   SQLiteTable table(*db);

   // List all device information
   if (type == ListAll) {
      table.Load("SELECT * FROM device_map;");      
      for (int i = 0; i < table.Rows(); ++i) {
         if (rec >= end) {
            *count = uint32(rec - buffer);
            return ErrorBufferTooSmall;
         }         

         FillDeviceRecord(rec, sizeof(*rec), table, i);
         ++rec;
      }

      *count = uint32(rec - buffer);
      return Success;
   }

   if (type == ByDevNumber) {
      table.Load("SELECT * FROM device_map WHERE DeviceNumber = %s;", key);
      if (!table.Rows()) {
         return ErrorDeviceNotExist;
      }

      FillDeviceRecord(rec, sizeof(*rec), table, 0);
      *count = 1;
      return Success;
   }

   //
   // Due the case sensitive reason, here we had manual compare the key with db.
   if (type == ByDevDesc) {
      table.Load("SELECT * FROM device_map;");      
      for (int i = 0; i < table.Rows(); ++i) {
         if (x_stricmp(table.Cell(i, DFI_DESC), key) == 0) {
            if (rec >= end) {
               *count = uint32(rec - buffer);
               return ErrorBufferTooSmall;
            } 

            FillDeviceRecord(rec, sizeof(*rec), table, i);
            ++rec;
         }
      }

      *count = uint32(rec - buffer);
      return Success;
   }

   if (type == ByDevPath) {
      table.Load("SELECT * FROM device_map;");      
      for (int i = 0; i < table.Rows(); ++i) {
#if defined(_WIN32) || defined(WIN32)
         // To compatible with DN3, skip the prefix "\\."
         if (x_stricmp(table.Cell(i, DFI_DEVPATH) + 3, key + 3) == 0)
#else
         if (x_stricmp(table.Cell(i, DFI_DEVPATH), key) == 0)
#endif
         {
            if (rec >= end) {
               *count = uint32(rec - buffer);
               return ErrorBufferTooSmall;
            } 

            FillDeviceRecord(rec, sizeof(*rec), table, i);
            ++rec;
         }
      }

      *count = uint32(rec - buffer);
      return Success;
   }

   return ErrorParamNotSpted;
}

// ----------------------------------------------------------------------------------------
// Update the device_map with device information
// Parameter:
//   deviceNumber : device number of the item to be updated, -1 for all items.
//
#if defined(_WIN32) || defined(WIN32)
BDAQ_API ErrorCode BDAQCALL UpdateDeviceInfoDB(DEVICE_HWINFO_EX const &hw, wchar_t const *devicePath)
{
   SQLiteScopedPtr db(DaqNaviConfigDB());
   int ret;

   TCHAR keyName[256];
   _sntprintf_s(keyName, _TRUNCATE, REG_PATH_BIODAQ_UM _T("\\") DEVX_KEY_NAME_FMT, hw.DeviceNumber);

   RegKey devKey(HKEY_LOCAL_MACHINE, keyName, KEY_READ | KEY_WRITE);
   assert(devKey.IsOpened());

   // Set locale information for string conversion
   std::setlocale(LC_CTYPE, "");          

   //
   // Update DN3 device information
   if (hw.DriverSpec < DRVSPEC_DN4) {

      WCHAR desc[MAX_DEVICE_DESC_LEN];
      DWORD termBD = WiringBoard;
      DWORD nchars;

      // Query Terminal board type
      devKey.QueryValue(ValTerminalBoardType, *(DWORD*)&termBD);

      // Query description
      nchars = 64;
      desc[0] = 0;
      devKey.QueryValue(ValDeviceDesc, desc, &nchars);

      ret = db->Exec("INSERT OR REPLACE INTO device_map("
         DFN_DEVNUM   "," DFN_PRODID  "," DFN_BOARDID "," DFN_BUSSLOT "," DFN_ACTIVE  "," 
         DFN_TERMBD   "," DFN_DRVSPEC "," DFN_DRVNAME "," DFN_DEVPATH "," DFN_DESC
         ") VALUES(%d, %d, %d, %d, %d, %d, %d, \"%ls\", \"%ls\", \"%ls\");",
               hw.DeviceNumber, hw.ProductId, hw.BoardId, hw.BusSlot, 1,
               termBD, 3, hw.DriverName, devicePath, desc);

      return ret == SQLITE_OK ? Success : ErrorConfigDataLost;
   } 

   //
   // Update DN4 device information DB
   ret = db->Exec("UPDATE device_map SET "
            DFN_BOARDID  " = %d,"
            DFN_BUSSLOT  " = %d,"
            DFN_ACTIVE   " = %d,"
            DFN_DEVPATH  " = \"%ls\","
            DFN_DESC     " = \"%ls\","
            DFN_TREEPATH " = \"%ls\""
            " WHERE DeviceNumber = %d;", 
            hw.BoardId, hw.BusSlot, 1, devicePath, hw.Description, hw.TreePath, hw.DeviceNumber);
   //
   // The mirror KEY 
   devKey.SetValue(ValBoardID, REG_DWORD, &hw.BoardId, sizeof(hw.BoardId));

   uint32 x = hw.BusSlot & 0xff;
   devKey.SetValue(ValBusNumber, REG_DWORD, &x, sizeof(x));

   x = hw.BusSlot >> 8;
   devKey.SetValue(ValSlotNumber, REG_DWORD, &x, sizeof(x));

   return ret == SQLITE_OK ? Success : ErrorConfigDataLost;
}

#else

BDAQ_API 
ErrorCode BDAQCALL UpdateDeviceInfoDB(DEVICE_HWINFO_EX const &hw, char const *devicePath)
{
   SQLiteScopedPtr db(DaqNaviConfigDB());
   int ret;

   // Note: 
   // Update ALL (DN3&DN4) device information.
   // At present, we haven't supported 'Fixed device number' mechanism for linux.
   /*if (hw.DriverSpec < DRVSPEC_DN4)*/ {
      ret = db->Exec("INSERT OR REPLACE INTO device_map("
               DFN_DEVNUM "," DFN_PRODID  "," DFN_BOARDID "," DFN_BUSSLOT "," DFN_ACTIVE  "," 
               DFN_TERMBD "," DFN_DRVSPEC "," DFN_DRVNAME "," DFN_DEVPATH
               ") VALUES(%d, %d, %d, %d, %d, %d, %d, \"%s\", \"%s\");",
               hw.DeviceNumber, hw.ProductId, hw.BoardId,     hw.BusSlot, 1, 
               WiringBoard, hw.DriverSpec,    hw.DriverName,  devicePath);

      return ret == SQLITE_OK ? Success : ErrorConfigDataLost;
   } 

   //
   // Update DN4 device information
   ret = db->Exec("UPDATE device_map SET "
      DFN_BOARDID  " = %d,"
      DFN_BUSSLOT  " = %d,"
      DFN_ACTIVE   " = %d,"
      DFN_DEVPATH  " = \"%s\","
      DFN_DESC     " = \"%s\","
      DFN_TREEPATH " = \"%s\""
      " WHERE DeviceNumber = %d;", 
      hw.BoardId, hw.BusSlot, 1, devicePath, hw.Description, hw.TreePath, hw.DeviceNumber);

   return Success;
}
#endif   

// ----------------------------------------------------------------------------------------
// Query the device_map for device information
// Parameter:
//   deviceNumber : device number of the item to be queried.
//
BDAQ_API
ErrorCode BDAQCALL QueryDeviceFieldDB(
   int32         deviceNumber, 
   char const    *field, 
   uint32        *strLen,
   wchar_t       *strBuf)
{
   XBuffer buffer;
   ErrorCode ret = QueryDeviceFieldADB(deviceNumber, field, strLen, buffer.Buffer(*strLen));
   if (ret == Success) {
      mbstowcs(strBuf, buffer.Buffer(), *strLen);
   } 

   return ret;
}

BDAQ_API
ErrorCode BDAQCALL QueryDeviceFieldADB(
   int32      deviceNumber, 
   char const *field, 
   uint32     *strLen,
   char       *strBuf)
{
   SQLiteScopedPtr db(DaqNaviConfigDB());
   db->Prepare("SELECT \"%s\" FROM device_map WHERE DeviceNumber = %d;", field, deviceNumber);

   if (db->Step() != SQLITE_ROW) {
      return ErrorDeviceNotExist;
   }

   char const *text = db->Column(0);
   uint32 len = (uint32)strlen(text);
   if (!text || !len) {
      return ErrorConfigDataLost;
   }

   if (*strLen < len + 1) {
      *strLen = len + 1;
      return ErrorBufferTooSmall;
   }

   *strLen = len + 1;
   x_strncpy(strBuf, text, len + 1);
   return Success;
}

BDAQ_API
ErrorCode BDAQCALL UpdateDeviceFieldDB(
   int32         deviceNumber, 
   char const    *field, 
   wchar_t const *value)
{
   XBuffer buffer;
   std::setlocale(LC_CTYPE,"");

   uint32 len = (uint32)wcslen(value) * 2 + 1;
   wcstombs(buffer.Buffer(len), value, len);

   return UpdateDeviceFieldADB(deviceNumber, field, buffer.Buffer());
}

BDAQ_API
ErrorCode BDAQCALL UpdateDeviceFieldADB(
   int32      deviceNumber, 
   char const *field, 
   char const *value)
{

   SQLiteScopedPtr db(DaqNaviConfigDB());

   if (deviceNumber != -1) {
      db->Prepare("UPDATE device_map SET %s = \"%s\" WHERE DeviceNumber = %d;", 
            field, value, deviceNumber);
   } else {
      db->Prepare("UPDATE device_map SET %s = \"%s\";", field, value);
   }

   return db->Step() == SQLITE_DONE ? Success : ErrorConfigDataLost;
}

// helper method
BDAQ_API
ErrorCode BDAQCALL QueryDeviceFieldI32DB(
   int32      deviceNumber, 
   char const *field, 
   int32      *value)
{
   SQLiteScopedPtr db(DaqNaviConfigDB());
   db->Prepare("SELECT \"%s\" FROM device_map WHERE DeviceNumber = %d;", field, deviceNumber);

   if (db->Step() == SQLITE_ROW) {
      *value = db->Column(0, int(0));
      return Success;
   }

   return ErrorDeviceNotExist;
}

BDAQ_API
ErrorCode BDAQCALL UpdateDeviceFieldI32DB(
   int32      deviceNumber, 
   char const *field, 
   int32      value)
{
   char buffer[64];
   sprintf(buffer, "%d", value);

   return UpdateDeviceFieldADB(deviceNumber, field, buffer);
}

//---------------------------------------------------------------------
// Access 'Misc' table(key/value)
// 
BDAQ_API
ErrorCode BDAQCALL QueryMiscField(
   char const *key,
   uint32     *strLen,
   char       *strBuf)
{
   SQLiteScopedPtr db(DaqNaviConfigDB());
   db->Prepare("SELECT Value FROM misc WHERE Name = \"%s\";", key);

   if (db->Step() != SQLITE_ROW) {
      return ErrorParamNotSpted;
   }

   char const *text = db->Column(0);
   uint32 len = (uint32)strlen(text);
   if (!text || !len) {
      return ErrorConfigDataLost;
   }

   if (*strLen < len + 1) {
      *strLen = len + 1;
      return ErrorBufferTooSmall;
   }

   *strLen = len + 1;
   x_strncpy(strBuf, text, len + 1);
   return Success;
}

//---------------------------------------------------------------------
// enum string information DB field access
//
BDAQ_API
ErrorCode BDAQCALL EnumStrEnumerateDB(
   char const         *category,
   char const         *field,
   EnumStrEnumCallback userFunc,
   void                *userParam)
{
   if (!userFunc) {
      return ErrorParamNotSpted;
   }

   SQLiteScopedPtr db(DaqNaviConfigDB());
   SQLiteTable table(*db);
   table.Load("SELECT %s FROM enum_string WHERE Category = '%s';", field, category);

   XBuffer valBuf;
   char const *val;

   for (int i = 0; i < table.Rows(); ++i) {
      // Copy the 'value' into local buffer due to it will 
      // be destroyed when the next sqlite3 command executes.
      val = table.Cell(i, 0);
      strcpy(valBuf.Buffer((uint32)strlen(val) + 1), val);
      
      userFunc(valBuf.Buffer(), userParam);
   }

   return Success;
}

//---------------------------------------------------------------------
// supported product information DB field access
//
BDAQ_API
ErrorCode BDAQCALL EnumSupportedProductDB(
   EnumSuppProductCallback userFunc,
   void                    *userParam)
{
   if (!userFunc) {
      return ErrorParamNotSpted;
   }

   SQLiteScopedPtr db(DaqNaviConfigDB());
   SQLiteTable table(*db);
   table.Load("SELECT * FROM supp_product;");
   
   PRODUCT_INFO info = {0};
   for (int i = 0; i < table.Rows(); ++i) {
      info.Name          = table.Cell(i, 0);
      info.Bus           = table.Cell(i, 1);
      info.HWID          = table.Cell(i, 2);
      info.Package       = table.Cell(i, 3);
      info.Model         = table.Cell(i, 4);
      info.Description   = table.Cell(i, 5);
      info.Manual        = table.Cell(i, 6);
      info.Feature       = table.Cell(i, 7);
      info.ManualPackage = table.Cell(i, 8);

      if (userFunc(&info, userParam) != 0) {
         break;
      }
   }

   return Success;
}

//
// New: Query device bus type
BDAQ_API
ErrorCode BDAQCALL QueryDeviceBusTypeDB(
   ProductId id,
   uint32   *strLen,
   char     *strBuf)
{
   SQLiteScopedPtr db(DaqNaviConfigDB());
   db->Prepare("SELECT Bus FROM supp_product WHERE Model = \"0x%X\";", id);

   if (db->Step() != SQLITE_ROW) {
      return ErrorParamNotSpted;
   }

   char const *text = db->Column(0);
   uint32 len = (uint32)strlen(text);
   if (!text || !len) {
      return ErrorConfigDataLost;
   }

   if (*strLen < len + 1) {
      *strLen = len + 1;
      return ErrorBufferTooSmall;
   }

   *strLen = len + 1;
   x_strncpy(strBuf, text, len + 1);
   return Success;
}
