/*******************************************************************************
              Copyright (c) 1983-2020 Advantech Co., Ltd.
********************************************************************************
THIS IS AN UNPUBLISHED WORK CONTAINING CONFIDENTIAL AND PROPRIETARY INFORMATION
WHICH IS THE PROPERTY OF ADVANTECH CORP., ANY DISCLOSURE, USE, OR REPRODUCTION,
WITHOUT WRITTEN AUTHORIZATION FROM ADVANTECH CORP., IS STRICTLY PROHIBITED. 

================================================================================
REVISION HISTORY
--------------------------------------------------------------------------------
$Log: $
--------------------------------------------------------------------------------
$NoKeywords:  $
*/

// ----------------------------------------------------------------------------
// File:        rtd.cpp
// Module:      UserLib.lib
// Environment: User Mode
// Reference:   Microsoft WDK
// Description: RTD functions
// ----------------------------------------------------------------------------
#include "std_decl.h"

//////////////////////////////////////////////////////////////////////////

BDAQ_API
void BDAQCALL RtdBuildMapFuncByCalParam(
   IN PCRTD_INFO   rtdInfo,
   IN uint32       gainParam,
   IN uint32       offsetParam,
   OUT SCALE_TABLE *table)
{
   /*
   * Note: Raw Data Conversion to Resistance Equation is:
   *
   *    Resistance = (data_binary - offsetParam) / (gainParam - offsetParam) * scaleFactor + offsetFactor
   *
   * where:
   *    lsb = scaleFactor / (gainParam - offsetParam)
   *    generally the scaleFactor and offsetFactor depend on input range, as shown in the Hardware Description,
   *
   * so:
   *     Resistance = (data_binary - offsetParam) * lsb + offsetFactor   
   *     Resistance = (data_binary - offsetParm + (offsetFactor / lsb)) * lsb
   *     Resistance = (data_binary - f) * lsb                                ----(1)
   *
   * where:
   *     f = offsetParm - (offsetFactor / lsb)
   *
   * Note: Resistance Conversion to Temperature Interpolation Equation is:
   *
   *   Temperature = a * R^4 + b * R^3 + c * R^2 + d * R + e                 ----(2)
   *
   * where:
   *   a\b\c\d\e: These are coefficients, and fixed value;
   *
   * SO, Combining Equation 1 and 2,the polynomial is:
   * Temperature = A * x^4 + B * x^3 + C * x^2 + D * x + E
   * where:
   * x : data_binary
   * A :  a * lsb^4
   * B : -a * lsb^4 * 4 * f   + b * lsb^3
   * C :  a * lsb^4 * 6 * f^2 - b * lsb^3 * 3 * f   + c * lsb^2
   * D : -a * lsb^4 * 4 * f^3 + b * lsb^3 * 3 * f^2 - c * lsb^2 * 2 * f   + d * lsb
   * E :  a * lsb^4 * 1 * f^4 - b * lsb^3 * 1 * f^3 + c * lsb^2 * 1 * f^2 - d * lsb * f + e
   */
   assert(rtdInfo && table);
   int32  piece_size = sizeof(MapFuncPiece) + sizeof(double) * 3; //  3 means: 3 extra coefficients
   int32  temp_lo    = rtdInfo->TempLow;
   int32  temp_hi    = rtdInfo->TempHigh;
   double resis_min  = rtdInfo->ResisMin;
   double resis_zero = rtdInfo->ResisZero;
   double resis_max  = rtdInfo->ResisMax;
   double of = rtdInfo->OffsetFactor;
   double sf = rtdInfo->ScaleFactor;

   double lsb = sf / (gainParam - offsetParam);
   double f = (offsetParam - (of / lsb));

   double a0_coef = rtdInfo->A0_Coef * pow(lsb, 4);
   double b0_coef = rtdInfo->B0_Coef * pow(lsb, 3);
   double c0_coef = rtdInfo->C0_Coef * pow(lsb, 2);
   double d0_coef = rtdInfo->D0_Coef * lsb;
   double e0_coef = rtdInfo->E0_Coef;

   double a1_coef = rtdInfo->A1_Coef * pow(lsb, 4);
   double b1_coef = rtdInfo->B1_Coef * pow(lsb, 3);
   double c1_coef = rtdInfo->C1_Coef * pow(lsb, 2);
   double d1_coef = rtdInfo->D1_Coef * lsb;
   double e1_coef = rtdInfo->E1_Coef;

   table->VrgType    = rtdInfo->VrgType;
   table->Gain       = rtdInfo->Gain;
   table->PieceCount = 4;

   //Piece 0
   table->Pieces[0].Size       = piece_size;
   table->Pieces[0].Degree     = 4;
   table->Pieces[0].UpperLimit = (resis_min - of) / lsb + offsetParam;

   table->Pieces[0].Coef[4] = 0;
   table->Pieces[0].Coef[3] = 0;
   table->Pieces[0].Coef[2] = 0;
   table->Pieces[0].Coef[1] = 0;
   table->Pieces[0].Coef[0] = (double)temp_lo;

   int64 ptr = (int64)table->Pieces;
   MapFuncPiece* piece1 = (MapFuncPiece*) (ptr + piece_size);

   //Piece 1
   piece1->Size       = piece_size;
   piece1->Degree     = 4;
   piece1->UpperLimit = (resis_zero - of) / lsb + offsetParam;

   piece1->Coef[4] =  a0_coef;
   piece1->Coef[3] = -a0_coef * 4 * f + b0_coef;
   piece1->Coef[2] =  a0_coef * 6 * pow(f, 2) - b0_coef * 3 * f + c0_coef;
   piece1->Coef[1] = -a0_coef * 4 * pow(f, 3) + b0_coef * 3 * pow(f, 2) - c0_coef * 2 * f + d0_coef;
   piece1->Coef[0] =  a0_coef * 1 * pow(f, 4) - b0_coef * 1 * pow(f, 3) + c0_coef * 1 * pow(f, 2) - d0_coef * f + e0_coef;

   MapFuncPiece* piece2 = (MapFuncPiece*) (ptr + piece_size * 2);

   //Piece 2
   piece2->Size       = piece_size;
   piece2->Degree     = 4;
   piece2->UpperLimit = (resis_max - of) / lsb + offsetParam;

   piece2->Coef[4] =  a1_coef;
   piece2->Coef[3] = -a1_coef * 4 * f + b1_coef;
   piece2->Coef[2] =  a1_coef * 6 * pow((double)offsetParam, 2) - b1_coef * 3 * offsetParam + c1_coef;
   piece2->Coef[1] = -a1_coef * 4 * pow((double)offsetParam, 3) + b1_coef * 3 * pow((double)offsetParam, 2) - c1_coef * 2 * offsetParam + d1_coef;
   piece2->Coef[0] =  a1_coef * 1 * pow((double)offsetParam, 4) - b1_coef * 1 * pow((double)offsetParam, 3) + c1_coef * 1 * pow((double)offsetParam, 2) - d1_coef * offsetParam + e1_coef;

   MapFuncPiece* piece3 = (MapFuncPiece*) (ptr + piece_size * 3);

   //Piece 3
   piece3->Size       = piece_size;
   piece3->Degree     = 4;
   piece3->UpperLimit = DBL_MAX;

   piece3->Coef[4] = 0;
   piece3->Coef[3] = 0;
   piece3->Coef[2] = 0;
   piece3->Coef[1] = 0;
   piece3->Coef[0] = (double)temp_hi;
}

//*************************************************************************
// Name: RtdFindItemByVrgType
// Description:
//       Helper function to retrieve a value range information entry 
//    according with the specified value range type.
// Parameters: 
//  [in]vrgType-- the value range type used to search.
//  [in]entries-- the value range info table to be searched.
//  [in]count  -- the count of the value range info table. 
//   
// return value: 
//    a const pointer to the AI value range info entry or NULL if not found.
//************************************************************************* 
BDAQ_API
PCRTD_INFO BDAQCALL RtdFindItemByVrgType(
   IN uint32     vrgType,
   IN PCRTD_INFO entries,
   IN uint32     count)
{
   for (uint32 i = 0; i < count; ++i) {
      if (entries[i].VrgType == vrgType) {
         return &entries[i];
      }
   }
   return NULL;
}