/*
 * BDaqTypes.h
 *
 *  Created on: 2011-8-18
 *      Author: rocky
 */

#ifndef _DAQNAVI_DEVICE_ADAPTOR_H
#define _DAQNAVI_DEVICE_ADAPTOR_H

#include "ioctl_helper.hpp" // for event create/destory

struct IntfType {
   enum { DN3_API, DN3_OBJ, DN4_OBJ};
};

struct PropFlag {
   enum { OpGet = 1, SysDB = 2, NtfNow = 4, UpdHW = 0x100, };
};

struct DevState { /*Device PnP State: (surprise) removed, arrived (reconnected)*/
   enum {Arrived, Removed};
};

//
// Special pointer value for multi-part buffer
#define MULTI_BUF_PTR   ((void *)1)

#define IsGet(flag)     (flag & PropFlag::OpGet)
#define IsDB(flag)      (flag & PropFlag::SysDB)
#define IsNtfNow(flag)  (flag & PropFlag::NtfNow)
#define IsUpdHW(flag)   (flag & PropFlag::UpdHW)

#ifdef _MSC_VER
#pragma warning(disable:4100)
#endif

typedef struct EventStatus {
   uint32 Size; // size of this structure, in bytes
   union {
      // General purpose event status, depends on the event type
      struct {
         uint32 lParam;
         uint32 rParam;
      } General;

      // Ai data ready / stopped event status
      struct {
         uint32 DataCount;
         uint32 DataOffset;
         uint32 MarkCount;
         uint32 Stopped;  
         uint32 RecordCount; // conversion stopped event, Record count generated at present.
      } AiDataReady;

      // Ai overrun event status
      struct {
         uint32 DataCount;
         uint32 DataOffset;
      } AiOverrun;

      struct {
         uint32 DataCount;
         uint32 DataOffset;
         uint32 Stopped;
      } CntrStoped;

      // Di data ready / stopped event status
      struct {
         uint32 DataCount;
         uint32 DataOffset;
         uint32 MarkCount;
         uint32 Stopped;  
         uint32 RecordCount; // conversion stopped event, Record count generated at present.
      } DiDataReady;

      // Di overrun event status
      struct {
         uint32 DataCount;
         uint32 DataOffset;
      } DiOverrun;

      //
      // Others event status can be appended from here...
      //
      // struct {
      // } xxx;
   };
} EventStatus;

class BDAQ_NO_VTABLE DaqModule
{
public:
   // !!!!!! default interface type !!!!!!
   // It will be overridden by DN3 classes
   virtual int32  BDAQCALL getIntfType() { return IntfType::DN4_OBJ; }

public:
   virtual ModuleType BDAQCALL getType()   = 0;
   virtual int32      BDAQCALL getNumber() = 0;    // device number or module index, depends on module type.
   virtual AccessMode BDAQCALL getMode()   = 0;

   // !!!!!! property & feature !!!!!!
   virtual ErrorCode BDAQCALL PropAccess(PropertyId id, uint32 &size, void *buffer, uint32 flag) = 0;

   // Return default attribute
   virtual ErrorCode BDAQCALL PropAttribute(PropertyId id, PropertyAttribute &attr)
   {
      // No special attribute, The caller should query the database to get the default attribute of the property.
      return ErrorFuncNotSpted; 
   }

   // Event access
   virtual ErrorCode BDAQCALL EventGetHandle(EventId id, HANDLE &handle) 
   { 
      return ErrorFuncNotSpted;  
   }
   virtual ErrorCode BDAQCALL EventGetStatus(EventId id, EventStatus &status) 
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL EventClearFlag(EventId id, EventStatus &flags)
   {
      return ErrorFuncNotSpted;
   }

public: // Helpers

   HANDLE InitializeEvent(DEVFD fd, EventId id)
   {
      HANDLE h = CreateEvent(NULL, FALSE, FALSE, NULL);
      if (h) {            
         RegisterEvent(fd, id, h);
      }
      return h;
   }

   void DeleteEvent(DEVFD fd, HANDLE handle)
   {
      if (handle) {
         CloseHandle(handle);
      }
   }

   ErrorCode RegisterEvent(DEVFD fd, EventId id, HANDLE handle)
   {
      EVENT_INFO info = { {id}, {handle} };
      return IoctlDevRegisterEvent(fd, &info);
   }
};

class BDAQ_NO_VTABLE DaqDevModule : public DaqModule
{
public:
   // Override: DaqMoudle
   virtual ModuleType BDAQCALL getType() { return DaqDevice; }

   // New method
   virtual void       BDAQCALL Close()             = 0;
   virtual ErrorCode  BDAQCALL Refresh()           = 0;
   virtual ErrorCode  BDAQCALL Reset(uint32 state) = 0;
   virtual ErrorCode  BDAQCALL GetModule(ModuleType type, uint32 index, DaqModule* &module) = 0;

public:
   // Undocumented functions
   virtual ErrorCode BDAQCALL ReadRegister(uint32 space, uint32 offset, uint32 length, void *buffer)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL WriteRegister(uint32 space, uint32 offset, uint32 length, void *buffer)
   {
      return ErrorFuncNotSpted;
   }

   virtual ErrorCode BDAQCALL ReadPrivRgn(uint32 signature, uint32 length, uint8 *buffer)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL WritePrivRgn(uint32 signature, uint32 length, uint8 *buffer)
   {
      return ErrorFuncNotSpted;
   }

public: // Timestamp
   virtual ErrorCode BDAQCALL SynchronizeTimebase()
   {
      return ErrorFuncNotSpted;
   }
   virtual double BDAQCALL CalculateAbsoluteTime(double const &relativeTime)
   {
      return relativeTime;
   }

public: // DN4 device surprise remove & reconnect
   virtual ErrorCode BDAQCALL OnDeviceStateChanged(int32 newState)
   {
      return ErrorFuncNotSpted;
   }   

public: // Build new description according the the current one.
   virtual ErrorCode BDAQCALL UpdateDescription(uint32 boardId, wchar_t const curDesc[], wchar_t newDesc[])
   {
      DeviceUpdateDescription(boardId, curDesc, newDesc);
      return Success;
   }

public: // Helpers
   static HANDLE CreateDevicePnpEvent(EventId evtId, uint32 prodId, uint32 devNum)
   {
      return DeviceCreatePnpEvent(evtId, prodId, devNum);
   }
};

class BDAQ_NO_VTABLE DaqAiModule : public DaqModule
{
public:
   // Override: DaqMoudle
   virtual ModuleType BDAQCALL getType()   { return DaqAi; }
   virtual int32      BDAQCALL getNumber() { return 0;     } // default module index

   // New method
   virtual ErrorCode BDAQCALL ReadSamples(uint32 chStart, uint32 chCount, void *rawBuf, double *scaledBuf) = 0;

   // Not all AI modules support buffered AI operation.
   virtual ErrorCode BDAQCALL BfdAiPrepare(uint32 &dataCount, void* &dataPtr)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdAiStart()
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdAiStop()
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdAiRelease()
   {
      return ErrorFuncNotSpted;
   }

   virtual ErrorCode BDAQCALL BfdAiGetData(
      uint32   &dataCount,      // IN/OUT, Necessary
      void     *rawBuf,         // OUT,    Optional  
      double   *dblBuf,         // OUT,    Optional
      double   *startTime,      // OUT,    Optional
      uint32   *markCount,      // IN/OUT, Optional
      DataMark *markBuf)        // OUT,    Optional
   {
      return ErrorFuncNotSpted;
   }

   // Sensor methods
   virtual ErrorCode BDAQCALL RunSensorCalibration(int32 ch, CalibrationData const *data)
   {
      return ErrorFuncNotSpted;
   }

public:
   // Undocumented functions
   virtual ErrorCode BDAQCALL GetScaleTable(uint32 ch, uint32 &size, MapFuncPiece *tables)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL SetScaleTable(uint32 ch, uint32 size, MapFuncPiece *tables)
   {
      return ErrorFuncNotSpted;
   }

public:
   // Undocumented functions: query FAI State, used by fusion driver
   // RETURN: a combination of one or more of DAQ_FN_xxx.
   virtual uint32 BDAQCALL BfdAiGetState()
   {
      return DAQ_FN_IDLE;
   }
};

class BDAQ_NO_VTABLE DaqAoModule : public DaqModule
{
public:
   // Override: DaqMoudle
   virtual ModuleType BDAQCALL getType()   { return DaqAo; }
   virtual int32      BDAQCALL getNumber() { return 0;     } // default module index

   // New method
   virtual ErrorCode BDAQCALL WriteSamples(uint32 chStart, uint32 chCount, void *rawBuf, double *scaledBuf) = 0;

   //
   // !!!CAUTION!!!
   // The driver must return 'MULTI_BUF_PTR' for parameter 'dataPtr' 
   // when it supports the new method 'BfdAoSetData'
   // !!!CAUTION!!!
   virtual ErrorCode BDAQCALL BfdAoPrepare(uint32 dataCount, void* &dataPtr)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdAoRunOnce(bool asynchronous)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdAoRun()
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdAoStop(uint32 action)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdAoRelease()
   {
      return ErrorFuncNotSpted;
   }

   //
   // !!!Deprecated method!!!
   // Any new driver should support the method 'BfdAoSetData'
   virtual ErrorCode BDAQCALL BfdAoScaleData(double *scaledBuf, void *rawBuf, uint32 count, uint32 *chanOffset)
   {
      return ErrorFuncNotSpted;
   }

public:
   // Undocumented functions
   virtual ErrorCode BDAQCALL GetScaleTable(uint32 ch, uint32 &size, MapFuncPiece *tables)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL SetScaleTable(uint32 ch, uint32 size, MapFuncPiece *tables)
   {
      return ErrorFuncNotSpted;
   }

public:
   // !!!CAUTION!!!
   // New method for buffered AO
   // The driver must return 'MULTI_BUF_PTR' for the parameter 'dataPtr' of method 'BfdAoPrepare' 
   // when it supports this method
   virtual ErrorCode BDAQCALL BfdAoSetData(
      uint32   &dataCount,      // IN/OUT, Necessary
      void     *rawBuf,         // IN,     Optional  
      double   *dblBuf)         // IN,     Optional
   {
      return ErrorFuncNotSpted;
   }

public:
   // Undocumented functions: query FAO State, used by fusion driver
   // RETURN: a combination of one or more of DAQ_FN_xxx.
   virtual uint32 BDAQCALL BfdAoGetState()
   {
      return DAQ_FN_IDLE;
   }
};

class BDAQ_NO_VTABLE DaqDioModule : public DaqModule
{
public:
   // Override: DaqMoudle
   virtual ModuleType BDAQCALL getType()   { return DaqDio; }
   virtual int32      BDAQCALL getNumber() { return 0;      } // default module index

   // New method
   // Not all DIO modules support all these operations.
   virtual ErrorCode BDAQCALL DiReadPort(uint32 portStart, uint32 portCount, uint8 buffer[])
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL DiReadBit(uint32 port, uint32 bit, uint8 &data)
   {
      uint8 status;
      ErrorCode res = DiReadPort(port, 1, &status);
      if (res == Success) { data = (status >> bit) & 0x1; }
      return res;
   }

   virtual ErrorCode BDAQCALL DoWritePort(uint32 portStart, uint32 portCount, uint8 buffer[])
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL DoWriteBit(uint32 port, uint32 bit, uint8 data)
   {
      uint8 status;
      ErrorCode res = DoReadPort(port, 1, &status);
      if (res != Success) { return res; }

      status = ((data & 0x1) << bit) | (~(1 << bit) & status);
      return DoWritePort(port, 1, &status);
   }

   virtual ErrorCode BDAQCALL DoReadPort(uint32 portStart, uint32 portCount, uint8 buffer[])
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL DoReadBit(uint32 port, uint32 bit, uint8 &data)
   {
      uint8 status;
      ErrorCode res = DoReadPort(port, 1, &status);
      if (res == Success) { data = (status >> bit) & 0x1; }

      return res;
   }

   virtual ErrorCode BDAQCALL DiSnapStart(EventId id, uint32 portStart, uint32 portCount, uint8 **dataPtr)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL DiSnapStop(EventId id)
   {
      return ErrorFuncNotSpted;
   }

   // buffered DI
   virtual ErrorCode BDAQCALL BfdDiPrepare(uint32 &dataCount, void* &dataPtr)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdDiStart()
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdDiStop()
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdDiRelease()
   {
      return ErrorFuncNotSpted;
   }

   virtual ErrorCode BDAQCALL BfdDiGetData(
      uint32   &dataCount,      // IN/OUT, Necessary
      void     *dataBuf,         // OUT,    Optional  
      double   *startTime,      // OUT,    Optional
      uint32   *markCount,      // IN/OUT, Optional
      DataMark *markBuf)        // OUT,    Optional
   {
      return ErrorFuncNotSpted;
   }

// buffered DO
   //
   // !!!CAUTION!!!
   // The driver must return 'MULTI_BUF_PTR' for parameter 'dataPtr' 
   // when it supports the new method 'BfdDoSetData'
   // !!!CAUTION!!!
   virtual ErrorCode BDAQCALL BfdDoPrepare(uint32 &dataCount, void* &dataPtr)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdDoRunOnce(bool asynchronous)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdDoRun()
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdDoStop(uint32 action)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdDoRelease()
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdDoSetData(uint32 &dataCount, void *dataBuf)
   {
      return ErrorFuncNotSpted;
   }
   
public:
   // Undocumented functions
   // 0 -- start, 1 -- stop, 2 -- feed
   enum WdtAction { WdtStart = 0, WdtStop, WdtFeed };
   virtual ErrorCode BDAQCALL WdtExec(WdtAction action)
   {
      return ErrorFuncNotSpted;
   }

public:
   // Undocumented functions: query FDI/FDO State, used by fusion driver
   // RETURN: a combination of one or more of DAQ_FN_xxx.
   virtual uint32 BDAQCALL BfdDiGetState()
   {
      return DAQ_FN_IDLE;
   }
   virtual uint32 BDAQCALL BfdDoGetState()
   {
      return DAQ_FN_IDLE;
   }
};

class BDAQ_NO_VTABLE DaqCtModule : public DaqModule
{
public:
   // Override: DaqMoudle
   virtual ModuleType BDAQCALL getType()   { return DaqCounter; }
   virtual int32      BDAQCALL getNumber() { return 0;     } // default module index

   // New method
   // Common methods
   virtual ErrorCode BDAQCALL CtReset(uint32 cntrStart, uint32 cntrCount) = 0;

   // Not all counter modules support the following operation.
   virtual ErrorCode BDAQCALL CtRead(uint32 cntrStart, uint32 cntrCount, uint32 loadVal[], uint32 holdVal[])
   {
      return ErrorFuncNotSpted;
   }

   // Primary counter
   virtual ErrorCode BDAQCALL CtStart(uint32 cntrStart, uint32 cntrCount)
   {
      return ErrorFuncNotSpted;
   }

   // Event Counting methods
   virtual ErrorCode BDAQCALL EcStart(uint32 cntrStart, uint32 cntrCount)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL EcRead(uint32 cntrStart, uint32 cntrCount, uint32 cntrValue[])
   {
      return ErrorFuncNotSpted;
   }

   // One-Shot methods
   virtual ErrorCode BDAQCALL OsStart(uint32 cntrStart, uint32 cntrCount)
   {
      return ErrorFuncNotSpted;
   }

   // Timer/Pulse methods
   virtual ErrorCode BDAQCALL TmrStart(uint32 cntrStart, uint32 cntrCount)
   {
      return ErrorFuncNotSpted;
   }

   // Frequency Measurement methods
   virtual ErrorCode BDAQCALL FmStart(uint32 cntrStart, uint32 cntrCount)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL FmRead(uint32 cntrStart, uint32 cntrCount, double frequency[])
   {
      return ErrorFuncNotSpted;
   }

   // Pulse width measurement
   virtual ErrorCode BDAQCALL PiStart(uint32 cntrStart, uint32 groupCount)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL PiRead(uint32 cntrStart, uint32 groupCount, double hiPeriod[], double lowPeriod[])
   {
      return ErrorFuncNotSpted;
   }

   // Pulse width modulation
   virtual ErrorCode BDAQCALL PoStart(uint32 cntrStart, uint32 cntrCount)
   {
      return ErrorFuncNotSpted;
   }

   // Up-down count
   virtual ErrorCode BDAQCALL UdStart(uint32 cntrStart, uint32 cntrCount, uint32 resetCounter)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL UdResetValue(uint32 cntrStart, uint32 cntrCount)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL UdRead(uint32 cntrStart, uint32 cntrCount, uint32 cntrValue[])
   {
      return ErrorFuncNotSpted;
   }

   // counter snap
   virtual ErrorCode BDAQCALL CtSnapStart(EventId id, uint32 cntrStart, uint32 cntrCount, uint32 **loadValues, uint32 **holdValues)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL CtSnapStop(EventId id)
   {
      return ErrorFuncNotSpted;
   }

   // continuous compare
   virtual ErrorCode BDAQCALL CCmpSetTable(uint32 counter, uint32 dataCount, uint32 buffer[])
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL CCmpSetInterval(uint32 counter, uint32 start, uint32 step, uint32 count)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL CCmpClear(uint32 counter)
   {
      return ErrorFuncNotSpted;
   }

   // buffered counter in/out
   virtual ErrorCode BDAQCALL BfdCtPrepare(uint32 &dataCount)
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdCtStart()
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdCtStop()
   {
      return ErrorFuncNotSpted;
   }
   virtual ErrorCode BDAQCALL BfdCtRelease()
   {
      return ErrorFuncNotSpted;
   }

   virtual ErrorCode BDAQCALL BfdCtGetData(
      uint32   cntr,            // IN, Necessary
      uint32   &dataCount,      // IN/OUT, Necessary
      void     *dataBuf,        // OUT,    Optional
      double   *startTime,      // OUT,    Optional
      void     *reserved)       // OUT,    Optional
   {
      return ErrorFuncNotSpted;
   }
   
   virtual ErrorCode BDAQCALL BfdCtSetData(uint32 cntr, void *dataBuf, uint32 &dataCount)
   {
      return ErrorFuncNotSpted;
   }

   // edge separation measurement
   virtual ErrorCode BDAQCALL EsStart(uint32 cntrStart, uint32 cntrCount)
   {
      return ErrorFuncNotSpted;
   }

   virtual ErrorCode BDAQCALL EsRead(uint32 cntrStart, uint32 cntrCount, double *timeSeparation)
   {
      return ErrorFuncNotSpted;
   }

public:
   // Undocumented functions: query FCT State, used by fusion driver
   // RETURN: a combination of one or more of DAQ_FN_xxx.
   virtual uint32 BDAQCALL BfdCtGetState()
   {
      return DAQ_FN_IDLE;
   }
};

// Rename to a More normative name
typedef DaqCtModule DaqCtrModule;

//
// New module for device calibration
class BDAQ_NO_VTABLE DaqCaliModule : public DaqModule
{
public:
   union ObjectId {
      uint32 xval;
      struct {
         uint8 sln;
         uint8 sect;
         uint8 subj;
      };
   };

public:
   // Override: DaqMoudle
   virtual ModuleType    BDAQCALL getType()                                                          { return DaqCali; }
   virtual int32         BDAQCALL getNumber()                                                        { return 0;       } // default module index
   virtual ErrorCode     BDAQCALL PropAccess(PropertyId id, uint32 &size, void *buffer, uint32 flag) { return ErrorPropNotSpted; }

   // Method: calibration
   virtual uint32        BDAQCALL getSolutionList(uint32 size, ObjectId buffer[]) = 0;
   
   // Method: solution object
   virtual uint32        BDAQCALL solution_getTypeName(ObjectId id, uint32 size, wchar_t buffer[])    { return 0; }
   virtual uint32        BDAQCALL solution_getDescription(ObjectId id, uint32 size, wchar_t buffer[]) { return 0; }
   virtual uint32        BDAQCALL solution_getInstruction(ObjectId id, uint32 size, wchar_t buffer[]) { return 0; }
   virtual uint8 const * BDAQCALL solution_getAuxImage(ObjectId id, uint32 *size)                     { return NULL; }
   virtual uint32        BDAQCALL solution_getSectionList(ObjectId id, uint32 size, ObjectId buffer[]){ return 0;    } 

   virtual ErrorCode     BDAQCALL solution_AdjCodesSave(ObjectId id)                                   { return ErrorFuncNotSpted; };
   virtual ErrorCode     BDAQCALL solution_AdjCodesLoad(ObjectId id)                                   { return ErrorFuncNotSpted; };
   virtual ErrorCode     BDAQCALL solution_AdjCodesSaveToGoldBackup(ObjectId id)                       { return ErrorFuncNotSpted; }
   virtual ErrorCode     BDAQCALL solution_AdjCodesLoadFromGoldBackup(ObjectId id, int32 action)       { return ErrorFuncNotSpted; }

   // Method: section object
   virtual uint32        BDAQCALL section_getTypeName(ObjectId id, uint32 size, wchar_t buffer[])     { return 0; }
   virtual uint32        BDAQCALL section_getDescription(ObjectId id, uint32 size, wchar_t buffer[])  { return 0; }
   virtual uint32        BDAQCALL section_getInstruction(ObjectId id, uint32 size, wchar_t buffer[])  { return 0; }
   virtual uint32        BDAQCALL section_getSubjectList(ObjectId id, uint32 size, ObjectId buffer[]) { return 0;    }

   virtual ErrorCode     BDAQCALL section_BatchStart(ObjectId id)                                     { return ErrorFuncNotSpted; }
   virtual void          BDAQCALL section_BatchTerminate(ObjectId id)                                 { };
   
   // Method: subject object
   virtual uint32        BDAQCALL subject_getDescription(ObjectId id, uint32 size, wchar_t buffer[])  { return 0; }
   virtual uint32        BDAQCALL subject_getInstruction(ObjectId id, uint32 size, wchar_t buffer[])  { return 0; }
   virtual uint32        BDAQCALL subject_getTargetRange(ObjectId id, uint32 size, wchar_t buffer[])  { return 0; }

   virtual CaliState     BDAQCALL subject_AdjCodeGetCurrentState(ObjectId id, double *value, uint32 *count, uint32 code[]) { return CaliIdle; }
   virtual void          BDAQCALL subject_AdjCodeGetRange(ObjectId id, uint32 lower[], uint32 upper[]){  }
   virtual ErrorCode     BDAQCALL subject_ManualAdjStart(ObjectId id)                                 { return ErrorFuncNotSpted; }
   virtual ErrorCode     BDAQCALL subject_ManualAdjStop(ObjectId id)                                  { return ErrorFuncNotSpted; }
   virtual ErrorCode     BDAQCALL subject_ManualAdjSetCode(ObjectId id, uint32 count, uint32 code[])  { return ErrorFuncNotSpted; }

   virtual ErrorCode     BDAQCALL section_Prepare(ObjectId id)                                        { return Success; }
};

//
// New module for device firmware download
class BDAQ_NO_VTABLE DaqFwModule : public DaqModule
{

public:
   // Override: DaqMoudle
   virtual ModuleType    BDAQCALL getType()                                                          { return DaqFw; }
   virtual int32         BDAQCALL getNumber()                                                        { return 0;     } // default module index
   virtual ErrorCode     BDAQCALL PropAccess(PropertyId id, uint32 &size, void *buffer, uint32 flag) { return ErrorPropNotSpted; }

   // Method: Firmware download
   virtual int32         BDAQCALL getCount()                                                         { return 1;     }  // default only one firmware item

   // Method: firmware object
   virtual uint32        BDAQCALL getDescription(int32 index, uint32 size, wchar_t buffer[])         { return 0;     }
   virtual uint32        BDAQCALL getNotice(int32 index, uint32 size, wchar_t buffer[])              { return 0;     }
   virtual uint32        BDAQCALL getActivationTitle(int32 index, uint32 size, wchar_t buffer[])     { return 0;     }
   virtual uint32        BDAQCALL getVersionOnDevice(int32 index, uint32 size, wchar_t buffer[])     { return 0;     }
   virtual uint32        BDAQCALL getVersionOnFile(int32 index, uint32 size, wchar_t buffer[])       { return 0;     }
   virtual int32         BDAQCALL getOverwriteProgressing(int32 index)                               { return -1000; }

   virtual ErrorCode     BDAQCALL LoadFwFile(int32 index, wchar_t const *filePath)                       { return ErrorFuncNotSpted; }
   virtual ErrorCode     BDAQCALL Overwrite(int32 index, FwAction action, wchar_t const *activationCode) { return ErrorFuncNotSpted; }
};

#ifdef _MSC_VER
#pragma warning(default:4100)
#endif

#endif 
