#ifndef _DAQNAVI_LOCK_H
#define _DAQNAVI_LOCK_H

#if defined(_WIN32) || defined(WIN32) 
#  include <windows.h>
#else
#  include <pthread.h>
#  include <semaphore.h>
#endif

#if defined(_WIN32) || defined(WIN32)
//
// atomic operation 
#  define ATOMIC_SUPPORTED  1

__inline int32 AtomicExchange(volatile int32 *ptr, int32 val) { return InterlockedExchange((long*)ptr, val); }
__inline int32 AtomicIncrement(volatile int32 *ptr)           { return InterlockedIncrement((long*)ptr); }
__inline int32 AtomicDecrement(volatile int32 *ptr)           { return InterlockedDecrement((long*)ptr); }

__inline int32 AtomicCompareExchange(volatile int32 *ptr, int32 val, int32 comperand) 
{ 
   return InterlockedCompareExchange((long*)ptr, val, comperand);
}

__inline int32 AtomicOr (volatile int32 *ptr, int32 val)
{
   int32 old;
   do {
      old = *ptr;
   } while (InterlockedCompareExchange((long*)ptr, old | val, old) != old);

   return old;
}

__inline int32 AtomicXor(volatile int32 *ptr, int32 val)
{
   int32 old;
   do { 
      old = *ptr;
   } while (InterlockedCompareExchange((long*)ptr, old ^ val, old) != old);

   return old;
}

__inline int32 AtomicAnd(volatile int32 *ptr, int32 val)
{
   int32 old;
   do { 
      old = *ptr;
   } while (InterlockedCompareExchange((long*)ptr, old & val, old) != old);

   return old;
}

__inline int32 AtomicBTS64(volatile int64 *ptr, int32 bit) 
{
#ifdef _WIN64
   return _interlockedbittestandset64(ptr, bit); 
#else 
   return bit < 32 ? InterlockedBitTestAndSet((long*)ptr, bit) \
                   : InterlockedBitTestAndSet((long*)ptr + 1, bit - 32);
#endif
}

__inline int32 AtomicBTR64(volatile int64 *ptr, int32 bit) 
{
#ifdef _WIN64
   return _interlockedbittestandreset64(ptr, bit); 
#else 
   return bit < 32 ? InterlockedBitTestAndReset((long*)ptr, bit) \
                   : InterlockedBitTestAndReset((long*)ptr + 1, bit - 32);
#endif
}

//
// fast mutex
typedef CRITICAL_SECTION FastMutex;
__inline void FastMutexInitialize(FastMutex * fm) { InitializeCriticalSection(fm);}
__inline void FastMutexDestroy(FastMutex *fm)     { DeleteCriticalSection(fm);    }
__inline void FastMutexLock(FastMutex *fm)        { EnterCriticalSection(fm);     }
__inline void FastMutexUnlock(FastMutex *fm)      { LeaveCriticalSection(fm);     }

#else
//
// atomic operation 
#  if (__GNUC__ > 4) || ((__GNUC__ == 4) && (__GNUC_MINOR__ >= 1)) // Need gcc >= 4.1.2
#  define ATOMIC_SUPPORTED  1
__inline int32 AtomicExchange(volatile int32 *ptr, int32 val) { return __sync_lock_test_and_set(ptr, val); }
__inline int32 AtomicIncrement(volatile int32 *ptr)           { return __sync_add_and_fetch(ptr, 1);  }
__inline int32 AtomicDecrement(volatile int32 *ptr)           { return __sync_sub_and_fetch(ptr, 1);  }
__inline int32 AtomicOr(volatile int32 *ptr, int32 val)       { return __sync_or_and_fetch(ptr, val); }
__inline int32 AtomicXor(volatile int32 *ptr, int32 val)      { return __sync_xor_and_fetch(ptr, val);}
__inline int32 AtomicAnd(volatile int32 *ptr, int32 val)      { return __sync_and_and_fetch(ptr, val);}

__inline int32 AtomicCompareExchange(volatile int32 *ptr, int32 val, int32 comperand) 
{ 
   return __sync_val_compare_and_swap(ptr, comperand, val); 
}

__inline int32 AtomicBTS64(volatile int64 *ptr, int32 bit)
{
   int64 old, k = 0x1 << bit;
   do {
      old = *ptr;
   } while (__sync_val_compare_and_swap(ptr, old, old | k) != old);

   return !!(old & k);
}

__inline int32 AtomicBTR64(volatile int64 *ptr, int32 bit)
{
   int64 old, k = ~(0x1 << bit);
   do {
      old = *ptr;
   } while (__sync_val_compare_and_swap(ptr, old, old & k) != old);

   return !!(old & k);
}

#  endif

// fast mutex
typedef pthread_mutex_t FastMutex;
__inline void FastMutexInitialize(FastMutex * fm) 
{  
   // 2017-11-16, initialize the posix mutex as a recursive mutex (default is non-recursive)
   pthread_mutexattr_t attr;
   pthread_mutexattr_init(&attr);
   pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
   pthread_mutex_init(fm, &attr); 
}
__inline void FastMutexDestroy(FastMutex *fm)     { pthread_mutex_destroy(fm);    }
__inline void FastMutexLock(FastMutex *fm)        { pthread_mutex_lock(fm);       }
__inline void FastMutexUnlock(FastMutex *fm)      { pthread_mutex_unlock(fm);     }

#endif

#ifdef __cplusplus

#if defined(_WIN32) || defined(WIN32)
//
// Named mutext
class NamedMutexLock
{
public:
   NamedMutexLock(TCHAR *name) { m_mutex = CreateMutex(NULL, FALSE, name); }
   ~NamedMutexLock()           {  CloseHandle(m_mutex);   }
   void Lock()                 {  WaitForSingleObject(m_mutex, INFINITE);   }
   void Unlock()               {  ReleaseMutex(m_mutex);   }
private:
   HANDLE m_mutex;
};

//
// Semaphore
class Semaphore
{
public:
   Semaphore(int initCount = 0)
   {
      m_sema = CreateSemaphore(NULL, initCount, 0x7FFFFFFF, NULL);
      if (m_sema == NULL) {
         throw "Create Semaphore failed!";
      }
   }
   virtual ~Semaphore()
   {
      CloseHandle(m_sema);
   }

   bool Post()
   {
      LONG prev;
      return ReleaseSemaphore(m_sema, 1, &prev) != 0;
   }

   bool Wait()
   {
      return WaitForSingleObject(m_sema, INFINITE) == WAIT_OBJECT_0;
   }

protected:
   HANDLE m_sema;
};

#else 

//
// POSIX Semaphore
class Semaphore
{
public:
   Semaphore(int initCount = 0)
   {
      if (sem_init(&m_sema, 0, initCount) == -1) {
         throw "Create Semaphore failed!";
      }
   }
   virtual ~Semaphore()
   {
      sem_destroy(&m_sema);
   }

   bool Post()
   {
      return sem_post(&m_sema) == 0;
   }

   bool Wait()
   {
      return sem_wait(&m_sema) == 0;
   }

protected:
   sem_t m_sema;
};

#endif

//-----------------------------------------------------------------------------------------
// A busy-wait lock to synchronize the access to shared data between user mode and kernel mode code.
//-----------------------------------------------------------------------------------------
#if defined(ATOMIC_SUPPORTED)
class BusyLock
{
public:
   BusyLock() : m_lock(0) { }
   void Lock()   {  while (AtomicExchange(&m_lock, 1)){};  }
   void Unlock() {  AtomicExchange(&m_lock, 0);  }
protected:
   volatile int32 m_lock;
};
#endif

class MutexLock
{
public:
   MutexLock()   { FastMutexInitialize(&m_mutex);  }
   ~MutexLock()  { FastMutexDestroy(&m_mutex);     }
   void Lock()   { FastMutexLock(&m_mutex);        }
   void Unlock() { FastMutexUnlock(&m_mutex);      }
private:
   FastMutex m_mutex;
};

template<class Lockable>
class AutoLock
{
public:
   AutoLock(Lockable & lock) : m_lock(lock) { lock.Lock(); }
   ~AutoLock()                              { m_lock.Unlock();  }
private:
   Lockable & m_lock;
};

#endif

#endif  // __cplusplus
