#ifndef _SQLITE_WRAPPER_H
#define _SQLITE_WRAPPER_H

#define SQLITE_MAX_RETRY 1000  // MAX. retry count if the db is locked
#define SQLITE_WAIT_TIME 10    // 10ms, Waiting time before we can retry the operation again

class SQLiteWrapper
{
public:
   SQLiteWrapper(char const * dbFile = NULL)
      : m_db(NULL), m_stmt(NULL), m_isLocked(false)
   { 
      if (dbFile) {
         Open(dbFile);
      }
   }

   virtual ~SQLiteWrapper()
   {
      Close();
   }

   int Open(char const * dbFile)
   {
      if (m_db) {
         sqlite3_close(m_db);
      }
      return sqlite3_open(dbFile, &m_db);
   }

   void Finalize()
   {
      if (m_stmt != NULL) {
         sqlite3_finalize(m_stmt);
         m_stmt = NULL;
      }
   }

   void Close()
   {
      Finalize();
      if (m_db != NULL) {
         sqlite3_close(m_db);
         m_db = NULL;
      }
   }

   void Lock() 
   { 
      m_dbLock.Lock(); 
      m_isLocked = true;
   }

   void Unlock() 
   { 
      m_isLocked = false;
      m_dbLock.Unlock(); 
   }

   sqlite3 * getHandle()
   {
      return m_db;
   }

public:

   bool IsOpened()
   {
      return m_db != NULL;
   }

   bool IsLocked()
   {
      return m_isLocked;
   }

   int Prepare(char const * format, ...)
   {
      assert(format != NULL);

      Finalize();

      va_list args;
      va_start(args, format);

      // build sql string
      int len = m_buffer.Format(format, args);
      if (len == -1) { return SQLITE_NOMEM; }

      // compile the sql string
      int rc = SQLITE_OK;
      for (int i = 0; i < SQLITE_MAX_RETRY; ++i) {
         rc = sqlite3_prepare_v2(m_db, m_buffer.Buffer(), len, &m_stmt, NULL);

         // DB access successfully or other errors
         if (rc != SQLITE_BUSY && rc != SQLITE_LOCKED) {
            break;
         }

         x_msleep(SQLITE_WAIT_TIME);
      }

      return rc;
   }

   int Step()
   {
      int rc = SQLITE_OK;
      for (int i = 0; i < SQLITE_MAX_RETRY; ++i) {
         rc = sqlite3_step(m_stmt);

         // DB access successfully or other errors
         if (rc != SQLITE_BUSY && rc != SQLITE_LOCKED) {
            break;
         }

         x_msleep(SQLITE_WAIT_TIME);
      }

      return rc;
   }

   int Exec(char const *format, ...)
   {
      assert(format != NULL);

      va_list args;
      va_start(args, format);

      // build sql string
      int len = m_buffer.Format(format, args);
      if (len == -1) { return SQLITE_NOMEM; }

      int rc = SQLITE_OK;
      for (int i = 0; i < SQLITE_MAX_RETRY; ++i) {
         rc = sqlite3_exec(m_db, m_buffer.Buffer(), NULL, NULL, NULL);

         // DB access successfully or other errors
         if (rc != SQLITE_BUSY && rc != SQLITE_LOCKED) {
            break;
         }

         x_msleep(SQLITE_WAIT_TIME);
      }

      return rc;
   }

   char const * ErrMsg()
   {
      return sqlite3_errmsg(m_db);
   }

   char const * Column(int col)
   {
      assert(m_stmt);
      char const *data = (char const *)sqlite3_column_text(m_stmt, col);
      return data ? data : "";
   }

   int Column(int col, int)
   {
      assert(m_stmt);
      return sqlite3_column_int(m_stmt, col);
   }

   double Column(int col, double)
   {
      assert(m_stmt);
      return sqlite3_column_double(m_stmt, col);
   }

   //
   // Transaction method
   int Begin()
   {
      return Exec("BEGIN TRANSACTION;");
   }

   int Commit()
   {
      return Exec("COMMIT TRANSACTION;");
   }

   int Rollback()
   {
      return Exec("ROLLBACK TRANSACTION;");
   }

private:
   // DISALLOW_COPY_AND_ASSIGN(SQLiteWrapper);
   SQLiteWrapper(const SQLiteWrapper&);
   void operator=(const SQLiteWrapper&);

private:
   sqlite3      *m_db;
   sqlite3_stmt *m_stmt;
   XBuffer       m_buffer;  
   MutexLock     m_dbLock;
   bool          m_isLocked;
};

class SQLiteTable
{
public:
   SQLiteTable(SQLiteWrapper &db) : m_db(db), m_data(NULL) 
   {
   }

   ~SQLiteTable() 
   {
      if (m_data) { sqlite3_free_table(m_data); }
   }

   int Load(char const *format, ...)
   {
      assert(format != NULL);

      // build SQL string
      XBuffer xbuf;
      va_list args;
      va_start(args, format);

      int len = xbuf.Format(format, args);
      if (len == -1) { return SQLITE_NOMEM; }

      if (m_data) { 
         sqlite3_free_table(m_data); 
         m_data = NULL;
      }

      int rc = SQLITE_OK;
      for (int i = 0; i < SQLITE_MAX_RETRY; ++i) {
         rc = sqlite3_get_table(m_db.getHandle(), xbuf.Buffer(), &m_data, &m_rows, &m_cols, NULL); 

         // DB access successfully or other errors
         if (rc != SQLITE_BUSY && rc != SQLITE_LOCKED) {
            break;
         }

         x_msleep(SQLITE_WAIT_TIME);
      }

      return rc;
   }

   int Rows()
   { 
      return m_rows; 
   }

   int Cols()
   { 
      return m_cols; 
   }
   
   char const * Column(int col)  
   { 
      assert(col < m_cols); 
      return m_data[col]; 
   }

   char const * Cell(int row, int col)
   {
      assert(row < m_rows && col < m_cols);
      char *data = m_data[(row + 1)* m_cols + col];
      return data ? data : "";
   }

   int Cell(int row, int col, int)
   {
      return atoi(Cell(row, col));
   }

   double Cell(int row, int col, double)
   {
      return atof(Cell(row, col));
   }

private:
   // DISALLOW_COPY_AND_ASSIGN(SQLiteTable);
   SQLiteTable(const SQLiteTable&);
   void operator=(const SQLiteTable&);

private:
   SQLiteWrapper &m_db;
   char ** m_data;
   int     m_rows;
   int     m_cols;
};

class SQLiteScopedPtr
{
public:
   SQLiteScopedPtr(SQLiteWrapper &db) : m_db(db) 
   { 
      // For bug#21444: the 'db' might have been locked before it is passed to us.
      if (!db.IsLocked()) {
         m_db.Lock();     
      }
   }
   ~SQLiteScopedPtr()
   { 
      m_db.Finalize();
      m_db.Unlock(); 
   }

   SQLiteWrapper & operator *()                  
   { 
      return m_db;     
   }
   SQLiteWrapper * operator ->()
   { 
      return &m_db;    
   }

private:
   // DISALLOW_COPY_AND_ASSIGN(SQLiteScopedPtr);
   SQLiteScopedPtr(const SQLiteScopedPtr&);
   void operator=(const SQLiteScopedPtr&);

private:
   SQLiteWrapper &m_db;
};


#endif
