//
// Thin wrapper of OS thread operation methods
//
#ifndef _DAQNAVI_X_THREAD_H
#define _DAQNAVI_X_THREAD_H

#if defined(_WIN32) || defined(WIN32) 
#  include <process.h>
#else
#  include <pthread.h>
#  include <limits.h>
#endif

typedef unsigned int (BDAQCALL * XThreadProc)(void *context);

//-------------------------------------------------------------------------
// Win/Posix Thread 
//-------------------------------------------------------------------------
class XThread
{
public:
   XThread() : m_isRunning(FALSE), m_thread(0)
   {
   }

   virtual ~XThread() 
   {
      Cleanup();
   }

   bool Start(XThreadProc proc, void *ctx)
   {
      AutoLock<MutexLock> lock(m_threadLock);

      // Busy!!! The previous thread is still running!!!
      if (m_isRunning) {
         return false; 
      }

      // Clean up
      Cleanup();
      
      // Start the new thread
      m_threadProc = proc;
      m_threadCtx  = ctx;
      m_isRunning  = true;

#if defined(_WIN32) || defined(WIN32) 
      m_thread     = (HANDLE)_beginthreadex(NULL, 0, Win32ThreadProc, this, 0, NULL);
      if (m_thread) {
         return true;
      }
#else
      if (pthread_create(&m_thread, NULL, PosixThreadProc, this) == 0) {
         return true;
      }
#endif

      // Start the new thread failed.
      m_isRunning = false;
      return false;     
   }

   bool Stop(int timeout_ms)
   {
      AutoLock<MutexLock> lock(m_threadLock);

      // The thread is not running (not started or had stopped)
      bool quited = true;
      if (m_isRunning) {
         quited = WaitForThread(timeout_ms);
         if (!quited) {
#if defined(_WIN32) || defined(WIN32) 
            TerminateThread(m_thread, -1);
#else
            daq_trace(("Wait for the posix thread exit failed\n"));
#endif
         }
      }

      Cleanup();
      m_isRunning = false;

      return quited;
   }

   bool IsRunning() 
   {
      return m_isRunning;
   }

private:

#if defined(_WIN32) || defined(WIN32) 
   //-------------------------------------------------------------------------
   // Win Thread Methods
   //-------------------------------------------------------------------------
   static unsigned BDAQCALL Win32ThreadProc(void *ctx) 
   { 
      XThread   *_this = static_cast<XThread *>(ctx);
      unsigned int ret = 0;

      try {
         ret = _this->m_threadProc(_this->m_threadCtx); 
      } catch(...) {
         daq_trace(("Unhandled Exception! handle = 0x%x\n", _this->m_thread));
      }

      _this->m_isRunning = false;
      return ret;
   }

   bool WaitForThread(int timeout_ms)
   {
      return WaitForSingleObject(m_thread, timeout_ms) == WAIT_OBJECT_0;
   }

   void Cleanup()
   {
      CloseHandle(m_thread);
      m_thread = NULL;
   }

   typedef HANDLE XHANDLE;

#else
   //-------------------------------------------------------------------------
   // POSIX Thread Methods
   //-------------------------------------------------------------------------
   static void * BDAQCALL PosixThreadProc(void *ctx) 
   { 
      XThread    *_this = static_cast<XThread *>(ctx);
      unsigned int  ret = 0;

      try {
         ret = _this->m_threadProc(_this->m_threadCtx);  
      } catch(...) {
         daq_trace(("Unhandled Exception! handle = 0x%x\n", _this->m_thread));
      }

      _this->m_isRunning = false;
      return (void *)(int64)ret;
   }

   bool WaitForThread(int timeout_ms)
   {
      if (timeout_ms < 0) {
         timeout_ms = INT_MAX;
      }

      int sec = timeout_ms / 1000;
      int ns  = (timeout_ms - sec * 1000) * 1000 * 1000;
      timespec ts;

      clock_gettime(CLOCK_REALTIME, &ts);

      ts.tv_sec  += sec;
      ts.tv_nsec += ns;
      return pthread_timedjoin_np(m_thread, NULL, &ts) == 0;
   }

   void Cleanup()
   {
      m_thread = 0;
   }

   typedef pthread_t XHANDLE;

#endif

private:
   bool        m_isRunning;   // whether the thread is running or not. 
   XHANDLE     m_thread;
   XThreadProc m_threadProc;
   void       *m_threadCtx;
   MutexLock   m_threadLock;
};

#endif
