/*
 * BioDio.h
 *
 * Created on: October 7, 2012
 * Author: 
 */

#ifndef _BDAQ_DIO_IMPL_H_
#define _BDAQ_DIO_IMPL_H_

#include "krnlstub.h"


EventId const s_dioSptedEvts[]     = { EvtDiintChannel000,  EvtDiintChannel001, EvtDiintChannel016,  EvtDiintChannel017 };  // DIO supported events
uint8   const s_dioPortType[]      = { PortDio, PortDio, PortDio, PortDio };       // DIO port type
uint8   const s_dioSptedTrigEdge[] = { RisingEdge, FallingEdge };                  // DI INT supported trigger edge
int     const s_dioDiintChlNum[]   = { 0, 1, 16, 17 };                             // DI INT supported channels


class BDaqDioImpl : public BDaqDio
{
public:
   // --------------------------------------------------------------
   // BDaqModule interface implementation
   // --------------------------------------------------------------
   virtual ModuleType getType()   { return DaqDio;   }
   virtual int32      getNumber() { return 0; }
   virtual AccessMode getMode()   { return m_kstubPtr->getMode(); }

   virtual ErrorCode EventGetHandle(EventId id, HANDLE &handle);
   virtual ErrorCode EventGetStatus(EventId id, uint32 *lParam, uint32 *rParam);
   virtual ErrorCode EventClearFlag(EventId id, uint32 lParam, uint32 rParam);

   // Property table
   BEGIN_PROPERTY_TABLE(BioDioImpl, PropAccess)
      // component specified properties
      // features
      PROP_RO_VAL(CFG_FeatureChannelNumberMax,    uint32,             DIO_CHL_COUNT - 1         )
      PROP_RO_PTR(CFG_FeaturePortsType,           s_dioPortType,      sizeof(s_dioPortType)      )
      PROP_RO_PTR(CFG_FeatureDiintTriggerEdges,   s_dioSptedTrigEdge, sizeof(s_dioSptedTrigEdge) )
      PROP_RO_PTR(CFG_FeatureDiSnapEventSources,  s_dioSptedEvts,     sizeof(s_dioSptedEvts)     )

      PROP_CUSTOM(CFG_FeatureDiintOfChannels,     ReadOnly,   PropAccessDiintSptedChan        )

      // DIO configuration
      PROP_CUSTOM(CFG_DiintTriggerEdgeOfChannels, Writable,   PropAccessDiintTrigEdge         )
      PROP_CUSTOM(CFG_DoInitialStateOfPorts,      Writable,   PropAccessDoPortInitState       )

   END_PROPERTY_TABLE()

public:
   // --------------------------------------------------------------
   // BDaqDio interface implementation
   // --------------------------------------------------------------
   virtual ErrorCode ReadDiPorts(uint32 portStart, uint32 portCount, uint8 buffer[]);
   virtual ErrorCode WriteDoPorts(uint32 portStart, uint32 portCount, uint8 buffer[]);
   virtual ErrorCode ReadDoPorts(uint32 portStart, uint32 portCount, uint8 buffer[]);
   virtual ErrorCode StartDiSnap(EventId id, uint32 portStart, uint32 portCount, uint8 **dataPtr);
   virtual ErrorCode StopDiSnap(EventId id);
   virtual ErrorCode WriteDoBit(uint32 port, uint32 bit, uint8 data);

public:
   // --------------------------------------------------------------
   // BDaqDioImpl interface implementation
   // --------------------------------------------------------------
   BDaqDioImpl()
   {
      m_kstubPtr = NULL;
      memset(m_dioEvents, 0, sizeof(m_dioEvents));
   }

   virtual ~BDaqDioImpl()
   {
   }

   ErrorCode Initialize(BioKrnlStub *kstub);
   ErrorCode Reset(uint32 state);

protected:
   ErrorCode PropAccessDiintSptedChan(uint32 id, uint32 &bufLen, void *buffer, uint32 opFlag);
   ErrorCode PropAccessDiintTrigEdge(uint32 id, uint32 &bufLen, void *buffer, uint32 opFlag);
   ErrorCode PropAccessDoPortInitState(uint32 id, uint32 &bufLen, void *buffer, uint32 opFlag);

   ErrorCode SetDiIntParam(uint32 setWhich, uint32 srcStart, uint32 srcCount, void *val);

private:
   BioKrnlStub  *m_kstubPtr;
   HANDLE        m_dioEvents[ARRAY_SIZE(s_dioSptedEvts)]; // DIO events
};

// -----------------------------------------------------------------------------------------
// inline functions
// -----------------------------------------------------------------------------------------
inline
ErrorCode BDaqDioImpl::Initialize(BioKrnlStub *kstub)
{
   assert(kstub != NULL);
   m_kstubPtr = kstub;
   return Success;
}

inline
ErrorCode BDaqDioImpl::Reset(uint32 state)
{
   CHK_PRVLG(m_kstubPtr->getMode() != ModeRead);

   // stop DI snap
   for (unsigned i = 0; i < ARRAY_SIZE(s_dioSptedEvts); ++i){
      StopDiSnap((EventId)s_dioSptedEvts[i]);
   }

   if (!state){
      return Success;
   }

   //-- reset the device to default state
   uint8 edges[DI_INT_SRC_COUNT];
   std::fill_n(edges, DI_INT_SRC_COUNT, DEF_DIINT_TRIGEDGE);
   SetDiIntParam(DIINT_SET_TRIGEDGE, 0, DI_INT_SRC_COUNT, edges);

   uint8 do_state[DIO_PORT_COUNT];
   std::fill_n(do_state, DIO_PORT_COUNT, DEF_DO_STATE);
   WriteDoPorts(0, DIO_PORT_COUNT, do_state);

   return Success;
}

inline
ErrorCode BDaqDioImpl::EventGetHandle(EventId id, HANDLE &handle)
{
   handle = NULL;

   // Look for the event in supported event list
   int index = ARRAY_INDEX_OF(s_dioSptedEvts, id);
   if (index == -1) {
      return ErrorEventNotSpted;
   }

   // Create the event if it hasn't been created.
   if (m_dioEvents[index] == NULL) {
      USER_EVENT_INFO info = { id };
      m_kstubPtr->Ioctl(IOCTL_DEVICE_REGISTER_USER_EVENT, &info);
      m_dioEvents[index] = info.Handle;
   }

   handle = m_dioEvents[index];
   return handle != NULL ? Success : ErrorMemoryNotEnough;
}

inline
ErrorCode BDaqDioImpl::EventGetStatus(EventId id, uint32 *lParam, uint32 *rParam)
{
   // Note: Don't disturb the kernel mode driver if we can handle the query here.
   uint32 kdx = GetEventKIndex(id);

   if (rParam){
      *rParam = 0;
   }

   if (kdx >= KdxDiBegin && kdx <= KdxDiEnd) {
      if (lParam){
         *lParam = (m_kstubPtr->getShared()->IsEvtSignaled[kdx]) != 0;
      }
   }  else { // other not supported event type
      return ErrorEventNotSpted;
   }

   return Success;
}

inline
ErrorCode BDaqDioImpl::EventClearFlag(EventId id, uint32 lParam, uint32 rParam)
{
   // Only the owner of the DI event can clear flag, others can only listen.
   CHK_PRVLG(m_kstubPtr->getMode() != ModeRead);

   // For performance reason, I had to implement the function for
   // clearing event flag here and don't disturb the kernel.
   // Note: using InterlockedXXX to synchronize tFeedhe access between user mode and kernel mode.
   uint32 kdx = GetEventKIndex(id);

   if (kdx >= KdxDiBegin && kdx <= KdxDiEnd) {
      *(uint32*)&m_kstubPtr->getShared()->IsEvtSignaled[kdx] = 0;
      return Success;
   }

   return ErrorEventNotSpted;
}

inline
ErrorCode BDaqDioImpl::PropAccessDiintSptedChan(uint32 id, uint32 &bufLen, void *buffer, uint32 opFlag)
{
   uint32 ch_count = DIO_CHL_COUNT;

   if (IsOpGet(opFlag)) {
      uint8 spted_chan[DIO_CHL_COUNT] = {0};

      for (unsigned i = 0; i < ARRAY_SIZE(s_dioDiintChlNum); ++i) {
    	  spted_chan[s_dioDiintChlNum[i]] = true;
      }

      return DefCopyPtr(spted_chan, ch_count * sizeof(spted_chan[0]), buffer, bufLen);
   }

   return ErrorPropReadOnly;
}

inline
ErrorCode BDaqDioImpl::PropAccessDiintTrigEdge(uint32 id, uint32 &bufLen, void *buffer, uint32 opFlag)
{
   uint32 ch_count = DIO_CHL_COUNT;

   if (IsOpGet(opFlag)) {
      uint8 trig_edge[DIO_CHL_COUNT] = {0};
      uint8 *ptr = m_kstubPtr->getShared()->DiintTrigEdge;


      for (unsigned i = 0; i < ARRAY_SIZE(s_dioDiintChlNum); ++i) {
    	  trig_edge[s_dioDiintChlNum[i]] = *ptr++;
      }

      return DefCopyPtr(trig_edge, sizeof(trig_edge), buffer, bufLen);
   } else {
      CHK_PRVLG(m_kstubPtr->getMode() != ModeRead);
      CHK_USER_BUF_LEN(buffer, bufLen >= sizeof(uint8) * ch_count);

      // Check user parameter
      uint8 edges[DI_INT_SRC_COUNT];
		for (unsigned i = 0; i < ARRAY_SIZE(s_dioDiintChlNum); ++i) {
		  edges[i] = ((uint8 *)buffer)[s_dioDiintChlNum[i]];
		  CHK_EXPR( edges[i] == RisingEdge || edges[i] == FallingEdge, ErrorPropValueNotSpted );
		}

      ErrorCode ret = SetDiIntParam(DIINT_SET_TRIGEDGE, 0, DI_INT_SRC_COUNT, edges);
      if (ret == Success && IsOpNtfNow(opFlag)) {
         m_kstubPtr->Ioctl(IOCTL_DEVICE_NOTIFY_PROP_CHGED, NULL);
      }

      return ret;
   }
}

inline
ErrorCode BDaqDioImpl::PropAccessDoPortInitState(uint32 id, uint32 &bufLen, void *buffer, uint32 opFlag)
{
   if (IsOpGet(opFlag)) {
      uint8 state[DIO_PORT_COUNT];
      std::fill_n(state, DIO_PORT_COUNT, DEF_DO_STATE);

      return DefCopyT(state, DIO_PORT_COUNT, (uint8 *)buffer, bufLen);
   } else {
      CHK_PRVLG(m_kstubPtr->getMode() != ModeRead);
      CHK_USER_BUF_LEN(buffer, bufLen >= sizeof(uint8) * DIO_PORT_COUNT);

      ErrorCode ret = Success;
      return ret;
   }
}

inline
ErrorCode BDaqDioImpl::ReadDiPorts(uint32 portStart, uint32 portCount, uint8 buffer[])
{
   CHK_USER_BUF(buffer);

   ErrorCode warning = Success;
   if (portStart >= DIO_PORT_COUNT) {
      portStart %= DIO_PORT_COUNT;
      SET_WARN(warning, WarningParamOutOfRange);
   }
   if (portCount > DIO_PORT_COUNT) {
      portCount = DIO_PORT_COUNT;
      SET_WARN(warning, WarningParamOutOfRange);
   }

   if (portCount == 0) {
      return Success;
   }

   DIO_RW_PORTS xbuf = { portStart, portCount, buffer };
   if (m_kstubPtr->Ioctl(IOCTL_DIO_READ_DI_PORTS, &xbuf)){
      return ErrorDeviceIoTimeOut;
   }
   return warning;
}

inline
ErrorCode BDaqDioImpl::WriteDoPorts(uint32 portStart, uint32 portCount, uint8 buffer[])
{
   CHK_PRVLG(m_kstubPtr->getMode() != ModeRead);
   CHK_USER_BUF(buffer);
   CHK_EXPR(portStart < DIO_PORT_COUNT, ErrorParamOutOfRange);
   CHK_EXPR(portCount <= DIO_PORT_COUNT, ErrorParamOutOfRange);

   if (portCount == 0) {
      return Success;
   }

   DIO_RW_PORTS xbuf = { portStart, portCount, buffer };
   if (m_kstubPtr->Ioctl(IOCTL_DIO_WRITE_DO_PORTS, &xbuf)){
      return ErrorDeviceIoTimeOut;
   }
   return Success;
}

inline
ErrorCode BDaqDioImpl::ReadDoPorts(uint32 portStart, uint32 portCount, uint8 buffer[])
{
   CHK_USER_BUF(buffer);

   ErrorCode warning = Success;
   if (portStart >= DIO_PORT_COUNT) {
     portStart %= DIO_PORT_COUNT;
     SET_WARN(warning, WarningParamOutOfRange);
   }
   if (portCount > DIO_PORT_COUNT) {
     portCount = DIO_PORT_COUNT;
     SET_WARN(warning, WarningParamOutOfRange);
   }

   if (portCount == 0) {
     return Success;
   }

   DIO_RW_PORTS xbuf = { portStart, portCount, buffer };
   if (m_kstubPtr->Ioctl(IOCTL_DIO_READ_DO_PORTS, &xbuf)){
     return ErrorDeviceIoTimeOut;
   }
   return warning;
}

inline
ErrorCode BDaqDioImpl::StartDiSnap(EventId id, uint32 portStart, uint32 portCount, uint8 **dataPtr)
{
   CHK_PRVLG(m_kstubPtr->getMode() != ModeRead);

   if (dataPtr) {
      *dataPtr = NULL;
   }

   uint32 kdx = GetEventKIndex(id);
   if (kdx < KdxDiBegin || kdx > KdxDiEnd) {
      return ErrorEventNotSpted;
   }

   ErrorCode warning = Success;
   if (portStart >= DIO_PORT_COUNT) {
      portStart %= DIO_PORT_COUNT;
      SET_WARN(warning, WarningParamOutOfRange);
   }
   if (portCount > DIO_PORT_COUNT) {
      portCount = DIO_PORT_COUNT;
      SET_WARN(warning, WarningParamOutOfRange);
   }

   DIO_START_DI_SNAP param = { id, portStart, portCount };
   if (m_kstubPtr->Ioctl(IOCTL_DIO_START_DI_SNAP, &param)){
      return ErrorDeviceIoTimeOut;
   }

   if (dataPtr) {
      *dataPtr = (uint8 *)m_kstubPtr->getShared()->DiSnap[kdx - KdxDiBegin].State;
   }

   return warning;
}

inline
ErrorCode BDaqDioImpl::StopDiSnap(EventId id)
{
   CHK_PRVLG(m_kstubPtr->getMode() != ModeRead);

   uint32 kdx = GetEventKIndex(id);
   if (kdx < KdxDiBegin || kdx > KdxDiEnd) {
      return ErrorEventNotSpted;
   }

   m_kstubPtr->Ioctl(IOCTL_DIO_STOP_DI_SNAP, (void*)id);
   return Success;
}

inline
ErrorCode BDaqDioImpl::SetDiIntParam(uint32 setWhich, uint32 srcStart, uint32 srcCount, void *val)
{
   CHK_USER_BUF(val);
   assert(setWhich == DIINT_SET_TRIGEDGE);

   DIO_SET_DIINT_CFG xbuf = { setWhich, srcStart, srcCount, (uint8 *)val };
   m_kstubPtr->Ioctl(IOCTL_DIO_SET_DIINT_CFG, &xbuf);

   return Success;
}

inline 
ErrorCode BDaqDioImpl::WriteDoBit(uint32 port, uint32 bit, uint8 data)
{
   CHK_PRVLG(m_kstubPtr->getMode() != ModeRead);
   CHK_EXPR(port < DIO_PORT_COUNT, ErrorParamOutOfRange);
   CHK_EXPR(bit < 8, ErrorParamOutOfRange);

   DIO_RW_BIT xbuf = { port, bit, data };
   if (m_kstubPtr->Ioctl(IOCTL_DIO_WRITE_DO_BIT, &xbuf)){
      return ErrorDeviceIoTimeOut;
   }
   return Success;
}

#endif /* _BDAQ_DIO_IMPL_H_ */
