#include <std_decl.h>
#include <malloc.h>
#include <wchar.h>
#include <bdaq_ctrl_ansi.h>

#include "daqnavi3_api.h"
#include "daqnavi_java.h"

#if defined(_WIN32) || defined(WIN32)
#  define x_stricmp    _stricmp
#else
#  define x_stricmp    strcasecmp
#endif

#define NATIVEBUF_CUR_POS(nbuf)  ((uint8*)nbuf.memory + (nbuf.elementSize / 8) * nbuf.position)

/*****************************************************************************/
/* Helper methods convert a wchar_t(UTF-16 or UTF-32) to jchar or vice versa */
/*****************************************************************************/
int UtfToJChar(jchar *out, wchar_t const *in)
{
   int len = 0;
   while ( (*out++ = (jchar)*in++) != 0) {
      ++len;
   }
   return len;
}

int JCharToUtf(wchar_t *out, jchar const *in)
{
   int len = 0;
   while ( (*out++ = (wchar_t)*in++) != 0) {
      ++len;
   }
   return len;
}

/************************************************************************/
/* JNI IDs used to access the JAVA object  and its helper accessors     */
/************************************************************************/
struct JavaByteByRef
{
   jfieldID fid_value;

   uint8 GetValue(JNIEnv * env, jobject jniObj)
   {
      return env->GetByteField(jniObj, fid_value);
   }
   void SetValue(JNIEnv * env, jobject jniObj, uint8 value)
   {
      env->SetByteField(jniObj, fid_value, value);
   }

   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/ByteByRef");
      fid_value = env->GetFieldID(clazz, "value", "B");
   }
};
static JavaByteByRef javaByteByRef;

struct JavaIntByRef
{
   jfieldID fid_value;

   jint GetValue(JNIEnv * env, jobject jniObj)
   {
      return env->GetIntField(jniObj, fid_value);
   }
   void SetValue(JNIEnv * env, jobject jniObj, jint value)
   {
      env->SetIntField(jniObj, fid_value, value);
   }

   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/IntByRef");
      fid_value = env->GetFieldID(clazz, "value", "I");
   }
};
static JavaIntByRef javaIntByRef;

struct JavaLongByRef
{
   jfieldID fid_value;

   jlong GetValue(JNIEnv * env, jobject jniObj)
   {
      return env->GetLongField(jniObj, fid_value);
   }
   void SetValue(JNIEnv * env, jobject jniObj, jlong value)
   {
      env->SetLongField(jniObj, fid_value, value);
   }

   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/LongByRef");
      fid_value = env->GetFieldID(clazz, "value", "J");
   }
};
static JavaLongByRef javaLongByRef;

struct JavaDoubleByRef
{
   jfieldID fid_value;

   jdouble GetValue(JNIEnv * env, jobject jniObj)
   {
      return env->GetDoubleField(jniObj, fid_value);
   }
   void SetValue(JNIEnv * env, jobject jniObj, jdouble value)
   {
      env->SetDoubleField(jniObj, fid_value, value);
   }

   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/DoubleByRef");
      fid_value = env->GetFieldID(clazz, "value", "D");
   }
};
static JavaDoubleByRef javaDoubleByRef;

struct JavaStringBuffer
{
   jmethodID mid_capacity;
   jmethodID mid_setLength;
   jmethodID mid_append;

   jint Capacity(JNIEnv * env, jobject jniObj)
   {
      return env->CallIntMethod(jniObj, mid_capacity);
   }
   void SetLength(JNIEnv * env, jobject jniObj, jint length)
   {
      // empty the string buffer 
      env->CallObjectMethod(jniObj, mid_setLength, length);
   }

   void Append(JNIEnv * env, jobject jniObj, jchar *value, jsize len)
   {
      // set new value into the string buffer
      jstring _jstring = env->NewString(value, len);
      env->CallObjectMethod (jniObj, mid_append, _jstring);
   }

   void Init(JNIEnv * env)
   {
      jclass clazz  = env->FindClass("java/lang/StringBuffer");
      mid_setLength = env->GetMethodID(clazz, "setLength", "(I)V");
      mid_append    = env->GetMethodID(clazz, "append", "(Ljava/lang/String;)Ljava/lang/StringBuffer;");
      mid_capacity  = env->GetMethodID(clazz, "capacity", "()I");
   }

   jchar * ToJChar(wchar_t * in, jsize & len)
   {
      len = (jsize)wcslen(in);
      if (sizeof(wchar_t) == 2){  // UTF-16 (windows)
         return (jchar*)in;
      } else {
         jchar * out = (jchar*)in;
         while ((*out++ = (jchar)*in++) != (jchar)'\0');
         return out - len;
      }
   }
};
static JavaStringBuffer javaStringBuffer;

struct JavaWaitHandle
{
   jfieldID fid_handle;
   jfieldID fid_eventId;

   HANDLE GetHandle(JNIEnv * env, jobject jniObj)
   {
      return (HANDLE)env->GetLongField(jniObj, fid_handle);
   }
   void SetHandle(JNIEnv * env, jobject jniObj, HANDLE value)
   {
      env->SetLongField(jniObj, fid_handle, (jlong)value);
   }

   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/WaitHandle");
      fid_handle = env->GetFieldID(clazz, "nativeHandle", "J");
   }
};
static JavaWaitHandle javaWaitHandle;

struct NativeBuffer
{
   void *   memory;
   int      elementSize;
   int      capacity;
   int      position;
   jboolean readonly;
};

struct JavaNativeBuffer
{
   jfieldID fid_memory;
   jfieldID fid_elementSize;
   jfieldID fid_capacity;
   jfieldID fid_position;
   jfieldID fid_readonly;

   NativeBuffer Get(JNIEnv * env, jobject jniObj)
   {
      NativeBuffer ntBuf;
      ntBuf.memory      = (void*)env->GetLongField(jniObj,  fid_memory);
      ntBuf.elementSize = env->GetIntField(jniObj,   fid_elementSize);
      ntBuf.capacity    = env->GetIntField(jniObj,   fid_capacity);
      ntBuf.position    = env->GetIntField(jniObj,   fid_position);
      ntBuf.readonly    = env->GetBooleanField(jniObj, fid_readonly);
      return ntBuf;
   }

   void Set(JNIEnv * env, jobject jniObj, NativeBuffer const &ntBuf)
   {
      env->SetLongField(jniObj,    fid_memory,      (jlong)ntBuf.memory);
      env->SetIntField(jniObj,     fid_elementSize, ntBuf.elementSize);
      env->SetIntField(jniObj,     fid_capacity,    ntBuf.capacity);
      env->SetIntField(jniObj,     fid_position,    ntBuf.position);
      env->SetBooleanField(jniObj, fid_readonly,    ntBuf.readonly);
   }

   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/NativeBuffer");
      fid_memory = env->GetFieldID(clazz, "memory", "J");
      fid_elementSize = env->GetFieldID(clazz, "elementSize", "I");
      fid_capacity = env->GetFieldID(clazz, "capacity", "I");
      fid_position = env->GetFieldID(clazz, "position", "I");
      fid_readonly = env->GetFieldID(clazz, "readonly", "Z");
   }

};
static JavaNativeBuffer javaNativeBuffer;

struct JavaMathInverval
{
   MathInterval GetValue(JNIEnv * env, jobject jniObj)
   {
      jclass clazz = env->GetObjectClass(jniObj);
      jfieldID fid;
      MathInterval value;

      // get: Type
      fid = env->GetFieldID(clazz, "Type", "I");
      value.Type = env->GetIntField(jniObj, fid);

      // get: Min
      fid = env->GetFieldID(clazz, "Min", "D");
      value.Min = env->GetDoubleField(jniObj, fid);

      // get: Max
      fid = env->GetFieldID(clazz, "Max", "D");
      value.Max = env->GetDoubleField(jniObj, fid);

      return value;
   }
   void SetValue(JNIEnv * env, jobject jniObj, MathInterval const &value)
   {
      jclass clazz = env->GetObjectClass(jniObj);
      jfieldID fid;

      // set: Type
      fid = env->GetFieldID(clazz, "Type", "I");
      env->SetIntField(jniObj, fid, value.Type);

      // set: Min
      fid = env->GetFieldID(clazz, "Min", "D");
      env->SetDoubleField(jniObj, fid, value.Min);

      // set: Max
      fid = env->GetFieldID(clazz, "Max", "D");
      env->SetDoubleField(jniObj, fid, value.Max);
   }

   void Init(JNIEnv * env)
   {
      ; // need do nothing, this type hardly be used.
   }
}; 
static JavaMathInverval javaMathInterval;

struct JavaRectangle
{
   RECT GetValue(JNIEnv * env, jobject jniObj)
   {
      jclass clazz = env->GetObjectClass(jniObj);
      jfieldID fid;
      RECT value;

      // get: left
      fid = env->GetFieldID(clazz, "x", "I");
      value.left = env->GetIntField(jniObj, fid);

      // get: top
      fid = env->GetFieldID(clazz, "y", "I");
      value.top = env->GetIntField(jniObj, fid);

      // get: right
      fid = env->GetFieldID(clazz, "width", "I");
      value.right = value.left + env->GetIntField(jniObj, fid);

      // get: bottom
      fid = env->GetFieldID(clazz, "height", "I");
      value.bottom = value.top + env->GetIntField(jniObj, fid);

      return value;
   }

   void Init(JNIEnv * env)
   {
      ; // need do nothing, this type hardly be used.
   }
};
static JavaRectangle javaRectangle;

struct JavaDeviceInformation
{
   jfieldID fid_deviceNumber;
   jfieldID fid_moduleIndex;
   jfieldID fid_deviceMode;
   jfieldID fid_description;

   DeviceInformation GetValue(JNIEnv * env, jobject jniObj)
   {
      jclass clazz = env->GetObjectClass(jniObj);
      jfieldID fid;
      DeviceInformation value;

      // get: DeviceNumber
      fid = env->GetFieldID(clazz, "DeviceNumber", "I");
      value.DeviceNumber = env->GetIntField(jniObj, fid);

      // get: ModuleIndex
      fid = env->GetFieldID(clazz, "ModuleIndex", "I");
      value.DeviceNumber = env->GetIntField(jniObj, fid);

      // get: AccessMode
      {
         fid = env->GetFieldID(clazz, "DeviceMode", "LAutomation/BDaq/AccessMode;");
         jobject mode = env->GetObjectField(jniObj, fid);
         jclass clazzMode = env->GetObjectClass(mode);
         jmethodID mid = env->GetMethodID(clazzMode, "toInt", "()I");
         value.DeviceMode = (AccessMode)env->CallIntMethod(mode, mid);
      }

      // get: Description
      {
         fid = env->GetFieldID(clazz, "Description", "Ljava/lang/String;");
         jstring desc = (jstring)env->GetObjectField(jniObj, fid);
         if (desc) {
            jchar const *jniStr = env->GetStringChars(desc, NULL);
            JCharToUtf(value.Description, jniStr);
            env->ReleaseStringChars(desc, jniStr);
         }
      }

      return value;
   }
   void SetValue(JNIEnv * env, jobject jniObj, DeviceInformation const &value)
   {
      jclass clazz = env->GetObjectClass(jniObj);
      jfieldID fid;

      // set: DeviceNumber
      fid = env->GetFieldID(clazz, "DeviceNumber", "I");
      env->SetIntField(jniObj, fid, value.DeviceNumber);

      // set: ModuleIndex
      fid = env->GetFieldID(clazz, "ModuleIndex", "I");
      env->SetIntField(jniObj, fid, value.ModuleIndex);

      // set: AccessMode
      {
         fid = env->GetFieldID(clazz, "DeviceMode", "LAutomation/BDaq/AccessMode;");
         jobject mode = env->GetObjectField(jniObj, fid);
         jclass c = env->GetObjectClass(mode);
         
         char const *name = NULL;
         switch (value.DeviceMode)
         {
         default:
         case ModeRead:           name = "ModeRead";  break;
         case ModeWrite:          name = "ModeWrite"; break;
         case ModeWriteWithReset: name = "ModeWriteWithReset"; break;
         }
         
         jfieldID sid = env->GetStaticFieldID(c, name, "LAutomation/BDaq/AccessMode;");
         env->SetObjectField(jniObj, fid, env->GetStaticObjectField(c, sid));
      }

      // set: Description
      {
         jsize len = (jsize)wcslen(value.Description);
         jchar *jch = new jchar[len + 1];
         UtfToJChar(jch, value.Description);
         jobject jstr = env->NewString(jch, len);
         delete jch;

         fid = env->GetFieldID(clazz, "Description", "Ljava/lang/String;");
         env->SetObjectField(jniObj, fid, jstr);
      }
   }

   void Init(JNIEnv * env)
   {
      ; // need do nothing, this type hardly be used.
   }                                            
}; 
static JavaDeviceInformation javaDeviceInformation;

struct JavaMapFuncPiece
{
   void GetValue(JNIEnv * env, jobject jniObj, MapFuncPiece &value)
   {
      jclass clazz = env->GetObjectClass(jniObj);
      jfieldID fid;

      // get: Size
      fid = env->GetFieldID(clazz, "Size", "I");
      value.Size = env->GetIntField(jniObj, fid);

      // get: Degree
      fid = env->GetFieldID(clazz, "Degree", "I");
      value.Degree = env->GetIntField(jniObj, fid);

      // get: UpperLimit
      fid = env->GetFieldID(clazz, "UpperLimit", "D");
      value.UpperLimit = env->GetDoubleField(jniObj, fid);

      // get: Coef[]
      {
         fid = env->GetFieldID(clazz, "Coef", "[D");
         jdoubleArray jarr = (jdoubleArray)env->GetObjectField(jniObj, fid);
         jdouble * arrBuf = env->GetDoubleArrayElements(jarr, NULL);
         
         if (arrBuf) {
            memcpy(value.Coef, arrBuf, value.Size - offset_of(MapFuncPiece, Coef));
            env->ReleaseDoubleArrayElements(jarr, arrBuf, JNI_ABORT);
         }
      }
   }
   void SetValue(JNIEnv * env, jobject jniObj, MapFuncPiece const &value)
   {
      jclass clazz = env->GetObjectClass(jniObj);
      jfieldID fid;

      // set: Size
      fid = env->GetFieldID(clazz, "Size", "I");
      int size = env->GetIntField(jniObj, fid);
      if (size < value.Size) {
         throw "Buffer Too Small!";
      }
      env->SetIntField(jniObj, fid, value.Size);

      // set: Degree
      fid = env->GetFieldID(clazz, "Degree", "I");
      env->SetIntField(jniObj, fid, value.Degree);

      // set: UpperLimit
      fid = env->GetFieldID(clazz, "UpperLimit", "D");
      env->SetDoubleField(jniObj, fid, value.UpperLimit);

      // set: Coef[]
      {
         fid = env->GetFieldID(clazz, "Coef", "[D");
         jdoubleArray jarr = (jdoubleArray)env->GetObjectField(jniObj, fid);
         jdouble * arrBuf = env->GetDoubleArrayElements(jarr, NULL);

         if (arrBuf) {
            memcpy(arrBuf, value.Coef, size - offset_of(MapFuncPiece, Coef));
            env->ReleaseDoubleArrayElements(jarr, arrBuf, 0);
         }
      }
   }

   void Init(JNIEnv * env)
   {
      ; // need do nothing, this type hardly be used.
   }
}; 
static JavaMapFuncPiece javaMapFuncPiece;

struct JavaDataMark
{
   jfieldID fid_dataIndex;
   jfieldID fid_srcId;

   DataMark GetValue(JNIEnv * env, jobject jniObj)
   {
      DataMark value;

      // get: DataIndex
      value.DataIndex = env->GetIntField(jniObj, fid_dataIndex);

      // get: SrcId
      value.SrcId = env->GetIntField(jniObj, fid_srcId);

      return value;
   }
   void SetValue(JNIEnv * env, jobject jniObj, DataMark const &value)
   {
      // set: DataIndex
      env->SetLongField(jniObj, fid_dataIndex, value.DataIndex);

      // set: SrcId
      env->SetIntField(jniObj, fid_srcId, value.SrcId);
   }

   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/DataMark");
      fid_dataIndex = env->GetFieldID(clazz, "DataIndex", "J");
      fid_srcId = env->GetFieldID(clazz, "SrcId", "I");
   }
}; 
static JavaDataMark javaDataMark;

struct JavaPulseWidth
{
   jfieldID fid_hiPeriod;
   jfieldID fid_loPeriod;

   PulseWidth GetValue(JNIEnv * env, jobject jniObj)
   {
      PulseWidth value;
      value.HiPeriod = env->GetDoubleField(jniObj, fid_hiPeriod);
      value.LoPeriod = env->GetDoubleField(jniObj, fid_loPeriod);

      return value;
   }
   void SetValue(JNIEnv * env, jobject jniObj, PulseWidth const &value)
   {
      env->SetDoubleField(jniObj, fid_hiPeriod, value.HiPeriod);
      env->SetDoubleField(jniObj, fid_loPeriod, value.LoPeriod);
   }

   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/PulseWidth");
      fid_hiPeriod = env->GetFieldID(clazz, "HiPeriod", "D");
      fid_loPeriod = env->GetFieldID(clazz, "LoPeriod", "D");
   }
}; 
static JavaPulseWidth javaPulseWidth;

struct JavaDeviceEventDelegate
{
   jmethodID mid_onNativeEvent;

   void Fire(JNIEnv *env, jobject d, DeviceEventArgs const *args)
   {
      env->CallVoidMethod(d, mid_onNativeEvent, args->Id);
   }

   void Init(JNIEnv *env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/TDeviceEventDelegate");
      mid_onNativeEvent = env->GetMethodID(clazz, "OnNativeEvent", "(I)V");
   }
};
static JavaDeviceEventDelegate javaDeviceEventDelegate;

struct JavaBfdAiEventDelegate
{
   jmethodID mid_onNativeEvent;

   void Fire(JNIEnv *env, jobject d, BfdAiEventArgs const *args)
   {
      env->CallVoidMethod(d, mid_onNativeEvent, args->Id, args->Count, args->Offset, args->MarkCount);
   }
   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/TBfdAiEventDelegate");
      mid_onNativeEvent = env->GetMethodID(clazz, "OnNativeEvent", "(IIII)V");
   }
};
static JavaBfdAiEventDelegate javaBfdAiEventDelegate;

struct JavaBfdAoEventDelegate
{
   jmethodID mid_onNativeEvent;

   void Fire(JNIEnv *env, jobject d, BfdAoEventArgs const *args)
   {
      env->CallVoidMethod(d, mid_onNativeEvent, args->Id, args->Count, args->Offset);
   }
   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/TBfdAoEventDelegate");
      mid_onNativeEvent = env->GetMethodID(clazz, "OnNativeEvent", "(III)V");
   }
};
static JavaBfdAoEventDelegate javaBfdAoEventDelegate;

struct JavaDiSnapEventDelegate
{
   jmethodID mid_onNativeEvent;
   jfieldID  fid_data;

   void Fire(JNIEnv *env, jobject d, DiSnapEventArgs const *args)
   {
      // Copy snapped data
      if (args->Length) {
         jbyteArray data = (jbyteArray)env->GetObjectField(d, fid_data);
         jbyte * rxBuf = env->GetByteArrayElements(data, NULL);
         memcpy(rxBuf, args->PortData, args->Length);
         env->ReleaseByteArrayElements(data, rxBuf, 0);
      }
      env->CallVoidMethod(d, mid_onNativeEvent, args->Id, args->SrcNum, args->Length);
   }
   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/TDiSnapEventDelegate");
      mid_onNativeEvent = env->GetMethodID(clazz, "OnNativeEvent", "(III)V");
      fid_data = env->GetFieldID(clazz, "Data", "[B");
   }
};
static JavaDiSnapEventDelegate javaDiSnapEventDelegate;

struct JavaBfdDiEventDelegate
{
   jmethodID mid_onNativeEvent;

   void Fire(JNIEnv *env, jobject d, BfdDiEventArgs const *args)                                     
   {
      env->CallVoidMethod(d, mid_onNativeEvent, args->Id, args->Count, args->Offset, args->MarkCount);
   }
   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/TBfdDiEventDelegate");
      mid_onNativeEvent = env->GetMethodID(clazz, "OnNativeEvent", "(IIII)V");
   }
};
static JavaBfdDiEventDelegate javaBfdDiEventDelegate;

struct JavaBfdDoEventDelegate
{
   jmethodID mid_onNativeEvent;

   void Fire(JNIEnv *env, jobject d, BfdDoEventArgs const *args)
   {
      env->CallVoidMethod(d, mid_onNativeEvent, args->Id, args->Count, args->Offset);
   }
   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/TBfdDoEventDelegate");
      mid_onNativeEvent = env->GetMethodID(clazz, "OnNativeEvent", "(III)V");
   }
};
static JavaBfdDoEventDelegate javaBfdDoEventDelegate;

struct JavaCntrEventDelegate
{
   jmethodID mid_onNativeEvent;

   void Fire(JNIEnv *env, jobject d, CntrEventArgs const *args)
   {
      env->CallVoidMethod(d, mid_onNativeEvent, args->Id, args->Channel);
   }
   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/TCntrEventDelegate");
      mid_onNativeEvent = env->GetMethodID(clazz, "OnNativeEvent", "(II)V");
   }
};
static JavaCntrEventDelegate javaCntrEventDelegate;

struct JavaUdCntrEventDelegate
{
   jmethodID mid_onNativeEvent;
   jfieldID  fid_data;

   void Fire(JNIEnv *env, jobject d, UdCntrEventArgs const *args)
   {
      // Copy snapped data
      if (args->Length) {
         jintArray data = (jintArray)env->GetObjectField(d, fid_data);
         jint * rxBuf = env->GetIntArrayElements(data, NULL);
         memcpy(rxBuf, args->Data, args->Length);
         env->ReleaseIntArrayElements(data, rxBuf, 0);
      }
      env->CallVoidMethod(d, mid_onNativeEvent, args->Id, args->Length);
   }
   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/TUdCntrEventDelegate");
      mid_onNativeEvent = env->GetMethodID(clazz, "OnNativeEvent", "(II)V");
      fid_data = env->GetFieldID(clazz, "Data", "[I");
   }
};
static JavaUdCntrEventDelegate javaUdCntrEventDelegate;

struct JavaBfdCntrEventDelegate
{
   jmethodID mid_onNativeEvent;

   void Fire(JNIEnv *env, jobject d, BfdCntrEventArgs const *args)
   {
      env->CallVoidMethod(d, mid_onNativeEvent, args->Id, args->Channel, args->Count, args->Offset);
   }
   void Init(JNIEnv * env)
   {
      jclass clazz = env->FindClass("Automation/BDaq/TBfdCntrEventDelegate");
      mid_onNativeEvent = env->GetMethodID(clazz, "OnNativeEvent", "(IIII)V");
   }
};
static JavaBfdCntrEventDelegate javaBfdCntrEventDelegate;

/************************************************************************/
/* Globals for JNI call back into JAVA in multi-thread environment      */
/************************************************************************/
JavaVM                *g_javaVM = NULL;
DaqThreadLocal<JNIEnv> g_jniTls;

//
// JNI thread initialize/clean routine
void JNI_Thread_Initialize()
{
   JNIEnv *env = NULL;
   g_javaVM->AttachCurrentThread((void **)&env, NULL);
   JNIEnv *old = g_jniTls.Set(env);
   assert(old == NULL);
}

void JNI_Thread_Cleanup()
{
   g_jniTls.Set(NULL);
   g_javaVM->DetachCurrentThread(); 
}

/************************************************************************/
/* initialization methods                                               */
/************************************************************************/
/*
 * Class:     automation_bdaq_BDaqApi
 * Method:    InitIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_BDaqApi_InitIDs(
   JNIEnv * env, 
   jclass   cls)
{
   // init: IntByRef
   javaIntByRef.Init(env);

   // init: LongByRef
   javaLongByRef.Init(env);

   // init: DoubleByRef
   javaDoubleByRef.Init(env);

   // init: StringBuffer
   javaStringBuffer.Init(env);

   // init: WaitHandle
   javaWaitHandle.Init(env);

   // init: native buffer
   javaNativeBuffer.Init(env);

   // init: MathInterval
   javaMathInterval.Init(env);

   // init: Rectangle
   javaRectangle.Init(env);
}

/*
 * Class:     automation_bdaq_BDaqApi
 * Method:    DaqNavi4InitIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_BDaqApi_DaqNavi4InitIDs(
   JNIEnv * env, 
   jclass   cls)
{
   // Get JVM instance
   env->GetJavaVM(&g_javaVM);

   // register JNI thread initiator
   KernEventDispatcher::initiator().Initialize = JNI_Thread_Initialize;
   KernEventDispatcher::initiator().Cleanup    = JNI_Thread_Cleanup;

   // init: ByteByRef
   javaByteByRef.Init(env);

   // init: IntByRef
   javaIntByRef.Init(env);

   // init: LongByRef
   javaLongByRef.Init(env);

   // init: DoubleByRef
   javaDoubleByRef.Init(env);

   // init: StringBuffer
   javaStringBuffer.Init(env);

   // init: MathInterval
   javaMathInterval.Init(env);

   // init: DeviceInformation
   javaDeviceInformation.Init(env);

   // init: MapFuncPiece
   javaMapFuncPiece.Init(env);

   // init: DataMarker
   javaDataMark.Init(env);

   // init: PulseWidth
   javaPulseWidth.Init(env);

   // init: DeviceEventDelegate
   javaDeviceEventDelegate.Init(env);

   // init: BfdAiEventDelegate
   javaBfdAiEventDelegate.Init(env);

   // init: BfdAoEventDelegate
   javaBfdAoEventDelegate.Init(env);

   // init: DiSnapEventDelegate
   javaDiSnapEventDelegate.Init(env);

   // init: BfdDiEventDelegate
   javaBfdDiEventDelegate.Init(env);

   // init: BfdDoEventDelegate
   javaBfdDoEventDelegate.Init(env);

   // init: CntrEventDelegate
   javaCntrEventDelegate.Init(env);

   // init: UdCntrEventDelegate
   javaUdCntrEventDelegate.Init(env);

   // init: BfdCntrEventDelegate
   javaBfdCntrEventDelegate.Init(env);
}

/************************************************************************/
/* Global Helper method                                                 */
/************************************************************************/
/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDeviceGetLinkageInfo
 * Signature: (IILAutomation/BDaq/IntByRef;Ljava/lang/StringBuffer;LAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDeviceGetLinkageInfo(
  JNIEnv * env, 
  jclass   cls, 
  jint     parent, 
  jint     index, 
  jobject  jniDevNumber, 
  jobject  jniDesc, 
  jobject  jniSubCount )
{
   if (jniDevNumber == NULL)
   {
      return ErrorBufferIsNull;
   }

   int32  devNumber;
   uint32 subCount;
   wchar_t desc[MAX_DEVICE_DESC_LEN];

   jint ret = AdxDeviceGetLinkageInfo(parent, index, &devNumber, 
               jniDesc != NULL ? desc : NULL, 
               jniSubCount != NULL ? &subCount : NULL );

   // save device number
   javaIntByRef.SetValue(env, jniDevNumber, devNumber);

   // save sub device count
   if (jniSubCount != NULL)
   {
      javaIntByRef.SetValue(env, jniSubCount, subCount);
   }

   // save device description
   if (jniDesc != NULL)
   {
      javaStringBuffer.SetLength(env, jniDesc, 0);
      jchar jcs[MAX_DEVICE_DESC_LEN];
      jsize len = UtfToJChar(jcs, desc);
      javaStringBuffer.Append(env, jniDesc, jcs, len);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxGetValueRangeInformation
 * Signature: (ILjava/lang/StringBuffer;LAutomation/BDaq/MathInterval;LAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxGetValueRangeInformation(
   JNIEnv * env, 
   jclass   cls, 
   jint     vrgType,
   jobject  jniDesc,
   jobject  jniInterval,
   jobject  jniUnit)
{
   jint    bufSize = 0;
   wchar_t desc[MAX_VRG_DESC_LEN];
   if (jniDesc != NULL)
   {
      javaStringBuffer.SetLength(env, jniDesc, 0);
      bufSize = javaStringBuffer.Capacity(env, jniDesc);
   }

   MathInterval interval = {0};
   uint32 unit = Volt;

   long ret = AdxGetValueRangeInformation(vrgType, bufSize,
               jniDesc == NULL ? NULL: desc,
               jniInterval == NULL ? NULL : &interval,
               jniUnit == NULL ? NULL : &unit);

   // save description
   if (jniDesc != NULL)
   {
      jchar jcs[MAX_VRG_DESC_LEN];
      jsize len = UtfToJChar(jcs, desc);
      javaStringBuffer.Append(env, jniDesc, jcs, len);
   }
   if (jniInterval != NULL)
   {
      javaMathInterval.SetValue(env, jniInterval, interval);
   }
   if (jniUnit != NULL)
   {
      javaIntByRef.SetValue(env, jniUnit, unit);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxGetSignalConnectionInformation
 * Signature: (ILjava/lang/StringBuffer;LAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxGetSignalConnectionInformation(
   JNIEnv * env, 
   jclass   cls, 
   jint     signalDrop,
   jobject  jniDesc,
   jobject  jniPos)
{
   jint  bufSize = 0;
   wchar_t desc[MAX_SIG_DROP_DESC_LEN];
   if (jniDesc != NULL)
   {
      javaStringBuffer.SetLength(env, jniDesc, 0);
      bufSize = javaStringBuffer.Capacity(env, jniDesc);
   }

   uint32 pos = InternalSig;
   long ret = AdxGetSignalConnectionInformation(signalDrop, bufSize,
               jniDesc== NULL ? NULL : desc,
               jniPos == NULL ? NULL : &pos);

   // save description
   if (jniDesc != NULL)
   {
      jchar jcs[MAX_SIG_DROP_DESC_LEN];
      jsize len = UtfToJChar(jcs, desc);
      javaStringBuffer.Append(env, jniDesc, jcs, len);
   }
   if (jniPos != NULL)
   {
      javaIntByRef.SetValue(env, jniPos, pos);
   }
   
   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxEnumToString
 * Signature: (Ljava/lang/String;IILjava/lang/StringBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxEnumToString(
   JNIEnv * env,
   jclass   cls,
   jstring  enumTypeName,
   jint     enumValue,
   jint     enumStringLength,
   jobject  enumString)
{
   if (enumTypeName == NULL) {
      return ErrorParamNotSpted;
   }
   if (enumString == NULL || enumStringLength == 0) {
      return ErrorBufferIsNull;
   }

   javaStringBuffer.SetLength(env, enumString, 0);

   // convert the enum type name to wchar_t
   jchar const *jniStr = env->GetStringChars(enumTypeName, NULL);
   wchar_t *fullName = new wchar_t[env->GetStringLength(enumTypeName) + 1];
   JCharToUtf(fullName, jniStr);
   env->ReleaseStringChars(enumTypeName, jniStr);

   // extract the type name
   wchar_t *typeName = wcsrchr(fullName, L'.');
   if (typeName == NULL) {
     typeName = fullName;
   } else {
     ++typeName;
   }

   // prepare the wchar_t type string buffer
   wchar_t *buffer  = new wchar_t[enumStringLength];

   // get the actual string
   jint ret = AdxEnumToString(typeName, enumValue, enumStringLength, buffer);

   // transfer back the string into jvm
   if (ret == Success){
      jchar *jcs = new jchar[enumStringLength];
      jsize len  = UtfToJChar(jcs, buffer);
      javaStringBuffer.Append(env, enumString, jcs, len);
      delete []jcs;
   }

   delete []buffer;
   delete []fullName;
   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxStringToEnum
 * Signature: (Ljava/lang/String;Ljava/lang/String;LAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxStringToEnum(
   JNIEnv * env,
   jclass   cls,
   jstring  enumTypeName,
   jstring  enumString,
   jobject  enumValue)
{
   if (enumTypeName == NULL || enumString == NULL) {
      return ErrorParamNotSpted;
   }
   if (enumValue == NULL) {
      return ErrorBufferIsNull;
   }

   // convert the enum type name to wchar_t
   jchar const *jniStr   = env->GetStringChars(enumTypeName, NULL);
   wchar_t *fullName = new wchar_t[env->GetStringLength(enumTypeName) + 1];
   JCharToUtf(fullName, jniStr);
   env->ReleaseStringChars(enumTypeName, jniStr);

   // extract the type name
   wchar_t *typeName = wcsrchr(fullName, L'.');
   if (typeName == NULL) {
     typeName = fullName;
   } else {
     ++typeName;
   }

   // convert the enum string to wchar_t
   jniStr = env->GetStringChars(enumString, NULL);
   wchar_t *enumStr = new wchar_t[env->GetStringLength(enumString) + 1];
   JCharToUtf(enumStr, jniStr);
   env->ReleaseStringChars(enumString, jniStr);

   // get actual value
   uint32 buffer = -1;
   jint ret = AdxStringToEnum(typeName, enumStr, &buffer);
   javaIntByRef.SetValue(env, enumValue, buffer);

   delete []fullName;
   delete []enumStr;
   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxTranslateTemperatureScale
 * Signature: (ID)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_BDaqApi_AdxTranslateTemperatureScale(
   JNIEnv * env, 
   jclass   cls,
   jint     degreeType,
   jdouble  degreeCelsius)
{
   return AdxTranslateTemperatureScale(degreeType, degreeCelsius);
}

//*************************************************************************************
// DAQNAVI JAVA STUB -- DAQNAVI3 API 
//*************************************************************************************

/************************************************************************/
/* Event methods                                                        */
/************************************************************************/
/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxEventGetHandle
 * Signature: (JILAutomation/BDaq/WaitHandle;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxEventGetHandle(
   JNIEnv * env, 
   jclass   cls, 
   jlong    handle, 
   jint     eventId,
   jobject  jniWaitHandle )
{
   if (jniWaitHandle == NULL)
   {
      return ErrorBufferIsNull;
   }
   HANDLE eventHandle = NULL;

   long ret = AdxEventGetHandle((HANDLE)handle, eventId, &eventHandle);
   javaWaitHandle.SetHandle(env, jniWaitHandle, eventHandle);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxEventGetLastStatus
 * Signature: (JILAutomation/BDaq/IntByRef;LAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxEventGetLastStatus(
   JNIEnv * env, 
   jclass   cls,
   jlong    handle, 
   jint     eventId,
   jobject  jniLParam,
   jobject  jniRParam)
{
   uint32 left, right;
   long ret = AdxEventGetLastStatus((HANDLE)handle, eventId, &left, &right);

   if (jniLParam)
   {
      javaIntByRef.SetValue(env, jniLParam, left);
   }
   if (jniRParam)
   {
      javaIntByRef.SetValue(env, jniRParam, right);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxEventClearFlag
 * Signature: (JIII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxEventClearFlag(
   JNIEnv * env, 
   jclass   cls,
   jlong    handle, 
   jint     eventId,
   jint     flagLParam,
   jint     flagRParam )
{
   return AdxEventClearFlag((HANDLE)handle, eventId, flagLParam, flagRParam);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxEventWaitSingle
 * Signature: (LAutomation/BDaq/WaitHandle;I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxEventWaitSingle(
   JNIEnv * env, 
   jclass   cls,
   jobject  jniHandle, 
   jint     timeout )
{
   int ret = WAIT_FAILED;
   if (jniHandle != NULL)
   {
      HANDLE eventHandle = javaWaitHandle.GetHandle(env, jniHandle);
      ret = WaitForMultipleObjects(1, &eventHandle, 0, timeout);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxEventWaitMultiple
 * Signature: ([LAutomation/BDaq/WaitHandle;IZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxEventWaitMultiple(
   JNIEnv     * env, 
   jclass       cls,
   jobjectArray jniHandles,
   jint         timeout,
   jboolean     waitAll )
{
   int    ret = WAIT_FAILED;
   HANDLE objs[MAXIMUM_WAIT_OBJECTS];
   if (jniHandles != NULL)
   {
      int count = env->GetArrayLength(jniHandles);
      if (count > 0 && count <= MAXIMUM_WAIT_OBJECTS)
      {
         for( int i = 0; i < count; ++i)
         {
            objs[i] = javaWaitHandle.GetHandle(env, env->GetObjectArrayElement(jniHandles, i));
         }

         ret = WaitForMultipleObjects(count, objs, waitAll, timeout);
       }
   }

   return ret;
}

/************************************************************************/
/* Property access methods                                              */
/************************************************************************/
/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyGetLength
 * Signature: (JILAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyGetLength(
   JNIEnv * env, 
   jclass   cls,
   jlong    handle, 
   jint     propId,
   jobject  jniLength )
{
   if (jniLength == NULL)
   {
      return ErrorBufferIsNull;
   }

   uint32 length = 0;
   long ret = AdxPropertyRead((HANDLE)handle, propId, 0, NULL, &length, NULL);
   javaIntByRef.SetValue(env, jniLength, length);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyGetAttribute
 * Signature: (JILAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyGetAttribute(
   JNIEnv * env, 
   jclass   cls,
   jlong    handle, 
   jint     propId,
   jobject  jniAttr )
{
   if (jniAttr == NULL)
   {
      return ErrorBufferIsNull;
   }

   uint32 length = 0;
   uint32 attribute= 0;
   long ret = AdxPropertyRead((HANDLE)handle, propId, 0, NULL, &length, &attribute);
   javaIntByRef.SetValue(env, jniAttr, attribute);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyRead
 * Signature: (JII[BLAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyRead__JII_3BLAutomation_BDaq_IntByRef_2(
   JNIEnv *   env, 
   jclass     cls,
   jlong      handle, 
   jint       propId, 
   jint       count,
   jbyteArray jniBuffer, 
   jobject    jniLength )
{
   uint32  length;
   jbyte* rxbuf = NULL;
   if (jniBuffer != NULL)
   {
      rxbuf = env->GetByteArrayElements(jniBuffer, NULL);
   }

   long ret = AdxPropertyRead((HANDLE)handle, propId, count * sizeof(jbyte), rxbuf, &length, NULL);

   if (jniBuffer != NULL)
   {
      env->ReleaseByteArrayElements(jniBuffer, rxbuf, 0);
   }
   if (jniLength != NULL)
   {
      javaIntByRef.SetValue(env, jniLength, length);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyRead
 * Signature: (JII[ILAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyRead__JII_3ILAutomation_BDaq_IntByRef_2(
  JNIEnv *   env, 
  jclass     cls,
  jlong      handle, 
  jint       propId,
  jint       count,
  jintArray  jniBuffer, 
  jobject    jniLength)
{
   uint32  length;
   jint*  rxbuf = NULL;
   if (jniBuffer != NULL)
   {
      rxbuf = env->GetIntArrayElements(jniBuffer, NULL);
   }

   long ret = AdxPropertyRead((HANDLE)handle, propId, count * sizeof(jint), rxbuf, &length, NULL);

   if (jniBuffer != NULL)
   {
      env->ReleaseIntArrayElements(jniBuffer, rxbuf, 0);
   }
   if (jniLength != NULL)
   {
      javaIntByRef.SetValue(env, jniLength, length);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyRead
 * Signature: (JII[DLAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyRead__JII_3DLAutomation_BDaq_IntByRef_2(
   JNIEnv *     env, 
   jclass       cls,
   jlong        handle, 
   jint         propId, 
   jint         count,
   jdoubleArray jniBuffer,
   jobject      jniLength )
{
   uint32     length;
   jdouble*  rxbuf = NULL;
   if (jniBuffer != NULL)
   {
      rxbuf = env->GetDoubleArrayElements(jniBuffer, NULL);
   }

   long ret = AdxPropertyRead((HANDLE)handle, propId, count * sizeof(jdouble), rxbuf, &length, NULL);

   if (jniBuffer != NULL)
   {
      env->ReleaseDoubleArrayElements(jniBuffer, rxbuf, 0);
   }
   if (jniLength != NULL)
   {
      javaIntByRef.SetValue(env, jniLength, length);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyRead
 * Signature: (JII[CLAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyRead__JII_3CLAutomation_BDaq_IntByRef_2(
   JNIEnv *    env, 
   jclass      cls,
   jlong       handle, 
   jint        propId,
   jint        count,
   jcharArray  jniBuffer, 
   jobject     jniLength)
{
   uint32  length;
   jchar   *jcs   = NULL;
   wchar_t *rxbuf = NULL;

   if (jniBuffer != NULL) {
      jcs = env->GetCharArrayElements(jniBuffer, NULL);
      if (sizeof(wchar_t) == 4) {// utf-32: linux, Mac OS, gcc
         rxbuf = new wchar_t[count];
      } else { // utf-16: Windows OS, java
         rxbuf = (wchar_t*)jcs;
      }
   }

   long ret = AdxPropertyRead((HANDLE)handle, propId, count * sizeof(wchar_t), rxbuf, &length, NULL);

   if (jniBuffer != NULL) {
      if (sizeof(wchar_t) == 4){
         UtfToJChar(jcs, rxbuf);
         delete []rxbuf;
      }
      env->ReleaseCharArrayElements(jniBuffer, jcs, 0);
   }

   if (jniLength != NULL) {
      javaIntByRef.SetValue(env, jniLength, length * sizeof(jchar) / sizeof(wchar_t));
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyRead
 * Signature: (JILAutomation/BDaq/MathInterval;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyRead__JILAutomation_BDaq_MathInterval_2(
   JNIEnv * env, 
   jclass   cls,
   jlong    handle, 
   jint     propId, 
   jobject  jniInterval )
{
   MathInterval interval = {0};
   long ret = AdxPropertyRead((HANDLE)handle, propId, sizeof(interval), &interval, NULL, NULL);
   if (jniInterval != NULL)
   {
      javaMathInterval.SetValue(env, jniInterval, interval);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyWrite
 * Signature: (JII[BI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyWrite__JII_3BI(
   JNIEnv *   env, 
   jclass     cls,
   jlong      handle, 
   jint       propId, 
   jint       count,
   jbyteArray jniBuffer, 
   jint       notifyNow )
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jbyte* txbuf = NULL;
   txbuf = env->GetByteArrayElements(jniBuffer, NULL);

   long ret = AdxPropertyWrite((HANDLE)handle, propId, count * sizeof(jbyte), txbuf, notifyNow);
   
   env->ReleaseByteArrayElements(jniBuffer, txbuf, JNI_ABORT);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyWrite
 * Signature: (JII[II)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyWrite__JII_3II(
   JNIEnv *  env, 
   jclass    cls,
   jlong     handle, 
   jint      propId, 
   jint      count,
   jintArray jniBuffer,
   jint      notifyNow )
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jint* txbuf = NULL;
   txbuf = env->GetIntArrayElements(jniBuffer, NULL);

   long ret = AdxPropertyWrite((HANDLE)handle, propId, count * sizeof(jint), txbuf, notifyNow);

   env->ReleaseIntArrayElements(jniBuffer, txbuf, JNI_ABORT);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyWrite
 * Signature: (JII[DI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyWrite__JII_3DI(
   JNIEnv *     env, 
   jclass       cls,
   jlong        handle, 
   jint         propId, 
   jint         count,
   jdoubleArray jniBuffer,
   jint         notifyNow)
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jdouble* txbuf = NULL;
   txbuf = env->GetDoubleArrayElements(jniBuffer, NULL);

   long ret = AdxPropertyWrite((HANDLE)handle, propId, count * sizeof(jdouble), txbuf, notifyNow);

   env->ReleaseDoubleArrayElements(jniBuffer, txbuf, JNI_ABORT);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyWrite
 * Signature: (JII[CI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyWrite__JII_3CI(
   JNIEnv *   env, 
   jclass     cls,
   jlong      handle, 
   jint       propId, 
   jint       count,
   jcharArray jniBuffer,
   jint       notifyNow )
{
   if (!jniBuffer) { return ErrorBufferIsNull; }

   jchar   *jcs = env->GetCharArrayElements(jniBuffer, NULL);
   wchar_t *txbuf;
   if (sizeof(wchar_t) == 4) {// utf-32: linux, Mac OS, gcc
      txbuf = new wchar_t[count];
      JCharToUtf(txbuf, jcs);
   } else { // utf-16: Windows OS, java
      txbuf = (wchar_t*)jcs;
   }

   long ret = AdxPropertyWrite((HANDLE)handle, propId, count * sizeof(wchar_t), txbuf, notifyNow);

   env->ReleaseCharArrayElements(jniBuffer, jcs, JNI_ABORT);

   if (sizeof(wchar_t) == 4){
      delete []txbuf;
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPropertyWrite
 * Signature: (JILAutomation/BDaq/MathInterval;I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPropertyWrite__JILAutomation_BDaq_MathInterval_2I(
   JNIEnv * env, 
   jclass   cls,
   jlong    handle, 
   jint     propId, 
   jobject  jniInterval, 
   jint     notifyNow )
{
   if (jniInterval == NULL)
   {
      return ErrorBufferIsNull;
   }

   MathInterval interval = javaMathInterval.GetValue(env, jniInterval);
   
   return AdxPropertyWrite((HANDLE)handle, propId, sizeof(MathInterval), &interval, notifyNow);
}

/************************************************************************/
/* Device method                                                        */
/************************************************************************/
/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDeviceOpen
 * Signature: (IILAutomation/BDaq/LongByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDeviceOpen(
  JNIEnv * env, 
  jclass   cls,
  jint     deviceNumber,
  jint     accessMode,
  jobject  jniHandle)
{
   if (jniHandle == NULL)
   {
      return ErrorBufferIsNull;
   }

   HANDLE device = NULL;
   long ret = AdxDeviceOpen(deviceNumber, accessMode, &device);
   javaLongByRef.SetValue(env, jniHandle, (jlong)device);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDeviceClose
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDeviceClose(
   JNIEnv * env, 
   jclass   cls,
   jlong    deviceHandle )
{
   return AdxDeviceClose((HANDLE)deviceHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDeviceReset
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDeviceReset(
   JNIEnv * env, 
   jclass   cls,
   jlong    deviceHandle,
   jint     state )
{
   return AdxDeviceReset((HANDLE)deviceHandle, state);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDeviceGetModuleHandle
 * Signature: (JIILAutomation/BDaq/LongByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDeviceGetModuleHandle(
   JNIEnv * env, 
   jclass   cls,
   jlong    deviceHandle,
   jint     moduleType, 
   jint     moduleIndex,
   jobject  jniHandle )
{
   if (jniHandle == NULL)
   {
      return ErrorBufferIsNull;
   }

   HANDLE module = NULL;
   long ret = AdxDeviceGetModuleHandle((HANDLE)deviceHandle, moduleType, moduleIndex, &module);
   javaLongByRef.SetValue(env, jniHandle, (jlong)module);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDeviceRefreshProperties
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDeviceRefreshProperties(
   JNIEnv * env, 
   jclass   cls,
   jlong    deviceHandle )
{
   return AdxDeviceRefreshProperties((HANDLE)deviceHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDeviceShowConfigDialogBox
 * Signature: (JJLjava/awt/Rectangle;IILAutomation/BDaq/LongByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDeviceShowConfigDialogBox(
   JNIEnv * env, 
   jclass   cls,
   jlong    deviceHandle, 
   jlong    parentWindow,
   jobject  jniRect,
   jint     dataSource,
   jint     modal,
   jobject  jniDialog )
{
   RECT rc;
   if (jniRect != NULL)
   {
      rc = javaRectangle.GetValue(env, jniRect);
   }

   HWND dialog = NULL;
   long ret = AdxDeviceShowConfigDialogBox((HANDLE)deviceHandle, (HWND)parentWindow, 
                  jniRect == NULL ? NULL : &rc, dataSource, modal, (HANDLE*)&dialog);
   if (jniDialog != NULL)
   {
      javaLongByRef.SetValue(env, jniDialog, (jlong)dialog);
   }

   return ret; 
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDeviceReadPorts
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDeviceReadPorts(
  JNIEnv     *env,
  jclass     cls,
  jlong      deviceHandle,
  jint       startAddress,
  jint       length,
  jbyteArray jniBuffer)
{
   jbyte* rxbuf = NULL;
   if (jniBuffer != NULL)
   {
      rxbuf = env->GetByteArrayElements(jniBuffer, NULL);
   }

   long ret = AdxDeviceReadPorts((HANDLE)deviceHandle, startAddress, length, rxbuf);

   if (jniBuffer != NULL)
   {
      env->ReleaseByteArrayElements(jniBuffer, rxbuf, 0);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDeviceWritePorts
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDeviceWritePorts(
   JNIEnv     *env,
   jclass     cls,
   jlong      deviceHandle,
   jint       startAddress,
   jint       length,
   jbyteArray jniBuffer)
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jbyte* txbuf = NULL;
   txbuf = env->GetByteArrayElements(jniBuffer, NULL);

   long ret = AdxDeviceWritePorts((HANDLE)deviceHandle, startAddress, length, txbuf);

   env->ReleaseByteArrayElements(jniBuffer, txbuf, JNI_ABORT);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDevicePrivateRegionRead
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDevicePrivateRegionRead(
   JNIEnv     *env,
   jclass     cls,
   jlong      deviceHandle,
   jint       signature,
   jint       length,
   jbyteArray jniBuffer)
{
   jbyte* rxbuf = NULL;
   if (jniBuffer != NULL)
   {
      rxbuf = env->GetByteArrayElements(jniBuffer, NULL);
   }

   long ret = AdxDevicePrivateRegionRead((HANDLE)deviceHandle, signature, length, (uint8*)rxbuf);

   if (jniBuffer != NULL)
   {
      env->ReleaseByteArrayElements(jniBuffer, rxbuf, 0);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDevicePrivateRegionWrite
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDevicePrivateRegionWrite(
   JNIEnv     *env,
   jclass     cls,
   jlong      deviceHandle,
   jint       signature,
   jint       length,
   jbyteArray jniBuffer)
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jbyte* txbuf = NULL;
   txbuf = env->GetByteArrayElements(jniBuffer, NULL);

   long ret = AdxDevicePrivateRegionWrite((HANDLE)deviceHandle, signature, length, (uint8*)txbuf);

   env->ReleaseByteArrayElements(jniBuffer, txbuf, JNI_ABORT);

   return ret;
}

/************************************************************************/
/* AI methods                                                           */
/************************************************************************/
/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxAiReadSamples
 * Signature: (JII[S[D)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxAiReadSamples__JII_3S_3D(
   JNIEnv *     env, 
   jclass       cls,
   jlong        aiHandle,
   jint         chStart,
   jint         chCount,
   jshortArray  jniRawBuf,
   jdoubleArray jniScaledBuf )
{
   jshort * rxRawBuf = NULL;
   if (jniRawBuf != NULL)
   {
      rxRawBuf = env->GetShortArrayElements(jniRawBuf, NULL);
   }

   jdouble * rxScaledBuf = NULL;
   if (jniScaledBuf != NULL)
   {
      rxScaledBuf = env->GetDoubleArrayElements(jniScaledBuf, NULL);
   }

   long ret = AdxAiReadSamples((HANDLE)aiHandle, chStart, chCount, rxRawBuf, rxScaledBuf);

   if (jniRawBuf != NULL)
   {
      env->ReleaseShortArrayElements(jniRawBuf, rxRawBuf, 0);
   }

   if (jniScaledBuf != NULL)
   {
      env->ReleaseDoubleArrayElements(jniScaledBuf, rxScaledBuf, 0);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxAiReadSamples
 * Signature: (JII[I[D)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxAiReadSamples__JII_3I_3D(
   JNIEnv *     env, 
   jclass       cls,
   jlong        aiHandle,
   jint         chStart,
   jint         chCount,
   jintArray    jniRawBuf,
   jdoubleArray jniScaledBuf )
{
   jint * rxRawBuf = NULL;
   if (jniRawBuf != NULL)
   {
      rxRawBuf = env->GetIntArrayElements(jniRawBuf, NULL);
   }

   jdouble * rxScaledBuf = NULL;
   if (jniScaledBuf != NULL)
   {
      rxScaledBuf = env->GetDoubleArrayElements(jniScaledBuf, NULL);
   }

   long ret = AdxAiReadSamples((HANDLE)aiHandle, chStart, chCount, rxRawBuf, rxScaledBuf);

   if (jniRawBuf != NULL)
   {
      env->ReleaseIntArrayElements(jniRawBuf, rxRawBuf, 0);
   }

   if (jniScaledBuf != NULL)
   {
      env->ReleaseDoubleArrayElements(jniScaledBuf, rxScaledBuf, 0);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAiPrepare
 * Signature: (JILAutomation/BDaq/NativeBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAiPrepare(
   JNIEnv * env, 
   jclass   cls,
   jlong    aiHandle,
   jint     dataCount, 
   jobject  jniDataPtr)
{
   if (jniDataPtr == NULL)
   {
      return ErrorBufferIsNull;
   }

   void *dataPtr = NULL;
   long ret = AdxBufferedAiPrepare((HANDLE)aiHandle, dataCount, &dataPtr);

   uint32 dataSize = 0;
   AdxPropertyRead((HANDLE)aiHandle, CFG_FeatureDataSize, sizeof(dataSize), &dataSize, NULL, NULL);

   // set the native buffer
   NativeBuffer nativeBuf = { dataPtr, (int)(dataSize * 8) /*in bits*/, dataCount, 0,  true };
   javaNativeBuffer.Set(env, jniDataPtr, nativeBuf);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAiRunOnce
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAiRunOnce(
   JNIEnv * env, 
   jclass   cls,
   jlong    aiHandle,
   jint     asynchronous )
{
   return AdxBufferedAiRunOnce((HANDLE)aiHandle, asynchronous);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAiRun
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAiRun(
   JNIEnv * env, 
   jclass   cls,
   jlong    aiHandle )
{
   return AdxBufferedAiRun((HANDLE)aiHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAiStop
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAiStop(
   JNIEnv * env, 
   jclass   cls,
   jlong    aiHandle )
{
   return AdxBufferedAiStop((HANDLE)aiHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAiRelease
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAiRelease(
   JNIEnv * env, 
   jclass   cls,
   jlong    aiHandle )
{
   return AdxBufferedAiRelease((HANDLE)aiHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAiScaleData
 * Signature: (JLAutomation/BDaq/NativeBuffer;[DILAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAiScaleData(
   JNIEnv *     env, 
   jclass       cls,
   jlong        aiHandle,
   jobject      jniRawBuf,
   jdoubleArray jniScaledBuf,
   jint         dataCount,
   jobject      jniChOffset )
{
   if (jniRawBuf == NULL || jniScaledBuf == NULL)
   {
      return ErrorBufferIsNull;
   }
   
   NativeBuffer raw = javaNativeBuffer.Get(env, jniRawBuf);
   if (raw.memory == NULL || raw.position >= raw.capacity || dataCount < 0)
   {
      return ErrorParamOutOfRange;
   }

   uint32 chOffset = 0;
   if (jniChOffset != NULL)
   {
      chOffset = javaIntByRef.GetValue(env, jniChOffset);
   }

   dataCount = x_min(x_min(dataCount, raw.capacity), env->GetArrayLength(jniScaledBuf));
   jdouble * scaled = env->GetDoubleArrayElements(jniScaledBuf, NULL);

   // first segment, it's from 'position' to 'buffer end', 'scaleLen' length
   int scaleLen = x_min(dataCount, raw.capacity - raw.position);
   long ret = AdxBufferedAiScaleData((HANDLE)aiHandle, NATIVEBUF_CUR_POS(raw), scaled, scaleLen, &chOffset );

   // copy the rest from 'buffer start'
   dataCount -= scaleLen;
   if (dataCount > 0)
   {
      AdxBufferedAiScaleData((HANDLE)aiHandle, raw.memory, scaled + scaleLen, dataCount, &chOffset );
   }

   env->ReleaseDoubleArrayElements(jniScaledBuf, scaled, 0);

   if (jniChOffset != NULL)
   {
      javaIntByRef.SetValue(env, jniChOffset, chOffset);
   }

   return ret;
}

/************************************************************************/
/* AO methods                                                           */
/************************************************************************/
/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxAoWriteSamples
 * Signature: (JII[S[D)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxAoWriteSamples__JII_3S_3D(
   JNIEnv *     env, 
   jclass       cls,
   jlong        aoHandle,
   jint         chStart,
   jint         chCount,
   jshortArray  jniRawBuf,
   jdoubleArray jniScaledBuf )
{
   jshort * txRawBuf = NULL;
   if (jniRawBuf != NULL)
   {
      txRawBuf = env->GetShortArrayElements(jniRawBuf, NULL);
   }

   jdouble * txScaledBuf = NULL;
   if (jniScaledBuf != NULL)
   {
      txScaledBuf = env->GetDoubleArrayElements(jniScaledBuf, NULL);
   }

   long ret = AdxAoWriteSamples((HANDLE)aoHandle, chStart, chCount, txRawBuf, txScaledBuf);

   if (jniRawBuf != NULL)
   {
      env->ReleaseShortArrayElements(jniRawBuf, txRawBuf, JNI_ABORT);
   }

   if (jniScaledBuf != NULL)
   {
      env->ReleaseDoubleArrayElements(jniScaledBuf, txScaledBuf, JNI_ABORT);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxAoWriteSamples
 * Signature: (JII[I[D)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxAoWriteSamples__JII_3I_3D(
   JNIEnv *     env, 
   jclass       cls,
   jlong        aoHandle,
   jint         chStart,
   jint         chCount,
   jintArray    jniRawBuf,
   jdoubleArray jniScaledBuf )
{
   jint * txRawBuf = NULL;
   if (jniRawBuf != NULL)
   {
      txRawBuf = env->GetIntArrayElements(jniRawBuf, NULL);
   }

   jdouble * txScaledBuf = NULL;
   if (jniScaledBuf != NULL)
   {
      txScaledBuf = env->GetDoubleArrayElements(jniScaledBuf, NULL);
   }

   long ret = AdxAoWriteSamples((HANDLE)aoHandle, chStart, chCount, txRawBuf, txScaledBuf);

   if (jniRawBuf != NULL)
   {
      env->ReleaseIntArrayElements(jniRawBuf, txRawBuf, JNI_ABORT);
   }

   if (jniScaledBuf != NULL)
   {
      env->ReleaseDoubleArrayElements(jniScaledBuf, txScaledBuf, JNI_ABORT);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAoPrepare
 * Signature: (JILAutomation/BDaq/NativeBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAoPrepare(
   JNIEnv * env, 
   jclass   cls,
   jlong    aoHandle,
   jint     dataCount, 
   jobject  jniDataPtr )
{
   if (jniDataPtr == NULL)
   {
      return ErrorBufferIsNull;
   }

   void *dataPtr = NULL;
   long ret = AdxBufferedAoPrepare((HANDLE)aoHandle, dataCount, &dataPtr);

   uint32 dataSize = 0;
   AdxPropertyRead((HANDLE)aoHandle, CFG_FeatureDataSize, sizeof(dataSize), &dataSize, NULL, NULL);

   // set the native buffer
   NativeBuffer nativeBuf = { dataPtr, (int)(dataSize * 8) /*in bits*/, dataCount, 0,  false };
   javaNativeBuffer.Set(env, jniDataPtr, nativeBuf);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAoRunOnce
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAoRunOnce(
   JNIEnv * env, 
   jclass   cls,
   jlong    aoHandle,
   jint     asynchronous )
{
   return AdxBufferedAoRunOnce((HANDLE)aoHandle, asynchronous);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAoRun
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAoRun(
   JNIEnv * env, 
   jclass   cls,
   jlong    aoHandle )
{
   return AdxBufferedAoRun((HANDLE)aoHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAoStop
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAoStop(
   JNIEnv * env, 
   jclass   cls,
   jlong    aoHandle, 
   jint     action )
{
   return AdxBufferedAoStop((HANDLE)aoHandle, action);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAoRelease
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAoRelease(
   JNIEnv * env, 
   jclass   cls,
   jlong    aoHandle )
{
   return AdxBufferedAoRelease((HANDLE)aoHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedAoScaleData
 * Signature: (J[DLAutomation/BDaq/NativeBuffer;ILAutomation/BDaq/IntByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedAoScaleData(
   JNIEnv *     env, 
   jclass       cls,
   jlong        aoHandle, 
   jdoubleArray jniScaledBuf, 
   jobject      jniRawBuf, 
   jint         dataCount, 
   jobject      jniChOffset )
{
   if (jniScaledBuf == NULL || jniRawBuf == NULL) {
      return ErrorBufferIsNull;
   }

   NativeBuffer raw = javaNativeBuffer.Get(env, jniRawBuf);
   if (raw.memory == NULL || raw.position >= raw.capacity || dataCount < 0) {
      return ErrorParamOutOfRange;
   }
   if (raw.readonly) {
      return ErrorParamNotSpted;
   }

   uint32 chOffset = 0;
   if (jniChOffset != NULL) {
      chOffset = javaIntByRef.GetValue(env, jniChOffset);
   }

   dataCount = x_min(x_min(dataCount, raw.capacity), env->GetArrayLength(jniScaledBuf));
   jdouble * scaled = env->GetDoubleArrayElements(jniScaledBuf, NULL);

   // first segment, it's from 'position' to 'buffer end', 'scaleLen' length
   int scaleLen = x_min(dataCount, raw.capacity - raw.position);
   long ret = AdxBufferedAoScaleData((HANDLE)aoHandle, scaled, NATIVEBUF_CUR_POS(raw), scaleLen, &chOffset );

   // copy the rest from 'buffer start'
   dataCount -= scaleLen;
   if (dataCount > 0)
   {
      AdxBufferedAoScaleData((HANDLE)aoHandle, scaled + scaleLen, raw.memory, dataCount, &chOffset );
   }

   env->ReleaseDoubleArrayElements(jniScaledBuf, scaled, JNI_ABORT);

   if (jniChOffset != NULL)
   {
      javaIntByRef.SetValue(env, jniChOffset, chOffset);
   }

   return ret;
}

/************************************************************************/
/* DIO methods                                                          */
/************************************************************************/
/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDiReadPorts
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDiReadPorts(
   JNIEnv *   env, 
   jclass     cls,
   jlong      dioHandle, 
   jint       portStart, 
   jint       portCount, 
   jbyteArray jniBuffer )
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jbyte * rxbuf = env->GetByteArrayElements(jniBuffer, NULL);
   long ret = AdxDiReadPorts((HANDLE)dioHandle, portStart, portCount, (uint8 *)rxbuf);
   env->ReleaseByteArrayElements(jniBuffer, rxbuf, 0);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDoWritePorts
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDoWritePorts(
   JNIEnv *   env, 
   jclass     cls,
   jlong      dioHandle, 
   jint       portStart, 
   jint       portCount, 
   jbyteArray jniBuffer )
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jbyte * txbuf = env->GetByteArrayElements(jniBuffer, NULL);
   long ret = AdxDoWritePorts((HANDLE)dioHandle, portStart, portCount, (uint8 *)txbuf);
   env->ReleaseByteArrayElements(jniBuffer, txbuf, JNI_ABORT);

   return ret;
}


/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDoReadBackPorts
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDoReadBackPorts(
   JNIEnv *   env, 
   jclass     cls,
   jlong      dioHandle, 
   jint       portStart, 
   jint       portCount, 
   jbyteArray jniBuffer )
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jbyte * rxbuf = env->GetByteArrayElements(jniBuffer, NULL);
   long ret = AdxDoReadBackPorts((HANDLE)dioHandle, portStart, portCount, (uint8 *)rxbuf);
   env->ReleaseByteArrayElements(jniBuffer, rxbuf, 0);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedDiPrepare
 * Signature: (JILAutomation/BDaq/NativeBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedDiPrepare(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle,
   jint     dataCount, 
   jobject  jniDataPtr )
{
   if (jniDataPtr == NULL)
   {
      return ErrorBufferIsNull;
   }

   void *dataPtr = NULL;
   long ret = AdxBufferedDiPrepare((HANDLE)dioHandle, dataCount, &dataPtr);

   // set the native buffer
   NativeBuffer nativeBuf = { dataPtr, 8 /*in bits*/, dataCount, 0,  true };
   javaNativeBuffer.Set(env, jniDataPtr, nativeBuf);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedDiRunOnce
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedDiRunOnce(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle,
   jint     asynchronous )
{
   return AdxBufferedDiRunOnce((HANDLE)dioHandle, asynchronous);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedDiRun
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedDiRun(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle )
{
   return AdxBufferedDiRun((HANDLE)dioHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedDiStop
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedDiStop(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle )
{
   return AdxBufferedDiStop((HANDLE)dioHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedDiRelease
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedDiRelease(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle )
{
   return AdxBufferedDiRelease((HANDLE)dioHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedDoPrepare
 * Signature: (JILAutomation/BDaq/NativeBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedDoPrepare(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle,
   jint     dataCount, 
   jobject  jniDataPtr )
{
   if (jniDataPtr == NULL)
   {
      return ErrorBufferIsNull;
   }

   void *dataPtr = NULL;
   long ret = AdxBufferedDoPrepare((HANDLE)dioHandle, dataCount, &dataPtr);

   // set the native buffer
   NativeBuffer nativeBuf = { dataPtr, 8 /*in bits*/, dataCount, 0,  false };
   javaNativeBuffer.Set(env, jniDataPtr, nativeBuf);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedDoRunOnce
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedDoRunOnce(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle,
   jint     asynchronous )
{
   return AdxBufferedDoRunOnce((HANDLE)dioHandle, asynchronous);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedDoRun
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedDoRun(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle )
{
   return AdxBufferedDoRun((HANDLE)dioHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedDoStop
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedDoStop(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle,
   jint     action )
{
   return AdxBufferedDoStop((HANDLE)dioHandle, action);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxBufferedDoRelease
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxBufferedDoRelease(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle )
{
   return AdxBufferedDoRelease((HANDLE)dioHandle);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDiSnapStart
 * Signature: (JIIILAutomation/BDaq/NativeBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDiSnapStart(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle, 
   jint     eventId, 
   jint     portStart, 
   jint     portCount, 
   jobject  jniDataPtr)
{
   uint8 * dataPtr = NULL;

   long ret = AdxDiSnapStart((HANDLE)dioHandle, eventId, portStart, portCount, &dataPtr);
   if (jniDataPtr != NULL)
   {
      uint32 total = 0;
      AdxPropertyRead((HANDLE)dioHandle, CFG_FeaturePortsCount, sizeof(total), &total, NULL, NULL);

      // set the native buffer
      NativeBuffer nativeBuf = { dataPtr, 8 /*in bits*/, (int)total, 0,  true };
      javaNativeBuffer.Set(env, jniDataPtr, nativeBuf);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxDiSnapStop
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxDiSnapStop(
   JNIEnv * env, 
   jclass   cls,
   jlong    dioHandle, 
   jint     eventId )
{
   return AdxDiSnapStop((HANDLE)dioHandle, eventId);
}


/************************************************************************/
/* Counter methods                                                      */
/************************************************************************/
/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxCounterReset
 * Signature: (JII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxCounterReset(
   JNIEnv * env, 
   jclass   cls,
   jlong    cntrHandle, 
   jint     cntrStart,
   jint     cntrCount )
{
   return AdxCounterReset((HANDLE)cntrHandle, cntrStart, cntrCount);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxEventCountStart
 * Signature: (JII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxEventCountStart(
   JNIEnv * env, 
   jclass   cls,
   jlong    cntrHandle, 
   jint     cntrStart,
   jint     cntrCount )
{
   return AdxEventCountStart((HANDLE)cntrHandle, cntrStart, cntrCount);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxEventCountRead
 * Signature: (JII[I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxEventCountRead(
   JNIEnv *  env, 
   jclass    cls,
   jlong     cntrHandle, 
   jint      cntrStart,
   jint      cntrCount,
   jintArray jniBuffer )
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jint * rxbuf = env->GetIntArrayElements(jniBuffer, NULL);
   long ret = AdxEventCountRead((HANDLE)cntrHandle, cntrStart, cntrCount, (uint32*)rxbuf);
   env->ReleaseIntArrayElements(jniBuffer, rxbuf, 0);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxOneShotStart
 * Signature: (JII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxOneShotStart(
   JNIEnv * env, 
   jclass   cls,
   jlong    cntrHandle, 
   jint     cntrStart,
   jint     cntrCount )
{
   return AdxOneShotStart((HANDLE)cntrHandle, cntrStart, cntrCount);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxTimerPulseStart
 * Signature: (JII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxTimerPulseStart(
   JNIEnv * env, 
   jclass   cls,
   jlong    cntrHandle, 
   jint     cntrStart,
   jint     cntrCount )
{
   return AdxTimerPulseStart((HANDLE)cntrHandle, cntrStart, cntrCount);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxFrequencyMeasureStart
 * Signature: (JII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxFrequencyMeasureStart(
   JNIEnv * env, 
   jclass   cls,
   jlong    cntrHandle, 
   jint     cntrStart,
   jint     cntrCount )
{
   return AdxFrequencyMeasureStart((HANDLE)cntrHandle, cntrStart, cntrCount);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxFrequencyMeasureRead
 * Signature: (JII[D)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxFrequencyMeasureRead(
   JNIEnv *     env, 
   jclass       cls,
   jlong        cntrHandle, 
   jint         cntrStart,
   jint         cntrCount,
   jdoubleArray jniBuffer )
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jdouble * rxbuf = env->GetDoubleArrayElements(jniBuffer, NULL);
   long ret = AdxFrequencyMeasureRead((HANDLE)cntrHandle, cntrStart, cntrCount, rxbuf);
   env->ReleaseDoubleArrayElements(jniBuffer, rxbuf, 0);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPwmInStart
 * Signature: (JII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPwmInStart(
   JNIEnv * env, 
   jclass   cls,
   jlong    cntrHandle, 
   jint     cntrStart,
   jint     cntrCount )
{
   return AdxPwmInStart((HANDLE)cntrHandle, cntrStart, cntrCount);
}


/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPwmInRead
 * Signature: (JII[D[D)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPwmInRead(
   JNIEnv *     env, 
   jclass       cls,
   jlong        cntrHandle, 
   jint         cntrStart,
   jint         cntrCount,
   jdoubleArray jniHiBuf, 
   jdoubleArray jniLoBuf )
{
   jdouble * rxHiBuf = NULL;
   if (jniHiBuf != NULL)
   {
      rxHiBuf = env->GetDoubleArrayElements(jniHiBuf, NULL);
   }

   jdouble * rxLoBuf = NULL;
   if (jniLoBuf != NULL)
   {
      rxLoBuf = env->GetDoubleArrayElements(jniLoBuf, NULL);
   }

   long ret = AdxPwmInRead((HANDLE)cntrHandle, cntrStart, cntrCount, rxHiBuf, rxLoBuf);

   if (jniHiBuf != NULL)
   {
      env->ReleaseDoubleArrayElements(jniHiBuf, rxHiBuf, 0);
   }
   if (jniLoBuf != NULL)
   {
      env->ReleaseDoubleArrayElements(jniLoBuf, rxLoBuf, 0);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPwmOutStart
 * Signature: (JII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPwmOutStart(
   JNIEnv * env, 
   jclass   cls,
   jlong    cntrHandle, 
   jint     cntrStart,
   jint     cntrCount )
{
   return AdxPwmOutStart((HANDLE)cntrHandle, cntrStart, cntrCount);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxUpDownCountStart
 * Signature: (JIII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxUpDownCountStart(
   JNIEnv * env,
   jclass   cls,
   jlong    cntrHandle,
   jint     cntrStart,
   jint     cntrCount,
   jint     resetCntr)
{
   return AdxUpDownCountStart((HANDLE)cntrHandle, cntrStart, cntrCount, resetCntr);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxUpDownCountValueReset
 * Signature: (JII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxUpDownCountValueReset(
   JNIEnv * env,
   jclass   cls,
   jlong    cntrHandle,
   jint     cntrStart,
   jint     cntrCount)
{
   return AdxUpDownCountValueReset((HANDLE)cntrHandle, cntrStart, cntrCount);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxUpDownCountRead
 * Signature: (JII[I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxUpDownCountRead(
   JNIEnv *  env,
   jclass    cls,
   jlong     cntrHandle,
   jint      cntrStart,
   jint      cntrCount,
   jintArray jniBuffer)
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jint * rxbuf = env->GetIntArrayElements(jniBuffer, NULL);
   long ret = AdxUpDownCountRead((HANDLE)cntrHandle, cntrStart, cntrCount, (uint32*)rxbuf);
   env->ReleaseIntArrayElements(jniBuffer, rxbuf, 0);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxCounterSnapStart
 * Signature: (JIIILAutomation/BDaq/NativeBuffer;LAutomation/BDaq/NativeBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxCounterSnapStart(
   JNIEnv * env,
   jclass   cls,
   jlong    cntrHandle,
   jint     eventId,
   jint     cntrStart,
   jint     cntrCount,
   jobject  jniLoadValues,
   jobject  jniHoldValues)
{
   uint32 * loadValues = NULL;
   uint32 * holdValues = NULL;

   long ret = AdxCounterSnapStart((HANDLE)cntrHandle, eventId, cntrStart, cntrCount, &loadValues, &holdValues);

   uint32 total = 0;
   AdxPropertyRead((HANDLE)cntrHandle, CFG_FeatureChannelNumberMax, sizeof(total), &total, NULL, NULL);
   ++total;

   if (jniLoadValues && loadValues)
   {
      // set the native buffer
      NativeBuffer nativeBuf = { loadValues, 32 /*in bits*/, (int)(total * sizeof(uint32)), 0, true };
      javaNativeBuffer.Set(env, jniLoadValues, nativeBuf);
   }

   if (jniHoldValues && holdValues)
   {
      // set the native buffer
      NativeBuffer nativeBuf = { holdValues, 32 /*in bits*/, (int)(total * sizeof(uint32)), 0, true };
      javaNativeBuffer.Set(env, jniHoldValues, nativeBuf);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxCounterSnapStop
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxCounterSnapStop(
   JNIEnv * env,
   jclass   cls,
   jlong    cntrHandle,
   jint     eventId)
{
   return AdxCounterSnapStop((HANDLE)cntrHandle, eventId);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxCounterCompareSetTable
 * Signature: (JII[I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxCounterCompareSetTable(
   JNIEnv *  env,
   jclass    cls,
   jlong     cntrHandle,
   jint      cntr,
   jint      dataCount,
   jintArray jniBuffer)
{
   if (jniBuffer == NULL)
   {
      return ErrorBufferIsNull;
   }

   jint * txBuf = env->GetIntArrayElements(jniBuffer, NULL);
   long ret = AdxCounterCompareSetTable((HANDLE)cntrHandle, cntr, dataCount, (uint32*)txBuf);
   env->ReleaseIntArrayElements(jniBuffer, txBuf, JNI_ABORT);

   return ret;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxCounterCompareSetInterval
 * Signature: (JIIII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxCounterCompareSetInterval(
   JNIEnv * env,
   jclass   cls,
   jlong    cntrHandle,
   jint     cntr,
   jint     firstValue,
   jint     increment,
   jint     count)
{
   return AdxCounterCompareSetInterval((HANDLE)cntrHandle, cntr, firstValue, increment, count);
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxCounterCompareClear
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxCounterCompareClear(
   JNIEnv * env,
   jclass   cls,
   jlong    cntrHandle,
   jint     cntr)
{
   return AdxCounterCompareClear((HANDLE)cntrHandle, cntr);
}
/************************************************************************/
/* Helper methods to exchange bulk data with JVM                        */
/************************************************************************/
/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxMemoryRead
 * Signature: ([BILAutomation/BDaq/NativeBuffer;I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxMemoryRead___3BILAutomation_BDaq_NativeBuffer_2I(
   JNIEnv *   env, 
   jclass     cls,
   jbyteArray jniDstBuf,
   jint       offset, 
   jobject    jniSrcBuf,
   jint       count )
{
   if (jniDstBuf == NULL || jniSrcBuf == NULL || offset < 0)
   {
      return 0;
   }

   NativeBuffer src = javaNativeBuffer.Get(env, jniSrcBuf);
   if (src.memory == NULL || src.elementSize != 8 || src.position >= src.capacity)
   {
      return 0;
   }

   count = x_min(x_min(count, env->GetArrayLength(jniDstBuf) - offset), src.capacity);
   if (count <= 0)
   {
      return 0;
   }

   jbyte * dest = env->GetByteArrayElements(jniDstBuf, NULL);

   // first segment, it's from 'position' to 'buffer end', 'scaleLen' length
   int copyLen = x_min(count, src.capacity - src.position);
   memcpy(dest + offset, NATIVEBUF_CUR_POS(src), copyLen * sizeof(jbyte));

   // copy the rest from 'buffer start'
   count -= copyLen;
   if (count > 0)
   {
      memcpy(dest + offset + copyLen, src.memory, count * sizeof(jbyte));
   }

   env->ReleaseByteArrayElements(jniDstBuf, dest, 0);

   return count;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxMemoryRead
 * Signature: ([SILAutomation/BDaq/NativeBuffer;I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxMemoryRead___3SILAutomation_BDaq_NativeBuffer_2I(
   JNIEnv *    env, 
   jclass      cls,
   jshortArray jniDstBuf,
   jint        offset, 
   jobject     jniSrcBuf,
   jint        count )
{
   if (jniDstBuf == NULL || jniSrcBuf == NULL || offset < 0)
   {
      return 0;
   }

   NativeBuffer src = javaNativeBuffer.Get(env, jniSrcBuf);
   if (src.memory == NULL || src.elementSize != 16 || src.position >= src.capacity)
   {
      return 0;
   }

   count = x_min(x_min(count, env->GetArrayLength(jniDstBuf) - offset), src.capacity);
   if (count <= 0)
   {
      return 0;
   }

   jshort * dest = env->GetShortArrayElements(jniDstBuf, NULL);

   // first segment, it's from 'position' to 'buffer end', 'scaleLen' length
   int copyLen = x_min(count, src.capacity - src.position);
   memcpy(dest + offset, NATIVEBUF_CUR_POS(src), copyLen * sizeof(jshort));

   // copy the rest from 'buffer start'
   count -= copyLen;
   if (count > 0)
   {
      memcpy(dest + offset + copyLen, src.memory, count * sizeof(jshort));
   }

   env->ReleaseShortArrayElements(jniDstBuf, dest, 0);

   return count;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxMemoryRead
 * Signature: ([IILAutomation/BDaq/NativeBuffer;I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxMemoryRead___3IILAutomation_BDaq_NativeBuffer_2I(
   JNIEnv *  env, 
   jclass    cls,
   jintArray jniDstBuf,
   jint      offset, 
   jobject   jniSrcBuf,
   jint      count )
{
   if (jniDstBuf == NULL || jniSrcBuf == NULL || offset < 0)
   {
      return 0;
   }

   NativeBuffer src = javaNativeBuffer.Get(env, jniSrcBuf);
   if (src.memory == NULL || src.elementSize != 32 || src.position >= src.capacity)
   {
      return 0;
   }

   count = x_min(x_min(count, env->GetArrayLength(jniDstBuf) - offset), src.capacity);
   if (count <= 0)
   {
      return 0;
   }

   jint * dest = env->GetIntArrayElements(jniDstBuf, NULL);

   // first segment, it's from 'position' to 'buffer end', 'scaleLen' length
   int copyLen = x_min(count, src.capacity - src.position);
   memcpy(dest + offset, NATIVEBUF_CUR_POS(src), copyLen * sizeof(jint));

   // copy the rest from 'buffer start'
   count -= copyLen;
   if (count > 0)
   {
      memcpy(dest + offset + copyLen, src.memory, count * sizeof(jint));
   }

   env->ReleaseIntArrayElements(jniDstBuf, dest, 0);

   return count;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxMemoryWrite
 * Signature: (LAutomation/BDaq/NativeBuffer;[BII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxMemoryWrite__LAutomation_BDaq_NativeBuffer_2_3BII(
   JNIEnv *   env, 
   jclass     cls,
   jobject    jniDstBuf,
   jbyteArray jniSrcBuf,
   jint       offset, 
   jint       count )
{
   if (jniDstBuf == NULL || jniSrcBuf == NULL || offset < 0)
   {
      return 0;
   }

   NativeBuffer dest = javaNativeBuffer.Get(env, jniDstBuf);
   if (dest.memory == NULL || dest.readonly || dest.elementSize != 8 || dest.position >= dest.capacity)
   {
      return 0;
   }

   count = x_min(x_min(count, env->GetArrayLength(jniSrcBuf) - offset), dest.capacity);
   if (count <= 0)
   {
      return 0;
   }

   jbyte * src = env->GetByteArrayElements(jniSrcBuf, NULL);

   // first segment, it's from 'position' to 'buffer end', 'scaleLen' length
   int copyLen = x_min(count, dest.capacity - dest.position);
   memcpy(NATIVEBUF_CUR_POS(dest), src + offset, copyLen * sizeof(jbyte));

   // copy the rest from 'buffer start'
   count -= copyLen;
   if (count > 0)
   {
      memcpy(dest.memory, src + offset + copyLen, count * sizeof(jbyte));
   }

   env->ReleaseByteArrayElements(jniSrcBuf, src, JNI_ABORT);

   return count;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxMemoryWrite
 * Signature: (LAutomation/BDaq/NativeBuffer;[SII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxMemoryWrite__LAutomation_BDaq_NativeBuffer_2_3SII(
   JNIEnv *    env, 
   jclass      cls,
   jobject     jniDstBuf,
   jshortArray jniSrcBuf,
   jint        offset, 
   jint        count )
{
   if (jniDstBuf == NULL || jniSrcBuf == NULL || offset < 0)
   {
      return 0;
   }

   NativeBuffer dest = javaNativeBuffer.Get(env, jniDstBuf);
   if (dest.memory == NULL || dest.readonly || dest.elementSize != 16 || dest.position >= dest.capacity)
   {
      return 0;
   }

   count = x_min(x_min(count, env->GetArrayLength(jniSrcBuf) - offset), dest.capacity);
   if (count <= 0)
   {
      return 0;
   }

   jshort * src = env->GetShortArrayElements(jniSrcBuf, NULL);

   // first segment, it's from 'position' to 'buffer end', 'scaleLen' length
   int copyLen = x_min(count, dest.capacity - dest.position);
   memcpy(NATIVEBUF_CUR_POS(dest), src + offset, copyLen * sizeof(jshort));

   // copy the rest from 'buffer start'
   count -= copyLen;
   if (count > 0)
   {
      memcpy(dest.memory, src + offset + copyLen, count * sizeof(jshort));
   }

   env->ReleaseShortArrayElements(jniSrcBuf, src, JNI_ABORT);

   return count;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxMemoryWrite
 * Signature: (LAutomation/BDaq/NativeBuffer;[III)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxMemoryWrite__LAutomation_BDaq_NativeBuffer_2_3III(
   JNIEnv *  env, 
   jclass    cls,
   jobject   jniDstBuf,
   jintArray jniSrcBuf,
   jint      offset, 
   jint      count )
{
   if (jniDstBuf == NULL || jniSrcBuf == NULL || offset < 0)
   {
      return 0;
   }

   NativeBuffer dest = javaNativeBuffer.Get(env, jniDstBuf);
   if (dest.memory == NULL || dest.readonly || dest.elementSize != 32 || dest.position >= dest.capacity)
   {
      return 0;
   }

   count = x_min(x_min(count, env->GetArrayLength(jniSrcBuf) - offset), dest.capacity);
   if (count <= 0)
   {
      return 0;
   }

   jint * src = env->GetIntArrayElements(jniSrcBuf, NULL);

   // first segment, it's from 'position' to 'buffer end', 'scaleLen' length
   int copyLen = x_min(count, dest.capacity - dest.position);
   memcpy(NATIVEBUF_CUR_POS(dest), src + offset, copyLen * sizeof(jint));

   // copy the rest from 'buffer start'
   count -= copyLen;
   if (count > 0)
   {
      memcpy(dest.memory, src + offset + copyLen, count * sizeof(jint));
   }
   env->ReleaseIntArrayElements(jniSrcBuf, src, JNI_ABORT);

   return count;
}

/*
 * Class:     Automation_BDaq_BDaqApi
 * Method:    AdxPrivateExchange
 * Signature: (Ljava/lang/String;Ljava/lang/Object;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_BDaqApi_AdxPrivateExchange(
   JNIEnv * env, 
   jclass   cls,
   jstring  jniCmd,
   jobject  jniRet )
{
   // hack method, do something which cannot be in the air.
   if (jniCmd == NULL)
   {
      return 0;
   }

   const char * xcmd = env->GetStringUTFChars(jniCmd, NULL);
#if defined(_WIN32) || defined(WIN32)
   __try {
#endif
      if (x_stricmp(xcmd, "CREVT") == 0)
      {
         if (jniRet != NULL && env->IsInstanceOf(jniRet, env->FindClass("Automation/BDaq/WaitHandle")))
         {
            javaWaitHandle.SetHandle(env, jniRet, CreateEvent(NULL, FALSE, FALSE, NULL));
         }
      } else if (x_stricmp(xcmd, "CLEVT") == 0){
         if (jniRet != NULL && env->IsInstanceOf(jniRet, env->FindClass("Automation/BDaq/WaitHandle")))
         {
            CloseHandle(javaWaitHandle.GetHandle(env, jniRet));
            javaWaitHandle.SetHandle(env, jniRet, NULL);
         }
      } else if (x_stricmp(xcmd, "STEVT") == 0){
         if (jniRet != NULL && env->IsInstanceOf(jniRet, env->FindClass("Automation/BDaq/WaitHandle")))
         {
           SetEvent(javaWaitHandle.GetHandle(env, jniRet));
         }
      }
#if defined(_WIN32) || defined(WIN32)
   }__except(EXCEPTION_CONTINUE_EXECUTION) {
      // need do nothing
   }
#endif
   env->ReleaseStringUTFChars(jniCmd, xcmd);
   
   return 0;
}


//*************************************************************************************
// DAQNAVI JAVA STUB -- DAQNAVI4 API 
//*************************************************************************************
/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TArray */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TArray
 * Method:    Dispose
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TArray_Dispose
  (JNIEnv *, jclass, jlong obj)
{
   TArray_Dispose((TArray)obj);
}

/*
 * Class:     Automation_BDaq_TArray
 * Method:    getLength
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TArray_getLength
  (JNIEnv *, jclass, jlong obj)
{
   return TArray_getLength((TArray)obj);
}

/*
 * Class:     Automation_BDaq_TArray
 * Method:    getItem
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TArray_getItem
  (JNIEnv *, jclass, jlong obj, jint index)
{
   return (jlong)TArray_getItem((TArray)obj, index);
}


/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TScanChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TScanChannel
 * Method:    getChannelStart
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanChannel_getChannelStart
  (JNIEnv *, jclass, jlong obj)
{
   return TScanChannel_getChannelStart((TScanChannel)obj);
}

/*
 * Class:     Automation_BDaq_TScanChannel
 * Method:    setChannelStart
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanChannel_setChannelStart
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TScanChannel_setChannelStart((TScanChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TScanChannel
 * Method:    getChannelCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanChannel_getChannelCount
  (JNIEnv *, jclass, jlong obj)
{
   return TScanChannel_getChannelCount((TScanChannel)obj);
}

/*
 * Class:     Automation_BDaq_TScanChannel
 * Method:    setChannelCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanChannel_setChannelCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TScanChannel_setChannelCount((TScanChannel)obj, value);
}
/*
 * Class:     Automation_BDaq_TScanChannel
 * Method:    getIntervalCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanChannel_getIntervalCount
  (JNIEnv *, jclass, jlong obj)
{
   return TScanChannel_getIntervalCount((TScanChannel)obj);
}

/*
 * Class:     Automation_BDaq_TScanChannel
 * Method:    setIntervalCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanChannel_setIntervalCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TScanChannel_setIntervalCount((TScanChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TScanChannel
 * Method:    getSamples
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanChannel_getSamples
  (JNIEnv *, jclass, jlong obj)
{
   return TScanChannel_getSamples((TScanChannel)obj);
}

/*
 * Class:     Automation_BDaq_TScanChannel
 * Method:    setSamples
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanChannel_setSamples
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TScanChannel_setSamples((TScanChannel)obj, value);
}


/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TConvertClock */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TConvertClock
 * Method:    getSource
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TConvertClock_getSource
  (JNIEnv *, jclass, jlong obj)
{
   return TConvertClock_getSource((TConvertClock)obj);
}

/*
 * Class:     Automation_BDaq_TConvertClock
 * Method:    setSource
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TConvertClock_setSource
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TConvertClock_setSource((TConvertClock)obj, (SignalDrop)value);
}

/*
 * Class:     Automation_BDaq_TConvertClock
 * Method:    getRate
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TConvertClock_getRate
  (JNIEnv *, jclass, jlong obj)
{
   return TConvertClock_getRate((TConvertClock)obj);
}

/*
 * Class:     Automation_BDaq_TConvertClock
 * Method:    setRate
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TConvertClock_setRate
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TConvertClock_setRate((TConvertClock)obj, value);
}


/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TScanClock */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TScanClock
 * Method:    getSource
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanClock_getSource
  (JNIEnv *, jclass, jlong obj)
{
   return TScanClock_getSource((TScanClock)obj);
}

/*
 * Class:     Automation_BDaq_TScanClock
 * Method:    setSource
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanClock_setSource
   (JNIEnv *, jclass, jlong obj, jint value)
{
   return TScanClock_setSource((TScanClock)obj, (SignalDrop)value);
}
/*
 * Class:     Automation_BDaq_TScanClock
 * Method:    getRate
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TScanClock_getRate
  (JNIEnv *, jclass, jlong obj)
{
   return TScanClock_getRate((TScanClock)obj);
}

/*
 * Class:     Automation_BDaq_TScanClock
 * Method:    setRate
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanClock_setRate
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TScanClock_setRate((TScanClock)obj, value);
}

/*
 * Class:     Automation_BDaq_TScanClock
 * Method:    getScanCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanClock_getScanCount
  (JNIEnv *, jclass, jlong obj)
{
   return TScanClock_getScanCount((TScanClock)obj);
}

/*
 * Class:     Automation_BDaq_TScanClock
 * Method:    setScanCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanClock_setScanCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TScanClock_setScanCount((TScanClock)obj, value);
}


/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TTrigger */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    getSource
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_getSource
  (JNIEnv *, jclass, jlong obj)
{
   return TTrigger_getSource((TTrigger)obj);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    setSource
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_setSource
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TTrigger_setSource((TTrigger)obj, (SignalDrop)value);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    getEdge
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_getEdge
  (JNIEnv *, jclass, jlong obj)
{
   return TTrigger_getEdge((TTrigger)obj);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    setEdge
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_setEdge
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TTrigger_setEdge((TTrigger)obj, (ActiveSignal)value);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    getLevel
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TTrigger_getLevel
  (JNIEnv *, jclass, jlong obj)
{
   return TTrigger_getLevel((TTrigger)obj);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    setLevel
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_setLevel
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TTrigger_setLevel((TTrigger)obj, value);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    getAction
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_getAction
  (JNIEnv *, jclass, jlong obj)
{
   return TTrigger_getAction((TTrigger)obj);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    setAction
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_setAction
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TTrigger_setAction((TTrigger)obj, (TriggerAction)value);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    getDelayCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_getDelayCount
  (JNIEnv *, jclass, jlong obj)
{
   return TTrigger_getDelayCount((TTrigger)obj);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    setDelayCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_setDelayCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TTrigger_setDelayCount((TTrigger)obj, value);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    getHysteresisIndex
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TTrigger_getHysteresisIndex
  (JNIEnv *, jclass, jlong obj)
{
   return TTrigger_getHysteresisIndex((TTrigger)obj);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    setHysteresisIndex
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_setHysteresisIndex
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TTrigger_setHysteresisIndex((TTrigger)obj, value);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    getFilterType
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_getFilterType
  (JNIEnv *, jclass, jlong obj)
{
   return TTrigger_getFilterType((TTrigger)obj);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    setFilterType
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_setFilterType
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TTrigger_setFilterType((TTrigger)obj, (FilterType)value);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    getFilterCutoffFreq
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TTrigger_getFilterCutoffFreq
  (JNIEnv *, jclass, jlong obj)
{
   return TTrigger_getFilterCutoffFreq((TTrigger)obj);
}

/*
 * Class:     Automation_BDaq_TTrigger
 * Method:    setFilterCutoffFreq
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTrigger_setFilterCutoffFreq
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TTrigger_setFilterCutoffFreq((TTrigger)obj, value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TConversion */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TConversion
 * Method:    getClockSource
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TConversion_getClockSource
  (JNIEnv *, jclass, jlong obj)
{
   return TConversion_getClockSource((TConversion)obj);
}

/*
 * Class:     Automation_BDaq_TConversion
 * Method:    setClockSource
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TConversion_setClockSource
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TConversion_setClockSource((TConversion)obj, (SignalDrop)value);
}

/*
 * Class:     Automation_BDaq_TConversion
 * Method:    getClockRate
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TConversion_getClockRate
  (JNIEnv *, jclass, jlong obj)
{
   return TConversion_getClockRate((TConversion)obj);
}

/*
 * Class:     Automation_BDaq_TConversion
 * Method:    setClockRate
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TConversion_setClockRate
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TConversion_setClockRate((TConversion)obj, value);
}

/*
 * Class:     Automation_BDaq_TConversion
 * Method:    getChannelStart
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TConversion_getChannelStart
   (JNIEnv *, jclass, jlong obj)
{
   return TConversion_getChannelStart((TConversion)obj);
}

/*
 * Class:     Automation_BDaq_TConversion
 * Method:    setChannelStart
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TConversion_setChannelStart
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TConversion_setChannelStart((TConversion)obj, value);
}

/*
 * Class:     Automation_BDaq_TConversion
 * Method:    getChannelCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TConversion_getChannelCount
  (JNIEnv *, jclass, jlong obj)
{
   return TConversion_getChannelCount((TConversion)obj);
}

/*
 * Class:     Automation_BDaq_TConversion
 * Method:    setChannelCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TConversion_setChannelCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TConversion_setChannelCount((TConversion)obj, value);
}

/*
 * Class:     Automation_BDaq_TConversion
 * Method:    getChannelMap
 * Signature: (JI[B)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TConversion_getChannelMap
  (JNIEnv *env, jclass, jlong obj, jint count, jbyteArray buffer)
{
   if (!buffer) { return; }

   jbyte* rxbuf = env->GetByteArrayElements(buffer, NULL);

   TConversion_getChannelMap((TConversion)obj, count, rxbuf);

   env->ReleaseByteArrayElements(buffer, rxbuf, 0);
}

/*
 * Class:     Automation_BDaq_TConversion
 * Method:    setChannelMap
 * Signature: (JI[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TConversion_setChannelMap
  (JNIEnv *env, jclass, jlong obj, jint count, jbyteArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jbyte* txbuf = env->GetByteArrayElements(buffer, NULL);

   ErrorCode ret = TConversion_setChannelMap((TConversion)obj, count, txbuf);

   env->ReleaseByteArrayElements(buffer, txbuf, JNI_ABORT);

   return ret;
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TRecord */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TRecord
 * Method:    getSectionLength
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TRecord_getSectionLength
  (JNIEnv *, jclass, jlong obj)
{
   return TRecord_getSectionLength((TRecord)obj);
}
/*
 * Class:     Automation_BDaq_TRecord
 * Method:    setSectionLength
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TRecord_setSectionLength
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TRecord_setSectionLength((TRecord)obj, value);
}

/*
 * Class:     Automation_BDaq_TRecord
 * Method:    getSectionCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TRecord_getSectionCount
  (JNIEnv *, jclass, jlong obj)
{
   return TRecord_getSectionCount((TRecord)obj);
}

/*
 * Class:     Automation_BDaq_TRecord
 * Method:    setSectionCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TRecord_setSectionCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TRecord_setSectionCount((TRecord)obj, value);
}

/*
 * Class:     Automation_BDaq_TRecord
 * Method:    getCycles
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TRecord_getCycles
  (JNIEnv *, jclass, jlong obj)
{
   return TRecord_getCycles((TRecord)obj);
}

/*
 * Class:     Automation_BDaq_TRecord
 * Method:    setCycles
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TRecord_setCycles
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TRecord_setCycles((TRecord)obj, value);
}


/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TScanPort */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    getPortStart
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanPort_getPortStart
  (JNIEnv *, jclass, jlong obj)
{
   return TScanPort_getPortStart((TScanPort)obj);
}  

/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    setPortStart
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanPort_setPortStart
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TScanPort_setPortStart((TScanPort)obj, value);
}

/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    getPortCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanPort_getPortCount
  (JNIEnv *, jclass, jlong obj)
{
   return TScanPort_getPortCount((TScanPort)obj);
}

/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    setPortCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanPort_setPortCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TScanPort_setPortCount((TScanPort)obj, value);
}

/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    getPortMap
 * Signature: (JI[Z)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TScanPort_getPortMap
  (JNIEnv *env, jclass, jlong obj, jint count, jbyteArray buffer)
{
   if (!buffer) { return; }

   jbyte* rxbuf = env->GetByteArrayElements(buffer, NULL);

   TScanPort_getPortMap((TScanPort)obj, count, rxbuf);

   env->ReleaseByteArrayElements(buffer, rxbuf, 0);
}

/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    setPortMap
 * Signature: (JI[Z)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanPort_setPortMap
  (JNIEnv *env, jclass, jlong obj, jint count, jbyteArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jbyte* txbuf = env->GetByteArrayElements(buffer, NULL);

   ErrorCode ret = TScanPort_setPortMap((TScanPort)obj, count, txbuf);

   env->ReleaseByteArrayElements(buffer, txbuf, JNI_ABORT);
   
   return ret;
}

/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    getSectionLength
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanPort_getSectionLength
  (JNIEnv *, jclass, jlong obj)
{
   return TScanPort_getSectionLength((TScanPort)obj);
}

/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    setSectionLength
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanPort_setSectionLength
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TScanPort_setSectionLength((TScanPort)obj, value);
}

/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    getSectionCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanPort_getSectionCount
  (JNIEnv *, jclass, jlong obj)
{
   return TScanPort_getSectionCount((TScanPort)obj);
}

/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    setSectionCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanPort_setSectionCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TScanPort_setSectionCount((TScanPort)obj, value);
}

/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    getCycles
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanPort_getCycles
  (JNIEnv *, jclass, jlong obj)
{
   return TScanPort_getCycles((TScanPort)obj);
}
/*
 * Class:     Automation_BDaq_TScanPort
 * Method:    setCycles
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TScanPort_setCycles
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TScanPort_setCycles((TScanPort)obj, value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TNosFltChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TNosFltChannel
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TNosFltChannel_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TNosFltChannel_getChannel((TNosFltChannel)obj);
}

/*
 * Class:     Automation_BDaq_TNosFltChannel
 * Method:    getEnabled
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TNosFltChannel_getEnabled
  (JNIEnv *, jclass, jlong obj)
{
   return TNosFltChannel_getEnabled((TNosFltChannel)obj);
}

/*
 * Class:     Automation_BDaq_TNosFltChannel
 * Method:    setEnabled
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TNosFltChannel_setEnabled
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TNosFltChannel_setEnabled((TNosFltChannel)obj, value);
}


/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TDeviceCtrl */
/*----------------------------------------------------------------------*/
static void BDAQCALL OnNativeDeviceEvent(void *sender, DeviceEventArgs *args, void *userParam)
{
   jobject gref = (jobject)userParam;
   JNIEnv *env = g_jniTls.Get();

   if (!gref || !env) { return; }

   javaDeviceEventDelegate.Fire(env, gref, args);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    addEventHandler
 * Signature: (JILAutomation/BDaq/TDeviceEventDelegate;)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_addEventHandler
  (JNIEnv *env, jclass, jlong obj, jint id, jobject jniDelegate)
{
   jobject gref = env->NewGlobalRef(jniDelegate);
   if (gref) {
      TDeviceCtrl_addEventHandler((TDeviceCtrl)obj, (EventId)id, OnNativeDeviceEvent, gref);
   }

   // return the global referenced object as the key used when unregister event handler
   return (jlong)gref; 
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    removeEventHandler
 * Signature: (JIJ)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDeviceCtrl_removeEventHandler
  (JNIEnv *env, jclass, jlong obj, jint id, jlong jniKey)
{
   jobject gref = (jobject)jniKey;
   if (gref) {
      TDeviceCtrl_removeEventHandler((TDeviceCtrl)obj, (EventId)id, OnNativeDeviceEvent, gref);
      env->DeleteGlobalRef(gref);
   }
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    Refresh
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_Refresh
  (JNIEnv *, jclass, jlong obj)
{
   return TDeviceCtrl_Refresh((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    ReadRegister
 * Signature: (JIII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_ReadRegister
  (JNIEnv *env, jclass, jlong obj, jint space, jint offset, jint length, jbyteArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jbyte* rxbuf = env->GetByteArrayElements(buffer, NULL);

   ErrorCode ret = TDeviceCtrl_ReadRegister((TDeviceCtrl)obj, space, offset, length, rxbuf);

   env->ReleaseByteArrayElements(buffer, rxbuf, 0);
   return ret;
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    WriteRegister
 * Signature: (JIII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_WriteRegister
  (JNIEnv *env, jclass, jlong obj, jint space, jint offset, jint length, jbyteArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jbyte* txbuf = env->GetByteArrayElements(buffer, NULL);

   ErrorCode ret = TDeviceCtrl_WriteRegister((TDeviceCtrl)obj, space, offset, length, txbuf);

   env->ReleaseByteArrayElements(buffer, txbuf, JNI_ABORT);

   return ret;
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    ReadPrivateRegion
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_ReadPrivateRegion
  (JNIEnv *env, jclass, jlong obj, jint signature, jint length, jbyteArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jbyte* rxbuf = env->GetByteArrayElements(buffer, NULL);

   ErrorCode ret = TDeviceCtrl_ReadPrivateRegion((TDeviceCtrl)obj, signature, length, (uint8*)rxbuf);

   env->ReleaseByteArrayElements(buffer, rxbuf, 0);
   return ret;

}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    WritePrivateRegion
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_WritePrivateRegion
  (JNIEnv *env, jclass, jlong obj, jint signature, jint length, jbyteArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jbyte* txbuf = env->GetByteArrayElements(buffer, NULL);

   ErrorCode ret = TDeviceCtrl_WritePrivateRegion((TDeviceCtrl)obj, signature, length, (uint8*)txbuf);

   env->ReleaseByteArrayElements(buffer, txbuf, JNI_ABORT);

   return ret;
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    SynchronizeTimebase
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_SynchronizeTimebase
  (JNIEnv *, jclass, jlong obj)
{
   return TDeviceCtrl_SynchronizeTimebase((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    CalculateAbsoluteTime
 * Signature: (JD)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TDeviceCtrl_CalculateAbsoluteTime
  (JNIEnv *, jclass, jlong obj, jdouble relativeTime)
{
   return TDeviceCtrl_CalculateAbsoluteTime((TDeviceCtrl)obj, relativeTime);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getDeviceNumber
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getDeviceNumber
  (JNIEnv *, jclass, jlong obj)
{
   return TDeviceCtrl_getDeviceNumber((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getDescription
 * Signature: (JILjava/lang/StringBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getDescription
  (JNIEnv *env, jclass, jlong obj, jint length, jobject buffer)
{
   if (buffer == NULL) { return ErrorBufferIsNull; }

   wchar_t desc[MAX_DEVICE_DESC_LEN];

   ErrorCode ret = TDeviceCtrl_getDescription((TDeviceCtrl)obj, MAX_DEVICE_DESC_LEN, desc);
   if (ret == Success) {
      jchar jcs[MAX_DEVICE_DESC_LEN];
      jsize len = UtfToJChar(jcs, desc);
      javaStringBuffer.SetLength(env, buffer, 0);
      javaStringBuffer.Append(env, buffer, jcs, len);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    setDescription
 * Signature: (JILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_setDescription
   (JNIEnv *env, jclass, jlong obj, jint length, jstring desc)
{
   if (desc == NULL) { return ErrorBufferIsNull; }
   if (length > MAX_DEVICE_DESC_LEN) { return ErrorDataLenExceedLimit; }

   wchar_t xbuf[MAX_DEVICE_DESC_LEN];
   jchar const *jstr = env->GetStringChars(desc, NULL);
   JCharToUtf(xbuf, jstr);
   env->ReleaseStringChars(desc, jstr);

   return TDeviceCtrl_setDescription((TDeviceCtrl)obj, length, xbuf);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getAccessMode
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getAccessMode
  (JNIEnv *, jclass, jlong obj)
{
   return TDeviceCtrl_getAccessMode((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getProductId
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getProductId
  (JNIEnv *, jclass, jlong obj)
{
   return TDeviceCtrl_getProductId((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getBoardId
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getBoardId
  (JNIEnv *, jclass, jlong obj)
{
   return TDeviceCtrl_getBoardId((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    setBoardId
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_setBoardId
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TDeviceCtrl_setBoardId((TDeviceCtrl)obj, value);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getBoardVersion
 * Signature: (JILjava/lang/StringBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getBoardVersion
  (JNIEnv *env, jclass, jlong obj, jint length, jobject buffer)
{
   if (buffer == NULL) { return ErrorBufferIsNull; }

   wchar_t ver[64];

   ErrorCode ret = TDeviceCtrl_getBoardVersion((TDeviceCtrl)obj, 64, ver);
   if (ret == Success) {
      jchar jcs[MAX_DEVICE_DESC_LEN];
      jsize len = UtfToJChar(jcs, ver);
      javaStringBuffer.SetLength(env, buffer, 0);
      javaStringBuffer.Append(env, buffer, jcs, len);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getDriverVersion
 * Signature: (JILjava/lang/StringBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getDriverVersion
   (JNIEnv *env, jclass, jlong obj, jint length, jobject buffer)
{
   if (buffer == NULL) { return ErrorBufferIsNull; }

   wchar_t ver[64];

   ErrorCode ret = TDeviceCtrl_getDriverVersion((TDeviceCtrl)obj, 64, ver);
   if (ret == Success) {
      jchar jcs[MAX_DEVICE_DESC_LEN];
      jsize len = UtfToJChar(jcs, ver);
      javaStringBuffer.SetLength(env, buffer, 0);
      javaStringBuffer.Append(env, buffer, jcs, len);
   }

   return ret;
}
/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getDllVersion
 * Signature: (JILjava/lang/StringBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getDllVersion
  (JNIEnv *env, jclass, jlong obj, jint length, jobject buffer)
{
   if (buffer == NULL) { return ErrorBufferIsNull; }

   wchar_t ver[64];

   ErrorCode ret = TDeviceCtrl_getDllVersion((TDeviceCtrl)obj, 64, ver);
   if (ret == Success) {
      jchar jcs[MAX_DEVICE_DESC_LEN];
      jsize len = UtfToJChar(jcs, ver);
      javaStringBuffer.SetLength(env, buffer, 0);
      javaStringBuffer.Append(env, buffer, jcs, len);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getLocation
 * Signature: (JILjava/lang/StringBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getLocation
  (JNIEnv *env, jclass, jlong obj, jint length, jobject buffer)
{
   if (buffer == NULL) { return ErrorBufferIsNull; }

   wchar_t loc[64];

   ErrorCode ret = TDeviceCtrl_getLocation((TDeviceCtrl)obj, 64, loc);
   if (ret == Success) {
      jchar jcs[MAX_DEVICE_DESC_LEN];
      jsize len = UtfToJChar(jcs, loc);
      javaStringBuffer.SetLength(env, buffer, 0);
      javaStringBuffer.Append(env, buffer, jcs, len);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getPrivateRegionLength
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getPrivateRegionLength
  (JNIEnv *, jclass, jlong obj)
{
   return TDeviceCtrl_getPrivateRegionLength((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getHotResetPreventable
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getHotResetPreventable
  (JNIEnv *, jclass, jlong obj)
{
   return TDeviceCtrl_getHotResetPreventable((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getBaseAddresses
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_getBaseAddresses
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDeviceCtrl_getBaseAddresses((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getInterrupts
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_getInterrupts
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDeviceCtrl_getInterrupts((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getSupportedTerminalBoard
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_getSupportedTerminalBoard
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDeviceCtrl_getSupportedTerminalBoard((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getSupportedEvents
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_getSupportedEvents
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDeviceCtrl_getSupportedEvents((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getTerminalBoard
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getTerminalBoard
  (JNIEnv *, jclass, jlong obj)
{
   return TDeviceCtrl_getTerminalBoard((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    setTerminalBoard
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_setTerminalBoard
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TDeviceCtrl_setTerminalBoard((TDeviceCtrl)obj, (TerminalBoard)value);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    setLocateEnabled
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_setLocateEnabled
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TDeviceCtrl_setLocateEnabled((TDeviceCtrl)obj, value);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getSupportedScenarios
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getSupportedScenarios
  (JNIEnv *, jclass, jlong obj)
{
   return TDeviceCtrl_getSupportedScenarios((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getDevice
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_getDevice
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDeviceCtrl_getDevice((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getHwSpecific
 * Signature: (JLjava/lang/String;LAutomation/BDaq/IntByRef;[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getHwSpecific
  (JNIEnv *env, jclass, jlong obj, jstring name, jobject size, jbyteArray buffer)
{
   if (buffer == NULL || size == NULL) { return ErrorBufferIsNull; }

   // Convert property name from jstring to wchar_t *
   int      name_len = env->GetStringLength(name);
   wchar_t  name_buf[128];
   wchar_t *x_name = name_buf;

   if (name_len > 128) {
      x_name = new wchar_t[name_len];
   } 

   jchar const *jstr = env->GetStringChars(name, NULL);
   JCharToUtf(x_name, jstr);
   env->ReleaseStringChars(name, jstr);

   // Transfer data 
   int32    x_size = javaIntByRef.GetValue(env, size);  
   jbyte   *x_data = env->GetByteArrayElements((jbyteArray)buffer, NULL);
   ErrorCode   ret = TDeviceCtrl_getHwSpecific((TDeviceCtrl)obj, x_name, &x_size, x_data);

   env->ReleaseByteArrayElements(buffer, x_data, 0);
   javaIntByRef.SetValue(env, size, x_size);

   if (x_name != name_buf) {
      delete[] x_name;
   }
   return ret;
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    setHwSpecific
 * Signature: (JLjava/lang/String;I[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_setHwSpecific
  (JNIEnv *env, jclass, jlong obj, jstring name, jint size, jbyteArray buffer)
{
   if (buffer == NULL) { return ErrorBufferIsNull; }

   // Convert property name from jstring to wchar_t *
   int      name_len = env->GetStringLength(name);
   wchar_t  name_buf[128];
   wchar_t *x_name = name_buf;

   if (name_len > 128) {
      x_name = new wchar_t[name_len];
   } 

   jchar const *jstr = env->GetStringChars(name, NULL);
   JCharToUtf(x_name, jstr);
   env->ReleaseStringChars(name, jstr);

   // Transfer data 
   jbyte   *x_data = env->GetByteArrayElements((jbyteArray)buffer, NULL);
   ErrorCode  ret  = TDeviceCtrl_setHwSpecific((TDeviceCtrl)obj, x_name, size, x_data);

   env->ReleaseByteArrayElements(buffer, x_data, JNI_ABORT);

   if (x_name != name_buf) {
      delete[] x_name;
   }
   return ret;
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getSupportedFusionTypes
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_getSupportedFusionTypes
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDeviceCtrl_getSupportedFusionTypes((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getProgrammableSignals
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_getProgrammableSignals
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDeviceCtrl_getProgrammableSignals((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getProgrammableSignalsDirection
 * Signature: (JI[I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getProgrammableSignalsDirection
  (JNIEnv *env, jclass, jlong obj, jint count, jintArray buffer)
{
   if (!count || !buffer) { return ErrorBufferIsNull; }

   jint *rxBuf = env->GetIntArrayElements(buffer, NULL);
   ErrorCode ret = TDeviceCtrl_getProgrammableSignalsDirection((TDeviceCtrl)obj, count, (SignalDirection *)rxBuf);
   env->ReleaseIntArrayElements(buffer, rxBuf, 0);

   return ret;
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    setProgrammableSignalsDirection
 * Signature: (JI[I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_setProgrammableSignalsDirection
  (JNIEnv *env, jclass, jlong obj, jint count, jintArray buffer)
{
   if (!count || !buffer) { return ErrorBufferIsNull; }

   jint *rxBuf = env->GetIntArrayElements(buffer, NULL);
   ErrorCode ret = TDeviceCtrl_setProgrammableSignalsDirection((TDeviceCtrl)obj, count, (SignalDirection *)rxBuf);
   env->ReleaseIntArrayElements(buffer, rxBuf, JNI_ABORT);

   return ret; 
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getRtSignalEndpoints
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_getRtSignalEndpoints
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDeviceCtrl_getRtSignalEndpoints((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getRtEndpointXSupportedSources
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_getRtEndpointXSupportedSources
  (JNIEnv *, jclass, jlong obj, jint epIndex)
{
   return (jlong)TDeviceCtrl_getRtEndpointXSupportedSources((TDeviceCtrl)obj, epIndex);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getRtEndpointXSource
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_getRtEndpointXSource
  (JNIEnv *, jclass, jlong obj, jint epIndex)
{
   return TDeviceCtrl_getRtEndpointXSource((TDeviceCtrl)obj, epIndex);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    setRtEndpointXSource
 * Signature: (JII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDeviceCtrl_setRtEndpointXSource
  (JNIEnv *, jclass, jlong obj, jint epIndex, jint source)
{
   return TDeviceCtrl_setRtEndpointXSource((TDeviceCtrl)obj, epIndex, (SignalDrop)source);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    Dispose
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDeviceCtrl_Dispose
  (JNIEnv *, jclass, jlong obj)
{
   return TDeviceCtrl_Dispose((TDeviceCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    Create
 * Signature: (ILjava/lang/String;I)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_Create
  (JNIEnv *env, jclass, jint deviceNumber, jstring description, jint mode)
{
   wchar_t *desc = NULL;

   if (description) {
      // convert to wchar_t
      jchar const *jniStr = env->GetStringChars(description, NULL);
      desc = new wchar_t[env->GetStringLength(description) + 1];
      JCharToUtf(desc, jniStr);
      env->ReleaseStringChars(description, jniStr);
   }

   // create control
   TDeviceCtrl ctrl = TDeviceCtrl_Create(deviceNumber, desc, (AccessMode)mode);

   // release buffer
   if (desc) { delete []desc; }

   return (jlong)ctrl;
}

/*
 * Class:     Automation_BDaq_TDeviceCtrl
 * Method:    getInstalledDevices
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDeviceCtrl_getInstalledDevices
  (JNIEnv *, jclass)
{
   return (jlong)TDeviceCtrl_getInstalledDevices();
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TDaqCtrlBase */
/*----------------------------------------------------------------------*/
static void BDAQCALL OnNativeBfdAiEvent (void *sender, BfdAiEventArgs  *args, void *userParam)
{
   jobject gref = (jobject)userParam;
   JNIEnv *env = g_jniTls.Get();

   if (!gref || !env) { return; }

   javaBfdAiEventDelegate.Fire(env, gref, args);
}
static void BDAQCALL OnNativeBfdAoEvent (void *sender, BfdAoEventArgs  *args, void *userParam)
{
   jobject gref = (jobject)userParam;
   JNIEnv *env = g_jniTls.Get();

   if (!gref || !env) { return; }

   javaBfdAoEventDelegate.Fire(env, gref, args);
}
static void BDAQCALL OnNativeDiSnapEvent(void *sender, DiSnapEventArgs *args, void *userParam)
{
   jobject gref = (jobject)userParam;
   JNIEnv *env = g_jniTls.Get();

   if (!gref || !env) { return; }

   javaDiSnapEventDelegate.Fire(env, gref, args);
}

static void BDAQCALL OnNativeBfdDiEvent (void *sender, BfdDiEventArgs  *args, void *userParam)
{
   jobject gref = (jobject)userParam;
   JNIEnv *env = g_jniTls.Get();

   if (!gref || !env) { return; }

   javaBfdDiEventDelegate.Fire(env, gref, args);
}
static void BDAQCALL OnNativeBfdDoEvent (void *sender, BfdDoEventArgs  *args, void *userParam)
{
   jobject gref = (jobject)userParam;
   JNIEnv *env = g_jniTls.Get();

   if (!gref || !env) { return; }

   javaBfdDoEventDelegate.Fire(env, gref, args);
}

static void BDAQCALL OnNativeCntrEvent(void *sender, CntrEventArgs   *args, void *userParam)
{
   jobject gref = (jobject)userParam;
   JNIEnv *env = g_jniTls.Get();

   if (!gref || !env) { return; }

   javaCntrEventDelegate.Fire(env, gref, args);
}
static void BDAQCALL OnNativeUdCntrEvent(void *sender, UdCntrEventArgs *args, void *userParam)
{
   jobject gref = (jobject)userParam;
   JNIEnv *env = g_jniTls.Get();

   if (!gref || !env) { return; }

   javaUdCntrEventDelegate.Fire(env, gref, args);
}
static void BDAQCALL OnNativeBfdCntrEvent(void *sender, BfdCntrEventArgs *args, void *userParam)
{
   jobject gref = (jobject)userParam;
   JNIEnv *env = g_jniTls.Get();

   if (!gref || !env) { return; }

   javaBfdCntrEventDelegate.Fire(env, gref, args);
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    addEventHandler
 * Signature: (JILjava/lang/Object;)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDaqCtrlBase_addEventHandler
  (JNIEnv *env, jclass, jlong obj, jint id, jobject jniDelegate)
{
   jobject gref = env->NewGlobalRef(jniDelegate);
   if (!gref) { return 0; }

   DaqEventProc proc = NULL;
   switch(((TDaqCtrlBase)obj)->ModType())
   {
   case DaqAi:  proc = (DaqEventProc)OnNativeBfdAiEvent;  break;
   case DaqAo:  proc = (DaqEventProc)OnNativeBfdAoEvent;  break;
   case DaqDio: 
      {         
         jclass clazz = env->FindClass("Automation/BDaq/TDiSnapEventDelegate");         
         if (env->IsInstanceOf(jniDelegate, clazz)) {
            proc = (DaqEventProc)OnNativeDiSnapEvent;
            break;
         }
         
         // Buffered DI event delegate for BufferedDiCtrl;
         clazz = env->FindClass("Automation/BDaq/TBfdDiEventDelegate");
         if (env->IsInstanceOf(jniDelegate, clazz)) {
            proc = (DaqEventProc)OnNativeBfdDiEvent;
            break;
         }

         // Buffered DO event delegate for BufferedDoCtrl;
         clazz = env->FindClass("Automation/BDaq/TBfdDoEventDelegate");
         if (env->IsInstanceOf(jniDelegate, clazz)) {
            proc = (DaqEventProc)OnNativeBfdDoEvent;
            break;
         }
      }
      //proc = (DaqEventProc)OnNativeDiSnapEvent; break;
   case DaqCounter:
      {
         // Ordinary counter event delegate for timer/pulse, one-shot, etc...
         jclass clazz = env->FindClass("Automation/BDaq/TCntrEventDelegate");
         if (env->IsInstanceOf(jniDelegate, clazz)) {
            proc = (DaqEventProc)OnNativeCntrEvent;
            break;
         }

         // Up/down counter event
         clazz = env->FindClass("Automation/BDaq/TUdCntrEventDelegate");
         if (env->IsInstanceOf(jniDelegate, clazz)) {
            proc = (DaqEventProc)OnNativeUdCntrEvent;
            break;
         }

         // Buffered counter event delegate for BufferedXXXXXXCtrl;
         clazz = env->FindClass("Automation/BDaq/TBfdCntrEventDelegate");
         if (env->IsInstanceOf(jniDelegate, clazz)) {
            proc = (DaqEventProc)OnNativeBfdCntrEvent;
            break;
         }
      }
      break;
   default:
      break;
   } 

   if (proc) { 
      TDaqCtrlBase_addEventHandler((TDaqCtrlBase)obj, (EventId)id, proc, gref);
   }
   // return the global referenced object as the key used when unregister event handler
   return (jlong)gref;
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    removeEventHandler
 * Signature: (JIJ)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDaqCtrlBase_removeEventHandler
  (JNIEnv *env, jclass, jlong obj, jint id, jlong jniKey)
{
   jobject gref = (jobject)jniKey;
   if (!gref) { return; }

   DaqEventProc proc = NULL;
   switch(((TDaqCtrlBase)obj)->ModType())
   {
   case DaqAi:  proc = (DaqEventProc)OnNativeBfdAiEvent;  break;
   case DaqAo:  proc = (DaqEventProc)OnNativeBfdAoEvent;  break;
   case DaqDio: proc = (DaqEventProc)OnNativeDiSnapEvent; break;
   case DaqCounter:
      {
         jclass clazz = env->FindClass("Automation/BDaq/TCntrEventDelegate");
         if (env->IsInstanceOf(gref, clazz)) {
            proc = (DaqEventProc)OnNativeCntrEvent;
         } else {
            clazz = env->FindClass("Automation/BDaq/TUdCntrEventDelegate");
            if (env->IsInstanceOf(gref, clazz)) {
               proc = (DaqEventProc)OnNativeUdCntrEvent;
            }
         }
      }
      break;
   default:
      break;
   } 
   if (proc) { 
      TDaqCtrlBase_removeEventHandler((TDaqCtrlBase)obj, (EventId)id, proc, gref);
      env->DeleteGlobalRef(gref);
   }
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    LoadProfile
 * Signature: (JLjava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDaqCtrlBase_LoadProfile
  (JNIEnv *env, jclass, jlong obj, jstring fileName)
{
   if (!fileName) {
      return ErrorBufferIsNull;
   }

   // convert to wchar_t
   jchar const *jniStr = env->GetStringChars(fileName, NULL);
   wchar_t *fn = new wchar_t[env->GetStringLength(fileName) + 1];
   assert(fn);

   JCharToUtf(fn, jniStr);
   env->ReleaseStringChars(fileName, jniStr);

   ErrorCode ret = TDaqCtrlBase_LoadProfile((TDaqCtrlBase)obj, fn);
   delete []fn;

   return ret;
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    Cleanup
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDaqCtrlBase_Cleanup
  (JNIEnv *, jclass, jlong obj)
{
   return TDaqCtrlBase_Cleanup((TDaqCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    getSelectedDevice
 * Signature: (JLAutomation/BDaq/DeviceInformation;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDaqCtrlBase_getSelectedDevice
  (JNIEnv *env, jclass, jlong obj, jobject devInfo)
{
   DeviceInformation x;
   TDaqCtrlBase_getSelectedDevice((TDaqCtrlBase)obj, &x);
   javaDeviceInformation.SetValue(env, devInfo, x);
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    setSelectedDevice
 * Signature: (JLAutomation/BDaq/DeviceInformation;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDaqCtrlBase_setSelectedDevice
  (JNIEnv *env, jclass, jlong obj, jobject devInfo)
{
   DeviceInformation info = javaDeviceInformation.GetValue(env, devInfo);
   return TDaqCtrlBase_setSelectedDevice((TDaqCtrlBase)obj, &info);
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    getState
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDaqCtrlBase_getState
  (JNIEnv *, jclass, jlong obj)
{
   return TDaqCtrlBase_getState((TDaqCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    getDevice
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDaqCtrlBase_getDevice
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDaqCtrlBase_getDevice((TDaqCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    getSupportedDevices
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDaqCtrlBase_getSupportedDevices
  (JNIEnv *, jclass, jlong obj)
{
   TArray devs = TDaqCtrlBase_getSupportedDevices((TDaqCtrlBase)obj);

   // Fix bug#16655: wrong description under linux platform due the wchar is different with jchar in size
   if (sizeof(jchar) != sizeof(wchar_t)) { // convert description from wchar to jchar
      for (int i = 0; i < devs->getLength(); ++i) {
         DeviceTreeNode *node = (DeviceTreeNode *)devs->getItem(i);
         UtfToJChar((jchar*)node->Description, node->Description);
      }
   }

   return (jlong)devs;
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    getSupportedModes
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDaqCtrlBase_getSupportedModes
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDaqCtrlBase_getSupportedModes((TDaqCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    getModule
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDaqCtrlBase_getModule
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDaqCtrlBase_getModule((TDaqCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    Dispose
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDaqCtrlBase_Dispose
  (JNIEnv *, jclass, jlong obj)
{
   return TDaqCtrlBase_Dispose((TDaqCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TDaqCtrlBase
 * Method:    Create
 * Signature: (I)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDaqCtrlBase_Create
  (JNIEnv *, jclass, jint scenario)
{
   return (jlong)TDaqCtrlBase_Create((Scenario)scenario);
}


/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TAiChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getChannel((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getLogicalNumber
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_getLogicalNumber
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getLogicalNumber((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getValueRange
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_getValueRange
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getValueRange((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setValueRange
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setValueRange
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TAiChannel_setValueRange((TAiChannel)obj, (ValueRange)value);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getSignalType
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_getSignalType
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getSignalType((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setSignalType
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setSignalType
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TAiChannel_setSignalType((TAiChannel)obj, (AiSignalType)value);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getBurnoutRetType
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_getBurnoutRetType
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getBurnoutRetType((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setBurnoutRetType
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setBurnoutRetType
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TAiChannel_setBurnoutRetType((TAiChannel)obj, (BurnoutRetType)value);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getBurnoutRetValue
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TAiChannel_getBurnoutRetValue
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getBurnoutRetValue((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setBurnoutRetValue
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setBurnoutRetValue
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TAiChannel_setBurnoutRetValue((TAiChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getBurnShortRetValue
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TAiChannel_getBurnShortRetValue
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getBurnShortRetValue((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setBurnShortRetValue
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setBurnShortRetValue
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TAiChannel_setBurnShortRetValue((TAiChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getFilterType
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_getFilterType
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getFilterType((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setFilterType
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setFilterType
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TAiChannel_setFilterType((TAiChannel)obj, (FilterType)value);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getFilterCutoffFreq
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TAiChannel_getFilterCutoffFreq
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getFilterCutoffFreq((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setFilterCutoffFreq
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setFilterCutoffFreq
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TAiChannel_setFilterCutoffFreq((TAiChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getFilterCutoffFreq1
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TAiChannel_getFilterCutoffFreq1
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getFilterCutoffFreq1((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setFilterCutoffFreq1
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setFilterCutoffFreq1
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TAiChannel_setFilterCutoffFreq1((TAiChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getCouplingType
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_getCouplingType
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getCouplingType((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setCouplingType
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setCouplingType
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TAiChannel_setCouplingType((TAiChannel)obj, (CouplingType)value);
}
/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getIepeType
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_getIepeType
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getIepeType((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setIepeType
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setIepeType
   (JNIEnv *, jclass, jlong obj, jint value)
{
   return TAiChannel_setIepeType((TAiChannel)obj, (IepeType)value);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getImpedanceType
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_getImpedanceType
  (JNIEnv *, jclass, jlong obj)
{
   return TAiChannel_getImpedanceType((TAiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setImpedanceType
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setImpedanceType
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TAiChannel_setImpedanceType((TAiChannel)obj, (ImpedanceType)value);
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getSensorDescription
 * Signature: (JLAutomation/BDaq/IntByRef;Ljava/lang/StringBuffer;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_getSensorDescription
  (JNIEnv *env, jclass, jlong obj, jobject length, jobject buffer)
{
   int32 sizeNeed = 0;
   ErrorCode ret = TAiChannel_getSensorDescription((TAiChannel)obj, &sizeNeed, NULL);
   if (ret != Success || sizeNeed == 0) {
      javaIntByRef.SetValue(env, length, 0);
      javaStringBuffer.SetLength(env, buffer, 0);
      return ret;
   }

   wchar_t *desc = new wchar_t[sizeNeed];
   if (!desc) { return ErrorMemoryNotEnough; }

   ret = TAiChannel_getSensorDescription((TAiChannel)obj, &sizeNeed, NULL);
   assert(ret == Success);

   jchar *jcs = new jchar[sizeNeed];
   if (!jcs) { return ErrorMemoryNotEnough; }

   jsize len = UtfToJChar(jcs, desc);
   javaStringBuffer.SetLength(env, buffer, 0);
   javaStringBuffer.Append(env, buffer, jcs, len);
   javaIntByRef.SetValue(env, length, len);

   delete[] desc;
   delete[] jcs;

   return ret;
}
/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setSensorDescription
 * Signature: (JILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setSensorDescription
  (JNIEnv * env, jclass, jlong obj, jint length, jstring desc)
{
   wchar_t *xbuf;
   wchar_t zero[] = {0};

   if (length && desc) {
      xbuf = new wchar_t[length + 1];

      jchar const *jstr = env->GetStringChars(desc, NULL);
      JCharToUtf(xbuf, jstr);
      env->ReleaseStringChars(desc, jstr);
   } else {
      xbuf   = zero;
      length = 0;
   }

   ErrorCode ret = TAiChannel_setSensorDescription((TAiChannel)obj, length, xbuf);
   if (length) {
      delete[] xbuf;
   }
   return ret;
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    getScaleTable
 * Signature: (JLAutomation/BDaq/IntByRef;[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_getScaleTable
  (JNIEnv *env, jclass, jlong obj, jobject size, jbyteArray buffer)
{
   int32 sizeNeed = 0;
   ErrorCode ret = TAiChannel_getScaleTable((TAiChannel)obj, &sizeNeed, NULL);
   if (ret != Success || sizeNeed == 0) {
      javaIntByRef.SetValue(env, size, 0);
      return ret;
   }

   int sizeCurr = javaIntByRef.GetValue(env, size);
   javaIntByRef.SetValue(env, size, sizeNeed);

   if (sizeCurr < sizeNeed || buffer == NULL) {
      return buffer ? ErrorBufferTooSmall : Success;
   }

   jbyte *rxBuf = env->GetByteArrayElements(buffer, NULL);
   ret = TAiChannel_getScaleTable((TAiChannel)obj, &sizeNeed, (MapFuncPiece*)rxBuf);
   env->ReleaseByteArrayElements(buffer, rxBuf, 0);
   
   return ret;
}

/*
 * Class:     Automation_BDaq_TAiChannel
 * Method:    setScaleTalbe
 * Signature: (JI[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiChannel_setScaleTable
  (JNIEnv *env, jclass, jlong obj, jint size, jbyteArray buffer)
{
   ErrorCode ret;
   if (size && buffer) {
      jbyte *txBuf = env->GetByteArrayElements(buffer, NULL);
      ret = TAiChannel_setScaleTable((TAiChannel)obj, size, (MapFuncPiece*)txBuf);
      env->ReleaseByteArrayElements(buffer, txBuf, JNI_ABORT);
   } else {
      ret = TAiChannel_setScaleTable((TAiChannel)obj, 0, NULL);
   }

   return ret;
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TCjcSetting */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TCjcSetting
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCjcSetting_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TCjcSetting_getChannel((TCjcSetting)obj);
}

/*
 * Class:     Automation_BDaq_TCjcSetting
 * Method:    setChannel
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCjcSetting_setChannel
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TCjcSetting_setChannel((TCjcSetting)obj, value);
}

/*
 * Class:     Automation_BDaq_TCjcSetting
 * Method:    getValue
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TCjcSetting_getValue
  (JNIEnv *, jclass, jlong obj)
{
   return TCjcSetting_getValue((TCjcSetting)obj);
}

/*
 * Class:     Automation_BDaq_TCjcSetting
 * Method:    setValue
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCjcSetting_setValue
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TCjcSetting_setValue((TCjcSetting)obj, value);
}

/*
 * Class:     Automation_BDaq_TCjcSetting
 * Method:    getUpdateFreq
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TCjcSetting_getUpdateFreq
  (JNIEnv *, jclass, jlong obj)
{
   return TCjcSetting_getUpdateFreq((TCjcSetting)obj);
}

/*
 * Class:     Automation_BDaq_TCjcSetting
 * Method:    setUpdateFreq
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCjcSetting_setUpdateFreq
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TCjcSetting_setUpdateFreq((TCjcSetting)obj, value);
}


/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TAiFeatures */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getResolution
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getResolution
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getResolution((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getDataSize
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getDataSize
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getDataSize((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getDataMask
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getDataMask
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getDataMask((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getTimestampResolution
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TAiFeatures_getTimestampResolution
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getTimestampResolution((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getChannelCountMax
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getChannelCountMax
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getChannelCountMax((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getChannelType
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getChannelType
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getChannelType((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getOverallValueRange
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TAiFeatures_getOverallValueRange
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getOverallValueRange((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getValueRanges
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getValueRanges
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiFeatures_getValueRanges((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getBurnoutReturnTypes
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getBurnoutReturnTypes
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiFeatures_getBurnoutReturnTypes((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getConnectionTypes
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getConnectionTypes
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiFeatures_getConnectionTypes((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getOverallConnection
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TAiFeatures_getOverallConnection
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getOverallConnection((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getCouplingTypes
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getCouplingTypes
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiFeatures_getCouplingTypes((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getIepeTypes
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getIepeTypes
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiFeatures_getIepeTypes((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getImpedanceTypes
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getImpedanceTypes
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiFeatures_getImpedanceTypes((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getFilterTypes
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getFilterTypes
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiFeatures_getFilterTypes((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getFilterCutoffFreqRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TAiFeatures_getFilterCutoffFreqRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TAiFeatures_getFilterCutoffFreqRange((TAiFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getFilterCutoffFreq1Range
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TAiFeatures_getFilterCutoffFreq1Range
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TAiFeatures_getFilterCutoffFreq1Range((TAiFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getThermoSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TAiFeatures_getThermoSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getThermoSupported((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getCjcChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getCjcChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiFeatures_getCjcChannels((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getBufferedAiSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TAiFeatures_getBufferedAiSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getBufferedAiSupported((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getSamplingMethod
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getSamplingMethod
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getSamplingMethod((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getChannelStartBase
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getChannelStartBase
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getChannelStartBase((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getChannelCountBase
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getChannelCountBase
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getChannelCountBase((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getConvertClockSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getConvertClockSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiFeatures_getConvertClockSources((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getConvertClockRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TAiFeatures_getConvertClockRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TAiFeatures_getConvertClockRange((TAiFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getBurstScanSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TAiFeatures_getBurstScanSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getBurstScanSupported((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getScanClockSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getScanClockSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiFeatures_getScanClockSources((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getScanClockRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TAiFeatures_getScanClockRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TAiFeatures_getScanClockRange((TAiFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getScanCountMax
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getScanCountMax
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getScanCountMax((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getTriggerCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getTriggerCount
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getTriggerCount((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getRetriggerable
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TAiFeatures_getRetriggerable
  (JNIEnv *, jclass, jlong obj)
{
   return TAiFeatures_getRetriggerable((TAiFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getTriggerFilterTypes
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getTriggerFilterTypes
  (JNIEnv *, jclass, jlong obj, jint reserved)
{
   return (jlong)TAiFeatures_getTriggerFilterTypes((TAiFeatures)obj, reserved);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getTriggerCutoffFreqRange
 * Signature: (JILAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TAiFeatures_getTriggerCutoffFreqRange
  (JNIEnv *env, jclass, jlong obj, jint reserved, jobject range)
{
   MathInterval x;
   TAiFeatures_getTriggerFilterCutoffFreqRange((TAiFeatures)obj, reserved, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getTriggerActions
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getTriggerActions
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TAiFeatures_getTriggerActions((TAiFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getTriggerDelayRange
 * Signature: (JILAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TAiFeatures_getTriggerDelayRange
  (JNIEnv *env, jclass, jlong obj, jint trigger, jobject range)
{
   MathInterval x;
   TAiFeatures_getTriggerDelayRange((TAiFeatures)obj, trigger, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getTriggerSources
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiFeatures_getTriggerSources
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TAiFeatures_getTriggerSources((TAiFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getTriggerSourceVrg
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getTriggerSourceVrg
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return TAiFeatures_getTriggerSourceVrg((TAiFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getTriggerHysteresisIndexMax
 * Signature: (JI)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TAiFeatures_getTriggerHysteresisIndexMax
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return TAiFeatures_getTriggerHysteresisIndexMax((TAiFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TAiFeatures
 * Method:    getTriggerHysteresisIndexStep
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiFeatures_getTriggerHysteresisIndexStep
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return TAiFeatures_getTriggerHysteresisIndexStep((TAiFeatures)obj, trigger);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TAiCtrlBase */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TAiCtrlBase
 * Method:    getFeatures
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiCtrlBase_getFeatures
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiCtrlBase_getFeatures((TAiCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TAiCtrlBase
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAiCtrlBase_getChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAiCtrlBase_getChannels((TAiCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TAiCtrlBase
 * Method:    getChannelCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAiCtrlBase_getChannelCount
  (JNIEnv *, jclass, jlong obj)
{
   return TAiCtrlBase_getChannelCount((TAiCtrlBase)obj);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TInstantAiCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TInstantAiCtrl
 * Method:    ReadAny
 * Signature: (JIILjava/lang/Object;[D)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantAiCtrl_ReadAny
  (JNIEnv *env, jclass, jlong obj, jint chStart, jint chCount, jobject rawBuf, jdoubleArray dblBuf)
{
   int dataSize = TAiFeatures_getDataSize(TAiCtrlBase_getFeatures((TAiCtrlBase)obj));
   void * rxRawBuf = NULL;
   if (rawBuf != NULL) {
      if (dataSize == sizeof(short)) {
         rxRawBuf = env->GetShortArrayElements((jshortArray)rawBuf, NULL);
      } else {
         rxRawBuf = env->GetIntArrayElements((jintArray)rawBuf, NULL);
      }
   }

   jdouble * rxDblBuf = NULL;
   if (dblBuf != NULL) {
      rxDblBuf = env->GetDoubleArrayElements(dblBuf, NULL);
   }

   ErrorCode ret = TInstantAiCtrl_ReadAny((TInstantAiCtrl)obj, chStart, chCount, rxRawBuf, rxDblBuf);

   if (rawBuf != NULL) {
      if (dataSize == sizeof(short)) {
         env->ReleaseShortArrayElements((jshortArray)rawBuf, (jshort*)rxRawBuf, 0);
      } else {
         env->ReleaseIntArrayElements((jintArray)rawBuf, (jint*)rxRawBuf, 0);
      }
   }

   if (dblBuf != NULL) {
      env->ReleaseDoubleArrayElements(dblBuf, rxDblBuf, 0);
   }

   return ret;
}

/*
 * Class:     Automation_BDaq_TInstantAiCtrl
 * Method:    getCjc
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TInstantAiCtrl_getCjc
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TInstantAiCtrl_getCjc((TInstantAiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TInstantAiCtrl
 * Method:    getAutoConvertClockRate
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TInstantAiCtrl_getAutoConvertClockRate
  (JNIEnv *, jclass, jlong obj)
{
   return TInstantAiCtrl_getAutoConvertClockRate((TInstantAiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TInstantAiCtrl
 * Method:    setAutoConvertClockRate
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantAiCtrl_setAutoConvertClockRate
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TInstantAiCtrl_setAutoConvertClockRate((TInstantAiCtrl)obj, value);
}

/*
 * Class:     Automation_BDaq_TInstantAiCtrl
 * Method:    getAutoConvertChannelStart
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantAiCtrl_getAutoConvertChannelStart
  (JNIEnv *, jclass, jlong obj)
{
   return TInstantAiCtrl_getAutoConvertChannelStart((TInstantAiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TInstantAiCtrl
 * Method:    setAutoConvertChannelStart
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantAiCtrl_setAutoConvertChannelStart
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TInstantAiCtrl_setAutoConvertChannelStart((TInstantAiCtrl)obj, value);
}

/*
 * Class:     Automation_BDaq_TInstantAiCtrl
 * Method:    getAutoConvertChannelCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantAiCtrl_getAutoConvertChannelCount
  (JNIEnv *, jclass, jlong obj)
{
   return TInstantAiCtrl_getAutoConvertChannelCount((TInstantAiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TInstantAiCtrl
 * Method:    setAutoConvertChannelCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantAiCtrl_setAutoConvertChannelCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TInstantAiCtrl_setAutoConvertChannelCount((TInstantAiCtrl)obj, value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TBufferedAiCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    Prepare
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAiCtrl_Prepare
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAiCtrl_Prepare((TBufferedAiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    RunOnce
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAiCtrl_RunOnce
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAiCtrl_RunOnce((TBufferedAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    Start
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAiCtrl_Start
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAiCtrl_Start((TBufferedAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    Stop
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAiCtrl_Stop
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAiCtrl_Stop((TBufferedAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    Release
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBufferedAiCtrl_Release
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAiCtrl_Release((TBufferedAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    GetBufferStatus
 * Signature: (JLAutomation/BDaq/IntByRef;LAutomation/BDaq/IntByRef;)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TBufferedAiCtrl_GetBufferStatus
  (JNIEnv *env, jclass, jlong obj, jobject count, jobject offset)
{
   int32 xc, xo;
   bool ovrn = !!TBufferedAiCtrl_GetBufferStatus((TBufferedAiCtrl)obj, &xc, &xo);
   if (count) { javaIntByRef.SetValue(env, count, xc); }
   if (offset) { javaIntByRef.SetValue(env, offset, xo); }

   return ovrn;
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    GetData
 * Signature: (JIILjava/lang/Object;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAiCtrl_GetData
  (JNIEnv *env, jclass, jlong obj, jint dt, jint count, jobject buffer)
{
   ErrorCode ret = ErrorParamNotSpted;
   void * rxBuf;
   if (dt == sizeof(short))    { 
      rxBuf = env->GetShortArrayElements((jshortArray)buffer, NULL); 
      ret   = TBufferedAiCtrl_GetData((TBufferedAiCtrl)obj, dt, count, rxBuf);
      env->ReleaseShortArrayElements((jshortArray)buffer, (jshort*)rxBuf, 0);
   } else if (dt == sizeof(int)) { 
      rxBuf = env->GetIntArrayElements((jintArray)buffer, NULL); 
      ret   = TBufferedAiCtrl_GetData((TBufferedAiCtrl)obj, dt, count, rxBuf);
      env->ReleaseIntArrayElements((jintArray)buffer, (jint*)rxBuf, 0);
   } else if (dt == sizeof(double)) {
      rxBuf = env->GetDoubleArrayElements((jdoubleArray)buffer, NULL); 
      ret   = TBufferedAiCtrl_GetData((TBufferedAiCtrl)obj, dt, count, rxBuf);
      env->ReleaseDoubleArrayElements((jdoubleArray)buffer, (jdouble*)rxBuf, 0);
   }
   return ret; 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    getBuffer
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedAiCtrl_getBuffer
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedAiCtrl_getBuffer((TBufferedAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    getBufferCapacity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAiCtrl_getBufferCapacity
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAiCtrl_getBufferCapacity((TBufferedAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    getScanChannel
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedAiCtrl_getScanChannel
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedAiCtrl_getScanChannel((TBufferedAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    getConvertClock
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedAiCtrl_getConvertClock
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedAiCtrl_getConvertClock((TBufferedAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    getScanClock
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedAiCtrl_getScanClock
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedAiCtrl_getScanClock((TBufferedAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    getTrigger
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedAiCtrl_getTrigger
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TBufferedAiCtrl_getTrigger((TBufferedAiCtrl)obj, trigger); 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    getStreaming
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TBufferedAiCtrl_getStreaming
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAiCtrl_getStreaming((TBufferedAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBufferedAiCtrl
 * Method:    setStreaming
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAiCtrl_setStreaming
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TBufferedAiCtrl_setStreaming((TBufferedAiCtrl)obj, !!value); 
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TWaveformAiCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TWaveformAiCtrl
 * Method:    Prepare
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TWaveformAiCtrl_Prepare
  (JNIEnv *, jclass, jlong obj)
{
   return TWaveformAiCtrl_Prepare((TWaveformAiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TWaveformAiCtrl
 * Method:    Start
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TWaveformAiCtrl_Start
  (JNIEnv *, jclass, jlong obj)
{
   return TWaveformAiCtrl_Start((TWaveformAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TWaveformAiCtrl
 * Method:    Stop
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TWaveformAiCtrl_Stop
  (JNIEnv *, jclass, jlong obj)
{
   return TWaveformAiCtrl_Stop((TWaveformAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TWaveformAiCtrl
 * Method:    Release
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TWaveformAiCtrl_Release
  (JNIEnv *, jclass, jlong obj)
{
   return TWaveformAiCtrl_Release((TWaveformAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TWaveformAiCtrl
 * Method:    GetBufferStatus
 * Signature: (JLAutomation/BDaq/IntByRef;LAutomation/BDaq/IntByRef;)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TWaveformAiCtrl_GetBufferStatus
  (JNIEnv *env, jclass, jlong obj, jobject count, jobject offset)
{
   int32 xc, xo;
   bool ovrn = !!TWaveformAiCtrl_GetBufferStatus((TWaveformAiCtrl)obj, &xc, &xo);
   if (count) { javaIntByRef.SetValue(env, count, xc); }
   if (offset) { javaIntByRef.SetValue(env, offset, xo); }

   return ovrn;

}

/*
 * Class:     Automation_BDaq_TWaveformAiCtrl
 * Method:    GetData
 * Signature: (JIILjava/lang/Object;ILAutomation/BDaq/IntByRef;LAutomation/BDaq/DoubleByRef;LAutomation/BDaq/IntByRef;[LAutomation/BDaq/DataMarker;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TWaveformAiCtrl_GetData
  (JNIEnv *env, jclass, jlong obj, jint dt, jint count, jobject buffer, jint timeout,
   jobject returned, jobject startTime, jobject markCount, jobjectArray markBuf)
{
   double xstime = 0;
   int32 xreted = 0, xmcount = 0, org_xmcount = 0;

   double *pxstime  = startTime ? &xstime : NULL;
   int32  *pxreted  = returned ? &xreted : NULL;
   int32  *pxmcount = NULL;
   DataMark *xmbuf  = NULL;

   // mark buffer
   if (markCount && markBuf) {
      xmcount = org_xmcount = javaIntByRef.GetValue(env, markCount);  
      if (xmcount) {
         pxmcount = &xmcount;
         xmbuf = new DataMark[xmcount];
         if (!xmbuf) { return ErrorMemoryNotEnough; }
      }
   }

   ErrorCode ret = ErrorParamNotSpted;
   void * rxBuf;
   if (dt == sizeof(short))    { 
      rxBuf = env->GetShortArrayElements((jshortArray)buffer, NULL); 
      ret   = TWaveformAiCtrl_GetData((TWaveformAiCtrl)obj, dt, 
                  count, rxBuf, timeout, pxreted, pxstime, pxmcount, xmbuf);
      env->ReleaseShortArrayElements((jshortArray)buffer, (jshort*)rxBuf, 0);
   } else if (dt == sizeof(int)) { 
      rxBuf = env->GetIntArrayElements((jintArray)buffer, NULL); 
      ret   = TWaveformAiCtrl_GetData((TWaveformAiCtrl)obj, dt, 
                  count, rxBuf, timeout, pxreted, pxstime, pxmcount, xmbuf);
      env->ReleaseIntArrayElements((jintArray)buffer, (jint*)rxBuf, 0);
   } else if (dt == sizeof(double)) {
      rxBuf = env->GetDoubleArrayElements((jdoubleArray)buffer, NULL); 
      ret   = TWaveformAiCtrl_GetData((TWaveformAiCtrl)obj, dt, 
                  count, rxBuf, timeout, pxreted, pxstime, pxmcount, xmbuf);
      env->ReleaseDoubleArrayElements((jdoubleArray)buffer, (jdouble*)rxBuf, 0);
   }

   if (xmcount) {
      for (int i = 0; i < x_min(xmcount, org_xmcount); ++i) {
         jobject item = env->GetObjectArrayElement(markBuf, i);
         javaDataMark.SetValue(env, item, xmbuf[i]);
      }
   }

   if (startTime){ javaDoubleByRef.SetValue(env, startTime, xstime); }
   if (returned) { javaIntByRef.SetValue(env, returned, xreted); }
   if (markCount){ javaIntByRef.SetValue(env, markCount, xmcount); }

   if (xmbuf) { delete []xmbuf; }

   return ret; 

}

/*
 * Class:     Automation_BDaq_TWaveformAiCtrl
 * Method:    getBuffer
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TWaveformAiCtrl_getBuffer
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TWaveformAiCtrl_getBuffer((TWaveformAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TWaveformAiCtrl
 * Method:    getBufferCapacity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TWaveformAiCtrl_getBufferCapacity
  (JNIEnv *, jclass, jlong obj)
{
   return TWaveformAiCtrl_getBufferCapacity((TWaveformAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TWaveformAiCtrl
 * Method:    getConversion
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TWaveformAiCtrl_getConversion
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TWaveformAiCtrl_getConversion((TWaveformAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TWaveformAiCtrl
 * Method:    getRecord
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TWaveformAiCtrl_getRecord
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TWaveformAiCtrl_getRecord((TWaveformAiCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TWaveformAiCtrl
 * Method:    getTrigger
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TWaveformAiCtrl_getTrigger
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TWaveformAiCtrl_getTrigger((TWaveformAiCtrl)obj, trigger); 
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TAoChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TAoChannel
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoChannel_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TAoChannel_getChannel((TAoChannel)obj);
}

/*
 * Class:     Automation_BDaq_TAoChannel
 * Method:    getValueRange
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoChannel_getValueRange
  (JNIEnv *, jclass, jlong obj)
{
   return TAoChannel_getValueRange((TAoChannel)obj); 
}

/*
 * Class:     Automation_BDaq_TAoChannel
 * Method:    setValueRange
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoChannel_setValueRange
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TAoChannel_setValueRange((TAoChannel)obj, (ValueRange)value); 
}

/*
 * Class:     Automation_BDaq_TAoChannel
 * Method:    getExtRefUnipolar
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TAoChannel_getExtRefUnipolar
  (JNIEnv *, jclass, jlong obj)
{
   return TAoChannel_getExtRefUnipolar((TAoChannel)obj); 
}

/*
 * Class:     Automation_BDaq_TAoChannel
 * Method:    setExtRefUnipolar
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoChannel_setExtRefUnipolar
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TAoChannel_setExtRefUnipolar((TAoChannel)obj, value); 
}

/*
 * Class:     Automation_BDaq_TAoChannel
 * Method:    getExtRefBipolar
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TAoChannel_getExtRefBipolar
  (JNIEnv *, jclass, jlong obj)
{
   return TAoChannel_getExtRefBipolar((TAoChannel)obj); 
}

/*
 * Class:     Automation_BDaq_TAoChannel
 * Method:    setExtRefBipolar
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoChannel_setExtRefBipolar
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TAoChannel_setExtRefBipolar((TAoChannel)obj, value); 
}

/*
 * Class:     Automation_BDaq_TAoChannel
 * Method:    getScaleTable
 * Signature: (JLAutomation/BDaq/IntByRef;[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoChannel_getScaleTable
  (JNIEnv *env, jclass, jlong obj, jobject size, jbyteArray buffer)
{
   int32 sizeNeed = 0;
   ErrorCode ret = TAoChannel_getScaleTable((TAoChannel)obj, &sizeNeed, NULL);
   if (ret != Success || sizeNeed == 0) {
      javaIntByRef.SetValue(env, size, 0);
      return ret;
   }

   int sizeCurr = javaIntByRef.GetValue(env, size);
   javaIntByRef.SetValue(env, size, sizeNeed);

   if (sizeCurr < sizeNeed || buffer == NULL) {
      return buffer ? ErrorBufferTooSmall : Success;
   }

   jbyte *rxBuf = env->GetByteArrayElements(buffer, NULL);
   ret = TAoChannel_getScaleTable((TAoChannel)obj, &sizeNeed, (MapFuncPiece*)rxBuf);
   env->ReleaseByteArrayElements(buffer, rxBuf, 0);
   
   return ret;
}

/*
 * Class:     Automation_BDaq_TAoChannel
 * Method:    setScaleTable
 * Signature: (JI[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoChannel_setScaleTable
  (JNIEnv *env, jclass, jlong obj, jint size, jbyteArray buffer)
{
   ErrorCode ret;
   if (size && buffer) {
      jbyte *txBuf = env->GetByteArrayElements(buffer, NULL);
      ret = TAoChannel_setScaleTable((TAoChannel)obj, size, (MapFuncPiece*)txBuf);
      env->ReleaseByteArrayElements(buffer, txBuf, JNI_ABORT);
   } else {
      ret = TAoChannel_setScaleTable((TAoChannel)obj, 0, NULL);
   }

   return ret;
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TAoFeatures */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getResolution
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoFeatures_getResolution
  (JNIEnv *, jclass, jlong obj)
{
   return TAoFeatures_getResolution((TAoFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getDataSize
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoFeatures_getDataSize
  (JNIEnv *, jclass, jlong obj)
{
   return TAoFeatures_getDataSize((TAoFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getDataMask
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoFeatures_getDataMask
  (JNIEnv *, jclass, jlong obj)
{
   return TAoFeatures_getDataMask((TAoFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getChannelCountMax
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoFeatures_getChannelCountMax
  (JNIEnv *, jclass, jlong obj)
{
   return TAoFeatures_getChannelCountMax((TAoFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getValueRanges
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAoFeatures_getValueRanges
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAoFeatures_getValueRanges((TAoFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getExternalRefAntiPolar
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TAoFeatures_getExternalRefAntiPolar
  (JNIEnv *, jclass, jlong obj)
{
   return TAoFeatures_getExternalRefAntiPolar((TAoFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getExternalRefRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TAoFeatures_getExternalRefRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TAoFeatures_getExternalRefRange((TAoFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getBufferedAoSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TAoFeatures_getBufferedAoSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TAoFeatures_getBufferedAoSupported((TAoFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getSamplingMethod
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoFeatures_getSamplingMethod
  (JNIEnv *, jclass, jlong obj)
{
   return TAoFeatures_getSamplingMethod((TAoFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getChannelStartBase
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoFeatures_getChannelStartBase
  (JNIEnv *, jclass, jlong obj)
{
   return TAoFeatures_getChannelStartBase((TAoFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getChannelCountBase
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoFeatures_getChannelCountBase
  (JNIEnv *, jclass, jlong obj)
{
   return TAoFeatures_getChannelCountBase((TAoFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getConvertClockSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAoFeatures_getConvertClockSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAoFeatures_getConvertClockSources((TAoFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getConvertClockRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TAoFeatures_getConvertClockRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TAoFeatures_getConvertClockRange((TAoFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getTriggerCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoFeatures_getTriggerCount
  (JNIEnv *, jclass, jlong obj)
{
   return TAoFeatures_getTriggerCount((TAoFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getRetriggerable
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TAoFeatures_getRetriggerable
  (JNIEnv *, jclass, jlong obj)
{
   return TAoFeatures_getRetriggerable((TAoFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getTriggerSources
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAoFeatures_getTriggerSources
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TAoFeatures_getTriggerSources((TAoFeatures)obj, trigger); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getTriggerActions
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAoFeatures_getTriggerActions
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TAoFeatures_getTriggerActions((TAoFeatures)obj, trigger); 
}

/*
 * Class:     Automation_BDaq_TAoFeatures
 * Method:    getTriggerDelayRange
 * Signature: (JILAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TAoFeatures_getTriggerDelayRange
  (JNIEnv *env, jclass, jlong obj, jint trigger, jobject range)
{
   MathInterval x;
   TAoFeatures_getTriggerDelayRange((TAoFeatures)obj, trigger, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TAoCtrlBase */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TAoCtrlBase
 * Method:    getFeatures
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAoCtrlBase_getFeatures
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAoCtrlBase_getFeatures((TAoCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TAoCtrlBase
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TAoCtrlBase_getChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TAoCtrlBase_getChannels((TAoCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TAoCtrlBase
 * Method:    getChannelCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoCtrlBase_getChannelCount
  (JNIEnv *, jclass, jlong obj)
{
   return TAoCtrlBase_getChannelCount((TAoCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TAoCtrlBase
 * Method:    getExtRefValueForUnipolar
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TAoCtrlBase_getExtRefValueForUnipolar
  (JNIEnv *, jclass, jlong obj)
{
   return TAoCtrlBase_getExtRefValueForUnipolar((TAoCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TAoCtrlBase
 * Method:    setExtRefValueForUnipolar
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoCtrlBase_setExtRefValueForUnipolar
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TAoCtrlBase_setExtRefValueForUnipolar((TAoCtrlBase)obj, value);
}

/*
 * Class:     Automation_BDaq_TAoCtrlBase
 * Method:    getExtRefValueForBipolar
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TAoCtrlBase_getExtRefValueForBipolar
  (JNIEnv *, jclass, jlong obj)
{
   return TAoCtrlBase_getExtRefValueForBipolar((TAoCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TAoCtrlBase
 * Method:    setExtRefValueForBipolar
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TAoCtrlBase_setExtRefValueForBipolar
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TAoCtrlBase_setExtRefValueForBipolar((TAoCtrlBase)obj, value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TInstantAoCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TInstantAoCtrl
 * Method:    WriteAny
 * Signature: (JIILjava/lang/Object;[D)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantAoCtrl_WriteAny
  (JNIEnv *env, jclass, jlong obj, jint chStart, jint chCount, jobject rawBuf, jdoubleArray dblBuf)
{
   int dataSize = TAoFeatures_getDataSize(TAoCtrlBase_getFeatures((TAoCtrlBase)obj));
   void * rxRawBuf = NULL;
   if (rawBuf != NULL) {
      if (dataSize == sizeof(short)) {
         rxRawBuf = env->GetShortArrayElements((jshortArray)rawBuf, NULL);
      } else {
         rxRawBuf = env->GetIntArrayElements((jintArray)rawBuf, NULL);
      }
   }

   jdouble * rxDblBuf = NULL;
   if (dblBuf != NULL) {
      rxDblBuf = env->GetDoubleArrayElements(dblBuf, NULL);
   }

   ErrorCode ret = TInstantAoCtrl_WriteAny((TInstantAoCtrl)obj, chStart, chCount, rxRawBuf, rxDblBuf);

   if (rawBuf != NULL) {
      if (dataSize == sizeof(short)) {
         env->ReleaseShortArrayElements((jshortArray)rawBuf, (jshort*)rxRawBuf, JNI_ABORT);
      } else {
         env->ReleaseIntArrayElements((jintArray)rawBuf, (jint*)rxRawBuf, JNI_ABORT);
      }
   }

   if (dblBuf != NULL) {
      env->ReleaseDoubleArrayElements(dblBuf, rxDblBuf, JNI_ABORT);
   }

   return ret; 
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TBufferedAoCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    Prepare
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAoCtrl_Prepare
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAoCtrl_Prepare((TBufferedAoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    RunOnce
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAoCtrl_RunOnce
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAoCtrl_RunOnce((TBufferedAoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    Start
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAoCtrl_Start
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAoCtrl_Start((TBufferedAoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    Stop
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAoCtrl_Stop
  (JNIEnv *, jclass, jlong obj, jint action)
{
   return TBufferedAoCtrl_Stop((TBufferedAoCtrl)obj, action);
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    Release
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBufferedAoCtrl_Release
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAoCtrl_Release((TBufferedAoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    GetBufferStatus
 * Signature: (JLAutomation/BDaq/IntByRef;LAutomation/BDaq/IntByRef;)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TBufferedAoCtrl_GetBufferStatus
  (JNIEnv *env, jclass, jlong obj, jobject count, jobject offset)
{
   int32 xc, xo;
   bool ovrn = !!TBufferedAoCtrl_GetBufferStatus((TBufferedAoCtrl)obj, &xc, &xo);
   if (count) { javaIntByRef.SetValue(env, count, xc); }
   if (offset) { javaIntByRef.SetValue(env, offset, xo); }

   return ovrn;
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    SetData
 * Signature: (JIILjava/lang/Object;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAoCtrl_SetData
  (JNIEnv *env, jclass, jlong obj, jint dt, jint count, jobject buffer)
{
   ErrorCode ret = ErrorParamNotSpted;
   void * rxBuf;
   if (dt == sizeof(short))    { 
      rxBuf = env->GetShortArrayElements((jshortArray)buffer, NULL); 
      ret   = TBufferedAoCtrl_SetData((TBufferedAoCtrl)obj, dt, count, rxBuf);
      env->ReleaseShortArrayElements((jshortArray)buffer, (jshort*)rxBuf, JNI_ABORT);
   } else if (dt == sizeof(int)) { 
      rxBuf = env->GetIntArrayElements((jintArray)buffer, NULL); 
      ret   = TBufferedAoCtrl_SetData((TBufferedAoCtrl)obj, dt, count, rxBuf);
      env->ReleaseIntArrayElements((jintArray)buffer, (jint*)rxBuf, JNI_ABORT);
   } else if (dt == sizeof(double)) {
      rxBuf = env->GetDoubleArrayElements((jdoubleArray)buffer, NULL); 
      ret   = TBufferedAoCtrl_SetData((TBufferedAoCtrl)obj, dt, count, rxBuf);
      env->ReleaseDoubleArrayElements((jdoubleArray)buffer, (jdouble*)rxBuf, JNI_ABORT);
   }
   return ret; 
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    getBuffer
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedAoCtrl_getBuffer
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedAoCtrl_getBuffer((TBufferedAoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    getBufferCapacity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAoCtrl_getBufferCapacity
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAoCtrl_getBufferCapacity((TBufferedAoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    getScanChannel
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedAoCtrl_getScanChannel
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedAoCtrl_getScanChannel((TBufferedAoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    getConvertClock
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedAoCtrl_getConvertClock
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedAoCtrl_getConvertClock((TBufferedAoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    getTrigger
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedAoCtrl_getTrigger
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TBufferedAoCtrl_getTrigger((TBufferedAoCtrl)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    getStreaming
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TBufferedAoCtrl_getStreaming
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedAoCtrl_getStreaming((TBufferedAoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedAoCtrl
 * Method:    setStreaming
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedAoCtrl_setStreaming
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TBufferedAoCtrl_setStreaming((TBufferedAoCtrl)obj, value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TDioPort */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TDioPort
 * Method:    getPort
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioPort_getPort
  (JNIEnv *, jclass, jlong obj)
{
   return TDioPort_getPort((TDioPort)obj);
}

/*
 * Class:     Automation_BDaq_TDioPort
 * Method:    getDirection
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioPort_getDirection
  (JNIEnv *, jclass, jlong obj)
{
   return TDioPort_getDirection((TDioPort)obj);
}

/*
 * Class:     Automation_BDaq_TDioPort
 * Method:    setDirection
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioPort_setDirection
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TDioPort_setDirection((TDioPort)obj, (DioPortDir)value);
}

/*
 * Class:     Automation_BDaq_TDioPort
 * Method:    getDiInversePort
 * Signature: (J)B
 */
JNIEXPORT jbyte JNICALL Java_Automation_BDaq_TDioPort_getDiInversePort
  (JNIEnv *, jclass, jlong obj)
{
   return TDioPort_getDiInversePort((TDioPort)obj);
}

/*
 * Class:     Automation_BDaq_TDioPort
 * Method:    setDiInversePort
 * Signature: (JB)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioPort_setDiInversePort
  (JNIEnv *, jclass, jlong obj, jbyte value)
{
   return TDioPort_setDiInversePort((TDioPort)obj, value);
}
/*
 * Class:     Automation_BDaq_TDioPort
 * Method:    getDiOpenState
 * Signature: (J)B
 */
JNIEXPORT jbyte JNICALL Java_Automation_BDaq_TDioPort_getDiOpenState
  (JNIEnv *, jclass, jlong obj)
{
   return TDioPort_getDiOpenState((TDioPort)obj);
}

/*
 * Class:     Automation_BDaq_TDioPort
 * Method:    setDiOpenState
 * Signature: (JB)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioPort_setDiOpenState
  (JNIEnv *, jclass, jlong obj, jbyte value)
{
   return TDioPort_setDiOpenState((TDioPort)obj, value);
}

/*
 * Class:     Automation_BDaq_TDioPort
 * Method:    getDoPresetValue
 * Signature: (J)B
 */
JNIEXPORT jbyte JNICALL Java_Automation_BDaq_TDioPort_getDoPresetValue
  (JNIEnv *, jclass, jlong obj)
{
   return TDioPort_getDoPresetValue((TDioPort)obj);
}

/*
 * Class:     Automation_BDaq_TDioPort
 * Method:    setDoPresetValue
 * Signature: (JB)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioPort_setDoPresetValue
  (JNIEnv *, jclass, jlong obj, jbyte value)
{
   return TDioPort_setDoPresetValue((TDioPort)obj, value);
}

/*
 * Class:     Automation_BDaq_TDioPort
 * Method:    getDoCircuitType
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioPort_getDoCircuitType
  (JNIEnv *, jclass, jlong obj)
{
   return TDioPort_getDoCircuitType((TDioPort)obj);
}

/*
 * Class:     Automation_BDaq_TDioPort
 * Method:    setDoCircuitType
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioPort_setDoCircuitType
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TDioPort_setDoCircuitType((TDioPort)obj, (DoCircuitType)value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TDiintChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TDiintChannel
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDiintChannel_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TDiintChannel_getChannel((TDiintChannel)obj);
}

/*
 * Class:     Automation_BDaq_TDiintChannel
 * Method:    getEnabled
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDiintChannel_getEnabled
  (JNIEnv *, jclass, jlong obj)
{
   return TDiintChannel_getEnabled((TDiintChannel)obj);
}

/*
 * Class:     Automation_BDaq_TDiintChannel
 * Method:    setEnabled
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDiintChannel_setEnabled
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TDiintChannel_setEnabled((TDiintChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TDiintChannel
 * Method:    getGated
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDiintChannel_getGated
  (JNIEnv *, jclass, jlong obj)
{
   return TDiintChannel_getGated((TDiintChannel)obj);
}

/*
 * Class:     Automation_BDaq_TDiintChannel
 * Method:    setGated
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDiintChannel_setGated
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TDiintChannel_setGated((TDiintChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TDiintChannel
 * Method:    getTrigEdge
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDiintChannel_getTrigEdge
  (JNIEnv *, jclass, jlong obj)
{
   return TDiintChannel_getTrigEdge((TDiintChannel)obj);
}

/*
 * Class:     Automation_BDaq_TDiintChannel
 * Method:    setTrigEdge
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDiintChannel_setTrigEdge
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TDiintChannel_setTrigEdge((TDiintChannel)obj, (ActiveSignal)value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TDiCosintPort */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TDiCosintPort
 * Method:    getPort
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDiCosintPort_getPort
  (JNIEnv *, jclass, jlong obj)
{
   return TDiCosintPort_getPort((TDiCosintPort)obj);
}

/*
 * Class:     Automation_BDaq_TDiCosintPort
 * Method:    getMask
 * Signature: (J)B
 */
JNIEXPORT jbyte JNICALL Java_Automation_BDaq_TDiCosintPort_getMask
  (JNIEnv *, jclass, jlong obj)
{
   return TDiCosintPort_getMask((TDiCosintPort)obj);
}

/*
 * Class:     Automation_BDaq_TDiCosintPort
 * Method:    setMask
 * Signature: (JB)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDiCosintPort_setMask
  (JNIEnv *, jclass, jlong obj, jbyte value)
{
   return TDiCosintPort_setMask((TDiCosintPort)obj, value);
}


/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TDiPmintPort */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TDiPmintPort
 * Method:    getPort
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDiPmintPort_getPort
  (JNIEnv *, jclass, jlong obj)
{
   return TDiPmintPort_getPort((TDiPmintPort)obj);
}

/*
 * Class:     Automation_BDaq_TDiPmintPort
 * Method:    getMask
 * Signature: (J)B
 */
JNIEXPORT jbyte JNICALL Java_Automation_BDaq_TDiPmintPort_getMask
  (JNIEnv *, jclass, jlong obj)
{
   return TDiPmintPort_getMask((TDiPmintPort)obj);
}

/*
 * Class:     Automation_BDaq_TDiPmintPort
 * Method:    setMask
 * Signature: (JB)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDiPmintPort_setMask
  (JNIEnv *, jclass, jlong obj, jbyte value)
{
   return TDiPmintPort_setMask((TDiPmintPort)obj, value);
}

/*
 * Class:     Automation_BDaq_TDiPmintPort
 * Method:    getPattern
 * Signature: (J)B
 */
JNIEXPORT jbyte JNICALL Java_Automation_BDaq_TDiPmintPort_getPattern
  (JNIEnv *, jclass, jlong obj)
{
   return TDiPmintPort_getPattern((TDiPmintPort)obj);
}

/*
 * Class:     Automation_BDaq_TDiPmintPort
 * Method:    setPattern
 * Signature: (JB)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDiPmintPort_setPattern
  (JNIEnv *, jclass, jlong obj, jbyte value)
{
   return TDiPmintPort_setPattern((TDiPmintPort)obj, value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TDioFeatures */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getPortProgrammable
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getPortProgrammable
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getPortProgrammable((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getChannelCountMax
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getChannelCountMax
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getChannelCountMax((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getPortCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getPortCount
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getPortCount((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getPortsType
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getPortsType
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getPortsType((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getDiSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDiSupported((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getDoSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDoSupported((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getResolution
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getResolution
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getResolution((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDataSize
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getDataSize
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDataSize((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiDataMask
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDiDataMask
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDiDataMask((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiNoiseFilterSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getDiNoiseFilterSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDiNoiseFilterSupported((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiNoiseFilterOfChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDiNoiseFilterOfChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDiNoiseFilterOfChannels((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiNoiseFilterBlockTimeRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDioFeatures_getDiNoiseFilterBlockTimeRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TDioFeatures_getDiNoiseFilterBlockTimeRange((TDioFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiintSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getDiintSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDiintSupported((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiintGateSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getDiintGateSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDiintGateSupported((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiCosintSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getDiCosintSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDiCosintSupported((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiPmintSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getDiPmintSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDiPmintSupported((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiintTriggerEdges
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDiintTriggerEdges
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDiintTriggerEdges((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiintOfChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDiintOfChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDiintOfChannels((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiintGateOfChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDiintGateOfChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDiintGateOfChannels((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiCosintOfPorts
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDiCosintOfPorts
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDiCosintOfPorts((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiPmintOfPorts
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDiPmintOfPorts
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDiPmintOfPorts((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiSnapEventSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDiSnapEventSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDiSnapEventSources((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getBufferedDiSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getBufferedDiSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getBufferedDiSupported((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiSamplingMethod
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getDiSamplingMethod
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDiSamplingMethod((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiConvertClockSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDiConvertClockSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDiConvertClockSources((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiConvertClockRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDioFeatures_getDiConvertClockRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TDioFeatures_getDiConvertClockRange((TDioFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiRetriggerable
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getDiRetriggerable
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDiRetriggerable((TDioFeatures)obj);          
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiTriggerCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getDiTriggerCount
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDiTriggerCount((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiTriggerActions
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDiTriggerActions
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TDioFeatures_getDiTriggerActions((TDioFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiTriggerDelayRange
 * Signature: (JILAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDioFeatures_getDiTriggerDelayRange
  (JNIEnv *env, jclass, jlong obj, jint trigger, jobject range)
{
   MathInterval x;
   TDioFeatures_getDiTriggerDelayRange((TDioFeatures)obj, trigger, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiTriggerSources
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDiTriggerSources
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TDioFeatures_getDiTriggerSources((TDioFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiTriggerSourceVrg
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getDiTriggerSourceVrg
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return TDioFeatures_getDiTriggerSourceVrg((TDioFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiTriggerHysteresisIndexMax
 * Signature: (JI)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TDioFeatures_getDiTriggerHysteresisIndexMax
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return TDioFeatures_getDiTriggerHysteresisIndexMax((TDioFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDiTriggerHysteresisIndexStep
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getDiTriggerHysteresisIndexStep
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return TDioFeatures_getDiTriggerHysteresisIndexStep((TDioFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoDataMask
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDoDataMask
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDoDataMask((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoFreezeSignalSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDoFreezeSignalSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDoFreezeSignalSources((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoReflectWdtFeedIntervalRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDioFeatures_getDoReflectWdtFeedIntervalRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TDioFeatures_getDoReflectWdtFeedIntervalRange((TDioFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoPresetValueDepository
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getDoPresetValueDepository
  (JNIEnv *, jclass, jlong obj)
{
   return (jint)TDioFeatures_getDoPresetValueDepository((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoCircuitSelectableTypes
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDoCircuitSelectableTypes
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDoCircuitSelectableTypes((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getBufferedDoSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getBufferedDoSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getBufferedDoSupported((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoSamplingMethod
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getDoSamplingMethod
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDoSamplingMethod((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoConvertClockSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDoConvertClockSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioFeatures_getDiConvertClockSources((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoConvertClockRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDioFeatures_getDoConvertClockRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TDioFeatures_getDiConvertClockRange((TDioFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoRetriggerable
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TDioFeatures_getDoRetriggerable
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDoRetriggerable((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoTriggerCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getDoTriggerCount
  (JNIEnv *, jclass, jlong obj)
{
   return TDioFeatures_getDoTriggerCount((TDioFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoTriggerActions
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDoTriggerActions
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TDioFeatures_getDoTriggerActions((TDioFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoTriggerDelayRange
 * Signature: (JILAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TDioFeatures_getDoTriggerDelayRange
  (JNIEnv *env, jclass, jlong obj, jint trigger, jobject range)
{
   MathInterval x;
   TDioFeatures_getDoTriggerDelayRange((TDioFeatures)obj, trigger, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoTriggerSources
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioFeatures_getDoTriggerSources
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TDioFeatures_getDoTriggerSources((TDioFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoTriggerSourceVrg
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getDoTriggerSourceVrg
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return TDioFeatures_getDoTriggerSourceVrg((TDioFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoTriggerHysteresisIndexMax
 * Signature: (JI)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TDioFeatures_getDoTriggerHysteresisIndexMax
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return TDioFeatures_getDoTriggerHysteresisIndexMax((TDioFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TDioFeatures
 * Method:    getDoTriggerHysteresisIndexStep
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioFeatures_getDoTriggerHysteresisIndexStep
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return TDioFeatures_getDoTriggerHysteresisIndexStep((TDioFeatures)obj, trigger);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TDioCtrlBase */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TDioCtrlBase
 * Method:    getFeatures
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioCtrlBase_getFeatures
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioCtrlBase_getFeatures((TDioCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TDioCtrlBase
 * Method:    getPortCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TDioCtrlBase_getPortCount
  (JNIEnv *, jclass, jlong obj)
{
   return TDioCtrlBase_getPortCount((TDioCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TDioCtrlBase
 * Method:    getPorts
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TDioCtrlBase_getPorts
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TDioCtrlBase_getPorts((TDioCtrlBase)obj);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TInstantDiCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TInstantDiCtrl
 * Method:    ReadAny
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantDiCtrl_ReadAny
  (JNIEnv * env, jclass, jlong obj, jint portStart, jint portCount, jbyteArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jbyte *rxBuf = env->GetByteArrayElements(buffer, NULL);
   ErrorCode ret = TInstantDiCtrl_ReadAny((TInstantDiCtrl)obj, portStart, portCount, (uint8*)rxBuf);
   env->ReleaseByteArrayElements(buffer, rxBuf, 0);

   return ret; 
}

/*
 * Class:     Automation_BDaq_TInstantDiCtrl
 * Method:    ReadBit
 * Signature: (JIILAutomation/BDaq/ByteByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantDiCtrl_ReadBit
  (JNIEnv *env, jclass, jlong obj, jint port, jint bit, jobject value)
{
   uint8 xv = 0;
   ErrorCode ret = TInstantDiCtrl_ReadBit((TInstantDiCtrl)obj, port, bit, &xv);
   javaByteByRef.SetValue(env, value, xv);

   return ret;
}

/*
 * Class:     Automation_BDaq_TInstantDiCtrl
 * Method:    SnapStart
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantDiCtrl_SnapStart
  (JNIEnv *, jclass, jlong obj)
{
   return TInstantDiCtrl_SnapStart((TInstantDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TInstantDiCtrl
 * Method:    SnapStop
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantDiCtrl_SnapStop
  (JNIEnv *, jclass, jlong obj)
{
   return TInstantDiCtrl_SnapStop((TInstantDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TInstantDiCtrl
 * Method:    getNoiseFilterBlockTime
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TInstantDiCtrl_getNoiseFilterBlockTime
  (JNIEnv *, jclass, jlong obj)
{
   return TInstantDiCtrl_getNoiseFilterBlockTime((TInstantDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TInstantDiCtrl
 * Method:    setNoiseFilterBlockTime
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantDiCtrl_setNoiseFilterBlockTime
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TInstantDiCtrl_setNoiseFilterBlockTime((TInstantDiCtrl)obj, value);
}

/*
 * Class:     Automation_BDaq_TInstantDiCtrl
 * Method:    getNoiseFilter
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TInstantDiCtrl_getNoiseFilter
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TInstantDiCtrl_getNoiseFilter((TInstantDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TInstantDiCtrl
 * Method:    getDiintChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TInstantDiCtrl_getDiintChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TInstantDiCtrl_getDiintChannels((TInstantDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TInstantDiCtrl
 * Method:    getDiCosintPorts
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TInstantDiCtrl_getDiCosintPorts
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TInstantDiCtrl_getDiCosintPorts((TInstantDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TInstantDiCtrl
 * Method:    getDiPmintPorts
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TInstantDiCtrl_getDiPmintPorts
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TInstantDiCtrl_getDiPmintPorts((TInstantDiCtrl)obj);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TInstantDoCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TInstantDoCtrl
 * Method:    WriteAny
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantDoCtrl_WriteAny
  (JNIEnv *env, jclass, jlong obj, jint portStart, jint portCount, jbyteArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jbyte *rxBuf = env->GetByteArrayElements(buffer, NULL);
   ErrorCode ret = TInstantDoCtrl_WriteAny((TInstantDoCtrl)obj, portStart, portCount, (uint8*)rxBuf);
   env->ReleaseByteArrayElements(buffer, rxBuf, JNI_ABORT);

   return ret; 
}

/*
 * Class:     Automation_BDaq_TInstantDoCtrl
 * Method:    ReadAny
 * Signature: (JII[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantDoCtrl_ReadAny
  (JNIEnv *env, jclass, jlong obj, jint portStart, jint portCount, jbyteArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jbyte *rxBuf = env->GetByteArrayElements(buffer, NULL);
   ErrorCode ret = TInstantDoCtrl_ReadAny((TInstantDoCtrl)obj, portStart, portCount, (uint8*)rxBuf);
   env->ReleaseByteArrayElements(buffer, rxBuf, 0);

   return ret; 
}

/*
 * Class:     Automation_BDaq_TInstantDoCtrl
 * Method:    WriteBit
 * Signature: (JIIB)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantDoCtrl_WriteBit
  (JNIEnv *, jclass, jlong obj, jint port, jint bit, jbyte value)
{
   return TInstantDoCtrl_WriteBit((TInstantDoCtrl)obj, port, bit, value);
}

/*
 * Class:     Automation_BDaq_TInstantDoCtrl
 * Method:    ReadBit
 * Signature: (JIILAutomation/BDaq/ByteByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TInstantDoCtrl_ReadBit
  (JNIEnv *env, jclass, jlong obj, jint port, jint bit, jobject value)
{
   uint8 xv = 0;
   ErrorCode ret = TInstantDoCtrl_ReadBit((TInstantDoCtrl)obj, port, bit, &xv);
   javaByteByRef.SetValue(env, value, xv);

   return ret;
}

/* Header for class Automation_BDaq_TBufferedDiCtrl */
/*
 * Class:     Automation_BDaq_TBufferedDiCtrl
 * Method:    Prepare
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedDiCtrl_Prepare
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedDiCtrl_Prepare((TBufferedDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDiCtrl
 * Method:    Start
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedDiCtrl_Start
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedDiCtrl_Start((TBufferedDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDiCtrl
 * Method:    Stop
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedDiCtrl_Stop
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedDiCtrl_Stop((TBufferedDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDiCtrl
 * Method:    Release
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBufferedDiCtrl_Release
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedDiCtrl_Release((TBufferedDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDiCtrl
 * Method:    GetBufferStatus
 * Signature: (JLAutomation/BDaq/IntByRef;LAutomation/BDaq/IntByRef;)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TBufferedDiCtrl_GetBufferStatus
  (JNIEnv *env, jclass, jlong obj, jobject count, jobject offset)
{
   int32 xc, xo;
   bool ovrn = !!TBufferedDiCtrl_GetBufferStatus((TBufferedDiCtrl)obj, &xc, &xo);
   if(count)  { javaIntByRef.SetValue(env, count, xc); }
   if(offset) { javaIntByRef.SetValue(env, count, xo); }

   return ovrn;
}

/*
 * Class:     Automation_BDaq_TBufferedDiCtrl
 * Method:    GetData
 * Signature: (JI[BILAutomation/BDaq/IntByRef;LAutomation/BDaq/DoubleByRef;LAutomation/BDaq/IntByRef;[LAutomation/BDaq/DataMark;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedDiCtrl_GetData
  (JNIEnv *env, jclass, jlong obj, jint count, jbyteArray buffer, jint timeout, 
   jobject returned, jobject startTime, jobject markCount, jobjectArray markBuf)
{
   double xstime = 0;
   int32 xreted = 0, xmcount = 0, org_xmcount = 0;

   double   *pxstime  = startTime ? &xstime : NULL;
   int32    *pxreted  = returned  ? &xreted : NULL;
   int32    *pxmcount = NULL;
   DataMark *xmbuf    = NULL;

   // mark buffer
   if (markCount && markBuf){
      xmcount = org_xmcount = javaIntByRef.GetValue(env, markCount);  
      if (xmcount) {
         pxmcount = &xmcount;
         xmbuf = new DataMark[xmcount];
         if (!xmbuf) { return ErrorMemoryNotEnough; }
      }
   }

   ErrorCode ret = ErrorParamNotSpted;
   void * rxBuf;

   rxBuf = env->GetByteArrayElements(buffer, NULL); 
   ret   = TBufferedDiCtrl_GetData((TBufferedDiCtrl)obj, count, rxBuf, timeout, pxreted, pxstime, pxmcount, xmbuf);
   env->ReleaseByteArrayElements(buffer, (jbyte*)rxBuf, 0);

   if (xmcount) {
      for (int i = 0; i < x_min(xmcount, org_xmcount); ++i) {
         jobject item = env->GetObjectArrayElement(markBuf, i);
         javaDataMark.SetValue(env, item, xmbuf[i]);
      }
   }

   if (startTime){ javaDoubleByRef.SetValue(env, startTime, xstime); }
   if (returned) { javaIntByRef.SetValue(env, returned, xreted);     }
   if (markCount){ javaIntByRef.SetValue(env, markCount, xmcount);   }

   if (xmbuf) { delete []xmbuf; }

   return ret; 
}

/*
 * Class:     Automation_BDaq_TBufferedDiCtrl
 * Method:    getBuffer
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedDiCtrl_getBuffer
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedDiCtrl_getBuffer((TBufferedDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDiCtrl
 * Method:    getBufferCapacity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedDiCtrl_getBufferCapacity
  (JNIEnv *, jclass, jlong obj)

{
   return TBufferedDiCtrl_getBufferCapacity((TBufferedDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDiCtrl
 * Method:    getConvertClock
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedDiCtrl_getConvertClock
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedDiCtrl_getConvertClock((TBufferedDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDiCtrl
 * Method:    getScanPort
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedDiCtrl_getScanPort
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedDiCtrl_getScanPort((TBufferedDiCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDiCtrl
 * Method:    getTrigger
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedDiCtrl_getTrigger
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TBufferedDiCtrl_getTrigger((TBufferedDiCtrl)obj, trigger);
}

/* Header for class Automation_BDaq_TBufferedDoCtrl */
/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    Prepare
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedDoCtrl_Prepare
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedDoCtrl_Prepare((TBufferedDoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    RunOnce
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedDoCtrl_RunOnce
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedDoCtrl_RunOnce((TBufferedDoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    Start
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedDoCtrl_Start
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedDoCtrl_Start((TBufferedDoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    Stop
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedDoCtrl_Stop
  (JNIEnv *, jclass, jlong obj, jint action)
{
   return TBufferedDoCtrl_Stop((TBufferedDoCtrl)obj, action);
}

/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    Release
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBufferedDoCtrl_Release
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedDoCtrl_Release((TBufferedDoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    GetBufferStatus
 * Signature: (JLAutomation/BDaq/IntByRef;LAutomation/BDaq/IntByRef;)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TBufferedDoCtrl_GetBufferStatus
  (JNIEnv *env, jclass, jlong obj, jobject count, jobject offset)
{
   int32 xc, xo;
   bool ovrn = !!TBufferedDoCtrl_GetBufferStatus((TBufferedDoCtrl)obj, &xc, &xo);
   if (count)  { javaIntByRef.SetValue(env, count, xc); }
   if (offset) { javaIntByRef.SetValue(env, count, xc); }

   return ovrn;
}

/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    SetData
 * Signature: (JI[B)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedDoCtrl_SetData
  (JNIEnv *env, jclass, jlong obj, jint count, jbyteArray buffer)
{
   ErrorCode ret = ErrorParamNotSpted;
   void * rxBuf;

   rxBuf = env->GetByteArrayElements(buffer, NULL);
   ret   = TBufferedDoCtrl_SetData((TBufferedDoCtrl)obj, count, rxBuf);
   env->ReleaseByteArrayElements(buffer, (jbyte*)rxBuf, JNI_ABORT);

   return ret;
}

/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    getBuffer
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedDoCtrl_getBuffer
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedDoCtrl_getBuffer((TBufferedDoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    getBufferCapacity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBufferedDoCtrl_getBufferCapacity
  (JNIEnv *, jclass, jlong obj)
{
   return TBufferedDoCtrl_getBufferCapacity((TBufferedDoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    getScanPort
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedDoCtrl_getScanPort
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedDoCtrl_getScanPort((TBufferedDoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    getConvertClock
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedDoCtrl_getConvertClock
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBufferedDoCtrl_getConvertClock((TBufferedDoCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBufferedDoCtrl
 * Method:    getTrigger
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBufferedDoCtrl_getTrigger
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TBufferedDoCtrl_getTrigger((TBufferedDoCtrl)obj, trigger);                 
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TCntrIndexer */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TCounterIndexer
 * Method:    Dispose
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCounterIndexer_Dispose
  (JNIEnv *, jclass, jlong obj)
{
   return TCounterIndexer_Dispose((TCounterIndexer)obj);
}

/*
 * Class:     Automation_BDaq_TCounterIndexer
 * Method:    getLength
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCounterIndexer_getLength
  (JNIEnv *, jclass, jlong obj)
{
   return TCounterIndexer_getLength((TCounterIndexer)obj);
}

/*
 * Class:     Automation_BDaq_TCounterIndexer
 * Method:    getItem
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCounterIndexer_getItem
  (JNIEnv *, jclass, jlong obj, jint index)
{
   return (jlong)TCounterIndexer_getItem((TCounterIndexer)obj, index);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TCntrFeatures */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getChannelCountMax
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrFeatures_getChannelCountMax
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrFeatures_getChannelCountMax((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getResolution
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrFeatures_getResolution
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrFeatures_getResolution((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getDataSize
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrFeatures_getDataSize
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrFeatures_getDataSize((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getCapabilities
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getCapabilities
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getCapabilities((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getNoiseFilterSupported
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TCntrFeatures_getNoiseFilterSupported
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrFeatures_getNoiseFilterSupported((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getNoiseFilterOfChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getNoiseFilterOfChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getNoiseFilterOfChannels((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getNoiseFilterBlockTimeRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getNoiseFilterBlockTimeRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getNoiseFilterBlockTimeRange((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getEcClockSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getEcClockSources
   (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getOsClockSources((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getEcClockPolarities
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getEcClockPolarities
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getEcClockPolarities((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getEcGatePolarities
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getEcGatePolarities
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getEcGatePolarities((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getEcGateControlOfChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getEcGateControlOfChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getEcGateControlOfChannels((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getFmMethods
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getFmMethods
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getFmMethods((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getOsClockSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getOsClockSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getOsClockSources((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getOsClockPolarities
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getOsClockPolarities
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getOsClockPolarities((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getOsGateSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getOsGateSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getOsGateSources((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getOsGatePolarities
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getOsGatePolarities
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getOsGatePolarities((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getOsOutSignals
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getOsOutSignals
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getOsOutSignals((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getOsDelayCountRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getOsDelayCountRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getOsDelayCountRange((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getTmrGateControlOfChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getTmrGateControlOfChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getTmrGateControlOfChannels((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getTmrGatePolarities
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getTmrGatePolarities
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getTmrGatePolarities((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getTmrOutSignals
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getTmrOutSignals
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getTmrOutSignals((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getTmrFrequencyRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getTmrFrequencyRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getTmrFrequencyRange((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPiCascadeGroup
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getPiCascadeGroup
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getPiCascadeGroup((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPoGateControlOfChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getPoGateControlOfChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getPoGateControlOfChannels((TCntrFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPoGatePolarities
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getPoGatePolarities
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getPoGatePolarities((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPoOutSignals
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getPoOutSignals
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getPoOutSignals((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPoHiPeriodRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getPoHiPeriodRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getPoHiPeriodRange((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPoLoPeriodRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getPoLoPeriodRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getPoLoPeriodRange((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPoOutCountRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getPoOutCountRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getPoOutCountRange((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPoTriggerCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrFeatures_getPoTriggerCount
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrFeatures_getTriggerCount((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPoTriggerSources
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getPoTriggerSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getPoTriggerSources((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPoTriggerActions
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getPoTriggerActions
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getPoTriggerActions((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPoTriggerDelayRange
 * Signature: (JILAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getPoTriggerDelayRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getPoTriggerDelayRange((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getUdCountingTypes
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getUdCountingTypes
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getUdCountingTypes((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getUdInitialValues
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getUdInitialValues
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getUdInitialValues((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getUdSnapEventSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getUdSnapEventSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getUdSnapEventSources((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getCcpGateControlOfChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getCcpGateControlOfChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getCcpGateControlOfChannels((TCntrFeatures)obj); 
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getCcpGatePolarities
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getCcpGatePolarities
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getCcpGatePolarities((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getCcpOutSignals
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getCcpOutSignals
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getCcpOutSignals((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getCcpHiPeriodRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getCcpHiPeriodRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getCcpHiPeriodRange((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getCcpLoPeriodRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getCcpLoPeriodRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getCcpLoPeriodRange((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getCcpOutCountRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getCcpOutCountRange
  (JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getCcpOutCountRange((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getMeasurementTimeoutRange
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getMeasurementTimeoutRange
(JNIEnv *env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getMeasurementTimeoutRange((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getUdValueResetTimes
 * Signature: (JLAutomation/BDaq/MathInterval;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TCntrFeatures_getUdValueResetTimes
  (JNIEnv * env, jclass, jlong obj, jobject range)
{
   MathInterval x;
   TCntrFeatures_getUdValueResetTimes((TCntrFeatures)obj, &x);
   javaMathInterval.SetValue(env, range, x);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getTriggerCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrFeatures_getTriggerCount
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrFeatures_getTriggerCount((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getTriggerSources
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getTriggerSources
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TCntrFeatures_getTriggerSources((TCntrFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getTriggerActions
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getTriggerActions
  (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TCntrFeatures_getTriggerActions((TCntrFeatures)obj, trigger);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getEcSampleClockSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getEcSampleClockSources
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getEcSampleClockSources((TCntrFeatures)obj);
}

/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPiSampleClockSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getPiSampleClockSources
   (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getPiSampleClockSources((TCntrFeatures)obj);
}


/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getPoSampleClockSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getPoSampleClockSources
   (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getPoSampleClockSources((TCntrFeatures)obj);
}


/*
 * Class:     Automation_BDaq_TCntrFeatures
 * Method:    getUdSampleClockSources
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrFeatures_getUdSampleClockSources
   (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrFeatures_getUdSampleClockSources((TCntrFeatures)obj);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TCntrCtrlBase */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    getFeatures
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrCtrlBase_getFeatures
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrCtrlBase_getFeatures((TCntrCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    getChannelStart
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrCtrlBase_getChannelStart
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrCtrlBase_getChannelStart((TCntrCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    setChannelStart
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrCtrlBase_setChannelStart
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TCntrCtrlBase_setChannelStart((TCntrCtrlBase)obj, value);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    getChannelCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrCtrlBase_getChannelCount
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrCtrlBase_getChannelCount((TCntrCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    setChannelCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrCtrlBase_setChannelCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TCntrCtrlBase_setChannelCount((TCntrCtrlBase)obj, value);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    getEnabled
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TCntrCtrlBase_getEnabled
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrCtrlBase_getEnabled((TCntrCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    setEnabled
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrCtrlBase_setEnabled
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TCntrCtrlBase_setEnabled((TCntrCtrlBase)obj, value);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    getRunning
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TCntrCtrlBase_getRunning
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrCtrlBase_getRunning((TCntrCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    getNoiseFilterBlockTime
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TCntrCtrlBase_getNoiseFilterBlockTime
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrCtrlBase_getNoiseFilterBlockTime((TCntrCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    setNoiseFilterBlockTime
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrCtrlBase_setNoiseFilterBlockTime
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TCntrCtrlBase_setNoiseFilterBlockTime((TCntrCtrlBase)obj, value);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    getNoiseFilter
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TCntrCtrlBase_getNoiseFilter
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TCntrCtrlBase_getNoiseFilter((TCntrCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    getMeasurementTimeout
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TCntrCtrlBase_getMeasurementTimeout
  (JNIEnv *, jclass, jlong obj)
{
   return TCntrCtrlBase_getMeasurementTimeout((TCntrCtrlBase)obj);
}

/*
 * Class:     Automation_BDaq_TCntrCtrlBase
 * Method:    setMeasurementTimeout
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TCntrCtrlBase_setMeasurementTimeout
   (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TCntrCtrlBase_setMeasurementTimeout((TCntrCtrlBase)obj, value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TEcChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TEcChannel
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TEcChannel_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TEcChannel_getChannel((TEcChannel)obj);
}

/*
 * Class:     Automation_BDaq_TEcChannel
 * Method:    getNoiseFiltered
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TEcChannel_getNoiseFiltered
  (JNIEnv *, jclass, jlong obj)
{
   return TEcChannel_getNoiseFiltered((TEcChannel)obj);
}

/*
 * Class:     Automation_BDaq_TEcChannel
 * Method:    setNoiseFiltered
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TEcChannel_setNoiseFiltered
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TEcChannel_setNoiseFiltered((TEcChannel)obj, !!value);
}

/*
 * Class:     Automation_BDaq_TEcChannel
 * Method:    getClockSource
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TEcChannel_getClockSource
  (JNIEnv *, jclass, jlong obj)
{
   return TEcChannel_getClockSource((TEcChannel)obj);
}

/*
 * Class:     Automation_BDaq_TEcChannel
 * Method:    setClockSource
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TEcChannel_setClockSource
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TEcChannel_setClockSource((TEcChannel)obj, (SignalDrop)value);
}

/*
 * Class:     Automation_BDaq_TEcChannel
 * Method:    getClockPolarity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TEcChannel_getClockPolarity
  (JNIEnv *, jclass, jlong obj)
{
   return TEcChannel_getClockPolarity((TEcChannel)obj);
}

/*
 * Class:     Automation_BDaq_TEcChannel
 * Method:    setClockPolarity
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TEcChannel_setClockPolarity
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TEcChannel_setClockPolarity((TEcChannel)obj, (SignalPolarity)value);
}

/*
 * Class:     Automation_BDaq_TEcChannel
 * Method:    getGatePolarity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TEcChannel_getGatePolarity
  (JNIEnv *, jclass, jlong obj)
{
   return TEcChannel_getGatePolarity((TEcChannel)obj);
}

/*
 * Class:     Automation_BDaq_TEcChannel
 * Method:    setGatePolarity
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TEcChannel_setGatePolarity
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TEcChannel_setGatePolarity((TEcChannel)obj, (SignalPolarity)value);
}

/*
 * Class:     Automation_BDaq_TEcChannel
 * Method:    getGated
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TEcChannel_getGated
  (JNIEnv *, jclass, jlong obj)
{
   return TEcChannel_getGated((TEcChannel)obj);
}

/*
 * Class:     Automation_BDaq_TEcChannel
 * Method:    setGated
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TEcChannel_setGated
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TEcChannel_setGated((TEcChannel)obj, !!value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TEventCounterCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TEventCounterCtrl
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TEventCounterCtrl_getChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TEventCounterCtrl_getChannels((TEventCounterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TEventCounterCtrl
 * Method:    Read
 * Signature: (JI[I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TEventCounterCtrl_Read
  (JNIEnv *env, jclass, jlong obj, jint count, jintArray buffer)
{
   if (!count || !buffer) { return ErrorBufferIsNull; }

   jint *rxBuf = env->GetIntArrayElements(buffer, NULL);
   ErrorCode ret = TEventCounterCtrl_Read((TEventCounterCtrl)obj, count, (int32*)rxBuf);
   env->ReleaseIntArrayElements(buffer, rxBuf, 0);

   return ret;
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TFmChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TFmChannel
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TFmChannel_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TFmChannel_getChannel((TFmChannel)obj);
}

/*
 * Class:     Automation_BDaq_TFmChannel
 * Method:    getNoiseFiltered
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TFmChannel_getNoiseFiltered
  (JNIEnv *, jclass, jlong obj)
{
   return TFmChannel_getNoiseFiltered((TFmChannel)obj);
}

/*
 * Class:     Automation_BDaq_TFmChannel
 * Method:    setNoiseFiltered
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TFmChannel_setNoiseFiltered
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TFmChannel_setNoiseFiltered((TFmChannel)obj, !!value);
}

/*
 * Class:     Automation_BDaq_TFmChannel
 * Method:    getFmMethod
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TFmChannel_getFmMethod
  (JNIEnv *, jclass, jlong obj)
{
   return TFmChannel_getFmMethod((TFmChannel)obj);
}

/*
 * Class:     Automation_BDaq_TFmChannel
 * Method:    setFmMethod
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TFmChannel_setFmMethod
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TFmChannel_setFmMethod((TFmChannel)obj, (FreqMeasureMethod)value);
}

/*
 * Class:     Automation_BDaq_TFmChannel
 * Method:    getCollectionPeriod
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TFmChannel_getCollectionPeriod
  (JNIEnv *, jclass, jlong obj)
{
   return TFmChannel_getCollectionPeriod((TFmChannel)obj);
}

/*
 * Class:     Automation_BDaq_TFmChannel
 * Method:    setCollectionPeriod
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TFmChannel_setCollectionPeriod
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TFmChannel_setCollectionPeriod((TFmChannel)obj, value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TFreqMeterCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TFreqMeterCtrl
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TFreqMeterCtrl_getChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TFreqMeterCtrl_getChannels((TFreqMeterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TFreqMeterCtrl
 * Method:    Read
 * Signature: (JI[D)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TFreqMeterCtrl_Read
  (JNIEnv *env, jclass, jlong obj, jint count, jdoubleArray buffer)
{
   if (!count || !buffer) { return ErrorBufferIsNull; }

   jdouble *rxBuf = env->GetDoubleArrayElements(buffer, NULL);
   ErrorCode ret = TFreqMeterCtrl_Read((TFreqMeterCtrl)obj, count, rxBuf);
   env->ReleaseDoubleArrayElements(buffer, rxBuf, 0);

   return ret;
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TOsChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TOsChannel_getChannel((TOsChannel)obj);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    getNoiseFiltered
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TOsChannel_getNoiseFiltered
  (JNIEnv *, jclass, jlong obj)
{
   return TOsChannel_getNoiseFiltered((TOsChannel)obj);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    setNoiseFiltered
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_setNoiseFiltered
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TOsChannel_setNoiseFiltered((TOsChannel)obj, !!value);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    getDelayCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_getDelayCount
  (JNIEnv *, jclass, jlong obj)
{
   return TOsChannel_getDelayCount((TOsChannel)obj);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    setDelayCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_setDelayCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TOsChannel_setDelayCount((TOsChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    getClockSource
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_getClockSource
  (JNIEnv *, jclass, jlong obj)
{
   return TOsChannel_getClockSource((TOsChannel)obj);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    setClockSource
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_setClockSource
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TOsChannel_setClockSource((TOsChannel)obj, (SignalDrop)value);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    getClockPolarity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_getClockPolarity
  (JNIEnv *, jclass, jlong obj)
{
   return TOsChannel_getClockPolarity((TOsChannel)obj);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    setClockPolarity
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_setClockPolarity
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TOsChannel_setClockPolarity((TOsChannel)obj, (SignalPolarity)value);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    getGateSource
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_getGateSource
  (JNIEnv *, jclass, jlong obj)
{
   return TOsChannel_getGateSource((TOsChannel)obj);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    setGateSource
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_setGateSource
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TOsChannel_setGateSource((TOsChannel)obj, (SignalDrop)value);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    getGatePolarity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_getGatePolarity
  (JNIEnv *, jclass, jlong obj)
{
   return TOsChannel_getGatePolarity((TOsChannel)obj);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    setGatePolarity
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_setGatePolarity
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TOsChannel_setGatePolarity((TOsChannel)obj, (SignalPolarity)value);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    getOutSignal
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_getOutSignal
  (JNIEnv *, jclass, jlong obj)
{
   return TOsChannel_getOutSignal((TOsChannel)obj);
}

/*
 * Class:     Automation_BDaq_TOsChannel
 * Method:    setOutSignal
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TOsChannel_setOutSignal
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TOsChannel_setOutSignal((TOsChannel)obj, (OutSignalType)value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TOneShotCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TOneShotCtrl
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TOneShotCtrl_getChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TOneShotCtrl_getChannels((TOneShotCtrl)obj);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TPiChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TPiChannel
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPiChannel_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TPiChannel_getChannel((TPiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPiChannel
 * Method:    getNoiseFiltered
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TPiChannel_getNoiseFiltered
  (JNIEnv *, jclass, jlong obj)
{
   return TPiChannel_getNoiseFiltered((TPiChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPiChannel
 * Method:    setNoiseFiltered
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPiChannel_setNoiseFiltered
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TPiChannel_setNoiseFiltered((TPiChannel)obj, !!value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TPwMeterCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TPwMeterCtrl
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TPwMeterCtrl_getChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TPwMeterCtrl_getChannels((TPwMeterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TPwMeterCtrl
 * Method:    Read
 * Signature: (JI[LAutomation/BDaq/PulseWidth;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPwMeterCtrl_Read
  (JNIEnv *env, jclass, jlong obj, jint count, jobjectArray buffer)
{
   if (!count || !buffer) { return ErrorBufferIsNull; }
   
   int32 max_count = TCntrCtrlBase_getChannelCount((TCntrCtrlBase)obj);
   if (count > max_count) {count = max_count;}

   PulseWidth rxBuf[MAX_CNTR_CH_COUNT];
   ErrorCode ret = TPwMeterCtrl_Read((TPwMeterCtrl)obj, count, rxBuf);
   if (ret == Success) {
      for (int i = 0; i < count; ++i) {
         jobject item = env->GetObjectArrayElement(buffer, i);
         javaPulseWidth.SetValue(env, item, rxBuf[i]);
      }
   }

   return ret;
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TPoChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TPoChannel_getChannel((TPoChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    getNoiseFiltered
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TPoChannel_getNoiseFiltered
  (JNIEnv *, jclass, jlong obj)
{
   return TPoChannel_getNoiseFiltered((TPoChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    setNoiseFiltered
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_setNoiseFiltered
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TPoChannel_setNoiseFiltered((TPoChannel)obj, !!value);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    getPulseWidth
 * Signature: (JLAutomation/BDaq/PulseWidth;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TPoChannel_getPulseWidth
  (JNIEnv *env, jclass, jlong obj, jobject value)
{
   PulseWidth x;
   TPoChannel_getPulseWidth((TPoChannel)obj, &x);
   javaPulseWidth.SetValue(env, value, x);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    setPulseWidth
 * Signature: (JLAutomation/BDaq/PulseWidth;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_setPulseWidth
  (JNIEnv *env, jclass, jlong obj, jobject value)
{
   if (!value) { return ErrorBufferIsNull; }
   PulseWidth x = javaPulseWidth.GetValue(env, value);
   return TPoChannel_setPulseWidth((TPoChannel)obj, &x);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    getGated
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TPoChannel_getGated
  (JNIEnv *, jclass, jlong obj)
{
   return TPoChannel_getGated((TPoChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    setGated
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_setGated
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TPoChannel_setGated((TPoChannel)obj, !!value);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    getGatePolarity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_getGatePolarity
  (JNIEnv *, jclass, jlong obj)
{
   return TPoChannel_getGatePolarity((TPoChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    setGatePolarity
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_setGatePolarity
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TPoChannel_setGatePolarity((TPoChannel)obj, (SignalPolarity)value);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    getOutSignal
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_getOutSignal
  (JNIEnv *, jclass, jlong obj)
{
   return TPoChannel_getOutSignal((TPoChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    setOutSignal
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_setOutSignal
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TPoChannel_setOutSignal((TPoChannel)obj, (OutSignalType)value);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    getOutCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_getOutCount
  (JNIEnv *, jclass, jlong obj)
{
   return TPoChannel_getOutCount((TPoChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    setOutCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_setOutCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TPoChannel_setOutCount((TPoChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    getTriggerSource
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_getTriggerSource
  (JNIEnv *, jclass, jlong obj)
{
   return TPoChannel_getTriggerSource((TPoChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    setTriggerSource
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_setTriggerSource
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TPoChannel_setTriggerSource((TPoChannel)obj, SignalDrop(value));
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    getTriggerEdge
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_getTriggerEdge
  (JNIEnv *, jclass, jlong obj)
{
   return TPoChannel_getTriggerEdge((TPoChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    setTriggerEdge
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_setTriggerEdge
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TPoChannel_setTriggerEdge((TPoChannel)obj, ActiveSignal(value));
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    getTriggerAction
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_getTriggerAction
  (JNIEnv *, jclass, jlong obj)
{
   return TPoChannel_getTriggerAction((TPoChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    setTriggerAction
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_setTriggerAction
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TPoChannel_setTriggerAction((TPoChannel)obj, TriggerAction(value));
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    getTriggerDelayCount
 * Signature: (J)I
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TPoChannel_getTriggerDelayCount
  (JNIEnv *, jclass, jlong obj)
{
   return TPoChannel_getTriggerDelayCount((TPoChannel)obj);
}

/*
 * Class:     Automation_BDaq_TPoChannel
 * Method:    setTriggerDelayCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TPoChannel_setTriggerDelayCount
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TPoChannel_setTriggerDelayCount((TPoChannel)obj, value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TPwModulatorCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TPwModulatorCtrl
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TPwModulatorCtrl_getChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TPwModulatorCtrl_getChannels((TPwModulatorCtrl)obj);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TTmrChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TTmrChannel
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTmrChannel_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TTmrChannel_getChannel((TTmrChannel)obj);
}

/*
 * Class:     Automation_BDaq_TTmrChannel
 * Method:    getNoiseFiltered
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TTmrChannel_getNoiseFiltered
  (JNIEnv *, jclass, jlong obj)
{
   return TTmrChannel_getNoiseFiltered((TTmrChannel)obj);
}

/*
 * Class:     Automation_BDaq_TTmrChannel
 * Method:    setNoiseFiltered
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTmrChannel_setNoiseFiltered
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TTmrChannel_setNoiseFiltered((TTmrChannel)obj, !!value);
}

/*
 * Class:     Automation_BDaq_TTmrChannel
 * Method:    getFrequency
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_Automation_BDaq_TTmrChannel_getFrequency
  (JNIEnv *, jclass, jlong obj)
{
   return TTmrChannel_getFrequency((TTmrChannel)obj);
}

/*
 * Class:     Automation_BDaq_TTmrChannel
 * Method:    setFrequency
 * Signature: (JD)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTmrChannel_setFrequency
  (JNIEnv *, jclass, jlong obj, jdouble value)
{
   return TTmrChannel_setFrequency((TTmrChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TTmrChannel
 * Method:    getGated
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TTmrChannel_getGated
  (JNIEnv *, jclass, jlong obj)
{
   return TTmrChannel_getGated((TTmrChannel)obj);
}

/*
 * Class:     Automation_BDaq_TTmrChannel
 * Method:    setGated
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTmrChannel_setGated
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TTmrChannel_setGated((TTmrChannel)obj, !!value);
}

/*
 * Class:     Automation_BDaq_TTmrChannel
 * Method:    getGatePolarity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTmrChannel_getGatePolarity
  (JNIEnv *, jclass, jlong obj)
{
   return TTmrChannel_getGatePolarity((TTmrChannel)obj);
}

/*
 * Class:     Automation_BDaq_TTmrChannel
 * Method:    setGatePolarity
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTmrChannel_setGatePolarity
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TTmrChannel_setGatePolarity((TTmrChannel)obj, (SignalPolarity)value);
}

/*
 * Class:     Automation_BDaq_TTmrChannel
 * Method:    getOutSignal
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTmrChannel_getOutSignal
  (JNIEnv *, jclass, jlong obj)
{
   return TTmrChannel_getOutSignal((TTmrChannel)obj);
}

/*
 * Class:     Automation_BDaq_TTmrChannel
 * Method:    setOutSignal
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TTmrChannel_setOutSignal
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TTmrChannel_setOutSignal((TTmrChannel)obj, (OutSignalType)value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TTimerPulseCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TTimerPulseCtrl
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TTimerPulseCtrl_getChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TTimerPulseCtrl_getChannels((TTimerPulseCtrl)obj);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TUdChannel */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    getChannel
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_getChannel
  (JNIEnv *, jclass, jlong obj)
{
   return TUdChannel_getChannel((TUdChannel)obj);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    getNoiseFiltered
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TUdChannel_getNoiseFiltered
  (JNIEnv *, jclass, jlong obj)
{
   return TUdChannel_getNoiseFiltered((TUdChannel)obj);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    setNoiseFiltered
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_setNoiseFiltered
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TUdChannel_setNoiseFiltered((TUdChannel)obj, !!value);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    getCountingType
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_getCountingType
  (JNIEnv *, jclass, jlong obj)
{
   return TUdChannel_getCountingType((TUdChannel)obj);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    setCountingType
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_setCountingType
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TUdChannel_setCountingType((TUdChannel)obj, (CountingType)value);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    getInitialValue
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_getInitialValue
  (JNIEnv *, jclass, jlong obj)
{
   return TUdChannel_getInitialValue((TUdChannel)obj);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    setInitialValue
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_setInitialValue
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TUdChannel_setInitialValue((TUdChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    getResetTimesByIndex
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_getResetTimesByIndex
  (JNIEnv *, jclass, jlong obj)
{
   return TUdChannel_getResetTimesByIndex((TUdChannel)obj);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    setResetTimesByIndex
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_setResetTimesByIndex
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TUdChannel_setResetTimesByIndex((TUdChannel)obj, value);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    getPulseWidth
 * Signature: (JLAutomation/BDaq/PulseWidth;)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TUdChannel_getPulseWidth
  (JNIEnv *env, jclass, jlong obj, jobject value)
{
   PulseWidth x;
   TUdChannel_getPulseWidth((TUdChannel)obj, &x);
   javaPulseWidth.SetValue(env, value, x);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    setPulseWidth
 * Signature: (JLAutomation/BDaq/PulseWidth;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_setPulseWidth
  (JNIEnv *env, jclass, jlong obj, jobject value)
{
   if (!value) { return ErrorBufferIsNull; }
   PulseWidth x = javaPulseWidth.GetValue(env, value);
   return TUdChannel_setPulseWidth((TUdChannel)obj, &x);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    getGated
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TUdChannel_getGated
  (JNIEnv *, jclass, jlong obj)
{
   return TUdChannel_getGated((TUdChannel)obj);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    setGated
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_setGated
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TUdChannel_setGated((TUdChannel)obj, !!value);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    getGatePolarity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_getGatePolarity
  (JNIEnv *, jclass, jlong obj)
{
   return TUdChannel_getGatePolarity((TUdChannel)obj);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    setGatePolarity
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_setGatePolarity
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TUdChannel_setGatePolarity((TUdChannel)obj, (SignalPolarity)value);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    getOutSignal
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_getOutSignal
  (JNIEnv *, jclass, jlong obj)
{
   return TUdChannel_getOutSignal((TUdChannel)obj);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    setOutSignal
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_setOutSignal
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TUdChannel_setOutSignal((TUdChannel)obj, (OutSignalType)value);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    getOutCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_getOutCount
  (JNIEnv *, jclass, jlong obj)
{
   return TUdChannel_getOutCount((TUdChannel)obj);
}

/*
 * Class:     Automation_BDaq_TUdChannel
 * Method:    setOutCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdChannel_setOutCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TUdChannel_setOutCount((TUdChannel)obj, value);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TUdCounterCtrl */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TUdCounterCtrl
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TUdCounterCtrl_getChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TUdCounterCtrl_getChannels((TUdCounterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TUdCounterCtrl
 * Method:    Read
 * Signature: (JI[I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdCounterCtrl_Read
  (JNIEnv *env, jclass, jlong obj, jint count, jintArray buffer)
{
   if (!count || !buffer) { return ErrorBufferIsNull; }

   jint *rxBuf = env->GetIntArrayElements(buffer, NULL);
   ErrorCode ret = TUdCounterCtrl_Read((TUdCounterCtrl)obj, count, (int32*)rxBuf);
   env->ReleaseIntArrayElements(buffer, rxBuf, 0);

   return ret;
}

/*
 * Class:     Automation_BDaq_TUdCounterCtrl
 * Method:    SnapStart
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdCounterCtrl_SnapStart
  (JNIEnv *, jclass, jlong obj, jint srcId)
{
   return TUdCounterCtrl_SnapStart((TUdCounterCtrl)obj, srcId);
}

/*
 * Class:     Automation_BDaq_TUdCounterCtrl
 * Method:    SnapStop
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdCounterCtrl_SnapStop
  (JNIEnv *, jclass, jlong obj, jint srcId)
{
   return TUdCounterCtrl_SnapStop((TUdCounterCtrl)obj, srcId);
}

/*
 * Class:     Automation_BDaq_TUdCounterCtrl
 * Method:    CompareSetTable
 * Signature: (JII[I)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdCounterCtrl_CompareSetTable
  (JNIEnv *env, jclass, jlong obj, jint channel, jint count, jintArray table)
{
   if (!count || !table) { return ErrorBufferIsNull; }

   jint * txBuf = env->GetIntArrayElements(table, NULL);
   ErrorCode ret = TUdCounterCtrl_CompareSetTable((TUdCounterCtrl)obj, channel, count, (int32*)txBuf);
   env->ReleaseIntArrayElements(table, txBuf, JNI_ABORT);
   return ret;
}

/*
 * Class:     Automation_BDaq_TUdCounterCtrl
 * Method:    CompareSetInterval
 * Signature: (JIIII)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdCounterCtrl_CompareSetInterval
  (JNIEnv *, jclass, jlong obj, jint channel, jint start, jint increment, jint count)
{
   return TUdCounterCtrl_CompareSetInterval((TUdCounterCtrl)obj, channel, start, increment, count);
}

/*
 * Class:     Automation_BDaq_TUdCounterCtrl
 * Method:    CompareClear
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdCounterCtrl_CompareClear
  (JNIEnv *, jclass, jlong obj, jint channel)
{
   return TUdCounterCtrl_CompareClear((TUdCounterCtrl)obj, channel);
}

/*
 * Class:     Automation_BDaq_TUdCounterCtrl
 * Method:    ValueReset
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TUdCounterCtrl_ValueReset
  (JNIEnv *, jclass, jlong obj)
{
   return TUdCounterCtrl_ValueReset((TUdCounterCtrl)obj);
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TBfdEcCtrl                            */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TBfdEcCtrl
 * Method:    getBufferCapacity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdEcCtrl_getBufferCapacity
  (JNIEnv *, jclass, jlong obj)
{
   return TBfdEcCtrl_getBufferCapacity((TBfdEventCounterCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBfdEcCtrl
 * Method:    getChannelMap
 * Signature: (JI[Z)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBfdEcCtrl_getChannelMap
  (JNIEnv *env, jclass, jlong obj, jint count, jbooleanArray buffer)
{
   if (!buffer) { return; }

   jboolean* rxbuf = env->GetBooleanArrayElements(buffer, NULL);
   TBfdEcCtrl_getChannelMap((TBfdEventCounterCtrl)obj, count, (int8*)rxbuf);
   env->ReleaseBooleanArrayElements(buffer, rxbuf, 0);
}
/*
 * Class:     Automation_BDaq_TBfdEcCtrl
 * Method:    setChannelMap
 * Signature: (JI[Z)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdEcCtrl_setChannelMap
  (JNIEnv *env, jclass, jlong obj, jint count, jbooleanArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jboolean* txbuf = env->GetBooleanArrayElements(buffer, NULL);
   ErrorCode ret = TBfdEcCtrl_setChannelMap((TBfdEventCounterCtrl)obj, count, (int8*)txbuf);
   env->ReleaseBooleanArrayElements(buffer, txbuf, JNI_ABORT);

   return ret;
}
/*
 * Class:     Automation_BDaq_TBfdEcCtrl
 * Method:    getTrigger
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBfdEcCtrl_getTrigger
   (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TBfdEcCtrl_getTrigger((TBfdEventCounterCtrl)obj, trigger); 
}
/*
 * Class:     Automation_BDaq_TBfdEcCtrl
 * Method:    getRecord
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBfdEcCtrl_getRecord
   (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBfdEcCtrl_getRecord((TBfdEventCounterCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBfdEcCtrl
 * Method:    getSampleClkSrc
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdEcCtrl_getSampleClkSrc
  (JNIEnv *, jclass, jlong obj)
{
   return TBfdEcCtrl_getSampleClkSrc((TBfdEventCounterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdEcCtrl
 * Method:    setSampleClkSrc
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdEcCtrl_setSampleClkSrc
   (JNIEnv *, jclass, jlong obj, jint value)
{
   return TBfdEcCtrl_setSampleClkSrc((TBfdEventCounterCtrl)obj, (SignalDrop)value);
}

/*
 * Class:     Automation_BDaq_TBfdEcCtrl
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBfdEcCtrl_getChannels
  (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBfdEcCtrl_getChannels((TBfdEventCounterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdEcCtrl
 * Method:    Prepare
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdEcCtrl_Prepare
   (JNIEnv *, jclass, jlong obj)
{
   return TBfdEcCtrl_Prepare((TBfdEventCounterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdEcCtrl
 * Method:    Release
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBfdEcCtrl_Release
  (JNIEnv *, jclass, jlong obj)
{
   TBfdEcCtrl_Release((TBfdEventCounterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdEcCtrl
 * Method:    GetData
 * Signature: (JII[IILAutomation/BDaq/IntByRef;LAutomation/BDaq/DoubleByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdEcCtrl_GetData
   (JNIEnv *env, jclass, jlong obj, jint ch, jint count, jintArray buffer,
   jint timeout,jobject returned, jobject startTime)
{
   if (!count || !buffer) { return ErrorBufferIsNull; }

   double xstime = 0;
   int32 xreted = 0;

   double *pxstime  = startTime ? &xstime : NULL;
   int32  *pxreted  = returned ? &xreted : NULL;

   ErrorCode ret = ErrorParamNotSpted;
   void * rxBuf;

   rxBuf = env->GetIntArrayElements(buffer, NULL); 
   ret   = TBfdEcCtrl_GetData((TBfdEventCounterCtrl)obj, ch, count, rxBuf, timeout, pxreted, pxstime);
   env->ReleaseIntArrayElements(buffer, (jint*)rxBuf, 0);

   if (startTime){ javaDoubleByRef.SetValue(env, startTime, xstime); }
   if (returned) { javaIntByRef.SetValue(env, returned, xreted); }

   return ret; 
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TBfdPwMeterCtrl                       */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TBfdPwMeterCtrl
 * Method:    getBufferCapacity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwMeterCtrl_getBufferCapacity
   (JNIEnv *, jclass, jlong obj)
{
   return TBfdPwMeterCtrl_getBufferCapacity((TBfdPwMeterCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBfdPwMeterCtrl
 * Method:    getChannelMap
 * Signature: (JI[Z)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBfdPwMeterCtrl_getChannelMap
   (JNIEnv *env, jclass, jlong obj, jint count, jbooleanArray buffer)
{
   if (!buffer) { return; }

   jboolean* rxbuf = env->GetBooleanArrayElements(buffer, NULL);
   TBfdPwMeterCtrl_getChannelMap((TBfdPwMeterCtrl)obj, count, (int8*)rxbuf);
   env->ReleaseBooleanArrayElements(buffer, rxbuf, 0);
}

/*
 * Class:     Automation_BDaq_TBfdPwMeterCtrl
 * Method:    setChannelMap
 * Signature: (JI[Z)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwMeterCtrl_setChannelMap
   (JNIEnv *env, jclass, jlong obj, jint count, jbooleanArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jboolean* txbuf = env->GetBooleanArrayElements(buffer, NULL);
   ErrorCode ret = TBfdPwMeterCtrl_setChannelMap((TBfdPwMeterCtrl)obj, count, (int8*)txbuf);
   env->ReleaseBooleanArrayElements(buffer, txbuf, JNI_ABORT);

   return ret;
}

/*
 * Class:     Automation_BDaq_TBfdPwMeterCtrl
 * Method:    getTrigger
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBfdPwMeterCtrl_getTrigger
   (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TBfdPwMeterCtrl_getTrigger((TBfdPwMeterCtrl)obj, trigger); 
}
/*
 * Class:     Automation_BDaq_TBfdPwMeterCtrl
 * Method:    getRecord
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBfdPwMeterCtrl_getRecord
   (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBfdPwMeterCtrl_getRecord((TBfdPwMeterCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBfdPwMeterCtrl
 * Method:    getSampleClkSrc
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwMeterCtrl_getSampleClkSrc
   (JNIEnv *, jclass, jlong obj)
{
   return TBfdPwMeterCtrl_getSampleClkSrc((TBfdPwMeterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdPwMeterCtrl
 * Method:    setSampleClkSrc
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwMeterCtrl_setSampleClkSrc
   (JNIEnv *, jclass, jlong obj, jint value)
{
   return TBfdPwMeterCtrl_setSampleClkSrc((TBfdPwMeterCtrl)obj, (SignalDrop)value);
}

/*
 * Class:     Automation_BDaq_TBfdPwMeterCtrl
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBfdPwMeterCtrl_getChannels
   (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBfdPwMeterCtrl_getChannels((TBfdPwMeterCtrl)obj);
}
/*
 * Class:     Automation_BDaq_TBfdPwMeterCtrl
 * Method:    Prepare
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwMeterCtrl_Prepare
  (JNIEnv *, jclass, jlong obj)
{
   return TBfdPwMeterCtrl_Prepare((TBfdPwMeterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdPwMeterCtrl
 * Method:    Release
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBfdPwMeterCtrl_Release
  (JNIEnv *, jclass, jlong obj)
{
   TBfdPwMeterCtrl_Release((TBfdPwMeterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdPwMeterCtrl
 * Method:    GetData
 * Signature: (JII[LAutomation/BDaq/PulseWidth;ILAutomation/BDaq/IntByRef;LAutomation/BDaq/DoubleByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwMeterCtrl_GetData
   (JNIEnv *env, jclass, jlong obj, jint ch, jint count, jobjectArray buffer,
    jint timeout,jobject returned, jobject startTime)
{
   if (!count || !buffer) { return ErrorBufferIsNull; }

   double xstime = 0;
   int32 xreted = 0;

   double *pxstime  = startTime ? &xstime : NULL;
   int32  *pxreted  = returned ? &xreted : NULL;

   const int STACKBUF_THRESHOLD = 1024 / sizeof(PulseWidth);  // 1024B stack buffer, for performance
   PulseWidth stackBuf[STACKBUF_THRESHOLD]; 
   PulseWidth *rxBuf = count <= STACKBUF_THRESHOLD ? stackBuf : new PulseWidth[count];
   if (rxBuf == NULL) {
      return ErrorMemoryNotEnough;
   }

   ErrorCode ret = TBfdPwMeterCtrl_GetData((TBfdPwMeterCtrl)obj, ch, count, rxBuf, timeout, pxreted, pxstime);
   if (startTime){ javaDoubleByRef.SetValue(env, startTime, xstime); }
   if (returned) { javaIntByRef.SetValue(env, returned, xreted); }
   if (ret == Success) {
      for (int i = 0; i < count; ++i) {
         jobject item = env->GetObjectArrayElement(buffer, i);
         javaPulseWidth.SetValue(env, item, rxBuf[i]);
      }
   }

   if (rxBuf != stackBuf) { 
      delete [] rxBuf;
   }

   return ret; 
}


/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TBfdUdCounterCtrl                     */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TBfdUdCounterCtrl
 * Method:    getBufferCapacity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdUdCounterCtrl_getBufferCapacity
   (JNIEnv *, jclass, jlong obj)
{
   return TBfdUdCounterCtrl_getBufferCapacity((TBfdUdCounterCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBfdUdCounterCtrl
 * Method:    getChannelMap
 * Signature: (JI[Z)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBfdUdCounterCtrl_getChannelMap
   (JNIEnv *env, jclass, jlong obj, jint count, jbooleanArray buffer)
{
   if (!buffer) { return; }

   jboolean* rxbuf = env->GetBooleanArrayElements(buffer, NULL);
   TBfdUdCounterCtrl_getChannelMap((TBfdUdCounterCtrl)obj, count, (int8*)rxbuf);
   env->ReleaseBooleanArrayElements(buffer, rxbuf, 0);
}

/*
 * Class:     Automation_BDaq_TBfdUdCounterCtrl
 * Method:    setChannelMap
 * Signature: (JI[Z)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdUdCounterCtrl_setChannelMap
   (JNIEnv *env, jclass, jlong obj, jint count, jbooleanArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jboolean* txbuf = env->GetBooleanArrayElements(buffer, NULL);
   ErrorCode ret = TBfdUdCounterCtrl_setChannelMap((TBfdUdCounterCtrl)obj, count, (int8*)txbuf);
   env->ReleaseBooleanArrayElements(buffer, txbuf, JNI_ABORT);

   return ret;
}

/*
 * Class:     Automation_BDaq_TBfdUdCounterCtrl
 * Method:    getTrigger
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBfdUdCounterCtrl_getTrigger
   (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TBfdUdCounterCtrl_getTrigger((TBfdUdCounterCtrl)obj, trigger); 
}
/*
 * Class:     Automation_BDaq_TBfdUdCounterCtrl
 * Method:    getRecord
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBfdUdCounterCtrl_getRecord
   (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBfdUdCounterCtrl_getRecord((TBfdUdCounterCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBfdUdCounterCtrl
 * Method:    getSampleClkSrc
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdUdCounterCtrl_getSampleClkSrc
  (JNIEnv *, jclass, jlong obj)
{
   return TBfdUdCounterCtrl_getSampleClkSrc((TBfdUdCounterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdUdCounterCtrl
 * Method:    setSampleClkSrc
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdUdCounterCtrl_setSampleClkSrc
   (JNIEnv *, jclass, jlong obj, jint value)
{
   return TBfdUdCounterCtrl_setSampleClkSrc((TBfdUdCounterCtrl)obj, (SignalDrop)value);
}

/*
 * Class:     Automation_BDaq_TBfdUdCounterCtrl
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBfdUdCounterCtrl_getChannels
   (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBfdUdCounterCtrl_getChannels((TBfdUdCounterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdUdCounterCtrl
 * Method:    Prepare
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdUdCounterCtrl_Prepare
   (JNIEnv *, jclass, jlong obj)
{
   return TBfdUdCounterCtrl_Prepare((TBfdUdCounterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdUdCounterCtrl
 * Method:    Release
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBfdUdCounterCtrl_Release
  (JNIEnv *, jclass, jlong obj)
{
   TBfdUdCounterCtrl_Release((TBfdUdCounterCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdUdCounterCtrl
 * Method:    GetData
 * Signature: (JII[IILAutomation/BDaq/IntByRef;LAutomation/BDaq/DoubleByRef;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdUdCounterCtrl_GetData
   (JNIEnv *env, jclass, jlong obj, jint ch, jint count, jintArray buffer,
    jint timeout,jobject returned, jobject startTime)
{
   if (!count || !buffer) { return ErrorBufferIsNull; }

   double xstime = 0;
   int32 xreted = 0;

   double *pxstime  = startTime ? &xstime : NULL;
   int32  *pxreted  = returned ? &xreted : NULL;

   ErrorCode ret = ErrorParamNotSpted;
   void * rxBuf;

   rxBuf = env->GetIntArrayElements(buffer, NULL); 
   ret   = TBfdUdCounterCtrl_GetData((TBfdUdCounterCtrl)obj, ch, count, rxBuf, timeout, pxreted, pxstime);
   env->ReleaseIntArrayElements(buffer, (jint*)rxBuf, 0);

   if (startTime){ javaDoubleByRef.SetValue(env, startTime, xstime); }
   if (returned) { javaIntByRef.SetValue(env, returned, xreted); }

   return ret; 
}

/*----------------------------------------------------------------------*/
/* Stub for class Automation_BDaq_TBfdPwModulatorCtrl                   */
/*----------------------------------------------------------------------*/
/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    getBufferCapacity
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_getBufferCapacity
   (JNIEnv *, jclass, jlong obj)
{
   return TBfdPwModulatorCtrl_getBufferCapacity((TBfdPwModulatorCtrl)obj); 
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    getChannelMap
 * Signature: (JI[Z)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_getChannelMap
   (JNIEnv *env, jclass, jlong obj, jint count, jbooleanArray buffer)
{
   if (!buffer) { return; }

   jboolean* rxbuf = env->GetBooleanArrayElements(buffer, NULL);
   TBfdPwModulatorCtrl_getChannelMap((TBfdPwModulatorCtrl)obj, count, (int8*)rxbuf);
   env->ReleaseBooleanArrayElements(buffer, rxbuf, 0);
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    setChannelMap
 * Signature: (JI[Z)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_setChannelMap
   (JNIEnv *env, jclass, jlong obj, jint count, jbooleanArray buffer)
{
   if (!buffer) { return ErrorBufferIsNull; }

   jboolean* txbuf = env->GetBooleanArrayElements(buffer, NULL);
   ErrorCode ret = TBfdPwModulatorCtrl_setChannelMap((TBfdPwModulatorCtrl)obj, count, (int8*)txbuf);
   env->ReleaseBooleanArrayElements(buffer, txbuf, JNI_ABORT);

   return ret;
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    getTrigger
 * Signature: (JI)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_getTrigger
   (JNIEnv *, jclass, jlong obj, jint trigger)
{
   return (jlong)TBfdPwModulatorCtrl_getTrigger((TBfdPwModulatorCtrl)obj, trigger); 
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    getSamples
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_getSamples
  (JNIEnv *, jclass, jlong obj)
{
   return TBfdPwModulatorCtrl_getSamples((TBfdPwModulatorCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    setSamples
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_setSamples
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TBfdPwModulatorCtrl_setSamples((TBfdPwModulatorCtrl)obj, value);
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    getIntervalCount
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_getIntervalCount
  (JNIEnv *, jclass, jlong obj)
{
   return TBfdPwModulatorCtrl_getIntervalCount((TBfdPwModulatorCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    setIntervalCount
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_setIntervalCount
  (JNIEnv *, jclass, jlong obj, jint value)
{
   return TBfdPwModulatorCtrl_setIntervalCount((TBfdPwModulatorCtrl)obj, value);
}


/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    getStreaming
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_getStreaming
  (JNIEnv *, jclass, jlong obj)
{
   return TBfdPwModulatorCtrl_getStreaming((TBfdPwModulatorCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    setStreaming
 * Signature: (JZ)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_setStreaming
  (JNIEnv *, jclass, jlong obj, jboolean value)
{
   return TBfdPwModulatorCtrl_setStreaming((TBfdPwModulatorCtrl)obj, value);
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    getSampleClkSrc
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_getSampleClkSrc
  (JNIEnv *, jclass, jlong obj)
{
   return TBfdPwModulatorCtrl_getSampleClkSrc((TBfdPwModulatorCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    setSampleClkSrc
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_setSampleClkSrc
   (JNIEnv *, jclass, jlong obj, jint value)
{
   return TBfdPwModulatorCtrl_setSampleClkSrc((TBfdPwModulatorCtrl)obj, (SignalDrop)value);
}
/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    getChannels
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_getChannels
   (JNIEnv *, jclass, jlong obj)
{
   return (jlong)TBfdPwModulatorCtrl_getChannels((TBfdPwModulatorCtrl)obj);
}


/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    Prepare
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_Prepare
  (JNIEnv *, jclass, jlong obj)
{
   return TBfdPwModulatorCtrl_Prepare((TBfdPwModulatorCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    Release
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_Release
  (JNIEnv *, jclass, jlong obj)
{
   TBfdPwModulatorCtrl_Release((TBfdPwModulatorCtrl)obj);
}

/*
 * Class:     Automation_BDaq_TBfdPwModulatorCtrl
 * Method:    SetData
 * Signature: (JII[LAutomation/BDaq/PulseWidth;)I
 */
JNIEXPORT jint JNICALL Java_Automation_BDaq_TBfdPwModulatorCtrl_SetData
  (JNIEnv *env, jclass, jlong obj, jint ch, jint count, jobjectArray buffer)
{
   if (!count || !buffer) { return ErrorBufferIsNull; }

   const int STACKBUF_THRESHOLD = 1024 / sizeof(PulseWidth);  // 1024B stack buffer, for performance
   PulseWidth stackBuf[STACKBUF_THRESHOLD]; 
   PulseWidth *rxBuf = count <= STACKBUF_THRESHOLD ? stackBuf : new PulseWidth[count];
   if (rxBuf == NULL) {
      return ErrorMemoryNotEnough;
   }

   for (int i = 0; i < count; ++i) {
      jobject item = env->GetObjectArrayElement(buffer, i);
      rxBuf[i] = javaPulseWidth.GetValue(env, item);
   }

   ErrorCode ret = TBfdPwModulatorCtrl_SetData((TBfdPwModulatorCtrl)obj, ch, count, rxBuf);

   if (rxBuf != stackBuf) { 
      delete [] rxBuf;
   }

   return ret; 
}
